@@ -34,17 +34,26 @@ using MonoDevelop.Ide.Tasks;
 using Mono.CSharp;
 using System.Linq;
 using ICSharpCode.NRefactory;
+using MonoDevelop.CSharp.Refactoring.CodeActions;
+using MonoDevelop.Core;
+using ICSharpCode.NRefactory.CSharp.Resolver;
 
 namespace MonoDevelop.CSharp.Parser
 {
-	public class TypeSystemParser : ITypeSystemParser
+	public class TypeSystemParser : MonoDevelop.Ide.TypeSystem.TypeSystemParser
 	{
-		public ParsedDocument Parse (bool storeAst, string fileName, System.IO.TextReader content, MonoDevelop.Projects.Project project = null)
+		public override ParsedDocument Parse (bool storeAst, string fileName, System.IO.TextReader content, MonoDevelop.Projects.Project project = null)
 		{
 			var parser = new ICSharpCode.NRefactory.CSharp.CSharpParser (GetCompilerArguments (project));
 			parser.GenerateTypeSystemMode = !storeAst;
 			var result = new ParsedDocumentDecorator ();
-			
+
+			if (project != null) {
+				var projectFile = project.Files.GetFile (fileName);
+				if (projectFile != null && !TypeSystemParserNode.IsCompileBuildAction (projectFile.BuildAction))
+					result.Flags |= ParsedDocumentFlags.NonSerializable;
+			}
+
 			var tagComments = CommentTag.SpecialCommentTags.Select (t => t.Tag).ToArray ();
 			
 			parser.CompilationUnitCallback = delegate (CompilerCompilationUnit top) {
@@ -53,8 +62,11 @@ namespace MonoDevelop.CSharp.Parser
 					if (comment != null) {
 						VisitComment (result, comment, tagComments);
 					} else {
-						if (storeAst)
-							VisitPreprocessorDirective (result, special as SpecialsBag.PreProcessorDirective);
+						if (storeAst) {
+							var ppd = special as SpecialsBag.PreProcessorDirective;
+							if  (ppd != null)
+								VisitPreprocessorDirective (result, ppd);
+						}
 					}
 				}
 			};
@@ -70,14 +82,19 @@ namespace MonoDevelop.CSharp.Parser
 
 			result.LastWriteTimeUtc = pf.LastWriteTime.Value;
 			result.ParsedFile = pf;
-			result.Add (GenerateFoldings (unit, result));
+			result.Add (GetSemanticTags (unit));
+
+			result.CreateRefactoringContext = (doc, token) => MDRefactoringContext.Create (doc, doc.Editor.Caret.Location, token).Result;
+			result.CreateRefactoringContextWithEditor = (data, resolver, token) => new MDRefactoringContext ((DotNetProject)project, data, result, (CSharpAstResolver)resolver, TextLocation.Empty, token);
+
 			if (storeAst) {
 				result.Ast = unit;
+				result.Add (GenerateFoldings (unit, result));
 			}
 			return result;
 		}
 		
-		IEnumerable<FoldingRegion> GenerateFoldings (CompilationUnit unit, ParsedDocument doc)
+		IEnumerable<FoldingRegion> GenerateFoldings (SyntaxTree unit, ParsedDocument doc)
 		{
 			foreach (var fold in doc.ConditionalRegions.ToFolds ())
 				yield return fold;
@@ -90,91 +107,155 @@ namespace MonoDevelop.CSharp.Parser
 			foreach (var fold in visitor.Foldings)
 				yield return fold;
 		}
-		
+
 		class FoldingVisitor : DepthFirstAstVisitor<object, object>
 		{
 			public readonly List<FoldingRegion> Foldings = new List<FoldingRegion> ();
-			
+
+			void AddUsings (AstNode parent)
+			{
+				var firstChild = parent.Children.FirstOrDefault (child => child is UsingDeclaration || child is UsingAliasDeclaration);
+				var node = firstChild;
+				while (node != null) {
+					var next = node.GetNextNode ();
+					if (next is UsingDeclaration || next is UsingAliasDeclaration) {
+						node = next;
+					} else {
+						break;
+					}
+				}
+				if (firstChild != node) {
+					Foldings.Add (new FoldingRegion (new DomRegion (firstChild.StartLocation, node.EndLocation), FoldType.Undefined));
+				}
+			}
+			public override object VisitSyntaxTree (SyntaxTree unit, object data)
+			{
+				AddUsings (unit);
+				return base.VisitSyntaxTree (unit, data);
+			}
+
+			static TextLocation CorrectEnd (AstNode token)
+			{
+				return new TextLocation (token.EndLocation.Line, token.EndLocation.Column + 1);
+			}
+
+			static bool LastToken(AstNode arg)
+			{
+				return !(arg.Role == Roles.NewLine || arg.Role == Roles.Whitespace || arg.Role == Roles.Comment);
+			}
+		
 			public override object VisitNamespaceDeclaration (NamespaceDeclaration namespaceDeclaration, object data)
 			{
-				if (!namespaceDeclaration.RBraceToken.IsNull)
-					Foldings.Add (new FoldingRegion (new DomRegion (namespaceDeclaration.LBraceToken.GetPrevNode ().EndLocation, namespaceDeclaration.RBraceToken.EndLocation), FoldType.Undefined));
+				AddUsings (namespaceDeclaration);
+				if (!namespaceDeclaration.RBraceToken.IsNull && namespaceDeclaration.LBraceToken.StartLocation.Line != namespaceDeclaration.RBraceToken.StartLocation.Line)
+					Foldings.Add (new FoldingRegion (new DomRegion (namespaceDeclaration.LBraceToken.GetPrevNode (LastToken).EndLocation, CorrectEnd (namespaceDeclaration.RBraceToken)), FoldType.Undefined));
 				return base.VisitNamespaceDeclaration (namespaceDeclaration, data);
 			}
 			
 			public override object VisitTypeDeclaration (TypeDeclaration typeDeclaration, object data)
 			{
-				if (!typeDeclaration.RBraceToken.IsNull)
-					Foldings.Add (new FoldingRegion (new DomRegion (typeDeclaration.LBraceToken.GetPrevNode ().EndLocation, typeDeclaration.RBraceToken.StartLocation), FoldType.Type));
+				if (!typeDeclaration.RBraceToken.IsNull && typeDeclaration.LBraceToken.StartLocation.Line != typeDeclaration.RBraceToken.StartLocation.Line)
+					Foldings.Add (new FoldingRegion (new DomRegion (typeDeclaration.LBraceToken.GetPrevNode (LastToken).EndLocation, CorrectEnd (typeDeclaration.RBraceToken)), FoldType.Type));
 				return base.VisitTypeDeclaration (typeDeclaration, data);
 			}
 			
 			public override object VisitMethodDeclaration (MethodDeclaration methodDeclaration, object data)
 			{
-				if (!methodDeclaration.Body.IsNull)
-					Foldings.Add (new FoldingRegion (new DomRegion (methodDeclaration.Body.LBraceToken.GetPrevNode ().EndLocation, methodDeclaration.Body.RBraceToken.StartLocation), FoldType.Member));
+				if (!methodDeclaration.Body.IsNull && methodDeclaration.Body.LBraceToken.StartLocation.Line != methodDeclaration.Body.RBraceToken.StartLocation.Line)
+					Foldings.Add (new FoldingRegion (new DomRegion (methodDeclaration.Body.LBraceToken.GetPrevNode (LastToken).EndLocation, CorrectEnd (methodDeclaration.Body.RBraceToken)), FoldType.Member));
 				return base.VisitMethodDeclaration (methodDeclaration, data);
 			}
 			
 			public override object VisitConstructorDeclaration (ConstructorDeclaration constructorDeclaration, object data)
 			{
-				if (!constructorDeclaration.Body.IsNull)
-					Foldings.Add (new FoldingRegion (new DomRegion (constructorDeclaration.Body.LBraceToken.GetPrevNode ().EndLocation, constructorDeclaration.Body.RBraceToken.StartLocation), FoldType.Member));
+				if (!constructorDeclaration.Body.IsNull && constructorDeclaration.Body.LBraceToken.StartLocation.Line != constructorDeclaration.Body.RBraceToken.StartLocation.Line)
+					Foldings.Add (new FoldingRegion (new DomRegion (constructorDeclaration.Body.LBraceToken.GetPrevNode (LastToken).EndLocation, CorrectEnd (constructorDeclaration.Body.RBraceToken)), FoldType.Member));
 				return base.VisitConstructorDeclaration (constructorDeclaration, data);
 			}
 			
 			public override object VisitDestructorDeclaration (DestructorDeclaration destructorDeclaration, object data)
 			{
-				if (!destructorDeclaration.Body.IsNull)
-					Foldings.Add (new FoldingRegion (new DomRegion (destructorDeclaration.Body.LBraceToken.GetPrevNode ().EndLocation, destructorDeclaration.Body.RBraceToken.StartLocation), FoldType.Member));
+				if (!destructorDeclaration.Body.IsNull && destructorDeclaration.Body.LBraceToken.StartLocation.Line != destructorDeclaration.Body.RBraceToken.StartLocation.Line)
+					Foldings.Add (new FoldingRegion (new DomRegion (destructorDeclaration.Body.LBraceToken.GetPrevNode (LastToken).EndLocation, CorrectEnd (destructorDeclaration.Body.RBraceToken)), FoldType.Member));
 				return base.VisitDestructorDeclaration (destructorDeclaration, data);
 			}
 			
 			public override object VisitOperatorDeclaration (OperatorDeclaration operatorDeclaration, object data)
 			{
-				if (!operatorDeclaration.Body.IsNull)
-					Foldings.Add (new FoldingRegion (new DomRegion (operatorDeclaration.Body.LBraceToken.GetPrevNode ().EndLocation, operatorDeclaration.Body.RBraceToken.StartLocation), FoldType.Member));
+				if (!operatorDeclaration.Body.IsNull && operatorDeclaration.Body.LBraceToken.StartLocation.Line != operatorDeclaration.Body.RBraceToken.StartLocation.Line)
+					Foldings.Add (new FoldingRegion (new DomRegion (operatorDeclaration.Body.LBraceToken.GetPrevNode (LastToken).EndLocation, CorrectEnd (operatorDeclaration.Body.RBraceToken)), FoldType.Member));
 				return base.VisitOperatorDeclaration (operatorDeclaration, data);
 			}
 			
 			public override object VisitPropertyDeclaration (PropertyDeclaration propertyDeclaration, object data)
 			{
-				if (!propertyDeclaration.LBraceToken.IsNull)
-					Foldings.Add (new FoldingRegion (new DomRegion (propertyDeclaration.LBraceToken.GetPrevNode ().EndLocation, propertyDeclaration.RBraceToken.StartLocation), FoldType.Member));
+				if (!propertyDeclaration.LBraceToken.IsNull && propertyDeclaration.LBraceToken.StartLocation.Line != propertyDeclaration.RBraceToken.StartLocation.Line)
+					Foldings.Add (new FoldingRegion (new DomRegion (propertyDeclaration.LBraceToken.GetPrevNode (LastToken).EndLocation, CorrectEnd (propertyDeclaration.RBraceToken)), FoldType.Member));
 				return base.VisitPropertyDeclaration (propertyDeclaration, data);
 			}
 			
 			public override object VisitIndexerDeclaration (IndexerDeclaration indexerDeclaration, object data)
 			{
-				if (!indexerDeclaration.LBraceToken.IsNull)
-					Foldings.Add (new FoldingRegion (new DomRegion (indexerDeclaration.LBraceToken.GetPrevNode ().EndLocation, indexerDeclaration.RBraceToken.StartLocation), FoldType.Member));
+				if (!indexerDeclaration.LBraceToken.IsNull && indexerDeclaration.LBraceToken.StartLocation.Line != indexerDeclaration.RBraceToken.StartLocation.Line)
+					Foldings.Add (new FoldingRegion (new DomRegion (indexerDeclaration.LBraceToken.GetPrevNode (LastToken).EndLocation, CorrectEnd (indexerDeclaration.RBraceToken)), FoldType.Member));
 				return base.VisitIndexerDeclaration (indexerDeclaration, data);
 			}
 			
 			public override object VisitCustomEventDeclaration (CustomEventDeclaration eventDeclaration, object data)
 			{
-				if (!eventDeclaration.LBraceToken.IsNull)
-					Foldings.Add (new FoldingRegion (new DomRegion (eventDeclaration.LBraceToken.GetPrevNode ().EndLocation, eventDeclaration.RBraceToken.StartLocation), FoldType.Member));
+				if (!eventDeclaration.LBraceToken.IsNull && eventDeclaration.LBraceToken.StartLocation.Line != eventDeclaration.RBraceToken.StartLocation.Line)
+					Foldings.Add (new FoldingRegion (new DomRegion (eventDeclaration.LBraceToken.GetPrevNode (LastToken).EndLocation, CorrectEnd (eventDeclaration.RBraceToken)), FoldType.Member));
 				return base.VisitCustomEventDeclaration (eventDeclaration, data);
 			}
 			
 			public override object VisitSwitchStatement (SwitchStatement switchStatement, object data)
 			{
-				if (!switchStatement.RBraceToken.IsNull)
-					Foldings.Add (new FoldingRegion (new DomRegion (switchStatement.LBraceToken.GetPrevNode ().EndLocation, switchStatement.RBraceToken.StartLocation), FoldType.Member));
+				if (!switchStatement.RBraceToken.IsNull && switchStatement.LBraceToken.StartLocation.Line != switchStatement.RBraceToken.StartLocation.Line)
+					Foldings.Add (new FoldingRegion (new DomRegion (switchStatement.LBraceToken.GetPrevNode (LastToken).EndLocation, CorrectEnd (switchStatement.RBraceToken)), FoldType.Member));
 				return base.VisitSwitchStatement (switchStatement, data);
 			}
 			
 			public override object VisitBlockStatement (BlockStatement blockStatement, object data)
 			{
 				if (!(blockStatement.Parent is EntityDeclaration) && blockStatement.EndLocation.Line - blockStatement.StartLocation.Line > 2) {
-					Foldings.Add (new FoldingRegion (new DomRegion (blockStatement.GetPrevNode ().EndLocation, blockStatement.RBraceToken.StartLocation), FoldType.Undefined));
+					Foldings.Add (new FoldingRegion (new DomRegion (blockStatement.GetPrevNode (LastToken).EndLocation, CorrectEnd (blockStatement.RBraceToken)), FoldType.Undefined));
 				}
 				
 				return base.VisitBlockStatement (blockStatement, data);
 			}
 		}
-		
+
+		static IEnumerable<Tag> GetSemanticTags (SyntaxTree unit)
+		{
+			var visitor = new SemanticTagVisitor ();
+			unit.AcceptVisitor (visitor);
+			foreach (var fold in visitor.Tags)
+				yield return fold;
+		}
+
+		public class SemanticTagVisitor : DepthFirstAstVisitor
+		{
+			public List<Tag> Tags =  new List<Tag> ();
+
+			public override void VisitThrowStatement (ThrowStatement throwStatement)
+			{
+				var createExpression = throwStatement.Expression as ObjectCreateExpression;
+				if (createExpression == null)
+					return;
+				var st = createExpression.Type as SimpleType;
+				var mt = createExpression.Type as MemberType;
+				if (st != null && st.Identifier == "NotImplementedException" ||
+				    mt != null && mt.MemberName == "NotImplementedException" && mt.Target.ToString () == "System") {
+
+					if (createExpression.Arguments.Any ()) {
+						Tags.Add (new Tag ("High", GettextCatalog.GetString ("NotImplementedException({0}) thrown.", createExpression.Arguments.First ().ToString ()), new DomRegion (throwStatement.StartLocation, throwStatement.EndLocation)));
+					} else {
+						Tags.Add (new Tag ("High", GettextCatalog.GetString ("NotImplementedException thrown."), new DomRegion (throwStatement.StartLocation, throwStatement.EndLocation)));
+					}
+				}
+			}
+		}
+
 		void VisitMcsUnit ()
 		{
 		}
@@ -201,11 +282,11 @@ namespace MonoDevelop.CSharp.Parser
 			}
 			cmt.Region = new DomRegion (comment.Line, comment.Col, comment.EndLine, comment.EndCol);
 			result.Comments.Add (cmt);
+			var trimmedContent = comment.Content.TrimStart ();
 			foreach (string tag in tagComments) {
-				int idx = comment.Content.IndexOf (tag);
-				if (idx < 0)
+				if (!trimmedContent.StartsWith (tag))
 					continue;
-				result.Add (new MonoDevelop.Ide.TypeSystem.Tag (tag, comment.Content, cmt.Region));
+				result.Add (new Tag (tag, comment.Content, cmt.Region));
 			}
 		}
 		
@@ -292,52 +373,65 @@ namespace MonoDevelop.CSharp.Parser
 				break;
 			}
 		}
-		
-		public static CompilerSettings GetCompilerArguments (MonoDevelop.Projects.Project project)
+
+		public static ICSharpCode.NRefactory.CSharp.CompilerSettings GetCompilerArguments (MonoDevelop.Projects.Project project)
 		{
-			var compilerArguments = new CompilerSettings ();
-			if (project == null || MonoDevelop.Ide.IdeApp.Workspace == null)
+			var compilerArguments = new ICSharpCode.NRefactory.CSharp.CompilerSettings ();
+	//		compilerArguments.TabSize = 1;
+
+			if (project == null || MonoDevelop.Ide.IdeApp.Workspace == null) {
+				compilerArguments.AllowUnsafeBlocks = true;
 				return compilerArguments;
+			}
 
 			var configuration = project.GetConfiguration (MonoDevelop.Ide.IdeApp.Workspace.ActiveConfiguration) as DotNetProjectConfiguration;
-			var par = configuration != null ? configuration.CompilationParameters as CSharpCompilerParameters : null;
-			
+			if (configuration == null)
+				return compilerArguments;
+
+			foreach (var sym in configuration.GetDefineSymbols ())
+				compilerArguments.ConditionalSymbols.Add (sym);
+
+			var par = configuration.CompilationParameters as CSharpCompilerParameters;
 			if (par == null)
 				return compilerArguments;
-				
-			if (!string.IsNullOrEmpty (par.DefineSymbols)) {
-				foreach (var sym in par.DefineSymbols.Split (';', ',', ' ', '\t').Where (s => !string.IsNullOrWhiteSpace (s)))
-					compilerArguments.AddConditionalSymbol (sym);
+			
+			compilerArguments.AllowUnsafeBlocks = par.UnsafeCode;
+			compilerArguments.LanguageVersion = ConvertLanguageVersion (par.LangVersion);
+			compilerArguments.CheckForOverflow = par.GenerateOverflowChecks;
+			compilerArguments.WarningLevel = par.WarningLevel;
+			compilerArguments.TreatWarningsAsErrors = par.TreatWarningsAsErrors;
+			if (!string.IsNullOrEmpty (par.NoWarnings)) {
+				foreach (var warning in par.NoWarnings.Split (';', ',', ' ', '\t')) {
+					int w;
+					try {
+						w = int.Parse (warning);
+					} catch (Exception) {
+						continue;
+					}
+					compilerArguments.DisabledWarnings.Add (w);
+				}
 			}
 			
-			compilerArguments.Unsafe = par.UnsafeCode;
-			compilerArguments.Version = ConvertLanguageVersion (par.LangVersion);
-			compilerArguments.Checked = par.GenerateOverflowChecks;
-			compilerArguments.TabSize = 1;
-			
-			// TODO: Warning options need to be set in the report object.
-			
-/*			compilerArguments.EnhancedWarnings = par.TreatWarningsAsErrors;
-			if (!string.IsNullOrEmpty (par.NoWarnings))
-				compilerArguments.Add ("-nowarn:" + string.Join (",", par.NoWarnings.Split (';', ',', ' ', '\t')));
-			
-			compilerArguments.Add ("-warn:" + par.WarningLevel);
-*/
-			
 			return compilerArguments;
 		}
 		
-		static Mono.CSharp.LanguageVersion ConvertLanguageVersion (LangVersion ver)
+		internal static Version ConvertLanguageVersion (LangVersion ver)
 		{
 			switch (ver) {
 			case LangVersion.Default:
-				return Mono.CSharp.LanguageVersion.Default;
+				return new Version (5, 0, 0, 0);
 			case LangVersion.ISO_1:
-				return Mono.CSharp.LanguageVersion.ISO_1;
+				return new Version (1, 0, 0, 0);
 			case LangVersion.ISO_2:
-				return Mono.CSharp.LanguageVersion.ISO_2;
+				return new Version (2, 0, 0, 0);
+			case LangVersion.Version3:
+				return new Version (3, 0, 0, 0);
+			case LangVersion.Version4:
+				return new Version (4, 0, 0, 0);
+			case LangVersion.Version5:
+				return new Version (5, 0, 0, 0);
 			}
-			return Mono.CSharp.LanguageVersion.Default;
+			return new Version (5, 0, 0, 0);;
 		}
 	}
 	
