@@ -42,11 +42,12 @@ using Mono.Cecil.Cil;
 using ICSharpCode.Decompiler;
 using ICSharpCode.Decompiler.Ast;
 using ICSharpCode.NRefactory.PatternMatching;
+using ICSharpCode.NRefactory.TypeSystem.Implementation;
 
 
 namespace MonoDevelop.CSharp.Refactoring
 {
-	public class CSharpCodeGenerator : CodeGenerator
+	class CSharpCodeGenerator : CodeGenerator
 	{
 		static CSharpAmbience ambience = new CSharpAmbience ();
 		
@@ -55,7 +56,7 @@ namespace MonoDevelop.CSharp.Refactoring
 		public MonoDevelop.CSharp.Formatting.CSharpFormattingPolicy Policy {
 			get {
 				if (policy == null) {
-					var types = MonoDevelop.Ide.DesktopService.GetMimeTypeInheritanceChain (CSharpFormatter.MimeType);
+					var types = MonoDevelop.Ide.DesktopService.GetMimeTypeInheritanceChain (MonoDevelop.CSharp.Formatting.CSharpFormatter.MimeType);
 					if (PolicyParent != null)
 						policy = PolicyParent.Get<CSharpFormattingPolicy> (types);
 					if (policy == null) {
@@ -73,7 +74,7 @@ namespace MonoDevelop.CSharp.Refactoring
 			}
 			set {
 				base.PolicyParent = value;
-				var types = MonoDevelop.Ide.DesktopService.GetMimeTypeInheritanceChain (CSharpFormatter.MimeType);
+				var types = MonoDevelop.Ide.DesktopService.GetMimeTypeInheritanceChain (MonoDevelop.CSharp.Formatting.CSharpFormatter.MimeType);
 				policy = value.Get<CSharpFormattingPolicy> (types);
 			}
 		}
@@ -84,6 +85,21 @@ namespace MonoDevelop.CSharp.Refactoring
 			public bool ExplicitDeclaration { get; set; }
 			public ITypeDefinition ImplementingType { get; set; }
 			public IUnresolvedTypeDefinition Part { get; set; }
+
+			public MonoDevelop.Ide.Gui.Document Document { get; set; }
+
+			public string GetShortType (string ns, string name, int typeArguments = 0)
+			{
+				if (Document == null || Document.ParsedDocument == null)
+					return ns + "." + name;
+				var typeDef = new GetClassTypeReference (ns, name, typeArguments).Resolve (Document.Compilation.TypeResolveContext);
+				if (typeDef == null)
+					return ns + "." + name;
+				var file = Document.ParsedDocument.ParsedFile as CSharpUnresolvedFile;
+				var csResolver = file.GetResolver (Document.Compilation, Document.Editor.Caret.Location);
+				var builder = new ICSharpCode.NRefactory.CSharp.Refactoring.TypeSystemAstBuilder (csResolver);
+				return OutputNode (Document, builder.ConvertType (typeDef));
+			}
 		}
 		
 		public override string WrapInRegions (string regionName, string text)
@@ -99,6 +115,35 @@ namespace MonoDevelop.CSharp.Refactoring
 			result.Append ("#endregion");
 			return result.ToString ();
 		}
+
+		void AppendObsoleteAttribute (StringBuilder result, CodeGenerationOptions options, IEntity entity)
+		{
+			string reason;
+			if (!entity.IsObsolete (out reason))
+				return;
+
+			var implementingType = options.Part;
+			var loc = implementingType.Region.End;
+			
+			var pf = implementingType.UnresolvedFile;
+			var file = pf as CSharpUnresolvedFile;
+
+			result.Append ("[");
+			var obsoleteRef = ReflectionHelper.ParseReflectionName ("System.ObsoleteAttribute");
+			var resolvedType = obsoleteRef.Resolve (options.ImplementingType.Compilation);
+			var shortType = resolvedType.Kind != TypeKind.Unknown ? CreateShortType (options.ImplementingType.Compilation, file, loc, resolvedType) : null;
+			var text = shortType != null ? shortType.ToString () : "System.Obsolete";
+			if (text.EndsWith ("Attribute", StringComparison.Ordinal))
+				text = text.Substring (0, text.Length - "Attribute".Length);
+			result.Append (text);
+			if (!string.IsNullOrEmpty (reason)) {
+				result.Append (" (\"");
+				result.Append (reason);
+				result.Append ("\")");
+			}
+			result.Append ("]");
+			result.AppendLine ();
+		}
 		
 		public override CodeGeneratorMemberResult CreateMemberImplementation (ITypeDefinition implementingType,
 		                                                                      IUnresolvedTypeDefinition part,
@@ -111,14 +156,20 @@ namespace MonoDevelop.CSharp.Refactoring
 				ImplementingType = implementingType,
 				Part = part
 			};
-			if (member is IMethod)
-				return GenerateCode ((IMethod)member, options);
-			if (member is IProperty)
-				return GenerateCode ((IProperty)member, options);
-			if (member is IField)
-				return GenerateCode ((IField)member, options);
-			if (member is IEvent)
-				return GenerateCode ((IEvent)member, options);
+			ITypeResolveContext ctx;
+
+			var doc = IdeApp.Workbench.GetDocument (part.Region.FileName);
+			ctx = new CSharpTypeResolveContext (implementingType.Compilation.MainAssembly, null, implementingType, null);
+			options.Document = doc;
+
+			if (member is IUnresolvedMethod)
+				return GenerateCode ((IMethod) ((IUnresolvedMethod)member).CreateResolved (ctx), options);
+			if (member is IUnresolvedProperty)
+				return GenerateCode ((IProperty) ((IUnresolvedProperty)member).CreateResolved (ctx), options);
+			if (member is IUnresolvedField)
+				return GenerateCode ((IField) ((IUnresolvedField)member).CreateResolved (ctx), options);
+			if (member is IUnresolvedEvent)
+				return GenerateCode ((IEvent) ((IUnresolvedEvent)member).CreateResolved (ctx), options);
 			throw new NotSupportedException ("member " +  member + " is not supported.");
 		}
 		
@@ -131,7 +182,8 @@ namespace MonoDevelop.CSharp.Refactoring
 			var options = new CodeGenerationOptions () {
 				ExplicitDeclaration = explicitDeclaration,
 				ImplementingType = implementingType,
-				Part = part
+				Part = part,
+				Document = IdeApp.Workbench.GetDocument (part.Region.FileName)
 			};
 			if (member is IMethod)
 				return GenerateCode ((IMethod)member, options);
@@ -220,17 +272,17 @@ namespace MonoDevelop.CSharp.Refactoring
 			var implementingType = options.Part;
 			var loc = implementingType.Region.End;
 			
-			var pf = implementingType.ParsedFile;
-			var file = pf as CSharpParsedFile;
+			var pf = implementingType.UnresolvedFile;
+			var file = pf as CSharpUnresolvedFile;
 			var resolved = type;
 			if (resolved.Kind == TypeKind.Unknown) {
-				result.Append (type.ToString ());
+				result.Append (type.FullName);
 				return;
 			}
 			var def = type.GetDefinition ();
 			if (def != null) {
 				using (var stringWriter = new System.IO.StringWriter ()) {
-					var formatter = new TextWriterOutputFormatter (stringWriter);
+					var formatter = new TextWriterTokenWriter (stringWriter);
 					stringWriter.NewLine = EolMarker; 
 					var visitor = new CSharpOutputVisitor (formatter, FormattingOptionsFactory.CreateMono ());
 					var shortType = CreateShortType (def.Compilation, file, loc, resolved);
@@ -269,7 +321,7 @@ namespace MonoDevelop.CSharp.Refactoring
 			result.Append (" ");
 			AppendReturnType (result, options, field.ReturnType);
 			result.Append (" ");
-			result.Append (field.Name);
+			result.Append (CSharpAmbience.FilterName (field.Name));
 			result.Append (";");
 			return new CodeGeneratorMemberResult (result.ToString (), -1, -1);
 		}
@@ -277,7 +329,7 @@ namespace MonoDevelop.CSharp.Refactoring
 		CodeGeneratorMemberResult GenerateCode (IEvent evt, CodeGenerationOptions options)
 		{
 			StringBuilder result = new StringBuilder ();
-			
+			AppendObsoleteAttribute (result, options, evt);
 			AppendModifiers (result, options, evt);
 			
 			result.Append ("event ");
@@ -287,7 +339,7 @@ namespace MonoDevelop.CSharp.Refactoring
 				result.Append (ambience.GetString (evt.DeclaringTypeDefinition, OutputFlags.IncludeGenerics));
 				result.Append (".");
 			}
-			result.Append (evt.Name);
+			result.Append (CSharpAmbience.FilterName (evt.Name));
 			if (options.ExplicitDeclaration) {
 				AppendBraceStart (result, Policy.EventBraceStyle);
 				AppendIndent (result);
@@ -319,21 +371,35 @@ namespace MonoDevelop.CSharp.Refactoring
 			AppendIndent (result);
 			bodyStartOffset = result.Length;
 			result.Append ("throw new ");
-			// TODO: Type system conversion.
-			result.Append ("System.NotImplementedException");
-//			AppendReturnType (result, options.ImplementingType, options.Ctx.GetTypeDefinition (typeof (System.NotImplementedException)));
+			result.Append (options.GetShortType ("System", "NotImplementedException"));
+			//			AppendReturnType (result, options.ImplementingType, options.Ctx.GetTypeDefinition (typeof (System.NotImplementedException)));
 			if (Policy.BeforeMethodCallParentheses)
 				result.Append (" ");
 			result.Append ("();");
 			bodyEndOffset = result.Length;
 			AppendLine (result);
 		}
-		
-		void AppendMonoTouchTodo (StringBuilder result, out int bodyStartOffset, out int bodyEndOffset)
+
+		internal static string[] MonoTouchComments = {
+			" NOTE: Don't call the base implementation on a Model class",
+			" see http://docs.xamarin.com/guides/ios/application_fundamentals/delegates,_protocols,_and_events"
+		};
+
+		void AppendMonoTouchTodo (StringBuilder result, CodeGenerationOptions options, out int bodyStartOffset, out int bodyEndOffset)
 		{
 			AppendIndent (result);
 			bodyStartOffset = result.Length;
-			result.Append ("// TODO: Implement - see: http://go-mono.com/docs/index.aspx?link=T%3aMonoTouch.Foundation.ModelAttribute");
+			foreach (var cmt in MonoTouchComments) {
+				result.AppendLine ("//" + cmt);
+				AppendIndent (result);
+			}
+			result.Append ("throw new ");
+			result.Append (options.GetShortType ("System", "NotImplementedException"));
+
+			if (Policy.BeforeMethodCallParentheses)
+				result.Append (" ");
+			result.Append ("();");
+
 			bodyEndOffset = result.Length;
 			AppendLine (result);
 		}
@@ -342,21 +408,25 @@ namespace MonoDevelop.CSharp.Refactoring
 		{
 			int bodyStartOffset = -1, bodyEndOffset = -1;
 			StringBuilder result = new StringBuilder ();
+			AppendObsoleteAttribute (result, options, method);
 			AppendModifiers (result, options, method);
+			if (method.IsPartial)
+				result.Append ("partial ");
 			AppendReturnType (result, options, method.ReturnType);
 			result.Append (" ");
 			if (options.ExplicitDeclaration) {
 				AppendReturnType (result, options, method.DeclaringType);
 				result.Append (".");
 			}
-			result.Append (method.Name);
+
+			result.Append (CSharpAmbience.FilterName (method.Name));
 			if (method.TypeParameters.Count > 0) {
 				result.Append ("<");
 				for (int i = 0; i < method.TypeParameters.Count; i++) {
 					if (i > 0)
 						result.Append (", ");
 					var p = method.TypeParameters [i];
-					result.Append (p.Name);
+					result.Append (CSharpAmbience.FilterName (p.Name));
 				}
 				result.Append (">");
 			}
@@ -381,7 +451,7 @@ namespace MonoDevelop.CSharp.Refactoring
 						result.Append (", ");
 					
 					typeParameterCount++;
-					result.Append (p.Name);
+					result.Append (CSharpAmbience.FilterName (p.Name));
 					result.Append (" : ");
 					int constraintCount = 0;
 					
@@ -430,7 +500,15 @@ namespace MonoDevelop.CSharp.Refactoring
 					if (options.ImplementingType.Properties.Any ()) 
 						result.Append (": ");
 					int i = 0;
+					var properties = new List<IProperty> ();
+
 					foreach (IProperty property in options.ImplementingType.Properties) {
+						if (properties.Any (p => p.Name == property.Name))
+							continue;
+						properties.Add (property); 
+					}
+
+					foreach (IProperty property in properties) {
 						if (property.IsStatic || !property.IsPublic)
 							continue;
 						if (i > 0)
@@ -441,7 +519,7 @@ namespace MonoDevelop.CSharp.Refactoring
 						result.Append ("}");
 					}
 					result.Append ("]\"");
-					foreach (IProperty property in options.ImplementingType.Properties) {
+					foreach (IProperty property in properties) {
 						if (property.IsStatic || !property.IsPublic)
 							continue;
 						result.Append (", ");
@@ -451,7 +529,7 @@ namespace MonoDevelop.CSharp.Refactoring
 					bodyEndOffset = result.Length;
 					AppendLine (result);
 				} else if (IsMonoTouchModelMember (method)) {
-					AppendMonoTouchTodo (result, out bodyStartOffset, out bodyEndOffset);
+					AppendMonoTouchTodo (result, options, out bodyStartOffset, out bodyEndOffset);
 				} else if (method.IsAbstract || !(method.IsVirtual || method.IsOverride) || method.DeclaringTypeDefinition.Kind == TypeKind.Interface) {
 					AppendNotImplementedException (result, options, out bodyStartOffset, out bodyEndOffset);
 				} else {
@@ -483,9 +561,9 @@ namespace MonoDevelop.CSharp.Refactoring
 										astBuilder.AddMethod (m);
 										
 										astBuilder.RunTransformations (o => false);
-										
+
 										var visitor = new ThrowsExceptionVisitor ();
-										astBuilder.CompilationUnit.AcceptVisitor (visitor);
+										astBuilder.SyntaxTree.AcceptVisitor (visitor);
 										skipBody = visitor.Throws;
 										if (skipBody)
 											break;
@@ -503,7 +581,7 @@ namespace MonoDevelop.CSharp.Refactoring
 						if (method.ReturnType.ReflectionName != typeof(void).FullName)
 							result.Append ("return ");
 						result.Append ("base.");
-						result.Append (method.Name);
+						result.Append (CSharpAmbience.FilterName (method.Name));
 						if (Policy.BeforeMethodCallParentheses)
 							result.Append (" ");
 						result.Append ("(");
@@ -516,7 +594,7 @@ namespace MonoDevelop.CSharp.Refactoring
 								result.Append ("out ");
 							if (p.IsRef)
 								result.Append ("ref ");
-							result.Append (p.Name);
+							result.Append (CSharpAmbience.FilterName (p.Name));
 						}
 						result.Append (");");
 					} else {
@@ -556,19 +634,63 @@ namespace MonoDevelop.CSharp.Refactoring
 					result.Append ("params ");
 				AppendReturnType (result, options, p.Type);
 				result.Append (" ");
-				result.Append (p.Name);
+				result.Append (CSharpAmbience.FilterName (p.Name));
+				if (p.ConstantValue != null) {
+					result.Append (" = ");
+					if (p.Type.Kind == TypeKind.Enum) {
+						bool found = false;
+						foreach (var literal in GetEnumLiterals(p.Type)) {
+							if (literal.ConstantValue.Equals (p.ConstantValue)) {
+								AppendReturnType (result, options, p.Type);
+								result.Append ("."+ literal.Name);
+								found = true;
+								break;
+							}
+						}
+						if (!found) {
+							result.Append ("(");
+							AppendReturnType (result, options, p.Type);
+							result.Append (")" + p.ConstantValue); 
+						}
+					} else if (p.ConstantValue is char) {
+						result.Append ("'" + p.ConstantValue + "'");
+					} else if (p.ConstantValue is string)  {
+						result.Append ("\"" + CSharpTextEditorIndentation.ConvertToStringLiteral ((string)p.ConstantValue) + "\"");
+					} else if (p.ConstantValue is bool)  {
+						result.Append ((bool)p.ConstantValue ? "true" : "false");
+					} else {
+						result.Append (p.ConstantValue);
+					}
+				} 
 			}
 		}
+
+		public IEnumerable<IField> GetEnumLiterals(IType type)
+		{
+			if (type.Kind != TypeKind.Enum)
+				throw new ArgumentException ("Type is no enum.");
+			foreach (var field in type.GetFields (f => f.IsConst && f.IsPublic))
+				yield return field;
+		}
 		
-		static string GetModifiers (IUnresolvedTypeDefinition implementingType, IMember member)
+		static string GetModifiers (ITypeDefinition implementingType, IUnresolvedTypeDefinition implementingPart, IMember member)
 		{
 			StringBuilder result = new StringBuilder ();
+
 			if (member.IsPublic || (member.DeclaringType != null && member.DeclaringTypeDefinition.Kind == TypeKind.Interface)) {
 				result.Append ("public ");
-			} else if (member.IsProtectedAndInternal) {
-				result.Append ("protected internal ");
 			} else if (member.IsProtectedOrInternal) {
-				result.Append ("internal protected ");
+				if (IdeApp.Workbench.ActiveDocument != null && member.DeclaringTypeDefinition.ParentAssembly != implementingType.ParentAssembly) {
+					result.Append ("protected ");
+				} else {
+					result.Append ("internal protected ");
+				}
+			} else if (member.IsProtectedAndInternal) {
+				if (IdeApp.Workbench.ActiveDocument != null && member.DeclaringTypeDefinition.ParentAssembly != implementingType.ParentAssembly) {
+					result.Append ("protected ");
+				} else {
+					result.Append ("protected internal ");
+				}
 			} else if (member.IsProtected) {
 				result.Append ("protected ");
 			} else if (member.IsInternal) {
@@ -586,8 +708,7 @@ namespace MonoDevelop.CSharp.Refactoring
 			AppendIndent (result);
 			if (options.ExplicitDeclaration || options.ImplementingType.Kind == TypeKind.Interface)
 				return;
-			result.Append (GetModifiers (options.Part, member));
-			
+			result.Append (GetModifiers (options.ImplementingType, options.Part, member));
 			bool isFromInterface = false;
 			if (member.DeclaringType != null && member.DeclaringTypeDefinition.Kind == TypeKind.Interface) {
 				isFromInterface = true;
@@ -596,22 +717,25 @@ namespace MonoDevelop.CSharp.Refactoring
 //					foreach (var type in options.ImplementingType.BaseTypes) {
 //						if (type.Kind == TypeKind.Interface)
 //							continue;
-//						if (type.Members.Any (m => m.Name == member.Name && member.EntityType == m.EntityType /* && DomMethod.ParameterListEquals (member.Parameters, m.Parameters)*/ )) {
+//						if (type.Members.Any (m => m.Name == member.Name && member.SymbolKind == m.SymbolKind /* && DomMethod.ParameterListEquals (member.Parameters, m.Parameters)*/ )) {
 //							isFromInterface = false;
 //							break;
 //						}
 //					}
 //				}
 			}
-			
+
 			if (!isFromInterface && member.IsOverridable)
 				result.Append ("override ");
+			if (member is IMethod && ((IMethod)member).IsAsync)
+				result.Append ("async ");
 		}
 		
 		CodeGeneratorMemberResult GenerateCode (IProperty property, CodeGenerationOptions options)
 		{
 			var regions = new List<CodeGeneratorBodyRegion> ();
 			var result = new StringBuilder ();
+			AppendObsoleteAttribute (result, options, property);
 			AppendModifiers (result, options, property);
 			AppendReturnType (result, options, property.ReturnType);
 			result.Append (" ");
@@ -624,7 +748,7 @@ namespace MonoDevelop.CSharp.Refactoring
 					result.Append (ambience.GetString (property.DeclaringType, OutputFlags.IncludeGenerics));
 					result.Append (".");
 				}
-				result.Append (property.Name);
+				result.Append (CSharpAmbience.FilterName (property.Name));
 			}
 			AppendBraceStart (result, Policy.PropertyBraceStyle);
 			if (property.CanGet) {
@@ -636,20 +760,20 @@ namespace MonoDevelop.CSharp.Refactoring
 				} else {
 					AppendBraceStart (result, Policy.PropertyGetBraceStyle);
 					if (IsMonoTouchModelMember (property)) {
-						AppendMonoTouchTodo (result, out bodyStartOffset, out bodyEndOffset);
+						AppendMonoTouchTodo (result, options, out bodyStartOffset, out bodyEndOffset);
 					} else if (property.IsAbstract || property.DeclaringTypeDefinition.Kind == TypeKind.Interface) {
 						AppendNotImplementedException (result, options, out bodyStartOffset, out bodyEndOffset);
 					} else {
 						AppendIndent (result);
 						bodyStartOffset = result.Length;
-						if (property.EntityType == EntityType.Indexer) {
+						if (property.SymbolKind == SymbolKind.Indexer) {
 							result.Append ("return base[");
 							if (property.Parameters.Count > 0)
-								result.Append (property.Parameters.First ().Name);
+								result.Append (CSharpAmbience.FilterName (property.Parameters.First ().Name));
 							result.Append ("];");
 						} else {
 							result.Append ("return base.");
-							result.Append (property.Name);
+							result.Append (CSharpAmbience.FilterName (property.Name));
 							result.Append (";");
 						}
 						bodyEndOffset = result.Length;
@@ -670,20 +794,20 @@ namespace MonoDevelop.CSharp.Refactoring
 				} else {
 					AppendBraceStart (result, Policy.PropertyGetBraceStyle);
 					if (IsMonoTouchModelMember (property)) {
-						AppendMonoTouchTodo (result, out bodyStartOffset, out bodyEndOffset);
+						AppendMonoTouchTodo (result, options, out bodyStartOffset, out bodyEndOffset);
 					} else if (property.IsAbstract || property.DeclaringTypeDefinition.Kind == TypeKind.Interface) {
 						AppendNotImplementedException (result, options, out bodyStartOffset, out bodyEndOffset);
 					} else {
 						AppendIndent (result);
 						bodyStartOffset = result.Length;
-						if (property.EntityType == EntityType.Indexer) {
+						if (property.SymbolKind == SymbolKind.Indexer) {
 							result.Append ("base[");
 							if (property.Parameters.Count > 0)
-								result.Append (property.Parameters.First ().Name);
+								result.Append (CSharpAmbience.FilterName (property.Parameters.First ().Name));
 							result.Append ("] = value;");
 						} else { 
 							result.Append ("base.");
-							result.Append (property.Name);
+							result.Append (CSharpAmbience.FilterName (property.Name));
 							result.Append (" = value;");
 						}
 						bodyEndOffset = result.Length;
@@ -698,7 +822,7 @@ namespace MonoDevelop.CSharp.Refactoring
 			return new CodeGeneratorMemberResult (result.ToString (), regions);
 		}
 		
-		static bool IsMonoTouchModelMember (IMember member)
+		internal static bool IsMonoTouchModelMember (IMember member)
 		{
 			if (member == null || member.DeclaringType == null)
 				return false;
@@ -721,7 +845,7 @@ namespace MonoDevelop.CSharp.Refactoring
 				ImplementingType = field.DeclaringTypeDefinition,
 				Part = implementingType
 			};
-			
+			result.Append ("public ");
 			AppendReturnType (result, options, field.ReturnType);
 			result.Append (" ");
 			result.Append (propertyName);
@@ -732,7 +856,7 @@ namespace MonoDevelop.CSharp.Refactoring
 			AppendBraceStart (result, Policy.PropertyGetBraceStyle);
 			AppendIndent (result);
 			result.Append ("return this.");
-			result.Append (field.Name);
+			result.Append (CSharpAmbience.FilterName (field.Name));
 			result.Append (";");
 			AppendLine (result);
 			AppendBraceEnd (result, Policy.PropertyGetBraceStyle);
@@ -743,7 +867,7 @@ namespace MonoDevelop.CSharp.Refactoring
 				result.Append ("set");
 				AppendBraceStart (result, Policy.PropertyGetBraceStyle);
 				AppendIndent (result);
-				result.Append (field.Name);
+				result.Append (CSharpAmbience.FilterName (field.Name));
 				result.Append (" = value;");
 				AppendLine (result);
 				AppendBraceEnd (result, Policy.PropertyGetBraceStyle);
@@ -766,6 +890,12 @@ namespace MonoDevelop.CSharp.Refactoring
 		}
 		
 		static bool InsertUsingAfter (AstNode node)
+		{
+			return node is NewLineNode && IsCommentOrUsing (node.GetNextSibling (s => !(s is NewLineNode))) ||
+				IsCommentOrUsing (node) || (node is PreProcessorDirective);
+		}
+
+		static bool IsCommentOrUsing (AstNode node)
 		{
 			return node is ICSharpCode.NRefactory.CSharp.Comment ||
 				node is UsingDeclaration ||
@@ -787,7 +917,7 @@ namespace MonoDevelop.CSharp.Refactoring
 		public override void AddGlobalNamespaceImport (MonoDevelop.Ide.Gui.Document doc, string nsName)
 		{
 			var parsedDocument = doc.ParsedDocument;
-			var unit = parsedDocument.GetAst<CompilationUnit> ();
+			var unit = parsedDocument.GetAst<SyntaxTree> ();
 			if (unit == null)
 				return;
 			
@@ -814,7 +944,7 @@ namespace MonoDevelop.CSharp.Refactoring
 			int offset = 0;
 			if (node != null) {
 				var loc = InsertUsingAfter (node) ? node.EndLocation : node.StartLocation;
-				offset = doc.Editor.LocationToOffset (loc);
+				offset = Math.Max (0, doc.Editor.LocationToOffset (loc));
 			}
 			
 			lines = policy.BlankLinesAfterUsings;
@@ -831,7 +961,7 @@ namespace MonoDevelop.CSharp.Refactoring
 		public override void AddLocalNamespaceImport (MonoDevelop.Ide.Gui.Document doc, string nsName, TextLocation caretLocation)
 		{
 			var parsedDocument = doc.ParsedDocument;
-			var unit = parsedDocument.GetAst<CompilationUnit> ();
+			var unit = parsedDocument.GetAst<SyntaxTree> ();
 			if (unit == null)
 				return;
 			
@@ -886,7 +1016,7 @@ namespace MonoDevelop.CSharp.Refactoring
 		
 		public override string GetShortTypeString (MonoDevelop.Ide.Gui.Document doc, IType type)
 		{
-			var shortType = CreateShortType (doc.Compilation, doc.ParsedDocument.ParsedFile as CSharpParsedFile, doc.Editor.Caret.Location, type);
+			var shortType = CreateShortType (doc.Compilation, doc.ParsedDocument.ParsedFile as CSharpUnresolvedFile, doc.Editor.Caret.Location, type);
 			return OutputNode (doc, shortType);
 		}
 		
@@ -894,7 +1024,7 @@ namespace MonoDevelop.CSharp.Refactoring
 		{
 			using (var stringWriter = new System.IO.StringWriter ()) {
 //				formatter.Indentation = indentLevel;
-				var formatter = new TextWriterOutputFormatter (stringWriter);
+				var formatter = new TextWriterTokenWriter (stringWriter);
 				stringWriter.NewLine = doc.Editor.EolMarker;
 				
 				var visitor = new CSharpOutputVisitor (formatter, doc.GetFormattingOptions ());
@@ -904,7 +1034,7 @@ namespace MonoDevelop.CSharp.Refactoring
 		}
 		
 		
-		public AstType CreateShortType (ICompilation compilation, CSharpParsedFile parsedFile, TextLocation loc, IType fullType)
+		public AstType CreateShortType (ICompilation compilation, CSharpUnresolvedFile parsedFile, TextLocation loc, IType fullType)
 		{
 			var csResolver = parsedFile.GetResolver (compilation, loc);
 			var builder = new ICSharpCode.NRefactory.CSharp.Refactoring.TypeSystemAstBuilder (csResolver);
@@ -913,55 +1043,11 @@ namespace MonoDevelop.CSharp.Refactoring
 		
 		public override void CompleteStatement (MonoDevelop.Ide.Gui.Document doc)
 		{
-			var file = doc.Editor;
-			var caretLocation = file.Caret.Location;
-			
-			int pos = file.LocationToOffset (caretLocation.Line + 1, 1);
-			var line = new StringBuilder ();
-			int lineNr = caretLocation.Line + 1, column = 1, maxColumn = 1, lastPos = pos;
-			if (true) 
-				while (lineNr == caretLocation.Line + 1) {
-					maxColumn = column;
-					lastPos = pos;
-					line.Append (file.GetCharAt (pos));
-					pos++;
-					var loc = file.OffsetToLocation (pos);
-					lineNr = loc.Line;
-					column = loc.Column;
-				}
-			string trimmedline = line.ToString ().Trim ();
-			string indent = line.ToString ().Substring (0, line.Length - line.ToString ().TrimStart (' ', '\t').Length);
-			if (trimmedline.EndsWith (";") || trimmedline.EndsWith ("{")) {
-				file.Caret.Location = caretLocation;
-				return;
+			var fixer = new ConstructFixer (doc.GetFormattingOptions (), doc.Editor.CreateNRefactoryTextEditorOptions ());
+			int newOffset;
+			if (fixer.TryFix (doc.Editor.Document, doc.Editor.Caret.Offset, out newOffset)) {
+				doc.Editor.Caret.Offset = newOffset;
 			}
-			int caretLine = caretLocation.Line;
-			int caretColumn = caretLocation.Column;
-			if (trimmedline.StartsWith ("if") || 
-				trimmedline.StartsWith ("while") ||
-				trimmedline.StartsWith ("switch") ||
-				trimmedline.StartsWith ("for") ||
-				trimmedline.StartsWith ("foreach")) {
-				if (!trimmedline.EndsWith (")")) {
-					file.Insert (lastPos, " () {" + file.EolMarker + indent + file.Options.IndentationString + file.EolMarker + indent + "}");
-					caretColumn = maxColumn + 1;
-				} else {
-					file.Insert (lastPos, " {" + file.EolMarker + indent + file.Options.IndentationString + file.EolMarker + indent + "}");
-					caretColumn = indent.Length + 1;
-					caretLine++;
-				}
-			} else if (trimmedline.StartsWith ("do")) {
-				file.Insert (lastPos, " {" + file.EolMarker + indent + file.Options.IndentationString + file.EolMarker + indent + "} while ();");
-				caretColumn = indent.Length + 1;
-				caretLine++;
-			} else {
-				file.Insert (lastPos, ";" + file.EolMarker + indent);
-				caretColumn = indent.Length;
-				caretLine++;
-			}
-			file.Caret.Location = new DocumentLocation (caretLine, caretColumn);
 		}
-		
-		
 	}
 }
\ No newline at end of file
