@@ -27,6 +27,7 @@
 using System;
 using System.Collections.Generic;
 using System.IO;
+using System.Linq;
 using System.Diagnostics;
 using System.Text.RegularExpressions;
 using System.Text;
@@ -38,6 +39,7 @@ using MonoDevelop.Core.Assemblies;
 using MonoDevelop.CSharp.Project;
 using System.Threading;
 using MonoDevelop.Ide;
+using MonoDevelop.Core.ProgressMonitoring;
 
 
 namespace MonoDevelop.CSharp
@@ -55,12 +57,13 @@ namespace MonoDevelop.CSharp
 
 		public static BuildResult Compile (ProjectItemCollection projectItems, DotNetProjectConfiguration configuration, ConfigurationSelector configSelector, IProgressMonitor monitor)
 		{
-			CSharpCompilerParameters compilerParameters = (CSharpCompilerParameters)configuration.CompilationParameters ?? new CSharpCompilerParameters ();
-			CSharpProjectParameters projectParameters = (CSharpProjectParameters)configuration.ProjectParameters ?? new CSharpProjectParameters ();
-			
-			FilePath outputName       = configuration.CompiledOutputName;
-			string responseFileName = Path.GetTempFileName();
+			var compilerParameters = (CSharpCompilerParameters)configuration.CompilationParameters ?? new CSharpCompilerParameters ();
+			var projectParameters = (CSharpProjectParameters)configuration.ProjectParameters ?? new CSharpProjectParameters ();
 			
+			FilePath outputName = configuration.CompiledOutputName;
+			string responseFileName = Path.GetTempFileName ();
+
+			//make sure that the output file is writable
 			if (File.Exists (outputName)) {
 				bool isWriteable = false;
 				int count = 0;
@@ -79,21 +82,56 @@ namespace MonoDevelop.CSharp
 					return null;
 				}
 			}
-			
+
+			//get the runtime
 			TargetRuntime runtime = MonoDevelop.Core.Runtime.SystemAssemblyService.DefaultRuntime;
 			DotNetProject project = configuration.ParentItem as DotNetProject;
 			if (project != null)
 				runtime = project.TargetRuntime;
 
-			StringBuilder sb = new StringBuilder ();
+			//get the compiler name
+			string compilerName;
+			try {
+				compilerName = GetCompilerName (runtime, configuration.TargetFramework);
+			} catch (Exception e) {
+				string message = "Could not obtain a C# compiler";
+				monitor.ReportError (message, e);
+				return null;
+			}
+
+			var sb = new StringBuilder ();
+
+			HashSet<string> alreadyAddedReference = new HashSet<string> ();
+
+			var monoRuntime = runtime as MonoTargetRuntime;
+			if (!compilerParameters.NoStdLib && (monoRuntime == null || monoRuntime.HasMultitargetingMcs)) {
+				string corlib = project.AssemblyContext.GetAssemblyFullName ("mscorlib", project.TargetFramework);
+				if (corlib != null) {
+					corlib = project.AssemblyContext.GetAssemblyLocation (corlib, project.TargetFramework);
+				}
+				if (corlib == null) {
+					var br = new BuildResult ();
+					br.AddError (string.Format ("Could not find mscorlib for framework {0}", project.TargetFramework.Id));
+					return br;
+				}
+				AppendQuoted (sb, "/r:", corlib);
+				sb.AppendLine ("-nostdlib");
+			}
+
 			List<string> gacRoots = new List<string> ();
 			sb.AppendFormat ("\"/out:{0}\"", outputName);
 			sb.AppendLine ();
 			
-			HashSet<string> alreadyAddedReference = new HashSet<string> ();
 			foreach (ProjectReference lib in projectItems.GetAll <ProjectReference> ()) {
-				if (lib.ReferenceType == ReferenceType.Project && !(lib.OwnerProject.ParentSolution.FindProjectByName (lib.Reference) is DotNetProject))
-					continue;
+				if (lib.ReferenceType == ReferenceType.Project) {
+					var ownerProject = lib.OwnerProject;
+					if (ownerProject != null) {
+						var parentSolution = ownerProject.ParentSolution;
+						if (parentSolution != null && !(parentSolution.FindProjectByName (lib.Reference) is DotNetProject))
+							continue;
+					}
+				} 
+				string refPrefix = string.IsNullOrEmpty (lib.Aliases) ? "" : lib.Aliases + "=";
 				foreach (string fileName in lib.GetReferencedFileNames (configSelector)) {
 					switch (lib.ReferenceType) {
 					case ReferenceType.Package:
@@ -105,7 +143,7 @@ namespace MonoDevelop.CSharp
 						}
 
 						if (alreadyAddedReference.Add (fileName))
-							AppendQuoted (sb, "/r:", fileName);
+							AppendQuoted (sb, "/r:", refPrefix + fileName);
 						
 						if (pkg.GacRoot != null && !gacRoots.Contains (pkg.GacRoot))
 							gacRoots.Add (pkg.GacRoot);
@@ -123,17 +161,24 @@ namespace MonoDevelop.CSharp
 						break;
 					default:
 						if (alreadyAddedReference.Add (fileName))
-							AppendQuoted (sb, "/r:", fileName);
+							AppendQuoted (sb, "/r:", refPrefix + fileName);
 						break;
 					}
 				}
 			}
-			
+
+			if (alreadyAddedReference.Any (reference => SystemAssemblyService.ContainsReferenceToSystemRuntime (reference))) {
+				LoggingService.LogInfo ("Found PCLv2 assembly.");
+				var facades = runtime.FindFacadeAssembliesForPCL (project.TargetFramework);
+				foreach (var facade in facades)
+					AppendQuoted (sb, "/r:", facade);
+			}
+
 			string sysCore = project.AssemblyContext.GetAssemblyFullName ("System.Core", project.TargetFramework);
-			if (sysCore != null) {
-				sysCore = project.AssemblyContext.GetAssemblyLocation (sysCore, project.TargetFramework);
-				if (sysCore != null && !alreadyAddedReference.Contains (sysCore))
-					AppendQuoted (sb, "/r:", sysCore);
+			if (sysCore != null && !alreadyAddedReference.Contains (sysCore)) {
+				var asm = project.AssemblyContext.GetAssemblyFromFullName (sysCore, null, project.TargetFramework);
+				if (asm != null)
+					AppendQuoted (sb, "/r:", asm.Location);
 			}
 			
 			sb.AppendLine ("/nologo");
@@ -143,27 +188,31 @@ namespace MonoDevelop.CSharp
 			if (configuration.SignAssembly) {
 				if (File.Exists (configuration.AssemblyKeyFile))
 					AppendQuoted (sb, "/keyfile:", configuration.AssemblyKeyFile);
+				if (configuration.DelaySign)
+					sb.AppendLine ("/delaySign");
 			}
-			
-			if (configuration.DebugMode) {
-//				sb.AppendLine ("/debug:+");
-				sb.AppendLine ("/debug:full");
+
+			var debugType = compilerParameters.DebugType;
+			if (string.IsNullOrEmpty (debugType)) {
+				debugType = configuration.DebugMode ? "full" : "none";
+			} else if (string.Equals (debugType, "pdbonly", StringComparison.OrdinalIgnoreCase)) {
+				//old Mono compilers don't support pdbonly
+				if (monoRuntime != null && !monoRuntime.HasMultitargetingMcs)
+					debugType = "full";
 			}
-			
-			switch (compilerParameters.LangVersion) {
-			case LangVersion.Default:
-				break;
-			case LangVersion.ISO_1:
-				sb.AppendLine ("/langversion:ISO-1");
-				break;
-			case LangVersion.ISO_2:
-				sb.AppendLine ("/langversion:ISO-2");
-				break;
-			default:
-				string message = "Invalid LangVersion enum value '" + compilerParameters.LangVersion.ToString () + "'";
-				monitor.ReportError (message, null);
-				LoggingService.LogError (message);
-				return null;
+			if (!string.Equals (debugType, "none", StringComparison.OrdinalIgnoreCase)) {
+					sb.AppendLine ("/debug:" + debugType);
+			}
+
+			if (compilerParameters.LangVersion != LangVersion.Default) {
+				var langVersionString = CSharpCompilerParameters.TryLangVersionToString (compilerParameters.LangVersion);
+				if (langVersionString == null) {
+					string message = "Invalid LangVersion enum value '" + compilerParameters.LangVersion.ToString () + "'";
+					monitor.ReportError (message, null);
+					LoggingService.LogError (message);
+					return null;
+				}
+				sb.AppendLine ("/langversion:" + langVersionString);
 			}
 			
 			// mcs default is + but others might not be
@@ -209,13 +258,10 @@ namespace MonoDevelop.CSharp
 				if (!string.IsNullOrEmpty (compilerParameters.WarningsNotAsErrors))
 					sb.AppendLine ("-warnaserror-:" + compilerParameters.WarningsNotAsErrors);
 			}
-			
-			if (compilerParameters.DefineSymbols.Length > 0) {
-				string define_str = string.Join (";", compilerParameters.DefineSymbols.Split (new char [] {',', ' ', ';'}, StringSplitOptions.RemoveEmptyEntries));
-				if (define_str.Length > 0) {
-					AppendQuoted (sb, "/define:", define_str);
-					sb.AppendLine ();
-				}
+
+			foreach (var define in configuration.GetDefineSymbols ()) {
+				AppendQuoted (sb, "/define:", define);
+				sb.AppendLine ();
 			}
 
 			CompileTarget ctarget = configuration.CompileTarget;
@@ -251,7 +297,7 @@ namespace MonoDevelop.CSharp
 						break;
 					case "EmbeddedResource":
 						string fname = finfo.Name;
-						if (string.Compare (Path.GetExtension (fname), ".resx", true) == 0)
+						if (string.Compare (Path.GetExtension (fname), ".resx", StringComparison.OrdinalIgnoreCase) == 0)
 							fname = Path.ChangeExtension (fname, ".resources");
 						sb.Append ('"');sb.Append ("/res:");
 						sb.Append (fname);sb.Append (',');sb.Append (finfo.ResourceId);
@@ -261,11 +307,8 @@ namespace MonoDevelop.CSharp
 						continue;
 				}
 			}
-			if (compilerParameters.GenerateXmlDocumentation) 
-				AppendQuoted (sb, "/doc:", Path.ChangeExtension (outputName, ".xml"));
-			
-			if (!string.IsNullOrEmpty (compilerParameters.AdditionalArguments)) 
-				sb.AppendLine (compilerParameters.AdditionalArguments);
+			if (!compilerParameters.DocumentationFile.IsNullOrEmpty) 
+				AppendQuoted (sb, "/doc:", compilerParameters.DocumentationFile);
 			
 			if (!string.IsNullOrEmpty (compilerParameters.NoWarnings)) 
 				AppendQuoted (sb, "/nowarn:", compilerParameters.NoWarnings);
@@ -279,37 +322,23 @@ namespace MonoDevelop.CSharp
 			string error  = "";
 			
 			File.WriteAllText (responseFileName, sb.ToString ());
-			
-			string compilerName;
-			try {
-				compilerName = GetCompilerName (runtime, configuration.TargetFramework);
-			} catch (Exception e) {
-				string message = "Could not obtain a C# compiler";
-				monitor.ReportError (message, e);
-				return null;
-			}
-			
-			monitor.Log.WriteLine (compilerName + " /noconfig " + sb.ToString ().Replace ('\n',' '));
-			
-			string workingDir = ".";
-			if (configuration.ParentItem != null) {
-				workingDir = configuration.ParentItem.BaseDirectory;
-				if (workingDir == null)
-					// Dummy projects created for single files have no filename
-					// and so no BaseDirectory.
-					// This is a workaround for a bug in 
-					// ProcessStartInfo.WorkingDirectory - not able to handle null
-					workingDir = ".";
-			}
 
-			LoggingService.LogInfo (compilerName + " " + sb.ToString ());
-			
+			monitor.Log.WriteLine (compilerName + " /noconfig " + sb.ToString ().Replace ('\n',' '));
+
+			// Dummy projects created for single files have no filename
+			// and so no BaseDirectory.
+			string workingDir = null;
+			if (configuration.ParentItem != null)
+				workingDir = configuration.ParentItem.BaseDirectory;
+
+			LoggingService.LogInfo (compilerName + " " + sb);
+
 			ExecutionEnvironment envVars = runtime.GetToolsExecutionEnvironment (project.TargetFramework);
 			string cargs = "/noconfig @\"" + responseFileName + "\"";
 
-			int exitCode = DoCompilation (compilerName, cargs, workingDir, envVars, gacRoots, ref output, ref error);
+			int exitCode = DoCompilation (monitor, compilerName, cargs, workingDir, envVars, gacRoots, ref output, ref error);
 			
-			BuildResult result = ParseOutput (output, error);
+			BuildResult result = ParseOutput (workingDir, output, error);
 			if (result.CompilerOutput.Trim ().Length != 0)
 				monitor.Log.WriteLine (result.CompilerOutput);
 			
@@ -342,7 +371,7 @@ namespace MonoDevelop.CSharp
 			}
 		}
 		
-		static BuildResult ParseOutput (string stdout, string stderr)
+		static BuildResult ParseOutput (string basePath, string stdout, string stderr)
 		{
 			BuildResult result = new BuildResult ();
 			
@@ -365,13 +394,13 @@ namespace MonoDevelop.CSharp
 					if (curLine.Length == 0) 
 						continue;
 					
-					if (curLine.StartsWith ("Unhandled Exception: System.TypeLoadException") || 
-					    curLine.StartsWith ("Unhandled Exception: System.IO.FileNotFoundException")) {
+					if (curLine.StartsWith ("Unhandled Exception: System.TypeLoadException", StringComparison.Ordinal) ||
+					    curLine.StartsWith ("Unhandled Exception: System.IO.FileNotFoundException", StringComparison.Ordinal)) {
 						result.ClearErrors ();
 						typeLoadException = true;
 					}
 					
-					BuildError error = CreateErrorFromString (curLine);
+					BuildError error = CreateErrorFromString (basePath, curLine);
 					
 					if (error != null)
 						result.Append (error);
@@ -389,10 +418,10 @@ namespace MonoDevelop.CSharp
 			return result;
 		}
 		
-		static int DoCompilation (string compilerName, string compilerArgs, string working_dir, ExecutionEnvironment envVars, List<string> gacRoots, ref string output, ref string error) 
+		static int DoCompilation (IProgressMonitor monitor, string compilerName, string compilerArgs, string working_dir, ExecutionEnvironment envVars, List<string> gacRoots, ref string output, ref string error)
 		{
-			output = Path.GetTempFileName();
-			error = Path.GetTempFileName();
+			output = Path.GetTempFileName ();
+			error = Path.GetTempFileName ();
 			
 			StreamWriter outwr = new StreamWriter (output);
 			StreamWriter errwr = new StreamWriter (error);
@@ -400,7 +429,9 @@ namespace MonoDevelop.CSharp
 			ProcessStartInfo pinfo = new ProcessStartInfo (compilerName, compilerArgs);
 			pinfo.StandardErrorEncoding = Encoding.UTF8;
 			pinfo.StandardOutputEncoding = Encoding.UTF8;
-			pinfo.WorkingDirectory = working_dir;
+
+			// The "." is a workaround for a bug in ProcessStartInfo.WorkingDirectory - not able to handle null
+			pinfo.WorkingDirectory = working_dir ?? ".";
 			
 			if (gacRoots.Count > 0) {
 				// Create the gac prefix string
@@ -416,20 +447,23 @@ namespace MonoDevelop.CSharp
 			pinfo.UseShellExecute = false;
 			pinfo.RedirectStandardOutput = true;
 			pinfo.RedirectStandardError = true;
-			
+
 			MonoDevelop.Core.Execution.ProcessWrapper pw = Runtime.ProcessService.StartProcess (pinfo, outwr, errwr, null);
-			pw.WaitForOutput();
+			using (var mon = new AggregatedOperationMonitor (monitor, pw)) {
+				pw.WaitForOutput ();
+			}
 			int exitCode = pw.ExitCode;
+			bool cancelRequested = pw.CancelRequested;
 			outwr.Close();
-			errwr.Close();
-			pw.Dispose ();
-			return exitCode;
+			errwr.Close ();
+			pw.Dispose();
+			return cancelRequested ? 0 : exitCode;
 		}
 		
 		// Snatched from our codedom code, with some changes to make it compatible with csc
 		// (the line+column group is optional is csc)
-		static Regex regexError = new Regex (@"^(\s*(?<file>[^\(]+)(\((?<line>\d*)(,(?<column>\d*[\+]*))?\))?:\s+)*(?<level>\w+)\s+(?<number>..\d+):\s*(?<message>.*)", RegexOptions.Compiled | RegexOptions.ExplicitCapture);
-		static BuildError CreateErrorFromString (string error_string)
+		static Regex regexError = new Regex (@"^(\s*(?<file>.+[^)])(\((?<line>\d*)(,(?<column>\d*[\+]*))?\))?:\s+)*(?<level>\w+)\s+(?<number>..\d+):\s*(?<message>.*)", RegexOptions.Compiled | RegexOptions.ExplicitCapture);
+		static BuildError CreateErrorFromString (string basePath, string error_string)
 		{
 			// When IncludeDebugInformation is true, prevents the debug symbols stats from braeking this.
 			if (error_string.StartsWith ("WROTE SYMFILE") ||
@@ -443,7 +477,14 @@ namespace MonoDevelop.CSharp
 				return null;
 			
 			BuildError error = new BuildError ();
-			error.FileName = match.Result ("${file}") ?? "";
+			FilePath filename = match.Result ("${file}");
+			if (filename.IsNullOrEmpty) {
+				filename = FilePath.Empty;
+			} else if (!filename.IsAbsolute && basePath != null) {
+				filename = filename.ToAbsolute (basePath);
+			}
+			error.FileName = filename;
+
 			
 			string line = match.Result ("${line}");
 			error.Line = !string.IsNullOrEmpty (line) ? Int32.Parse (line) : 0;
