@@ -49,6 +49,10 @@ using ICSharpCode.NRefactory.TypeSystem.Implementation;
 using Mono.TextEditor.Theatrics;
 using MonoDevelop.SourceEditor;
 using XmlDocIdLib;
+using MonoDevelop.Ide.Gui;
+using MonoDevelop.Components;
+using System.Threading.Tasks;
+using System.Threading;
 
 namespace MonoDevelop.AssemblyBrowser
 {
@@ -56,6 +60,12 @@ namespace MonoDevelop.AssemblyBrowser
 	[System.ComponentModel.ToolboxItem(true)]
 	partial class AssemblyBrowserWidget : Gtk.Bin
 	{
+		Gtk.Button buttonBack;
+		Gtk.Button buttonForeward;
+		Gtk.ComboBox comboboxVisibilty;
+		MonoDevelop.Components.SearchEntry searchentry1;
+		Gtk.ComboBox languageCombobox;
+
 		public AssemblyBrowserTreeView TreeView {
 			get;
 			private set;
@@ -73,7 +83,6 @@ namespace MonoDevelop.AssemblyBrowser
 		}
 		
 		DocumentationPanel documentationPanel = new DocumentationPanel ();
-		TextEditorContainer textEditorContainer;
 		readonly TextEditor inspectEditor;
 
 		public class AssemblyBrowserTreeView : ExtensibleTreeView
@@ -88,8 +97,12 @@ namespace MonoDevelop.AssemblyBrowser
 					if (publicApiOnly == value)
 						return;
 					publicApiOnly = value;
-					GetRootNode ().Options ["PublicApiOnly"] = publicApiOnly;
-					RefreshTree ();
+					var root = GetRootNode ();
+					if (root != null) {
+						do {
+							RefreshNode (root);
+						} while (root.MoveNext ());
+					}
 				}
 			}
 
@@ -101,6 +114,56 @@ namespace MonoDevelop.AssemblyBrowser
 		public AssemblyBrowserWidget ()
 		{
 			this.Build ();
+
+			buttonBack = new Gtk.Button (ImageService.GetImage ("md-breadcrumb-prev", Gtk.IconSize.Menu));
+			buttonBack.Clicked += OnNavigateBackwardActionActivated;
+
+			buttonForeward = new Gtk.Button (ImageService.GetImage ("md-breadcrumb-next", Gtk.IconSize.Menu));
+			buttonForeward.Clicked += OnNavigateForwardActionActivated;
+
+			comboboxVisibilty = ComboBox.NewText ();
+			comboboxVisibilty.InsertText (0, GettextCatalog.GetString ("Only public members"));
+			comboboxVisibilty.InsertText (1, GettextCatalog.GetString ("All members"));
+			comboboxVisibilty.Active = 0;
+			comboboxVisibilty.Changed += delegate {
+				TreeView.PublicApiOnly = comboboxVisibilty.Active == 0;
+				FillInspectLabel ();
+			};
+
+			searchentry1 = new MonoDevelop.Components.SearchEntry ();
+			searchentry1.Ready = true;
+			searchentry1.HasFrame = true;
+			searchentry1.WidthRequest = 200;
+			searchentry1.Visible = true;
+			searchentry1.EmptyMessage = GettextCatalog.GetString ("Search for types or members");
+			searchentry1.InnerEntry.Changed += SearchEntryhandleChanged;
+
+			CheckMenuItem checkMenuItem = this.searchentry1.AddFilterOption (0, GettextCatalog.GetString ("Types"));
+			checkMenuItem.Active = true;
+			checkMenuItem.Toggled += delegate {
+				if (checkMenuItem.Active) {
+					searchMode = AssemblyBrowserWidget.SearchMode.Type;
+					CreateColumns ();
+					StartSearch ();
+				}
+			};
+			
+			CheckMenuItem checkMenuItem1 = this.searchentry1.AddFilterOption (1, GettextCatalog.GetString ("Members"));
+			checkMenuItem1.Toggled += delegate {
+				if (checkMenuItem1.Active) {
+					searchMode = AssemblyBrowserWidget.SearchMode.Member;
+					CreateColumns ();
+					StartSearch ();
+				}
+			};
+
+			languageCombobox = Gtk.ComboBox.NewText ();
+			//languageCombobox.AppendText (GettextCatalog.GetString ("Summary"));
+			languageCombobox.AppendText (GettextCatalog.GetString ("IL"));
+			languageCombobox.AppendText (GettextCatalog.GetString ("C#"));
+			languageCombobox.Active = Math.Min (1, PropertyService.Get ("AssemblyBrowser.InspectLanguage", 1));
+			languageCombobox.Changed += LanguageComboboxhandleChanged;
+
 			loader = new CecilLoader (true);
 			loader.IncludeInternalMembers = true;
 			TreeView = new AssemblyBrowserTreeView (new NodeBuilder[] { 
@@ -109,7 +172,7 @@ namespace MonoDevelop.AssemblyBrowser
 				new AssemblyNodeBuilder (this),
 				new ModuleReferenceNodeBuilder (),
 				new AssemblyReferenceNodeBuilder (this),
-				new AssemblyReferenceFolderNodeBuilder (this),
+				//new AssemblyReferenceFolderNodeBuilder (this),
 				new AssemblyResourceFolderNodeBuilder (),
 				new ResourceNodeBuilder (),
 				new NamespaceBuilder (this),
@@ -120,27 +183,24 @@ namespace MonoDevelop.AssemblyBrowser
 				new DomPropertyNodeBuilder (this),
 				new BaseTypeFolderNodeBuilder (this),
 				new BaseTypeNodeBuilder (this)
-				}, new TreePadOption [] {
-				new TreePadOption ("PublicApiOnly", GettextCatalog.GetString ("Show public members only"), true)
-			});
+				}, new TreePadOption [0]);
 			TreeView.Tree.Selection.Mode = Gtk.SelectionMode.Single;
 			TreeView.Tree.CursorChanged += HandleCursorChanged;
-			TreeView.ShadowType = ShadowType.In;
+			TreeView.ShadowType = ShadowType.None;
+			TreeView.BorderWidth = 1;
+			TreeView.ShowBorderLine = false;
+			TreeView.Zoom = 1.0;
 			treeViewPlaceholder.Add (TreeView);
-			treeViewPlaceholder.ShowAll ();
-			
+
 //			this.descriptionLabel.ModifyFont (Pango.FontDescription.FromString ("Sans 9"));
-			this.documentationLabel.ModifyFont (Pango.FontDescription.FromString ("Sans 12"));
-			this.documentationLabel.ModifyBg (Gtk.StateType.Normal, new Gdk.Color (255, 255, 225));
-			this.documentationLabel.Wrap = true;
+//			this.documentationLabel.ModifyFont (Pango.FontDescription.FromString ("Sans 12"));
+//			this.documentationLabel.ModifyBg (Gtk.StateType.Normal, new Gdk.Color (255, 255, 225));
+//			this.documentationLabel.Wrap = true;
 			
 			var options = new MonoDevelop.Ide.Gui.CommonTextEditorOptions () {
 				ShowFoldMargin = false,
 				ShowIconMargin = false,
-				ShowInvalidLines = false,
 				ShowLineNumberMargin = false,
-				ShowSpaces = false,
-				ShowTabs = false,
 				HighlightCaretLine = true,
 			};
 			inspectEditor = new TextEditor (new TextDocument (), options);
@@ -178,93 +238,62 @@ namespace MonoDevelop.AssemblyBrowser
 				return referencedSegment.Reference.ToString ();
 			};
 			this.inspectEditor.LinkRequest += InspectEditorhandleLinkRequest;
-			var scrolledWindow = new SmartScrolledWindow ();
-			scrolledWindow.Show ();
-			textEditorContainer = new TextEditorContainer (inspectEditor);
-			scrolledWindow.Child = textEditorContainer;
-			notebookInspection.Add (scrolledWindow);
-			var notebookChild = ((Notebook.NotebookChild)(notebookInspection [scrolledWindow]));
-			notebookChild.Position = 1;
+			documentationScrolledWindow.Add (inspectEditor);
 
-//			this.inspectLabel.ModifyBg (Gtk.StateType.Normal, new Gdk.Color (255, 255, 250));
-			
-//			this.vpaned1.ExposeEvent += VPaneExpose;
 			this.hpaned1.ExposeEvent += HPaneExpose;
-/*			this.notebook1.SwitchPage += delegate {
-				// Hack for the switch page select all bug.
-//				this.inspectLabel.Selectable = false;
-			};*/
+			hpaned1 = hpaned1.ReplaceWithWidget (new HPanedThin (), true);
+			hpaned1.Position = 271;
 
-			languageCombobox.AppendText (GettextCatalog.GetString ("Summary"));
-			languageCombobox.AppendText (GettextCatalog.GetString ("IL"));
-			languageCombobox.AppendText (GettextCatalog.GetString ("C#"));
-			languageCombobox.Active = PropertyService.Get ("AssemblyBrowser.InspectLanguage", 2);
-			languageCombobox.Changed += LanguageComboboxhandleChanged;
-			searchentry1.Ready = true;
-			searchentry1.WidthRequest = 200;
-			searchentry1.Visible = true;
-			searchentry1.EmptyMessage = GettextCatalog.GetString ("Search for types or members");
-			searchentry1.InnerEntry.Changed += SearchEntryhandleChanged;
-			
-			CheckMenuItem checkMenuItem = this.searchentry1.AddFilterOption (0, GettextCatalog.GetString ("Types"));
-			checkMenuItem.Active = true;
-			checkMenuItem.Toggled += delegate {
-				if (checkMenuItem.Active) {
-					searchMode = AssemblyBrowserWidget.SearchMode.Type;
-					CreateColumns ();
-					StartSearch ();
-				}
-			};
-			
-			CheckMenuItem checkMenuItem1 = this.searchentry1.AddFilterOption (1, GettextCatalog.GetString ("Members"));
-			checkMenuItem1.Toggled += delegate {
-				if (checkMenuItem1.Active) {
-					searchMode = AssemblyBrowserWidget.SearchMode.Member;
-					CreateColumns ();
-					StartSearch ();
-				}
-			};
-			comboboxVisibilty.InsertText (0, GettextCatalog.GetString ("Only public members"));
-			comboboxVisibilty.InsertText (1, GettextCatalog.GetString ("All members"));
-			comboboxVisibilty.Active = 0;
-			comboboxVisibilty.Changed += delegate {
-				TreeView.PublicApiOnly = comboboxVisibilty.Active == 0;
-
-				FillInspectLabel ();
-			};
-			/*
-			this.searchInCombobox.Active = 0;
-			this.searchInCombobox.Changed += SearchInComboboxhandleChanged;
-			*/
 			this.notebook1.SetTabLabel (this.documentationScrolledWindow, new Label (GettextCatalog.GetString ("Documentation")));
-			this.notebook1.SetTabLabel (this.notebookInspection, new Label (GettextCatalog.GetString ("Inspect")));
 			this.notebook1.SetTabLabel (this.searchWidget, new Label (GettextCatalog.GetString ("Search")));
+			notebook1.Page = 0;
 			//this.searchWidget.Visible = false;
 				
-			typeListStore = new Gtk.ListStore (typeof(Gdk.Pixbuf), // type image
+			typeListStore = new Gtk.ListStore (typeof(Xwt.Drawing.Image), // type image
 			                                   typeof(string), // name
 			                                   typeof(string), // namespace
 			                                   typeof(string), // assembly
 				                               typeof(IMember)
 			                                  );
 			
-			memberListStore = new Gtk.ListStore (typeof(Gdk.Pixbuf), // member image
+			memberListStore = new Gtk.ListStore (typeof(Xwt.Drawing.Image), // member image
 			                                   typeof(string), // name
 			                                   typeof(string), // Declaring type full name
 			                                   typeof(string), // assembly
 				                               typeof(IMember)
 			                                  );
 			CreateColumns ();
-			SetInspectWidget ();
 //			this.searchEntry.Changed += SearchEntryhandleChanged;
 			this.searchTreeview.RowActivated += SearchTreeviewhandleRowActivated;
-			this.searchentry1.ShowAll ();
-			this.buttonBack.Clicked += this.OnNavigateBackwardActionActivated;
-			this.buttonForeward.Clicked += this.OnNavigateForwardActionActivated;
 			this.notebook1.ShowTabs = false;
-			this.notebookInspection.ShowTabs = false;
 			this.ShowAll ();
 		}
+
+		internal void SetToolbar (DocumentToolbar toolbar)
+		{
+			toolbar.Add (buttonBack);
+
+			toolbar.Add (buttonForeward);
+
+			toolbar.Add (new VSeparator ());
+
+			Gtk.Label la = new Label (GettextCatalog.GetString ("Visibility"));
+			toolbar.Add (la);
+
+			toolbar.Add (comboboxVisibilty);
+
+			la = new Label ("");
+			toolbar.Add (la, true);
+
+			toolbar.Add (searchentry1);
+
+			la = new Label (GettextCatalog.GetString ("Language"));
+			toolbar.Add (la);
+
+			toolbar.Add (languageCombobox);
+
+			toolbar.ShowAll ();
+		}
 		
 		[CommandHandler (EditCommands.Copy)]
 		protected void OnCopyCommand ()
@@ -285,7 +314,7 @@ namespace MonoDevelop.AssemblyBrowser
 			var menuSet = new CommandEntrySet ();
 			menuSet.AddItem (EditCommands.SelectAll);
 			menuSet.AddItem (EditCommands.Copy);
-			IdeApp.CommandService.ShowContextMenu (menuSet, this);
+			IdeApp.CommandService.ShowContextMenu (this, args.Event, menuSet, this);
 		}
 
 		void SearchTreeviewhandleRowActivated (object o, RowActivatedArgs args)
@@ -310,12 +339,13 @@ namespace MonoDevelop.AssemblyBrowser
 		{
 			this.notebook1.Page = 0;
 			PropertyService.Set ("AssemblyBrowser.InspectLanguage", this.languageCombobox.Active);
-			SetInspectWidget ();
 			FillInspectLabel ();
 		}
 
 		void InspectEditorhandleLinkRequest (object sender, Mono.TextEditor.LinkEventArgs args)
 		{
+			var loader = (AssemblyLoader)this.TreeView.GetSelectedNode ().GetParentDataItem (typeof(AssemblyLoader), true);
+
 			if (args.Button == 2 || (args.Button == 1 && (args.ModifierState & Gdk.ModifierType.ShiftMask) == Gdk.ModifierType.ShiftMask)) {
 				AssemblyBrowserViewContent assemblyBrowserView = new AssemblyBrowserViewContent ();
 				foreach (var cu in definitions) {
@@ -324,7 +354,7 @@ namespace MonoDevelop.AssemblyBrowser
 				IdeApp.Workbench.OpenDocument (assemblyBrowserView, true);
 				((AssemblyBrowserWidget)assemblyBrowserView.Control).Open (args.Link);
 			} else {
-				this.Open (args.Link);
+				this.Open (args.Link, loader);
 			}
 		}
 		
@@ -346,33 +376,61 @@ namespace MonoDevelop.AssemblyBrowser
 			
 		ITreeNavigator SearchMember (string helpUrl)
 		{
-			return SearchMember (TreeView.GetRootNode (), helpUrl);
+			var nav = SearchMember (TreeView.GetRootNode (), helpUrl);
+			if (nav != null)
+				return nav;
+			// Constructor may be a generated default without implementation.
+			var ctorIdx = helpUrl.IndexOf (".#ctor", StringComparison.Ordinal);
+			if (helpUrl.StartsWith ("M:", StringComparison.Ordinal) && ctorIdx > 0) {
+				return SearchMember ("T" + helpUrl.Substring (1, ctorIdx - 1));
+			}
+			return null;
 		}
 		
 		static void AppendTypeReference (StringBuilder result, ITypeReference type)
 		{
+			if (type is ByReferenceTypeReference) {
+				var brtr = (ByReferenceTypeReference)type;
+				AppendTypeReference (result, brtr.ElementType);
+				return;
+			}
+
 			if (type is ArrayTypeReference) {
 				var array = (ArrayTypeReference)type;
 				AppendTypeReference (result, array.ElementType);
 				result.Append ("[");
-				result.Append (new string (',', array.Dimensions));
+				result.Append (new string (',', array.Dimensions - 1));
 				result.Append ("]");
 				return;
 			}
-			
+
 			if (type is PointerTypeReference) {
 				var ptr = (PointerTypeReference)type;
 				AppendTypeReference (result, ptr.ElementType);
 				result.Append ("*");
 				return;
 			}
-			
-			if (type is IUnresolvedTypeDefinition)
+
+			if (type is GetClassTypeReference) {
+				var r = (GetClassTypeReference)type;
+				var n = r.FullTypeName.TopLevelTypeName;
+				result.Append (n.Namespace + "." + n.Name);
+				return;
+			}
+
+			if (type is IUnresolvedTypeDefinition) {
 				result.Append (((IUnresolvedTypeDefinition)type).FullName);
+			}
+
+			if (type is TypeParameterReference) {
+				result.Append ("`" +((TypeParameterReference)type).Index);
+			}
 		}
 		
 		static void AppendHelpParameterList (StringBuilder result, IList<IUnresolvedParameter> parameters)
 		{
+			if (parameters == null || parameters.Count == 0)
+				return;
 			result.Append ('(');
 			if (parameters != null) {
 				for (int i = 0; i < parameters.Count; i++) {
@@ -381,36 +439,41 @@ namespace MonoDevelop.AssemblyBrowser
 					var p = parameters [i];
 					if (p == null)
 						continue;
-					if (p.IsRef || p.IsOut)
-						result.Append ("&");
 					AppendTypeReference (result, p.Type);
+					if (p.IsRef)
+						result.Append ("&");
+					if (p.IsOut) {
+						result.Append ("@");
+					}
 				}
 			}
 			result.Append (')');
 		}
 		
-		static string GetIdString (IUnresolvedEntity member)
+		internal static string GetIdString (IUnresolvedEntity member)
 		{
 			StringBuilder sb;
 			
-			switch (member.EntityType) {
-			case EntityType.TypeDefinition:
+			switch (member.SymbolKind) {
+			case SymbolKind.TypeDefinition:
 				var type = member as IUnresolvedTypeDefinition;
 				if (type.TypeParameters.Count == 0)
 					return "T:" + type.FullName;
 				return "T:" + type.FullName + "`" + type.TypeParameters.Count;
-			case EntityType.Method:
+			case SymbolKind.Method:
 				var method = (IUnresolvedMethod)member;
 				sb = new StringBuilder ();
 				sb.Append ("M:");
-				sb.Append (method.FullName);
+				sb.Append (method.DeclaringTypeDefinition.ReflectionName);
+				sb.Append (".");
+				sb.Append (method.Name);
 				if (method.TypeParameters.Count > 0) {
 					sb.Append ("`");
 					sb.Append (method.TypeParameters.Count);
 				}
 				AppendHelpParameterList (sb, method.Parameters);
 				return sb.ToString ();
-			case EntityType.Constructor:
+			case SymbolKind.Constructor:
 				var constructor = (IUnresolvedMethod)member;
 				sb = new StringBuilder ();
 				sb.Append ("M:");
@@ -418,23 +481,38 @@ namespace MonoDevelop.AssemblyBrowser
 				sb.Append (".#ctor");
 				AppendHelpParameterList (sb, constructor.Parameters);
 				return sb.ToString ();
-			case EntityType.Destructor: // todo
+			case SymbolKind.Destructor: // todo
 				return "todo";
-			case EntityType.Property:
-				return "P:" + member.FullName;
-			case EntityType.Indexer:
+			case SymbolKind.Property:
+				sb = new StringBuilder ();
+				sb.Append ("P:");
+				sb.Append (member.DeclaringTypeDefinition.ReflectionName);
+				sb.Append (".");
+				sb.Append (member.Name);
+				return sb.ToString ();
+			case SymbolKind.Indexer:
 				var indexer = (IUnresolvedProperty)member;
 				sb = new StringBuilder ();
 				sb.Append ("P:");
-				sb.Append (indexer.DeclaringTypeDefinition.FullName);
+				sb.Append (indexer.DeclaringTypeDefinition.ReflectionName);
 				sb.Append (".Item");
 				AppendHelpParameterList (sb, indexer.Parameters);
 				return sb.ToString ();
-			case EntityType.Field:
-				return "F:" + member.FullName;
-			case EntityType.Event:
-				return "E:" + member.FullName;
-			case EntityType.Operator: // todo
+			case SymbolKind.Field:
+				sb = new StringBuilder ();
+				sb.Append ("F:");
+				sb.Append (member.DeclaringTypeDefinition.ReflectionName);
+				sb.Append (".");
+				sb.Append (member.Name);
+				return sb.ToString ();
+			case SymbolKind.Event:
+				sb = new StringBuilder ();
+				sb.Append ("E:");
+				sb.Append (member.DeclaringTypeDefinition.ReflectionName);
+				sb.Append (".");
+				sb.Append (member.Name);
+				return sb.ToString ();
+			case SymbolKind.Operator: // todo
 				return "todo";
 			}
 			return "unknown entity: " + member;
@@ -470,8 +548,9 @@ namespace MonoDevelop.AssemblyBrowser
 				if (member is IUnresolvedTypeDefinition)
 					return GetIdString (member) == helpUrl;
 			} else {
-				if (member is IUnresolvedMember)
+				if (member is IUnresolvedMember) {
 					return GetIdString (member) == helpUrl;
+				}
 			}
 			return false;
 		}
@@ -505,11 +584,15 @@ namespace MonoDevelop.AssemblyBrowser
 		
 		ITreeNavigator SearchMember (ITreeNavigator nav, string helpUrl)
 		{
-			bool searchType = helpUrl.StartsWith ("T:");
+			if (nav == null)
+				return null;
+			bool searchType = helpUrl.StartsWith ("T:", StringComparison.Ordinal);
 			do {
 				if (IsMatch (nav, helpUrl, searchType)) {
+					inspectEditor.ClearSelection ();
 					nav.ExpandToNode ();
 					nav.Selected = nav.Expanded = true;
+					nav.ScrollToNode ();
 					return nav;
 				}
 				if (!SkipChildren (nav, helpUrl, searchType) && nav.HasChildren ()) {
@@ -522,8 +605,7 @@ namespace MonoDevelop.AssemblyBrowser
 						return null;
 					}
 					try {
-						if (nav.DataItem is TypeDefinition && nav.Options ["PublicApiOnly"]) {
-							nav.Options ["PublicApiOnly"] = PublicApiOnly;
+						if (nav.DataItem is TypeDefinition && PublicApiOnly) {
 							nav.MoveToFirstChild ();
 							result = SearchMember (nav, helpUrl);
 							if (result != null)
@@ -562,11 +644,11 @@ namespace MonoDevelop.AssemblyBrowser
 			case SearchMode.Decompiler:
 				col = new TreeViewColumn ();
 				col.Title = GettextCatalog.GetString ("Member");
-				crp = new Gtk.CellRendererPixbuf ();
+				crp = new CellRendererImage ();
 				crt = new Gtk.CellRendererText ();
 				col.PackStart (crp, false);
 				col.PackStart (crt, true);
-				col.AddAttribute (crp, "pixbuf", 0);
+				col.AddAttribute (crp, "image", 0);
 				col.AddAttribute (crt, "text", 1);
 				searchTreeview.AppendColumn (col);
 				col.Resizable = true;
@@ -579,11 +661,11 @@ namespace MonoDevelop.AssemblyBrowser
 			case SearchMode.Type:
 				col = new TreeViewColumn ();
 				col.Title = GettextCatalog.GetString ("Type");
-				crp = new Gtk.CellRendererPixbuf ();
+				crp = new CellRendererImage ();
 				crt = new Gtk.CellRendererText ();
 				col.PackStart (crp, false);
 				col.PackStart (crt, true);
-				col.AddAttribute (crp, "pixbuf", 0);
+				col.AddAttribute (crp, "image", 0);
 				col.AddAttribute (crt, "text", 1);
 				searchTreeview.AppendColumn (col);
 				col.Resizable = true;
@@ -672,7 +754,7 @@ namespace MonoDevelop.AssemblyBrowser
 							foreach (var member in kv.Value) {
 								if (worker.CancellationPending)
 									return;
-								memberListStore.AppendValues (ImageService.GetPixbuf (member.GetStockIcon (), Gtk.IconSize.Menu),
+								memberListStore.AppendValues (ImageService.GetIcon (member.GetStockIcon (), Gtk.IconSize.Menu),
 								                              member.Name,
 								                              member.DeclaringTypeDefinition.FullName,
 								                              kv.Key.Assembly.FullName,
@@ -707,7 +789,7 @@ namespace MonoDevelop.AssemblyBrowser
 							foreach (var member in kv.Value) {
 								if (worker.CancellationPending)
 									return;
-								memberListStore.AppendValues ("", //iImageService.GetPixbuf (member.StockIcon, Gtk.IconSize.Menu),
+								memberListStore.AppendValues ("", //iImageService.GetIcon (member.StockIcon, Gtk.IconSize.Menu),
 								                              member.Name,
 								                              member.DeclaringTypeDefinition.FullName,
 								                              kv.Key.Assembly.FullName,
@@ -737,7 +819,7 @@ namespace MonoDevelop.AssemblyBrowser
 							foreach (var member in kv.Value) {
 								if (worker.CancellationPending)
 									return;
-								memberListStore.AppendValues ("", //ImageService.GetPixbuf (member.StockIcon, Gtk.IconSize.Menu),
+								memberListStore.AppendValues ("", //ImageService.GetIcon (member.StockIcon, Gtk.IconSize.Menu),
 								                              member.Name,
 								                              member.DeclaringTypeDefinition.FullName,
 								                              kv.Key.Assembly.FullName,
@@ -758,14 +840,13 @@ namespace MonoDevelop.AssemblyBrowser
 								typeList.Add (type);
 						}
 						typeDict [unit] = typeList;
-						Console.WriteLine (unit.UnresolvedAssembly.AssemblyName);
 					}
 					Gtk.Application.Invoke (delegate {
 						foreach (var kv in typeDict) {
 							foreach (var type in kv.Value) {
 								if (worker.CancellationPending)
 									return;
-								typeListStore.AppendValues (ImageService.GetPixbuf (type.GetStockIcon (), Gtk.IconSize.Menu),
+								typeListStore.AppendValues (ImageService.GetIcon (type.GetStockIcon (), Gtk.IconSize.Menu),
 								                            type.Name,
 								                            type.Namespace,
 								                            kv.Key.Assembly.FullName,
@@ -985,15 +1066,6 @@ namespace MonoDevelop.AssemblyBrowser
 			return result.ToString ();
 		}
 		
-		void SetInspectWidget ()
-		{
-			if (this.languageCombobox.Active <= 0) {
-				notebookInspection.Page = 0;
-			} else {
-				notebookInspection.Page = 1;
-			}
-		}
-		
 		List<ReferenceSegment> ReferencedSegments = new List<ReferenceSegment>();
 		List<UnderlineMarker> underlineMarkers = new List<UnderlineMarker> ();
 		
@@ -1017,7 +1089,7 @@ namespace MonoDevelop.AssemblyBrowser
 				string text = inspectEditor.GetTextAt (seg);
 				if (text != null && text.Length == 1 && !(char.IsLetter (text [0]) || text [0] == '…'))
 					continue;
-				var marker = new UnderlineMarker ("blue", 1 + seg.Offset - line.Offset, 1 + seg.EndOffset - line.Offset);
+				var marker = new UnderlineMarker (new Cairo.Color (0, 0, 1.0), 1 + seg.Offset - line.Offset, 1 + seg.EndOffset - line.Offset);
 				marker.Wave = false;
 				underlineMarkers.Add (marker);
 				inspectEditor.Document.AddMarker (line, marker);
@@ -1039,16 +1111,11 @@ namespace MonoDevelop.AssemblyBrowser
 			inspectEditor.Document.ClearFoldSegments ();
 			switch (this.languageCombobox.Active) {
 			case 0:
-				inspectEditor.Options.ShowFoldMargin = false;
-				this.inspectEditor.Document.MimeType = "text/x-csharp";
-				this.documentationPanel.Markup = builder.GetDocumentationMarkup (nav);
-				break;
-			case 1:
 				inspectEditor.Options.ShowFoldMargin = true;
 				this.inspectEditor.Document.MimeType = "text/x-ilasm";
 				SetReferencedSegments (builder.Disassemble (inspectEditor.GetTextEditorData (), nav));
 				break;
-			case 2:
+			case 1:
 				inspectEditor.Options.ShowFoldMargin = true;
 				this.inspectEditor.Document.MimeType = "text/x-csharp";
 				SetReferencedSegments (builder.Decompile (inspectEditor.GetTextEditorData (), nav, PublicApiOnly));
@@ -1084,7 +1151,7 @@ namespace MonoDevelop.AssemblyBrowser
 					} catch (Exception) {
 					}
 				}
-				this.documentationLabel.Markup = documentation;
+//				this.documentationLabel.Markup = documentation;
 /*				IAssemblyBrowserNodeBuilder builder = nav.TypeNodeBuilder as IAssemblyBrowserNodeBuilder;
 				if (builder != null) {
 					this.descriptionLabel.Markup  = builder.GetDescription (nav);
@@ -1114,25 +1181,107 @@ namespace MonoDevelop.AssemblyBrowser
 			this.hpaned1.Position = Math.Min (350, this.Allocation.Width * 2 / 3);
 		}
 		
-		public void Open (string url)
+		public void Open (string url, AssemblyLoader currentAssembly = null)
 		{
 			ITreeNavigator nav = SearchMember (url);
 			if (definitions == null) // we've been disposed
 				return;
-			if (nav == null) {
-				foreach (var definition in definitions.ToArray ()) {
-					var cecilObject = loader.GetCecilObject (definition.UnresolvedAssembly);
-					if (cecilObject == null)
-						continue;
-					foreach (var assemblyNameReference in cecilObject.MainModule.AssemblyReferences) {
-						AddReferenceByAssemblyName (assemblyNameReference);
+			if (nav != null)
+				return;
+			try {
+				if (currentAssembly != null) {
+					OpenFromAssembly (url, currentAssembly);
+				} else {
+					OpenFromAssemblyNames (url);
+				}
+			} catch (Exception e) {
+				LoggingService.LogError ("Error while opening the assembly browser with id:" + url, e); 
+			}
+		}
+
+		void OpenFromAssembly (string url, AssemblyLoader currentAssembly)
+		{
+			var cecilObject = loader.GetCecilObject (currentAssembly.UnresolvedAssembly);
+			if (cecilObject == null)
+				return;
+
+			int i = 0;
+			System.Action loadNext = null;
+			var references = cecilObject.MainModule.AssemblyReferences;
+			loadNext = () => {
+				var reference = references [i];
+				string fileName = currentAssembly.LookupAssembly (reference.FullName);
+				if (string.IsNullOrEmpty (fileName)) {
+					LoggingService.LogWarning ("Assembly browser: Can't find assembly: " + reference.FullName + ".");
+					if (++i == references.Count)
+						LoggingService.LogError ("Assembly browser: Can't find: " + url + ".");
+					else
+						loadNext ();
+					return;
+				}
+				var result = AddReferenceByFileName (fileName, true);
+				result.LoadingTask.ContinueWith (t2 => {
+					t2.Wait ();
+					if (definitions == null) // disposed
+						return;
+					Application.Invoke (delegate {
+						var nav = SearchMember (url);
+						if (nav == null) {
+							if (++i == references.Count)
+								LoggingService.LogError ("Assembly browser: Can't find: " + url + ".");
+							else
+								loadNext ();
+						}
+					});
+				}, TaskScheduler.Current);
+			};
+		}
+
+		void OpenFromAssemblyNames (string url)
+		{
+			var tasks = new List<Task> ();
+			foreach (var definition in definitions.ToArray ()) {
+				var cecilObject = loader.GetCecilObject (definition.UnresolvedAssembly);
+				if (cecilObject == null) {
+					LoggingService.LogWarning ("Assembly browser: Can't find assembly: " + definition.UnresolvedAssembly.FullAssemblyName + ".");
+					continue;
+				}
+				foreach (var assemblyNameReference in cecilObject.MainModule.AssemblyReferences) {
+					var result = AddReferenceByAssemblyName (assemblyNameReference);
+					if (result == null) {
+						LoggingService.LogWarning ("Assembly browser: Can't find assembly: " + assemblyNameReference.FullName + ".");
+					} else {
+						tasks.Add (result.LoadingTask);
 					}
 				}
-				nav = SearchMember (url);
-			}
-			if (nav == null) {
-				LoggingService.LogError ("Can't open: " + url + " (not found).");
 			}
+			if (tasks.Count == 0) {
+				var nav = SearchMember (url);
+				if (nav == null) {
+					LoggingService.LogError ("Assembly browser: Can't find: " + url + ".");
+				}
+				return;
+			};
+			Task.Factory.ContinueWhenAll (tasks.ToArray (), tarr => {
+				var exceptions = tarr.Where (t => t.IsFaulted).Select (t => t.Exception).ToArray ();
+				if (exceptions != null) {
+					var ex = new AggregateException (exceptions).Flatten ();
+					if (ex.InnerExceptions.Count > 0) {
+						foreach (var inner in ex.InnerExceptions) {
+							LoggingService.LogError ("Error while loading assembly in the browser.", inner);
+						}
+						throw ex;
+					}
+				}
+				if (definitions == null) // disposed
+					return;
+				Application.Invoke (delegate {
+					var nav = SearchMember (url);
+					if (nav == null) {
+						LoggingService.LogError ("Assembly browser: Can't find: " + url + ".");
+					}
+				});
+			}, CancellationToken.None, TaskContinuationOptions.None, TaskScheduler.Current);
 		}
 		
 		public void SelectAssembly (string fileName)
@@ -1150,6 +1299,7 @@ namespace MonoDevelop.AssemblyBrowser
 				if (nav.DataItem == cu) {
 					nav.ExpandToNode ();
 					nav.Selected = true;
+					nav.ScrollToNode ();
 					return;
 				}
 			} while (nav.MoveNext());
@@ -1173,6 +1323,7 @@ namespace MonoDevelop.AssemblyBrowser
 		
 		protected override void OnDestroyed ()
 		{
+			ClearReferenceSegment ();
 			if (searchBackgoundWorker != null && searchBackgoundWorker.IsBusy) {
 				searchBackgoundWorker.CancelAsync ();
 				searchBackgoundWorker.Dispose ();
@@ -1180,13 +1331,15 @@ namespace MonoDevelop.AssemblyBrowser
 			}
 			
 			if (this.TreeView != null) {
-			//	Dispose (TreeView.GetRootNode ());
+				//	Dispose (TreeView.GetRootNode ());
 				TreeView.Tree.CursorChanged -= HandleCursorChanged;
 				this.TreeView.Clear ();
 				this.TreeView = null;
 			}
 			
 			if (definitions != null) {
+				foreach (var def in definitions)
+					def.Dispose ();
 				definitions.Clear ();
 				definitions = null;
 			}
@@ -1207,6 +1360,7 @@ namespace MonoDevelop.AssemblyBrowser
 				documentationPanel = null;
 			}
 			if (inspectEditor != null) {
+				inspectEditor.TextViewMargin.GetLink = null;
 				inspectEditor.LinkRequest -= InspectEditorhandleLinkRequest;
 				inspectEditor.Destroy ();
 			}
@@ -1215,11 +1369,22 @@ namespace MonoDevelop.AssemblyBrowser
 				this.UIManager.Dispose ();
 				this.UIManager = null;
 			}
+
+			this.loader = null;
 			this.languageCombobox.Changed -= LanguageComboboxhandleChanged;
 //			this.searchInCombobox.Changed -= SearchInComboboxhandleChanged;
 //			this.searchEntry.Changed -= SearchEntryhandleChanged;
 			this.searchTreeview.RowActivated -= SearchTreeviewhandleRowActivated;
 			hpaned1.ExposeEvent -= HPaneExpose;
+			if (NavigateBackwardAction != null) {
+				this.NavigateBackwardAction.Dispose ();
+				this.NavigateBackwardAction = null;
+			}
+
+			if (NavigateForwardAction != null) {
+				this.NavigateForwardAction.Dispose ();
+				this.NavigateForwardAction = null;
+			}
 			base.OnDestroyed ();
 		}
 		
@@ -1242,15 +1407,14 @@ namespace MonoDevelop.AssemblyBrowser
 		List<AssemblyLoader> definitions = new List<AssemblyLoader> ();
 		List<Project> projects = new List<Project> ();
 		
-		public AssemblyLoader AddReferenceByAssemblyName (AssemblyNameReference reference, bool selectReference = false)
+		public AssemblyLoader AddReferenceByAssemblyName (AssemblyNameReference reference, bool selectReference = true)
 		{
 			return AddReferenceByAssemblyName (reference.Name, selectReference);
 		}
 		
-		public AssemblyLoader AddReferenceByAssemblyName (string assemblyFullName, bool selectReference = false)
+		public AssemblyLoader AddReferenceByAssemblyName (string assemblyFullName, bool selectReference = true)
 		{
 			string assemblyFile = Runtime.SystemAssemblyService.DefaultAssemblyContext.GetAssemblyLocation (assemblyFullName, null);
-			
 			if (assemblyFile == null || !System.IO.File.Exists (assemblyFile)) {
 				foreach (var wrapper in definitions) {
 					assemblyFile = wrapper.LookupAssembly (assemblyFullName);
@@ -1264,30 +1428,60 @@ namespace MonoDevelop.AssemblyBrowser
 			return AddReferenceByFileName (assemblyFile, selectReference);
 		}
 		
-		public AssemblyLoader AddReferenceByFileName (string fileName, bool selectReference = false)
+		public AssemblyLoader AddReferenceByFileName (string fileName, bool selectReference = true)
 		{
 			var result = definitions.FirstOrDefault (d => d.FileName == fileName);
-			if (result != null)
+			if (result != null) {
+				// Select the result.
+				if (selectReference) {
+					ITreeNavigator navigator = TreeView.GetNodeAtObject (result);
+					if (navigator != null) {
+						navigator.Selected = true;
+					} else {
+						LoggingService.LogWarning (result + " could not be found.");
+					}
+				}
+
 				return result;
+			}
 			result = new AssemblyLoader (this, fileName);
 			
 			definitions.Add (result);
-			ITreeBuilder builder;
-			if (definitions.Count + projects.Count == 1) {
-				builder = TreeView.LoadTree (result);
-			} else {
-				builder = TreeView.AddChild (result);
+			result.LoadingTask.ContinueWith (delegate {
+				Application.Invoke (delegate {
+					if (definitions == null)
+						return;
+					try {
+						ITreeBuilder builder;
+						if (definitions.Count + projects.Count == 1) {
+							builder = TreeView.LoadTree (result);
+						} else {
+							builder = TreeView.AddChild (result);
+						}
+						builder.Selected = builder.Expanded = selectReference;
+					} catch (Exception e) {
+						LoggingService.LogError ("Error while adding assembly to the assembly list", e);
+					}
+				});
 			}
-			builder.Selected = builder.Expanded = selectReference;
+			);
 			return result;
 		}
 		
-		public void AddProject (Project project, bool selectReference = false)
+		public void AddProject (Project project, bool selectReference = true)
 		{
 			if (project == null)
 				throw new ArgumentNullException ("project");
-			if (projects.Contains (project))
+
+			if (projects.Contains (project)) {
+				// Select the project.
+				if (selectReference) {
+					ITreeNavigator navigator = TreeView.GetNodeAtObject (project);
+					navigator.Selected = true;
+				}
+
 				return;
+			}
 			projects.Add (project);
 			ITreeBuilder builder;
 			if (definitions.Count + projects.Count == 1) {
@@ -1326,7 +1520,7 @@ namespace MonoDevelop.AssemblyBrowser
 						inspectEditor.GrabFocus ();
 				};
 				searchAndReplaceWidget.UpdateSearchPattern ();
-				textEditorContainer.AddAnimatedWidget (popupWidgetFrame, 300, Easing.ExponentialInOut, Blocking.Downstage, inspectEditor.Allocation.Width - 400, -searchAndReplaceWidget.Allocation.Height);
+				inspectEditor.AddAnimatedWidget (popupWidgetFrame, 300, Mono.TextEditor.Theatrics.Easing.ExponentialInOut, Blocking.Downstage, inspectEditor.Allocation.Width - 400, -searchAndReplaceWidget.Allocation.Height);
 				searchAndReplaceWidget.IsReplaceMode = false;
 			}
 			
@@ -1363,7 +1557,7 @@ namespace MonoDevelop.AssemblyBrowser
 					if (inspectEditor.IsRealized)
 						inspectEditor.GrabFocus ();
 				};
-				textEditorContainer.AddAnimatedWidget (popupWidgetFrame, 300, Easing.ExponentialInOut, Blocking.Downstage, inspectEditor.Allocation.Width - 400, -gotoLineNumberWidget.Allocation.Height);
+				inspectEditor.AddAnimatedWidget (popupWidgetFrame, 300, Mono.TextEditor.Theatrics.Easing.ExponentialInOut, Blocking.Downstage, inspectEditor.Allocation.Width - 400, -gotoLineNumberWidget.Allocation.Height);
 			}
 			
 			gotoLineNumberWidget.Focus ();
@@ -1391,9 +1585,10 @@ namespace MonoDevelop.AssemblyBrowser
 		
 		void UpdateNavigationActions ()
 		{
-			buttonBack.Sensitive = navigationBackwardHistory.Count != 0;
-			buttonForeward.Sensitive = navigationForwardHistory.Count != 0;
-			
+			if (buttonBack != null) {
+				buttonBack.Sensitive = navigationBackwardHistory.Count != 0;
+				buttonForeward.Sensitive = navigationForwardHistory.Count != 0;
+			}
 		}
 		
 		protected virtual void OnNavigateBackwardActionActivated (object sender, System.EventArgs e)
