@@ -43,12 +43,11 @@ using MonoDevelop.Ide.Gui.Components;
 using MonoDevelop.Ide;
 using System.ComponentModel;
 using System.Threading;
+using MonoDevelop.Components;
 
 namespace MonoDevelop.Gettext
 {
-	[System.ComponentModel.Category("widget")]
-	[System.ComponentModel.ToolboxItem(true)]
-	public partial class POEditorWidget : Gtk.Bin, IUndoHandler
+	partial class POEditorWidget : Gtk.Bin, IUndoHandler
 	{
 		TranslationProject project;
 		CatalogHeadersWidget headersEditor;
@@ -62,6 +61,12 @@ namespace MonoDevelop.Gettext
 		
 		static List<POEditorWidget> widgets = new List<POEditorWidget> (); 
 		
+		//simple escaping to make the messages display in single lines in the treeview
+		static string EscapeForTreeView (string message)
+		{
+			return StringEscaping.ToGettextFormat (message);
+		}
+		
 		public Catalog Catalog {
 			get {
 				return catalog;
@@ -90,6 +95,15 @@ namespace MonoDevelop.Gettext
 		{
 			this.project = project;
 			this.Build ();
+			
+			//FIXME: avoid unnecessary creation of old treeview
+			scrolledwindow1.Remove (treeviewEntries);
+			treeviewEntries.Destroy ();
+			treeviewEntries = new MonoDevelop.Components.ContextMenuTreeView ();
+			treeviewEntries.ShowAll ();
+			scrolledwindow1.Add (treeviewEntries);
+			((MonoDevelop.Components.ContextMenuTreeView)treeviewEntries).DoPopupMenu = ShowPopup;
+			
 			this.headersEditor = new CatalogHeadersWidget ();
 			this.notebookPages.AppendPage (headersEditor, new Gtk.Label ());
 			
@@ -109,7 +123,7 @@ namespace MonoDevelop.Gettext
 			fuzzyColumn.SortColumnId = 0;
 				
 			fuzzyColumn.Title = GettextCatalog.GetString ("Fuzzy");
-			var iconRenderer = new CellRendererPixbuf ();
+			var iconRenderer = new CellRendererImage ();
 			fuzzyColumn.PackStart (iconRenderer, false);
 			fuzzyColumn.SetCellDataFunc (iconRenderer, CatalogIconDataFunc);
 			
@@ -145,13 +159,13 @@ namespace MonoDevelop.Gettext
 			treeviewEntries.Selection.Changed += OnEntrySelected;
 			
 			// found in tree view
-			foundInStore = new ListStore (typeof(string), typeof(string), typeof(string), typeof(Pixbuf));
+			foundInStore = new ListStore (typeof(string), typeof(string), typeof(string), typeof(Xwt.Drawing.Image));
 			this.treeviewFoundIn.Model = foundInStore;
 			
 			TreeViewColumn fileColumn = new TreeViewColumn ();
-			var pixbufRenderer = new CellRendererPixbuf ();
+			var pixbufRenderer = new CellRendererImage ();
 			fileColumn.PackStart (pixbufRenderer, false);
-			fileColumn.SetAttributes (pixbufRenderer, "pixbuf", FoundInColumns.Pixbuf);
+			fileColumn.SetAttributes (pixbufRenderer, "image", FoundInColumns.Pixbuf);
 			
 			CellRendererText textRenderer = new CellRendererText ();
 			fileColumn.PackStart (textRenderer, true);
@@ -205,7 +219,7 @@ namespace MonoDevelop.Gettext
 				if (this.isUpdating)
 					return;
 				if (this.currentEntry != null) {
-					string[] lines = StringEscaping.FromGettextFormat (textviewComments.Buffer.Text).Split (new string[] { System.Environment.NewLine }, System.StringSplitOptions.None);
+					string[] lines =  textviewComments.Buffer.Text.Split (CatalogParser.LineSplitStrings, System.StringSplitOptions.None);
 					for (int i = 0; i < lines.Length; i++) {
 						if (!lines[i].StartsWith ("#"))
 							lines[i] = "# " + lines[i];
@@ -215,15 +229,6 @@ namespace MonoDevelop.Gettext
 				UpdateProgressBar ();
 			};
 			
-			this.treeviewEntries.PopupMenu += delegate {
-				ShowPopup ();
-			};
-			
-			this.treeviewEntries.ButtonReleaseEvent += delegate(object sender, Gtk.ButtonReleaseEventArgs e) {
-				if (e.Event.Button == 3)
-					ShowPopup ();
-			};
-			
 			searchEntryFilter.Ready = true;
 			searchEntryFilter.Visible = true;
 			searchEntryFilter.ForceFilterButtonVisible = true;
@@ -237,9 +242,7 @@ namespace MonoDevelop.Gettext
 			options.ShowLineNumberMargin = false;
 			options.ShowFoldMargin = false;
 			options.ShowIconMargin = false;
-			options.ShowInvalidLines = false;
-			options.ShowSpaces = options.ShowTabs = options.ShowEolMarkers = false;
-			options.ColorScheme = PropertyService.Get ("ColorScheme", "Default");
+			options.ColorScheme = IdeApp.Preferences.ColorScheme;
 			options.FontName = PropertyService.Get<string> ("FontName");
 			
 			this.scrolledwindowOriginal.Child = this.texteditorOriginal;
@@ -267,7 +270,7 @@ namespace MonoDevelop.Gettext
 		void CatalogIconDataFunc (Gtk.TreeViewColumn column, Gtk.CellRenderer cell, Gtk.TreeModel model, Gtk.TreeIter iter)
 		{
 			CatalogEntry entry = (CatalogEntry)model.GetValue (iter, 0);
-			((CellRendererPixbuf)cell).Pixbuf = ImageService.GetPixbuf (GetStockForEntry (entry), IconSize.Menu);
+			((CellRendererImage)cell).Image = ImageService.GetIcon (GetStockForEntry (entry), IconSize.Menu);
 			cell.CellBackgroundGdk = GetRowColorForEntry (entry);
 		}
 		
@@ -281,7 +284,7 @@ namespace MonoDevelop.Gettext
 		void OriginalTextDataFunc (Gtk.TreeViewColumn column, Gtk.CellRenderer cell, Gtk.TreeModel model, Gtk.TreeIter iter)
 		{
 			CatalogEntry entry = (CatalogEntry)model.GetValue (iter, 0);
-			((CellRendererText)cell).Text = StringEscaping.ToGettextFormat (entry.String);
+			((CellRendererText)cell).Text = EscapeForTreeView (entry.String);
 			cell.CellBackgroundGdk = GetRowColorForEntry (entry);
 			((CellRendererText)cell).ForegroundGdk = GetForeColorForEntry (entry);
 		}
@@ -289,14 +292,13 @@ namespace MonoDevelop.Gettext
 		void TranslationTextDataFunc (Gtk.TreeViewColumn column, Gtk.CellRenderer cell, Gtk.TreeModel model, Gtk.TreeIter iter)
 		{
 			CatalogEntry entry = (CatalogEntry)model.GetValue (iter, 0);
-			((CellRendererText)cell).Text = StringEscaping.ToGettextFormat (entry.GetTranslation (0));
+			((CellRendererText)cell).Text = EscapeForTreeView (entry.GetTranslation (0));
 			cell.CellBackgroundGdk = GetRowColorForEntry (entry);
 			((CellRendererText)cell).ForegroundGdk = GetForeColorForEntry (entry);
 		}
 		
 		void CheckbuttonWhiteSpacesToggled (object sender, EventArgs e)
 		{
-			options.ShowSpaces = options.ShowTabs = checkbuttonWhiteSpaces.Active;
 			texteditorOriginal.QueueDraw ();
 			texteditorPlural.QueueDraw ();
 			for (int i = this.notebookTranslated.NPages - 1; i >= 0; i--) {
@@ -491,7 +493,7 @@ namespace MonoDevelop.Gettext
 					return;
 				try {
 					if (this.currentEntry != null) {
-						string escapedText = StringEscaping.FromGettextFormat (textView.Document.Text);
+						string escapedText = textView.Document.Text;
 						string oldText     = this.currentEntry.GetTranslation (index);
 						this.currentEntry.SetTranslation (escapedText, index);
 						AddChange (this.currentEntry, oldText, escapedText, index);
@@ -513,11 +515,11 @@ namespace MonoDevelop.Gettext
 			this.notebookTranslated.AppendPage (window, label);
 		}
 		
-		void ShowPopup ()
+		void ShowPopup (EventButton evt)
 		{
 			Gtk.Menu contextMenu = CreateContextMenu ();
 			if (contextMenu != null)
-				contextMenu.Popup ();
+				GtkWorkarounds.ShowContextMenu (contextMenu, this, evt);
 		}
 		
 		Gtk.Menu CreateContextMenu ()
@@ -615,8 +617,8 @@ namespace MonoDevelop.Gettext
 			this.isUpdating = true;
 			try {
 				currentEntry = entry;
-				this.texteditorOriginal.Caret.Offset = 0;
-				this.texteditorOriginal.Document.Text = entry != null ? StringEscaping.ToGettextFormat (entry.String) : "";
+				this.texteditorOriginal.Caret.Location = new DocumentLocation (1, 1);
+				this.texteditorOriginal.Document.Text = entry != null ? entry.String : "";
 				this.texteditorOriginal.VAdjustment.Value = this.texteditorOriginal.HAdjustment.Value = 0;
 				
 //				if (GtkSpell.IsSupported && !gtkSpellSet.ContainsKey (this.textviewOriginal)) {
@@ -628,8 +630,8 @@ namespace MonoDevelop.Gettext
 				this.notebookTranslated.ShowTabs = entry != null && entry.HasPlural;
 				
 				if (entry != null && entry.HasPlural) {
-					this.texteditorPlural.Caret.Offset = 0;
-					this.texteditorPlural.Document.Text = StringEscaping.ToGettextFormat (entry.PluralString);
+					this.texteditorPlural.Caret.Location = new DocumentLocation (1, 1);
+					this.texteditorPlural.Document.Text = entry.PluralString;
 					this.texteditorPlural.VAdjustment.Value = this.texteditorPlural.HAdjustment.Value = 0;
 //					if (GtkSpell.IsSupported && !gtkSpellSet.ContainsKey (this.textviewOriginalPlural)) {
 //						GtkSpell.Attach (this.textviewOriginalPlural, "en");
@@ -651,7 +653,7 @@ namespace MonoDevelop.Gettext
 						if (textView == null)
 							continue;
 						textView.ClearSelection ();
-						textView.Document.Text = entry != null ? StringEscaping.ToGettextFormat (entry.GetTranslation (i)) : "";
+						textView.Document.Text = entry != null ?  entry.GetTranslation (i) : "";
 						textView.Caret.Offset = textView.Document.Text.Length;
 						textView.VAdjustment.Value = textView.HAdjustment.Value = 0;
 						textView.Document.CommitUpdateAll ();
@@ -669,11 +671,11 @@ namespace MonoDevelop.Gettext
 							line = "?";
 						}
 						string fullName = System.IO.Path.Combine (System.IO.Path.GetDirectoryName (this.poFileName), file);
-						this.foundInStore.AppendValues (file, line, fullName, DesktopService.GetPixbufForFile (fullName, IconSize.Menu));
+						this.foundInStore.AppendValues (file, line, fullName, DesktopService.GetIconForFile (fullName, IconSize.Menu));
 					}
 				}
 				
-				this.textviewComments.Buffer.Text = entry != null ? StringEscaping.ToGettextFormat (entry.Comment) : null;
+				this.textviewComments.Buffer.Text = entry != null ?  entry.Comment : null;
 				
 /*				if (GtkSpell.IsSupported) {
 					foreach (TextView view in this.gtkSpellSet.Keys)
@@ -701,9 +703,9 @@ namespace MonoDevelop.Gettext
 			return entry.IsFuzzy ? iconFuzzy : entry.IsTranslated ? iconValid : iconMissing;
 		}
 		
-		static string iconFuzzy   = "silk_error";// "md-translation-fuzzy";
-		static string iconValid   = "silk_accept";//"md-translation-valid";
-		static string iconMissing = "silk_exclamation";//"md-translation-missing";
+		static string iconFuzzy   = "error-light-16.png";// "md-translation-fuzzy";
+		static string iconValid   = "done-light-16.png";//"md-translation-valid";
+		static string iconMissing = "warning-light-16.png";//"md-translation-missing";
 		
 //		static Color translated   = new Color (255, 255, 255);
 		static Color untranslated = new Color (234, 232, 227);
@@ -901,8 +903,8 @@ namespace MonoDevelop.Gettext
 			}
 			store.AppendValues (GetStockForEntry (entry), 
 			                    entry.IsFuzzy,
-			                    StringEscaping.ToGettextFormat (entry.String), 
-			                    StringEscaping.ToGettextFormat (entry.GetTranslation (0)), 
+			                    EscapeForTreeView (entry.String), 
+			                    EscapeForTreeView (entry.GetTranslation (0)), 
 			                    entry,
 			                    GetRowColorForEntry (entry),
 			                    GetTypeSortIndicator (entry),
@@ -1261,12 +1263,16 @@ namespace MonoDevelop.Gettext
 			undoStack.Push (change);
 		}
 		
-		public void BeginAtomicUndo ()
+		class DisposeStub : IDisposable
 		{
+			public void Dispose ()
+			{
+			}
 		}
 		
-		public void EndAtomicUndo ()
+		public IDisposable OpenUndoGroup ()
 		{
+			return new DisposeStub ();
 		}
 		
 		public bool EnableUndo {
@@ -1282,4 +1288,4 @@ namespace MonoDevelop.Gettext
 		}
 		#endregion
 	}
-}
+}
\ No newline at end of file
