@@ -28,29 +28,28 @@
 
 using System;
 using System.IO;
-using System.Reflection;
 using System.Collections;
 using System.CodeDom;
 using System.CodeDom.Compiler;
 
 using MonoDevelop.Ide.Gui;
-using MonoDevelop.Ide.Gui.Content;
 using MonoDevelop.Projects;
-using MonoDevelop.Projects.Dom;
-using MonoDevelop.Projects.Dom.Parser;
-using MonoDevelop.Projects.CodeGeneration;
-using MonoDevelop.Core.Gui;
 using MonoDevelop.Projects.Text;
 using MonoDevelop.Core;
 using MonoDevelop.Core.Execution;
 using MonoDevelop.Deployment;
-using Mono.Cecil;
+using MonoDevelop.Projects.Policies;
+using MonoDevelop.Ide;
+using MonoDevelop.Ide.Gui.Content;
+using MonoDevelop.Ide.TypeSystem;
+
 
 namespace MonoDevelop.GtkCore.GuiBuilder
 {
 	class GuiBuilderService
 	{
-		static string GuiBuilderLayout = "GUI Builder";
+		static string GuiBuilderLayout = "Visual Design";
+		static int loadedGuiProjects;
 		
 #if DUMMY_STRINGS_FOR_TRANSLATION_DO_NOT_COMPILE
 		private void DoNotCompile ()
@@ -65,8 +64,6 @@ namespace MonoDevelop.GtkCore.GuiBuilder
 		static Stetic.Application steticApp;
 		
 		static bool generating;
-		static Stetic.CodeGenerationResult generationResult = null;
-		static Exception generatedException = null;
 		
 		static Stetic.IsolationMode IsolationMode = Stetic.IsolationMode.None;
 //		static Stetic.IsolationMode IsolationMode = Stetic.IsolationMode.ProcessUnix;
@@ -75,8 +72,6 @@ namespace MonoDevelop.GtkCore.GuiBuilder
 		{
 			if (IdeApp.Workbench == null)
 				return;
-			IdeApp.Workbench.ActiveDocumentChanged += new EventHandler (OnActiveDocumentChanged);
-			IdeApp.ProjectOperations.EndBuild += OnProjectCompiled;
 //			IdeApp.Workspace.ParserDatabase.AssemblyInformationChanged += (AssemblyInformationEventHandler) DispatchService.GuiDispatch (new AssemblyInformationEventHandler (OnAssemblyInfoChanged));
 			
 			IdeApp.Exited += delegate {
@@ -86,9 +81,37 @@ namespace MonoDevelop.GtkCore.GuiBuilder
 				}
 			};
 		}
-		
+
+		static void InitializeService ()
+		{
+			if (IdeApp.IsInitialized) {
+				IdeApp.ProjectOperations.EndBuild += OnProjectCompiled;
+				IdeApp.Workbench.ActiveDocumentChanged += OnActiveDocumentChanged;
+			}
+		}
+
+		static void ShutdownService ()
+		{
+			if (IdeApp.IsInitialized) {
+				IdeApp.ProjectOperations.EndBuild -= OnProjectCompiled;
+				IdeApp.Workbench.ActiveDocumentChanged -= OnActiveDocumentChanged;
+			}
+			if (steticApp != null) {
+				StoreConfiguration ();
+				steticApp.Dispose ();
+				steticApp = null;
+			}
+		}
+
+		public static GuiBuilderProject CreateBuilderProject (DotNetProject project, string fileName)
+		{
+			return new GuiBuilderProject (project, fileName);
+		}
+
 		public static Stetic.Application SteticApp {
 			get {
+				// Stetic is not thread safe, so all has to be done in the gui thread
+				DispatchService.AssertGuiThread ();
 				if (steticApp == null) {
 					steticApp = Stetic.ApplicationFactory.CreateApplication (Stetic.IsolationMode.None);
 					steticApp.AllowInProcLibraries = false;
@@ -100,20 +123,32 @@ namespace MonoDevelop.GtkCore.GuiBuilder
 				return steticApp;
 			}
 		}
+
+		internal static void NotifyGuiProjectLoaded ()
+		{
+			if (++loadedGuiProjects == 1)
+				InitializeService ();
+		}
+
+		internal static void NotifyGuiProjectUnloaded ()
+		{
+			if (--loadedGuiProjects == 0)
+				ShutdownService ();
+		}
 		
 		static string OnAssemblyResolve (string assemblyName)
 		{
-			return Runtime.SystemAssemblyService.GetAssemblyLocation (assemblyName);
+			return Runtime.SystemAssemblyService.DefaultAssemblyContext.GetAssemblyLocation (assemblyName, null);
 		}
 		
 		static string OnMimeResolve (string url)
 		{
-			return MonoDevelop.Core.Gui.Services.PlatformService.GetMimeTypeForUri (url);
+			return DesktopService.GetMimeTypeForUri (url);
 		}
 		
 		static void OnShowUrl (string url)
 		{
-			MonoDevelop.Core.Gui.Services.PlatformService.ShowUrl (url);
+			DesktopService.ShowUrl (url);
 		}
 		
 		internal static void StoreConfiguration ()
@@ -175,7 +210,7 @@ namespace MonoDevelop.GtkCore.GuiBuilder
 		static void SetDesignerLayout ()
 		{
 			if (AutoSwitchGuiLayout && IdeApp.Workbench.CurrentLayout != GuiBuilderLayout) {
-				bool exists = Array.IndexOf (IdeApp.Workbench.Layouts, GuiBuilderLayout) != -1;
+				bool exists = IdeApp.Workbench.Layouts.Contains (GuiBuilderLayout);
 				defaultLayout = IdeApp.Workbench.CurrentLayout;
 				IdeApp.Workbench.CurrentLayout = GuiBuilderLayout;
 				if (!exists) {
@@ -202,23 +237,15 @@ namespace MonoDevelop.GtkCore.GuiBuilder
 				// being used by the IDE. This will avoid unnecessary updates.
 				if (IdeApp.Workspace.IsOpen) {
 					foreach (Project prj in IdeApp.Workspace.GetAllProjects ()) {
-						GtkDesignInfo info = GtkDesignInfo.FromProject (prj);
-						if (!HasOpenDesigners (prj, false)) {
+						if (!HasOpenDesigners (prj, false) && GtkDesignInfo.HasDesignedObjects (prj)) {
+							GtkDesignInfo info = GtkDesignInfo.FromProject (prj);
 							info.ReloadGuiBuilderProject ();
 						}
 					}
 				}
-				
+
 				SteticApp.UpdateWidgetLibraries (false);
 			}
-			else {
-				// Some gtk# packages don't include the .pc file unless you install gtk-sharp-devel
-				if (Runtime.SystemAssemblyService.GetPackage ("gtk-sharp-2.0") == null) {
-					string msg = GettextCatalog.GetString ("ERROR: MonoDevelop could not find the Gtk# 2.0 development package. Compilation of projects depending on Gtk# libraries will fail. You may need to install development packages for gtk-sharp-2.0.");
-					args.ProgressMonitor.Log.WriteLine ();
-					args.ProgressMonitor.Log.WriteLine (msg);
-				}
-			}
 		}
 		
 		internal static bool HasOpenDesigners (Project project, bool modifiedOnly)
@@ -286,7 +313,7 @@ namespace MonoDevelop.GtkCore.GuiBuilder
 			int i = name.LastIndexOf ('.');
 			if (i != -1) {
 				ns = name.Substring (0, i);
-				name = name.Substring (i+1);
+				name = name.Substring (i + 1);
 			}
 			
 			if (saveToFile && !overwrite && File.Exists (fileName))
@@ -319,8 +346,7 @@ namespace MonoDevelop.GtkCore.GuiBuilder
 						)
 					);
 				}
-			}
-			else {
+			} else {
 				if (!saveToFile)
 					return fileName;
 				CodeNamespace cns = new CodeNamespace ();
@@ -332,175 +358,208 @@ namespace MonoDevelop.GtkCore.GuiBuilder
 			if (provider == null)
 				throw new UserException ("Code generation not supported for language: " + project.LanguageName);
 			
-			TextWriter fileStream;
-			if (saveToFile)
-				fileStream = new StreamWriter (fileName);
-			else
-				fileStream = new StringWriter ();
+			string text;
+			var pol = project.Policies.Get<TextStylePolicy> ();
+			using (var fileStream = new StringWriter ()) {
+				var options = new CodeGeneratorOptions () {
+					IndentString = pol.TabsToSpaces? new string (' ', pol.TabWidth) : "\t",
+					BlankLinesBetweenMembers = true,
+				};
+				provider.GenerateCodeFromCompileUnit (cu, fileStream, options);
+				text = fileStream.ToString ();
+				text = FormatGeneratedFile (fileName, text, project, provider);
+			}
 			
-			try {
-				provider.GenerateCodeFromCompileUnit (cu, fileStream, new CodeGeneratorOptions ());
-			} finally {
-				fileStream.Close ();
-			}
-
-			if (ProjectDomService.HasDom (project)) {
-				// Only update the parser database if the project is actually loaded in the IDE.
-				if (saveToFile) {
-					ProjectDomService.Parse (project, fileName, "");
-					FileService.NotifyFileChanged (fileName);
-				}
-				else
-					ProjectDomService.Parse (project, fileName, "", ((StringWriter)fileStream).ToString ());
-			}
+			if (saveToFile)
+				File.WriteAllText (fileName, text);
+			TypeSystemService.ParseFile (project, fileName);
+//			
+//			if (ProjectDomService.HasDom (project)) {
+//				// Only update the parser database if the project is actually loaded in the IDE.
+//				ProjectDomService.Parse (project, fileName, text);
+//				if (saveToFile) 
+//					FileService.NotifyFileChanged (fileName);
+//			}
 
 			return fileName;
 		}
 		
 		
-		public static Stetic.CodeGenerationResult GenerateSteticCode (IProgressMonitor monitor, DotNetProject project, string configuration)
+		public static Stetic.CodeGenerationResult GenerateSteticCode (IProgressMonitor monitor, DotNetProject project, ConfigurationSelector configuration)
 		{
 			if (generating || !GtkDesignInfo.HasDesignedObjects (project))
 				return null;
+			
+			using (var timer = Counters.SteticFileGeneratedTimer.BeginTiming ()) {
+				
+				timer.Trace ("Checking references");
+				GtkDesignInfo info = GtkDesignInfo.FromProject (project);
+	
+				DateTime last_gen_time = File.Exists (info.SteticGeneratedFile) ? File.GetLastWriteTime (info.SteticGeneratedFile) : DateTime.MinValue;
+				
+				bool ref_changed = false;
 
-			GtkDesignInfo info = GtkDesignInfo.FromProject (project);
+				// Disabled check for changes in referenced assemblies, since it cause too much
+				// regeneration of code. If a component has changed in a referenced project, this
+				// project may not build, but this can be solved by editing some file in the
+				// designer and saving.
 
-			DateTime last_gen_time = File.Exists (info.SteticGeneratedFile) ? File.GetLastWriteTime (info.SteticGeneratedFile) : DateTime.MinValue;
-			
-			bool ref_changed = false;
-			foreach (ProjectReference pref in project.References) {
-				if (!pref.IsValid)
-					continue;
-				foreach (string filename in pref.GetReferencedFileNames (configuration)) {
-					if (File.GetLastWriteTime (filename) > last_gen_time) {
-						ref_changed = true;
-						break;
-					}
-				}
-				if (ref_changed)
-					break;
-			}
-
-			// Check if generated code is already up to date.
-			if (!ref_changed && last_gen_time >= File.GetLastWriteTime (info.SteticFile))
-				return null;
-			
-			if (info.GuiBuilderProject.HasError) {
-				monitor.ReportError (GettextCatalog.GetString ("GUI code generation failed for project '{0}'. The file '{1}' could not be loaded.", project.Name, info.SteticFile), null);
-				monitor.AsyncOperation.Cancel ();
-				return null;
-			}
-			
-			if (info.GuiBuilderProject.IsEmpty) 
-				return null;
-
-			monitor.Log.WriteLine (GettextCatalog.GetString ("Generating GUI code for project '{0}'...", project.Name));
-			
-			// Make sure the referenced assemblies are up to date. It is necessary to do
-			// it now since they may contain widget libraries.
-			project.CopySupportFiles (monitor, configuration);
-			
-			info.GuiBuilderProject.UpdateLibraries ();
-			
-			ArrayList projects = new ArrayList ();
-			projects.Add (info.GuiBuilderProject.File);
-			
-			generating = true;
-			generationResult = null;
-			generatedException = null;
-			
-			bool canGenerateInProcess = IsolationMode != Stetic.IsolationMode.None || info.GuiBuilderProject.SteticProject.CanGenerateCode;
-			
-			if (!canGenerateInProcess) {
-				// Run the generation in another thread to avoid freezing the GUI
-				System.Threading.ThreadPool.QueueUserWorkItem ( delegate {
-					try {
-						// Generate the code in another process if stetic is not isolated
-						CodeGeneratorProcess cob = (CodeGeneratorProcess) Runtime.ProcessService.CreateExternalProcessObject (typeof (CodeGeneratorProcess), false);
-						using (cob) {
-							generationResult = cob.GenerateCode (projects, info.GenerateGettext, info.GettextClass, project.UsePartialTypes);
+/*				foreach (ProjectReference pref in project.References) {
+					if (!pref.IsValid)
+						continue;
+					foreach (string filename in pref.GetReferencedFileNames (configuration)) {
+						if (File.GetLastWriteTime (filename) > last_gen_time) {
+							ref_changed = true;
+							break;
 						}
+					}
+					if (ref_changed)
+						break;
+				}*/
+	
+				// Check if generated code is already up to date.
+				if (!ref_changed && last_gen_time >= File.GetLastWriteTime (info.SteticFile))
+					return null;
+				
+				if (info.GuiBuilderProject.HasError) {
+					monitor.ReportError (GettextCatalog.GetString ("GUI code generation failed for project '{0}'. The file '{1}' could not be loaded.", project.Name, info.SteticFile), null);
+					monitor.AsyncOperation.Cancel ();
+					return null;
+				}
+				
+				if (info.GuiBuilderProject.IsEmpty) 
+					return null;
+	
+				monitor.Log.WriteLine (GettextCatalog.GetString ("Generating GUI code for project '{0}'...", project.Name));
+				
+				timer.Trace ("Copy support files");
+				
+				// Make sure the referenced assemblies are up to date. It is necessary to do
+				// it now since they may contain widget libraries.
+				project.CopySupportFiles (monitor, configuration);
+				
+				timer.Trace ("Update libraries");
+				
+				info.GuiBuilderProject.UpdateLibraries ();
+				
+				ArrayList projects = new ArrayList ();
+				projects.Add (info.GuiBuilderProject.File);
+				
+				generating = true;
+				Stetic.CodeGenerationResult generationResult = null;
+				Exception generatedException = null;
+				
+				bool canGenerateInProcess = IsolationMode != Stetic.IsolationMode.None || info.GuiBuilderProject.SteticProject.CanGenerateCode;
+				
+				if (!canGenerateInProcess) {
+					timer.Trace ("Generating out of process");
+				
+					// Run the generation in another thread to avoid freezing the GUI
+					System.Threading.ThreadPool.QueueUserWorkItem (delegate {
+						try {
+							// Generate the code in another process if stetic is not isolated
+							CodeGeneratorProcess cob = (CodeGeneratorProcess)Runtime.ProcessService.CreateExternalProcessObject (typeof(CodeGeneratorProcess), false);
+							using (cob) {
+								generationResult = cob.GenerateCode (projects, info.GenerateGettext, info.GettextClass, info.ImageResourceLoaderClass, project.UsePartialTypes);
+							}
+						} catch (Exception ex) {
+							generatedException = ex;
+						} finally {
+							generating = false;
+						}
+					});
+				
+					while (generating) {
+						DispatchService.RunPendingEvents ();
+						System.Threading.Thread.Sleep (100);
+					}
+				} else {
+					timer.Trace ("Generating in-process");
+					// No need to create another process, since stetic has its own backend process
+					// or the widget libraries have no custom wrappers
+					try {
+						Stetic.GenerationOptions options = new Stetic.GenerationOptions ();
+						options.UseGettext = info.GenerateGettext;
+						options.GettextClass = info.GettextClass;
+						options.ImageResourceLoaderClass = info.ImageResourceLoaderClass;
+						options.UsePartialClasses = project.UsePartialTypes;
+						options.GenerateSingleFile = false;
+						options.GenerateModifiedOnly = true;
+						generationResult = SteticApp.GenerateProjectCode (options, info.GuiBuilderProject.SteticProject);
+						info.GuiBuilderProject.SteticProject.ResetModifiedWidgetFlags ();
 					} catch (Exception ex) {
 						generatedException = ex;
-					} finally {
-						generating = false;
 					}
-				});
-			
-				while (generating) {
-					DispatchService.RunPendingEvents ();
-					System.Threading.Thread.Sleep (100);
+					generating = false;
 				}
-			} else {
-				// No need to create another process, since stetic has its own backend process
-				// or the widget libraries have no custom wrappers
-				try {
-					Stetic.GenerationOptions options = new Stetic.GenerationOptions ();
-					options.UseGettext = info.GenerateGettext;
-					options.GettextClass = info.GettextClass;
-					options.UsePartialClasses = project.UsePartialTypes;
-					options.GenerateSingleFile = false;
-					generationResult = SteticApp.GenerateProjectCode (options, info.GuiBuilderProject.SteticProject);
-				} catch (Exception ex) {
-					generatedException = ex;
-				}
-				generating = false;
-			}
-			
-			if (generatedException != null) {
-				LoggingService.LogError ("GUI code generation failed", generatedException);
-				throw new UserException ("GUI code generation failed: " + generatedException.Message);
-			}
-			
-			if (generationResult == null)
-				return null;
+				timer.Trace ("Writing code units");
 				
-			CodeDomProvider provider = project.LanguageBinding.GetCodeDomProvider ();
-			if (provider == null)
-				throw new UserException ("Code generation not supported for language: " + project.LanguageName);
-			
-			string basePath = Path.GetDirectoryName (info.SteticGeneratedFile);
-			string ext = Path.GetExtension (info.SteticGeneratedFile);
-			
-			foreach (Stetic.SteticCompilationUnit unit in generationResult.Units) {
-				string fname;
-				if (unit.Name.Length == 0)
-					fname = info.SteticGeneratedFile;
-				else
-					fname = Path.Combine (basePath, unit.Name) + ext;
-				StringWriter sw = new StringWriter ();
-				try {
-					provider.GenerateCodeFromCompileUnit (unit, sw, new CodeGeneratorOptions ());
-					// Remove the runtime version number from the file. It may generate unnecessary
-					// version control changes when upgrading the mono version.
-					string content = sw.ToString ();
-					content = content.Replace ("Mono Runtime Version: " + Environment.Version, "");
-					File.WriteAllText (fname, content);
-				} finally {
-					FileService.NotifyFileChanged (fname);
+				if (generatedException != null) {
+					LoggingService.LogError ("GUI code generation failed", generatedException);
+					throw new UserException ("GUI code generation failed: " + generatedException.Message);
 				}
+				
+				if (generationResult == null)
+					return null;
+					
+				CodeDomProvider provider = project.LanguageBinding.GetCodeDomProvider ();
+				if (provider == null)
+					throw new UserException ("Code generation not supported for language: " + project.LanguageName);
+				
+				string basePath = Path.GetDirectoryName (info.SteticGeneratedFile);
+				string ext = Path.GetExtension (info.SteticGeneratedFile);
+				
+				var pol = project.Policies.Get<TextStylePolicy> ();
+				var codeGeneratorOptions = new CodeGeneratorOptions () {
+					IndentString = pol.TabsToSpaces? new string (' ', pol.TabWidth) : "\t",
+					BlankLinesBetweenMembers = true
+				};
+
+				foreach (Stetic.SteticCompilationUnit unit in generationResult.Units) {
+					string fname;
+					if (unit.Name.Length == 0)
+						fname = info.SteticGeneratedFile;
+					else
+						fname = Path.Combine (basePath, unit.Name) + ext;
+					StringWriter sw = new StringWriter ();
+					try {
+						foreach (CodeNamespace ns in unit.Namespaces)
+							ns.Comments.Add (new CodeCommentStatement ("This file has been generated by the GUI designer. Do not modify."));
+						timer.Trace ("Generating code for " + unit.Name);
+						provider.GenerateCodeFromCompileUnit (unit, sw, codeGeneratorOptions);
+						string content = sw.ToString ();
+								
+						timer.Trace ("Formatting code");
+						content = FormatGeneratedFile (fname, content, project, provider);
+						timer.Trace ("Writing code");
+						File.WriteAllText (fname, content);
+					} finally {
+						timer.Trace ("Notifying changes");
+						FileService.NotifyFileChanged (fname);
+					}
+				}
+				
+				timer.Trace ("Updating GTK folder");
+				
+				// Make sure the generated files are added to the project
+				if (info.UpdateGtkFolder ()) {
+					Gtk.Application.Invoke (delegate {
+						IdeApp.ProjectOperations.Save (project);
+					});
+				}
+				
+				return generationResult;
 			}
-			
-			// Make sure the generated files are added to the project
-			if (info.UpdateGtkFolder ()) {
-				Gtk.Application.Invoke (delegate {
-					IdeApp.ProjectOperations.Save (project);
-				});
-			}
-			
-			return generationResult;
 		}
 		
 		internal static string ImportFile (Project prj, string file)
 		{
 			ProjectFile pfile = prj.Files.GetFile (file);
 			if (pfile == null) {
-				string[] files = IdeApp.ProjectOperations.AddFilesToProject (prj, new string[] { file }, prj.BaseDirectory);
-				if (files.Length == 0)
+				var files = IdeApp.ProjectOperations.AddFilesToProject (prj, new string[] { file }, prj.BaseDirectory);
+				if (files.Count == 0 || files[0] == null)
 					return null;
-				if (files [0] == null)
-					return null;
-				pfile = prj.Files.GetFile (files[0]);
+				pfile = files [0];
 			}
 			if (pfile.BuildAction == BuildAction.EmbeddedResource) {
 				AlertButton embedButton = new AlertButton (GettextCatalog.GetString ("_Use as Source"));
@@ -513,12 +572,84 @@ namespace MonoDevelop.GtkCore.GuiBuilder
 			return pfile.FilePath;
 		}
 		
+		static string FormatGeneratedFile (string file, string content, Project project, CodeDomProvider provider)
+		{
+			content = StripHeaderAndBlankLines (content, provider);
+
+			string mt = DesktopService.GetMimeTypeForUri (file);
+			var formatter = MonoDevelop.Ide.CodeFormatting.CodeFormatterService.GetFormatter (mt);
+			if (formatter != null)
+				content = formatter.FormatText (PolicyService.InvariantPolicies, content);
+			
+			// The project policies should be taken for generated files (windows git eol problem)
+			var pol = project.Policies.Get<TextStylePolicy> (DesktopService.GetMimeTypeForUri (file));
+			string eol = pol.GetEolMarker ();
+			if (Environment.NewLine != eol)
+				content = content.Replace (Environment.NewLine, eol);
+			
+			return content;
+		}
+		
+		static string StripHeaderAndBlankLines (string text, CodeDomProvider provider)
+		{
+			Mono.TextEditor.TextDocument doc = new Mono.TextEditor.TextDocument ();
+			doc.Text = text;
+			int realStartLine = 0;
+			for (int i = 1; i <= doc.LineCount; i++) {
+				string lineText = doc.GetTextAt (doc.GetLine (i));
+				// Microsoft.NET generates "auto-generated" tags where Mono generates "autogenerated" tags.
+				if (lineText.Contains ("</autogenerated>") || lineText.Contains ("</auto-generated>")) {
+					realStartLine = i + 2;
+					break;
+				}
+			}
+			
+			// The Mono provider inserts additional blank lines, so strip them out
+			// But blank lines might actually be significant in other languages.
+			// We reformat the C# generated output to the user's coding style anyway, but the reformatter preserves blank lines
+			if (provider is Microsoft.CSharp.CSharpCodeProvider) {
+				bool previousWasBlank = false;
+				for (int i = 1; i <= doc.LineCount; i++) {
+					Mono.TextEditor.DocumentLine line = doc.GetLine (i);
+					bool isBlank, isBracket;
+					CheckLine (doc, line, out isBlank, out isBracket);
+					if (isBlank && previousWasBlank && line.LengthIncludingDelimiter > 0) {
+						doc.Remove (line.Offset, line.LengthIncludingDelimiter);
+						i--;
+					}
+					previousWasBlank = isBlank || isBracket;
+				}
+			}
+			
+			int offset = doc.GetLine (realStartLine).Offset;
+			return doc.GetTextAt (offset, doc.TextLength - offset);
+		}
+
+		static void CheckLine (Mono.TextEditor.TextDocument doc, Mono.TextEditor.DocumentLine line, out bool isBlank, out bool isBracket)
+		{
+			isBlank = true;
+			isBracket = false;
+			for (int i = 0; i < line.LengthIncludingDelimiter; i++) {
+				char c = doc.GetCharAt (line.Offset + i);
+				if (c == '{') {
+					isBracket = true;
+					isBlank = false;
+				}
+				else if (!Char.IsWhiteSpace (c)) {
+					isBlank = false;
+					if (isBracket) {
+						isBracket = false;
+						break;
+					}
+				}
+			}
+		}
 	}
 
 
 	public class CodeGeneratorProcess: RemoteProcessObject
 	{
-		public Stetic.CodeGenerationResult GenerateCode (ArrayList projectFiles, bool useGettext, string gettextClass, bool usePartialClasses)
+		public Stetic.CodeGenerationResult GenerateCode (ArrayList projectFiles, bool useGettext, string gettextClass, string resourceLoaderClass, bool usePartialClasses)
 		{
 			Gtk.Application.Init ();
 			
@@ -533,6 +664,7 @@ namespace MonoDevelop.GtkCore.GuiBuilder
 			Stetic.GenerationOptions options = new Stetic.GenerationOptions ();
 			options.UseGettext = useGettext;
 			options.GettextClass = gettextClass;
+			options.ImageResourceLoaderClass = resourceLoaderClass;
 			options.UsePartialClasses = usePartialClasses;
 			options.GenerateSingleFile = false;
 			
