@@ -82,8 +82,7 @@ namespace MonoDevelop.SourceEditor.OptionPanels
 		void HandleButtonNewClicked (object sender, EventArgs e)
 		{
 			var newShemeDialog = new NewColorShemeDialog ();
-			MessageService.RunCustomDialog (newShemeDialog, dialog);
-			newShemeDialog.Destroy ();
+			MessageService.ShowCustomDialog (newShemeDialog, dialog);
 			ShowStyles ();
 		}
 
@@ -99,7 +98,7 @@ namespace MonoDevelop.SourceEditor.OptionPanels
 			if (sheme == null)
 				return;
 			this.buttonExport.Sensitive = true;
-			string fileName = Mono.TextEditor.Highlighting.SyntaxModeService.GetFileNameForStyle (sheme);
+			string fileName = sheme.FileName;
 			if (fileName == null)
 				return;
 			this.removeButton.Sensitive = true;
@@ -111,23 +110,20 @@ namespace MonoDevelop.SourceEditor.OptionPanels
 			TreeIter selectedIter;
 			if (styleTreeview.Selection.GetSelected (out selectedIter)) {
 				var editor = new ColorShemeEditor (this);
-				editor.SetSheme ((Mono.TextEditor.Highlighting.ColorScheme)this.styleStore.GetValue (selectedIter, 1));
-				MessageService.RunCustomDialog (editor, dialog);
-				editor.Destroy ();
+				var colorScheme = (Mono.TextEditor.Highlighting.ColorScheme)this.styleStore.GetValue (selectedIter, 1);
+				editor.SetSheme (colorScheme);
+				MessageService.ShowCustomDialog (editor, dialog);
 			}
 		}
 		
 		Mono.TextEditor.Highlighting.ColorScheme LoadStyle (string styleName, bool showException = true)
 		{
 			try {
-				return Mono.TextEditor.Highlighting.SyntaxModeService.GetColorStyle (Style, styleName);
+				return Mono.TextEditor.Highlighting.SyntaxModeService.GetColorStyle (styleName);
 			} catch (Exception e) {
 				if (showException)
-					MessageService.ShowError ("Error while importing color style.", e.InnerException.Message);
-				var result = new Mono.TextEditor.Highlighting.DefaultStyle (Style);
-				result.Name = styleName;
-				result.Description = "Error";
-				return result;
+					MessageService.ShowError ("Error while importing color style " + styleName, (e.InnerException ?? e).Message);
+				return Mono.TextEditor.Highlighting.SyntaxModeService.DefaultColorStyle;
 			}
 		
 		}
@@ -143,7 +139,7 @@ namespace MonoDevelop.SourceEditor.OptionPanels
 				string name = style.Name ?? "";
 				string description = style.Description ?? "";
 				// translate only build-in sheme names
-				if (string.IsNullOrEmpty (Mono.TextEditor.Highlighting.SyntaxModeService.GetFileNameForStyle (style))) {
+				if (string.IsNullOrEmpty (style.FileName)) {
 					try {
 						name = GettextCatalog.GetString (name);
 						if (!string.IsNullOrEmpty (description))
@@ -163,11 +159,11 @@ namespace MonoDevelop.SourceEditor.OptionPanels
 			TreeIter selectedIter;
 			if (!styleTreeview.Selection.GetSelected (out selectedIter)) 
 				return;
-			var sheme = (Mono.TextEditor.Highlighting.ColorScheme)this.styleStore.GetValue (selectedIter, 1);
+			var sheme = (ColorScheme)this.styleStore.GetValue (selectedIter, 1);
 			
-			string fileName = Mono.TextEditor.Highlighting.SyntaxModeService.GetFileNameForStyle (sheme);
+			string fileName = sheme.FileName;
 			
-			if (fileName != null && fileName.StartsWith (SourceEditorDisplayBinding.SyntaxModePath)) {
+			if (fileName != null && fileName.StartsWith (SourceEditorDisplayBinding.SyntaxModePath, StringComparison.Ordinal)) {
 				Mono.TextEditor.Highlighting.SyntaxModeService.Remove (sheme);
 				File.Delete (fileName);
 				ShowStyles ();
@@ -179,13 +175,16 @@ namespace MonoDevelop.SourceEditor.OptionPanels
 			var dialog = new SelectFileDialog (GettextCatalog.GetString ("Highlighting Scheme"), Gtk.FileChooserAction.Save) {
 				TransientFor = this.Toplevel as Gtk.Window,
 			};
-			dialog.AddFilter (null, "*.xml");
+			dialog.AddFilter (GettextCatalog.GetString ("Color schemes"), "*.json");
 			if (!dialog.Run ())
 				return;
 			TreeIter selectedIter;
 			if (styleTreeview.Selection.GetSelected (out selectedIter)) {
 				var sheme = (Mono.TextEditor.Highlighting.ColorScheme)this.styleStore.GetValue (selectedIter, 1);
-				sheme.Save (dialog.SelectedFile);
+				var selectedFile = dialog.SelectedFile.ToString ();
+				if (!selectedFile.EndsWith (".json", StringComparison.Ordinal))
+					selectedFile += ".json";
+				sheme.Save (selectedFile);
 			}
 
 		}
@@ -195,45 +194,23 @@ namespace MonoDevelop.SourceEditor.OptionPanels
 			var dialog = new SelectFileDialog (GettextCatalog.GetString ("Highlighting Scheme"), Gtk.FileChooserAction.Open) {
 				TransientFor = this.Toplevel as Gtk.Window,
 			};
-			dialog.AddFilter (null, "*.xml");
+			dialog.AddFilter (GettextCatalog.GetString ("Color schemes"), "*.json");
+			dialog.AddFilter (GettextCatalog.GetString ("Visual Studio .NET settings"), "*.vssettings");
 			if (!dialog.Run ())
 				return;
-			
-			System.Collections.Generic.List<System.Xml.Schema.ValidationEventArgs> validationResult;
+
+			string newFileName = SourceEditorDisplayBinding.SyntaxModePath.Combine (dialog.SelectedFile.FileName);
+
+			bool success = true;
 			try {
-				validationResult = Mono.TextEditor.Highlighting.SyntaxModeService.ValidateStyleFile (dialog.SelectedFile);
-			} catch (Exception) {
-				MessageService.ShowError (GettextCatalog.GetString ("Validation of style file failed."));
-				return;
+				File.Copy (dialog.SelectedFile.FullPath, newFileName);
+			} catch (Exception e) {
+				success = false;
+				LoggingService.LogError ("Can't copy syntax mode file.", e);
 			}
-			
-			if (validationResult.Count == 0) {
-				string newFileName = SourceEditorDisplayBinding.SyntaxModePath.Combine (dialog.SelectedFile.FileName);
-				if (!newFileName.EndsWith ("Style.xml"))
-					newFileName = SourceEditorDisplayBinding.SyntaxModePath.Combine (dialog.SelectedFile.FileNameWithoutExtension + "Style.xml");
-				bool success = true;
-				try {
-					File.Copy (dialog.SelectedFile, newFileName);
-				} catch (Exception e) {
-					success = false;
-					LoggingService.LogError ("Can't copy syntax mode file.", e);
-				}
-				if (success) {
-					SourceEditorDisplayBinding.LoadCustomStylesAndModes ();
-					ShowStyles ();
-				}
-			} else {
-				StringBuilder errorMessage = new StringBuilder ();
-				errorMessage.AppendLine (GettextCatalog.GetString ("Validation of style file failed."));
-				int count = 0;
-				foreach (System.Xml.Schema.ValidationEventArgs vArg in validationResult) {
-					errorMessage.AppendLine (vArg.Message);
-					if (count++ > 5) {
-						errorMessage.AppendLine ("...");
-						break;
-					}
-				}
-				MessageService.ShowError (errorMessage.ToString ());
+			if (success) {
+				SourceEditorDisplayBinding.LoadCustomStylesAndModes ();
+				ShowStyles ();
 			}
 		}
 		
@@ -241,11 +218,26 @@ namespace MonoDevelop.SourceEditor.OptionPanels
 		{
 			this.enableSemanticHighlightingCheckbutton.Sensitive = this.enableHighlightingCheckbutton.Active;
 		}
+
+		internal static void UpdateActiveDocument ()
+		{
+			if (IdeApp.Workbench.ActiveDocument != null) {
+				IdeApp.Workbench.ActiveDocument.UpdateParseDocument ();
+				var editor = IdeApp.Workbench.ActiveDocument.Editor;
+				if (editor != null) {
+					editor.Parent.TextViewMargin.PurgeLayoutCache ();
+					editor.Parent.QueueDraw ();
+				}
+			}
+		}
 		
 		public virtual void ApplyChanges ()
 		{
 			DefaultSourceEditorOptions.Instance.EnableSyntaxHighlighting = this.enableHighlightingCheckbutton.Active;
-			DefaultSourceEditorOptions.Instance.EnableSemanticHighlighting = this.enableSemanticHighlightingCheckbutton.Active;
+			if (DefaultSourceEditorOptions.Instance.EnableSemanticHighlighting != this.enableSemanticHighlightingCheckbutton.Active) {
+				DefaultSourceEditorOptions.Instance.EnableSemanticHighlighting = this.enableSemanticHighlightingCheckbutton.Active;
+				UpdateActiveDocument ();
+			}
 			TreeIter selectedIter;
 			if (styleTreeview.Selection.GetSelected (out selectedIter)) {
 				ColorScheme sheme = ((Mono.TextEditor.Highlighting.ColorScheme)this.styleStore.GetValue (selectedIter, 1));
