@@ -29,6 +29,7 @@ using System.Collections.Generic;
 using System.IO;
 using System.Threading;
 using MonoDevelop.Core;
+using Gtk;
 
 namespace MonoDevelop.SourceEditor
 {
@@ -66,7 +67,15 @@ namespace MonoDevelop.SourceEditor
 			if (!autoSaveEnabled)
 				return false;
 			try {
-				return File.Exists (GetAutoSaveFileName (fileName));
+				var autoSaveFilename = GetAutoSaveFileName (fileName);
+				bool autoSaveExists = File.Exists (autoSaveFilename);
+				if (autoSaveExists) {
+					if (File.GetLastWriteTimeUtc (autoSaveFilename) < File.GetLastWriteTimeUtc (fileName)) {
+						File.Delete (autoSaveFilename);
+						return false;
+					}
+				}
+				return autoSaveExists;
 			} catch (Exception e) {
 				LoggingService.LogError ("Error in auto save - disableing.", e);
 				DisableAutoSave ();
@@ -136,16 +145,21 @@ namespace MonoDevelop.SourceEditor
 				resetEvent.WaitOne ();
 				while (queue.Count > 0) {
 					var content = queue.Dequeue ();
-					lock (contentLock) {
+					// Don't create an auto save for unsaved files.
+					if (string.IsNullOrEmpty (content.FileName))
+						continue;
+					Application.Invoke (delegate {
 						string text;
 						try {
 							text = content.Content.Text;
 						} catch (Exception e) {
 							LoggingService.LogError ("Exception in auto save thread.", e);
-							continue;
+							return;
 						}
 						CreateAutoSave (content.FileName, text);
 					}
+					);
+					
 				}
 			}
 		}
