@@ -33,10 +33,7 @@ using Mono.TextEditor;
 using MonoDevelop.Core;
 using MonoDevelop.Ide.Gui;
 using MonoDevelop.Ide.Gui.Content;
-using MonoDevelop.Projects.Dom;
-using MonoDevelop.Projects.Dom.Parser;
 using MonoDevelop.Ide.CodeCompletion;
-using MonoDevelop.Projects.CodeGeneration;
 using MonoDevelop.Components.Commands;
 using Mono.TextEditor.Highlighting;
 using MonoDevelop.Ide.CodeTemplates;
@@ -45,6 +42,10 @@ using MonoDevelop.Projects.Text;
 using MonoDevelop.Ide;
 using MonoDevelop.Ide.CodeFormatting;
 using MonoDevelop.SourceEditor.Extension;
+using ICSharpCode.NRefactory.TypeSystem;
+using MonoDevelop.Ide.TypeSystem;
+using ICSharpCode.NRefactory.Semantics;
+using MonoDevelop.Components;
 
 namespace MonoDevelop.SourceEditor
 {
@@ -54,8 +55,7 @@ namespace MonoDevelop.SourceEditor
 		
 		SourceEditorView view;
 		ExtensionContext extensionContext;
-		
-		Cairo.Point menuPopupLocation;
+		Adjustment cachedHAdjustment, cachedVAdjustment;
 		
 		public ITextEditorExtension Extension {
 			get;
@@ -65,11 +65,34 @@ namespace MonoDevelop.SourceEditor
 		public new ISourceEditorOptions Options {
 			get { return (ISourceEditorOptions)base.Options; }
 		}
+
+		static ExtensibleTextEditor ()
+		{
+			var icon = Xwt.Drawing.Image.FromResource ("gutter-bookmark-light-15.png");
+
+			BookmarkMarker.DrawBookmarkFunc = delegate(TextEditor editor, Cairo.Context cr, DocumentLine lineSegment, double x, double y, double width, double height) {
+				if (!lineSegment.IsBookmarked)
+					return;
+				cr.DrawImage (
+					editor, 
+					icon, 
+					Math.Floor (x + (width - icon.Width) / 2), 
+					Math.Floor (y + (height - icon.Height) / 2)
+				);
+			};
+
+		}
 		
-		public ExtensibleTextEditor (SourceEditorView view, ISourceEditorOptions options, Mono.TextEditor.Document doc) : base(doc, options)
+		public ExtensibleTextEditor (SourceEditorView view, ISourceEditorOptions options, Mono.TextEditor.TextDocument doc) : base(doc, options)
 		{
 			Initialize (view);
 		}
+
+		void HandleParseOperationFinished (object sender, EventArgs e)
+		{
+			resolveResult = null;
+			resolveRegion = DomRegion.Empty;
+		}
 		
 		public ExtensibleTextEditor (SourceEditorView view)
 		{
@@ -95,37 +118,24 @@ namespace MonoDevelop.SourceEditor
 			};
 			
 			Document.TextReplaced += HandleSkipCharsOnReplace;
-			
-			Document.TextReplaced += delegate(object sender, ReplaceEventArgs args) {
-				if (Extension != null) {
-					try {
-						Extension.TextChanged (args.Offset, args.Offset + Math.Max (args.Count, args.Value != null ? args.Value.Length : 0));
-					} catch (Exception ex) {
-						ReportExtensionError (ex);
-					}
-				}
-			};
-			
+			TypeSystemService.ParseOperationFinished += HandleParseOperationFinished;
+
 			UpdateEditMode ();
-			this.GetTextEditorData ().Paste += HandleTextPaste;
-			
-			this.ButtonPressEvent += OnPopupMenu;
+			this.DoPopupMenu = ShowPopup;
 		}
 		
-		void HandleSkipCharsOnReplace (object sender, ReplaceEventArgs args)
+		void HandleSkipCharsOnReplace (object sender, DocumentChangeEventArgs args)
 		{
+			var skipChars = GetTextEditorData ().SkipChars;
 			for (int i = 0; i < skipChars.Count; i++) {
-				SkipChar sc = skipChars[i];
-				if (args.Offset < sc.Start || args.Offset > sc.Offset) {
+				var sc = skipChars [i];
+				if (args.Offset > sc.Offset) {
 					skipChars.RemoveAt (i);
 					i--;
 					continue;
 				}
 				if (args.Offset <= sc.Offset) {
-					sc.Offset -= args.Count;
-					if (!string.IsNullOrEmpty (args.Value)) {
-						sc.Offset += args.Value.Length;
-					}
+					sc.Offset += args.ChangeDelta;
 				}
 			}
 		}
@@ -137,7 +147,7 @@ namespace MonoDevelop.SourceEditor
 			set {
 				if (extensionContext != null) {
 					extensionContext.RemoveExtensionNodeHandler ("MonoDevelop/SourceEditor2/TooltipProviders", OnTooltipProviderChanged);
-					TooltipProviders.Clear ();
+					ClearTooltipProviders ();
 				}
 				extensionContext = value;
 				if (extensionContext != null)
@@ -167,44 +177,51 @@ namespace MonoDevelop.SourceEditor
 			} else {
 		//		if (!(CurrentMode is SimpleEditMode)){
 					SimpleEditMode simpleMode = new SimpleEditMode ();
-					simpleMode.KeyBindings [EditMode.GetKeyCode (Gdk.Key.Tab)] = new TabAction (this).Action;
-					simpleMode.KeyBindings [EditMode.GetKeyCode (Gdk.Key.BackSpace)] = EditActions.AdvancedBackspace;
+					simpleMode.KeyBindings [EditMode.GetKeyCode (Gdk.Key.Tab)] = new TabAction (this).Action;
+					simpleMode.KeyBindings [EditMode.GetKeyCode (Gdk.Key.BackSpace)] = EditActions.AdvancedBackspace;
 					CurrentMode = simpleMode;
 		//		}
 			}
 		}
 
+		void UnregisterAdjustments ()
+		{
+			if (cachedHAdjustment != null)
+				cachedHAdjustment.ValueChanged -= HAdjustment_ValueChanged;
+			if (cachedVAdjustment != null)
+				cachedVAdjustment.ValueChanged -= VAdjustment_ValueChanged;
+			cachedHAdjustment = null;
+			cachedVAdjustment = null;
+		}
+
 		protected override void OnDestroyed ()
 		{
+			TypeSystemService.ParseOperationFinished -= HandleParseOperationFinished;
+			UnregisterAdjustments ();
+			resolveResult = null;
+			Extension = null;
 			ExtensionContext = null;
 			view = null;
-			this.ButtonPressEvent -= OnPopupMenu;
-
 			base.OnDestroyed ();
+			if (Options != null) {
+				Options.Dispose ();
+				base.Options = null;
+			}
 		}
 		
 		void OnTooltipProviderChanged (object s, ExtensionNodeEventArgs a)
 		{
-			if (a.Change == ExtensionChange.Add) {
-				TooltipProviders.Add ((ITooltipProvider) a.ExtensionObject);
-			} else {
-				TooltipProviders.Remove ((ITooltipProvider) a.ExtensionObject);
+			TooltipProvider provider;
+			try {
+				provider = (TooltipProvider) a.ExtensionObject;
+			} catch (Exception e) {
+				LoggingService.LogError ("Can't create tooltip provider:"+ a.ExtensionNode, e);
+				return;
 			}
-		}
-
-		void OnPopupMenu (object sender, Gtk.ButtonPressEventArgs args)
-		{
-			if (args.Event.Button == 3) {
-				int textEditorXOffset = (int)args.Event.X - (int)this.TextViewMargin.XOffset;
-				if (textEditorXOffset < 0)
-					return;
-				this.menuPopupLocation = new Cairo.Point ((int)args.Event.X, (int)args.Event.Y);
-				DocumentLocation loc= PointToLocation (textEditorXOffset, (int)args.Event.Y);
-				if (!this.IsSomethingSelected || !this.SelectionRange.Contains (Document.LocationToOffset (loc)))
-					Caret.Location = loc;
-				
-				this.ShowPopup ();
-				base.ResetMouseState ();
+			if (a.Change == ExtensionChange.Add) {
+				AddTooltipProvider (provider);
+			} else {
+				RemoveTooltipProvider (provider);
 			}
 		}
 		
@@ -215,26 +232,19 @@ namespace MonoDevelop.SourceEditor
 		
 		protected override void OptionsChanged (object sender, EventArgs args)
 		{
-			if (view.Control != null) {
+			if (view != null && view.Control != null) {
 				if (!Options.ShowFoldMargin)
 					this.Document.ClearFoldSegments ();
 			}
 			UpdateEditMode ();
 			base.OptionsChanged (sender, args);
 		}
+		
 		bool isInKeyStroke = false;
 		protected override bool OnKeyPressEvent (Gdk.EventKey evnt)
 		{
 			isInKeyStroke = true;
 			try {
-				// Handle keyboard menu popup
-				if (evnt.Key == Gdk.Key.Menu || (evnt.Key == Gdk.Key.F10 && (evnt.State & Gdk.ModifierType.ShiftMask) == Gdk.ModifierType.ShiftMask)) {
-					this.menuPopupLocation = LocationToPoint (this.Caret.Location);
-					this.menuPopupLocation.Y += (int)this.TextViewMargin.LineHeight;
-					this.ShowPopup ();
-					return true;
-				}
-				
 				// Handle keyboard toolip popup
 	/*			if ((evnt.Key == Gdk.Key.F1 && (evnt.State & Gdk.ModifierType.ControlMask) == Gdk.ModifierType.ControlMask)) {
 					Gdk.Point p = this.TextViewMargin.LocationToDisplayCoordinates (this.Caret.Location);
@@ -252,10 +262,13 @@ namespace MonoDevelop.SourceEditor
 		
 		bool ExtensionKeyPress (Gdk.Key key, uint ch, Gdk.ModifierType state)
 		{
+			isInKeyStroke = true;
 			try {
 				return Extension.KeyPress (key, (char)ch, state);
 			} catch (Exception ex) {
 				ReportExtensionError (ex);
+			} finally {
+				isInKeyStroke = false;
 			}
 			return false;
 		}
@@ -268,11 +281,11 @@ namespace MonoDevelop.SourceEditor
 		
 		IEnumerable<char> TextWithoutCommentsAndStrings {
 			get {
-				return from p in GetTextWithoutCommentsAndStrings (Document, 0, Document.Length) select p.Key;
+				return from p in GetTextWithoutCommentsAndStrings (Document, 0, Document.TextLength) select p.Key;
 			}
 		}
 		
-		static IEnumerable<KeyValuePair <char, int>> GetTextWithoutCommentsAndStrings (Mono.TextEditor.Document doc, int start, int end) 
+		static IEnumerable<KeyValuePair <char, int>> GetTextWithoutCommentsAndStrings (Mono.TextEditor.TextDocument doc, int start, int end) 
 		{
 			bool isInString = false, isInChar = false;
 			bool isInLineComment = false, isInBlockComment = false;
@@ -288,7 +301,7 @@ namespace MonoDevelop.SourceEditor
 						if (isInBlockComment) {
 							if (pos > 0 && doc.GetCharAt (pos - 1) == '*') 
 								isInBlockComment = false;
-						} else  if (!isInString && !isInChar && pos + 1 < doc.Length) {
+						} else  if (!isInString && !isInChar && pos + 1 < doc.TextLength) {
 							char nextChar = doc.GetCharAt (pos + 1);
 							if (nextChar == '/')
 								isInLineComment = true;
@@ -312,26 +325,6 @@ namespace MonoDevelop.SourceEditor
 			}
 		}
 		
-		class SkipChar {
-			public int Start { get; set; }
-			public int Offset { get; set; }
-			public char Char  { get; set; }
-			
-			public override string ToString ()
-			{
-				return string.Format ("[SkipChar: Start={0}, Offset={1}, Char={2}]", Start, Offset, Char);
-			}
-		}
-		
-		List<SkipChar> skipChars = new List<SkipChar> ();
-		void SetInsertionChar (int offset, char ch)
-		{
-			skipChars.Add (new SkipChar () {
-				Start = offset - 1,
-				Offset = offset,
-				Char = ch
-			});
-		}
 		
 		protected override bool OnIMProcessedKeyPressEvent (Gdk.Key key, uint ch, Gdk.ModifierType state)
 		{
@@ -342,67 +335,70 @@ namespace MonoDevelop.SourceEditor
 					view.SourceEditorWidget.RemoveSearchWidget ();
 					return true;
 				}
-				return false;
+				return false; 
 			}
 
 			if (Document == null)
 				return true;
 
 			bool inStringOrComment = false;
-			LineSegment line = Document.GetLine (Caret.Line);
+			bool isString = false;
+			DocumentLine line = Document.GetLine (Caret.Line);
 			if (line == null)
 				return true;
-			bool inChar = false;
-			bool inComment = false;
-			bool inString = false;
-//			string escape = "\"";
-			var stack = line.StartSpan.Clone();
-			Mono.TextEditor.Highlighting.SyntaxModeService.ScanSpans (Document, Document.SyntaxMode, Document.SyntaxMode, stack, line.Offset, Caret.Offset);
+			//			string escape = "\"";
+			var stack = line.StartSpan.Clone ();
+			var sm = Document.SyntaxMode as SyntaxMode;
+			if (sm != null)
+				Mono.TextEditor.Highlighting.SyntaxModeService.ScanSpans (Document, sm, sm, stack, line.Offset, Caret.Offset);
+
 			foreach (Span span in stack) {
 				if (string.IsNullOrEmpty (span.Color))
 					continue;
-				if (span.Color == "string.other") {
-					inStringOrComment = inChar = inString = true;
-					break;
-				}
-				if (span.Color == "string.single" || span.Color == "string.double" || span.Color.StartsWith ("comment")) {
+				if (span.Color.StartsWith ("String", StringComparison.Ordinal) ||
+				    span.Color.StartsWith ("Comment", StringComparison.Ordinal) ||
+				    span.Color.StartsWith ("Xml Attribute Value", StringComparison.Ordinal)) {
+					//Treat "Xml Attribute Value" as "String" so quotes in SkipChars works in Xml
+					if (span.Color.StartsWith ("Comment", StringComparison.Ordinal)) {
+						isString = false;
+					} else {
+						isString = true;
+					}
 					inStringOrComment = true;
-					inChar |= span.Color == "string.single";
-					inComment |= span.Color.StartsWith ("comment");
-					inString = !inChar && !inComment;
-					//escape = span.Escape;
 					break;
 				}
 			}
-			if (Caret.Offset > 0) {
-				char c = GetCharAt (Caret.Offset - 1);
-				if (c == '"' || c == '\'')
-					inStringOrComment = inChar = inString = true;
-			}
-			Document.BeginAtomicUndo ();
-
 			// insert template when space is typed (currently disabled - it's annoying).
 			bool templateInserted = false;
 			//!inStringOrComment && (key == Gdk.Key.space) && DoInsertTemplate ();
-			bool returnBetweenBraces = key == Gdk.Key.Return && (state & (Gdk.ModifierType.ControlMask | Gdk.ModifierType.ShiftMask)) == Gdk.ModifierType.None && Caret.Offset > 0 && Caret.Offset < Document.Length && Document.GetCharAt (Caret.Offset - 1) == '{' && Document.GetCharAt (Caret.Offset) == '}' && !inStringOrComment;
+			bool returnBetweenBraces = key == Gdk.Key.Return && (state & (Gdk.ModifierType.ControlMask | Gdk.ModifierType.ShiftMask)) == Gdk.ModifierType.None && Caret.Offset > 0 && Caret.Offset < Document.TextLength && Document.GetCharAt (Caret.Offset - 1) == '{' && Document.GetCharAt (Caret.Offset) == '}' && !inStringOrComment;
 //			int initialOffset = Caret.Offset;
 			const string openBrackets = "{[('\"";
 			const string closingBrackets = "}])'\"";
 			int braceIndex = openBrackets.IndexOf ((char)ch);
-			SkipChar skipChar = skipChars.Find (sc => sc.Char == (char)ch && sc.Offset == Caret.Offset);
-			
+			var skipChars = GetTextEditorData ().SkipChars;
+			var skipChar = skipChars.Find (sc => sc.Char == (char)ch && sc.Offset == Caret.Offset);
+//			bool startedAtomicOperation = false;
 
 			// special handling for escape chars inside ' and "
 			if (Caret.Offset > 0) {
 				char charBefore = Document.GetCharAt (Caret.Offset - 1);
-				if (inStringOrComment && (ch == '"' || (inChar && ch == '\'')) && charBefore == '\\')
-					skipChar = null;
+				if (ch == '"') {
+					if (!inStringOrComment && charBefore == '"' || 
+						isString && charBefore == '\\' ) {
+						skipChar = null;
+						braceIndex = -1;
+					}
+				}
+
 			}
 			char insertionChar = '\0';
-			if (skipChar == null && Options.AutoInsertMatchingBracket && braceIndex >= 0) {
+			bool insertMatchingBracket = false;
+			IDisposable undoGroup = null;
+			if (skipChar == null && Options.AutoInsertMatchingBracket && braceIndex >= 0 && !IsSomethingSelected) {
 				if (!inStringOrComment) {
-					char closingBrace = closingBrackets[braceIndex];
-					char openingBrace = openBrackets[braceIndex];
+					char closingBrace = closingBrackets [braceIndex];
+					char openingBrace = openBrackets [braceIndex];
 
 					int count = 0;
 					foreach (char curCh in TextWithoutCommentsAndStrings) {
@@ -414,51 +410,58 @@ namespace MonoDevelop.SourceEditor
 					}
 
 					if (count >= 0) {
-						GetTextEditorData ().EnsureCaretIsNotVirtual ();
-						
-						int offset = Caret.Offset;
+						insertMatchingBracket = true;
 						insertionChar = closingBrace;
-						Insert (offset, closingBrace.ToString ());
-						Caret.Offset = offset;
 					}
 				} else {
 					char charBefore = Document.GetCharAt (Caret.Offset - 1);
-					if (!inString && !inComment && !inChar && ch == '"' && charBefore != '\\') {
-						GetTextEditorData ().EnsureCaretIsNotVirtual ();
+					if (!inStringOrComment && ch == '"' && charBefore != '\\') {
+						insertMatchingBracket = true;
 						insertionChar = '"';
-						int offset = Caret.Offset;
-						Insert (Caret.Offset, "\"");
-						Caret.Offset = offset;
 					}
 				}
 			}
 			
 			//Console.WriteLine (Caret.Offset + "/" + insOff);
+			if (insertMatchingBracket)
+				undoGroup = Document.OpenUndoGroup ();
+
+			var oldMode = Caret.IsInInsertMode;
 			if (skipChar != null) {
-				Caret.Offset++;
+				Caret.IsInInsertMode = false;
 				skipChars.Remove (skipChar);
-			} else {
-				if (Extension != null) {
-					if (ExtensionKeyPress (key, ch, state)) 
-						result = base.OnIMProcessedKeyPressEvent (key, ch, state);
-					if (returnBetweenBraces)
-						 HitReturn ();
+			}
+			if (Extension != null) {
+				if (!DefaultSourceEditorOptions.Instance.GenerateFormattingUndoStep) {
+					using (var undo = Document.OpenUndoGroup ()) {
+						if (ExtensionKeyPress (key, ch, state))
+							result = base.OnIMProcessedKeyPressEvent (key, ch, state);
+					}
 				} else {
-					result = base.OnIMProcessedKeyPressEvent (key, ch, state);
-					if (returnBetweenBraces)
-						 HitReturn ();
+					if (ExtensionKeyPress (key, ch, state))
+						result = base.OnIMProcessedKeyPressEvent (key, ch, state);
 				}
+				if (returnBetweenBraces)
+					HitReturn ();
+			} else {
+				result = base.OnIMProcessedKeyPressEvent (key, ch, state);
+				if (returnBetweenBraces)
+					HitReturn ();
 			}
-			if (insertionChar != '\0')
-				SetInsertionChar (Caret.Offset, insertionChar);
-			
-			if (templateInserted) {
-				Document.EndAtomicUndo ();
-				return true;
+			if (skipChar != null) {
+				Caret.IsInInsertMode = oldMode;
 			}
-				
-			Document.EndAtomicUndo ();
-			return result;
+
+			if (insertMatchingBracket) {
+				GetTextEditorData ().EnsureCaretIsNotVirtual ();
+				int offset = Caret.Offset;
+				Caret.AutoUpdatePosition = false;
+				Insert (offset, insertionChar.ToString ());
+				Caret.AutoUpdatePosition = true;
+				GetTextEditorData ().SetSkipChar (offset, insertionChar);
+				undoGroup.Dispose ();
+			}
+			return templateInserted || result;
 		}
 		
 		void HitReturn ()
@@ -473,7 +476,7 @@ namespace MonoDevelop.SourceEditor
 		internal string GetErrorInformationAt (int offset)
 		{
 			DocumentLocation location = Document.OffsetToLocation (offset);
-			LineSegment line = Document.GetLine (location.Line);
+			DocumentLine line = Document.GetLine (location.Line);
 			if (line == null)
 				return null;
 			var error = line.Markers.FirstOrDefault (m => m is ErrorMarker) as ErrorMarker;
@@ -488,29 +491,44 @@ namespace MonoDevelop.SourceEditor
 			return null;
 		}
 		
-		public ProjectDom ProjectDom {
+		internal ParsedDocument ParsedDocument {
 			get {
-				MonoDevelop.Ide.Gui.Document doc = IdeApp.Workbench.ActiveDocument;
+				var doc = IdeApp.Workbench.ActiveDocument;
 				if (doc != null) 
-					return doc.Dom;
+					return doc.ParsedDocument;
+				return null;
+			}
+		}
+		
+		public MonoDevelop.Projects.Project Project {
+			get {
+				var doc = IdeApp.Workbench.ActiveDocument;
+				if (doc != null) 
+					return doc.Project;
 				return null;
 			}
 		}
 		
 		int           oldOffset = -1;
 		ResolveResult resolveResult = null;
-		public ResolveResult GetLanguageItem (int offset)
+		DomRegion     resolveRegion = DomRegion.Empty;
+		public ResolveResult GetLanguageItem (int offset, out DomRegion region)
 		{
-			
 			// we'll cache old results.
-			if (offset == oldOffset)
+			if (offset == oldOffset) {
+				region = this.resolveRegion;
 				return this.resolveResult;
+			}
+			
 			oldOffset = offset;
 			
 			if (textEditorResolverProvider != null) {
-				this.resolveResult = textEditorResolverProvider.GetLanguageItem (this.ProjectDom, GetTextEditorData (), offset);
+				this.resolveResult = textEditorResolverProvider.GetLanguageItem (view.WorkbenchWindow.Document, offset, out region);
+				this.resolveRegion = region;
 			} else {
+				region = DomRegion.Empty;
 				this.resolveResult = null;
+				this.resolveRegion = region;
 			}
 			
 			return this.resolveResult;
@@ -519,15 +537,8 @@ namespace MonoDevelop.SourceEditor
 		public CodeTemplateContext GetTemplateContext ()
 		{
 			if (IsSomethingSelected) {
-				string fileName = view.ContentName ?? view.UntitledName;
-				IParser parser = ProjectDomService.GetParser (fileName);
-				if (parser == null)
-					return CodeTemplateContext.Standard;
-
-				IExpressionFinder expressionFinder = parser.CreateExpressionFinder (ProjectDom);
-				if (expressionFinder == null) 
-					return CodeTemplateContext.Standard;
-				if (expressionFinder.IsExpression (Document.GetTextAt (SelectionRange)))
+				var result = GetLanguageItem (Caret.Offset, Document.GetTextAt (SelectionRange));
+				if (result != null && !result.IsError)
 					return CodeTemplateContext.InExpression;
 			}
 			return CodeTemplateContext.Standard;
@@ -545,7 +556,7 @@ namespace MonoDevelop.SourceEditor
 			oldOffset = offset;
 			
 			if (textEditorResolverProvider != null) {
-				this.resolveResult = textEditorResolverProvider.GetLanguageItem (this.ProjectDom, GetTextEditorData (), offset, expression);
+				this.resolveResult = textEditorResolverProvider.GetLanguageItem (view.WorkbenchWindow.Document, offset, expression);
 			} else {
 				this.resolveResult = null;
 			}
@@ -553,30 +564,22 @@ namespace MonoDevelop.SourceEditor
 			return this.resolveResult;
 		}
 
-		public string GetExpression (int offset)
-		{
-			string fileName = View.ContentName;
-			if (fileName == null)
-				fileName = View.UntitledName;
-			
-			IExpressionFinder expressionFinder = ProjectDomService.GetExpressionFinder (fileName);
-			string expression = expressionFinder == null ? GetExpressionBeforeOffset (offset) : expressionFinder.FindFullExpression (GetTextEditorData () , offset).Expression;
-			
-			if (expression == null)
-				return string.Empty;
-			else
-				return expression.Trim ();
-		}
+//		public string GetExpression (int offset)
+//		{
+//			if (textEditorResolverProvider != null) 
+//				return textEditorResolverProvider.GetExpression (view.WorkbenchWindow.Document, offset);
+//			return string.Empty;
+//		}
 		
 		string GetExpressionBeforeOffset (int offset)
 		{
 			int start = offset;
 			while (start > 0 && IsIdChar (Document.GetCharAt (start)))
 				start--;
-			while (offset < Document.Length && IsIdChar (Document.GetCharAt (offset)))
+			while (offset < Document.TextLength && IsIdChar (Document.GetCharAt (offset)))
 				offset++;
 			start++;
-			if (offset - start > 0 && start < Document.Length)
+			if (offset - start > 0 && start < Document.TextLength)
 				return Document.GetTextAt (start, offset - start);
 			else
 				return string.Empty;
@@ -590,35 +593,49 @@ namespace MonoDevelop.SourceEditor
 		protected override bool OnFocusOutEvent (Gdk.EventFocus evnt)
 		{
 			CompletionWindowManager.HideWindow ();
-			ParameterInformationWindowManager.HideWindow (view);
+			ParameterInformationWindowManager.HideWindow (null, view);
 			return base.OnFocusOutEvent (evnt); 
 		}
 		
-		void ShowPopup ()
+		void ShowPopup (Gdk.EventButton evt)
 		{
+			view.FireCompletionContextChanged ();
+			CompletionWindowManager.HideWindow ();
+			ParameterInformationWindowManager.HideWindow (null, view);
 			HideTooltip ();
-			CommandEntrySet cset = IdeApp.CommandService.CreateCommandEntrySet (ExtensionContext ?? AddinManager.AddinEngine, "/MonoDevelop/SourceEditor2/ContextMenu/Editor");
-			Gtk.Menu menu = IdeApp.CommandService.CreateMenu (cset);
-			menu.Append (new SeparatorMenuItem ());
-			menu.Append (CreateInputMethodMenuItem (GettextCatalog.GetString ("_Input Methods")));
-			menu.Destroyed += delegate {
-				this.QueueDraw ();
-			};
-			
-			menu.Popup (null, null, new Gtk.MenuPositionFunc (PositionPopupMenu), 0, Gtk.Global.CurrentEventTime);
+			const string menuPath = "/MonoDevelop/SourceEditor2/ContextMenu/Editor";
+			var ctx = ExtensionContext ?? AddinManager.AddinEngine;
+
+			CommandEntrySet cset = IdeApp.CommandService.CreateCommandEntrySet (ctx, menuPath);
+
+			if (Platform.IsMac) {
+				IdeApp.CommandService.ShowContextMenu (this, evt, cset, this);
+			} else {
+				Gtk.Menu menu = IdeApp.CommandService.CreateMenu (cset);
+				var imMenu = CreateInputMethodMenuItem (GettextCatalog.GetString ("_Input Methods"));
+				if (imMenu != null) {
+					menu.Append (new SeparatorMenuItem ());
+					menu.Append (imMenu);
+				}
+
+				menu.Hidden += HandleMenuHidden;
+				if (evt != null) {
+					GtkWorkarounds.ShowContextMenu (menu, this, evt);
+				} else {
+					var pt = LocationToPoint (this.Caret.Location);
+					GtkWorkarounds.ShowContextMenu (menu, this, new Gdk.Rectangle (pt.X, pt.Y, 1, (int)LineHeight));
+				}
+			}
 		}
-		
-		void PositionPopupMenu (Menu menu, out int x, out int y, out bool pushIn)
-		{
-			this.GdkWindow.GetOrigin (out x, out y);
-			x += this.menuPopupLocation.X;
-			y += this.menuPopupLocation.Y;
-			Requisition request = menu.SizeRequest ();
-			Gdk.Rectangle geometry = Screen.GetMonitorGeometry (Screen.GetMonitorAtPoint (x, y));
-			
-			y = Math.Max (geometry.Top, Math.Min (y, geometry.Bottom - request.Height));
-			x = Math.Max (geometry.Left, Math.Min (x, geometry.Right - request.Width));
-			pushIn = true;
+
+		void HandleMenuHidden (object sender, EventArgs e)
+		{	
+			var menu = sender as Gtk.Menu;
+			menu.Hidden -= HandleMenuHidden; 
+			GLib.Timeout.Add (10, delegate {
+				menu.Destroy ();
+				return false;
+			});
 		}
 		
 #region Templates
@@ -663,102 +680,54 @@ namespace MonoDevelop.SourceEditor
 			return false;
 		}
 		
-		void HandleTextPaste (int insertionOffset, string text)
-		{
-			if (PropertyService.Get ("OnTheFlyFormatting", false)) {
-				var prettyPrinter = CodeFormatterService.GetFormatter (Document.MimeType);
-				if (prettyPrinter != null && ProjectDom != null && text != null) {
-					try {
-						var policies = ProjectDom != null && ProjectDom.Project != null ? ProjectDom.Project.Policies : null;
-						string newText = prettyPrinter.FormatText (policies, Document.Text, insertionOffset, insertionOffset + text.Length);
-						if (!string.IsNullOrEmpty (newText)) {
-							Replace (insertionOffset, text.Length, newText);
-							Caret.Offset = insertionOffset + newText.Length;
-						}
-					} catch (Exception e) {
-						LoggingService.LogError ("Error formatting pasted text", e);
-					}
-				}
-			}
-		}
-		
+
 		internal void InsertTemplate (CodeTemplate template, MonoDevelop.Ide.Gui.Document document)
 		{
-			Document.BeginAtomicUndo ();
-			var result = template.InsertTemplateContents (document);
-			var tle = new TextLinkEditMode (this, result.InsertPosition, result.TextLinks);
-			
-			if (PropertyService.Get ("OnTheFlyFormatting", false)) {
-				var prettyPrinter = CodeFormatterService.GetFormatter (Document.MimeType);
-				if (prettyPrinter != null) {
-					int endOffset = result.InsertPosition + result.Code.Length;
-					string oldText = Document.GetTextAt (result.InsertPosition, result.Code.Length);
-					var policies = document.Project != null ? document.Project.Policies : null;
-					string text = prettyPrinter.FormatText (policies, Document.Text, result.InsertPosition, endOffset);
-					
-					if (text != null)
-						Replace (result.InsertPosition, result.Code.Length, text);
-					else
-						//if formatting failed, just use the unformatted text
-						text = oldText;
-					
-					Caret.Offset = result.InsertPosition + TranslateOffset (oldText, text, Caret.Offset - result.InsertPosition);
-					foreach (TextLink textLink in tle.Links) {
-						foreach (ISegment segment in textLink.Links) {
-							segment.Offset = TranslateOffset (oldText, text, segment.Offset);
-						}
-					}
+			using (var undo = Document.OpenUndoGroup ()) {
+				var result = template.InsertTemplateContents (document);
+
+				var links = result.TextLinks;
+
+				var tle = new TextLinkEditMode (this, result.InsertPosition, links);
+				tle.TextLinkMode = TextLinkMode.General;
+				if (tle.ShouldStartTextLinkMode) {
+					tle.OldMode = CurrentMode;
+					tle.StartMode ();
+					CurrentMode = tle;
 				}
 			}
-			
-			if (tle.ShouldStartTextLinkMode) {
-				tle.OldMode = CurrentMode;
-				tle.StartMode ();
-				CurrentMode = tle;
-			}
-			Document.EndAtomicUndo ();
 		}
-		
-		static int TranslateOffset (string baseInput, string formattedInput, int offset)
+
+		protected override void OnScrollAdjustmentsSet()
 		{
-			int i = 0;
-			int j = 0;
-			while (i < baseInput.Length && j < formattedInput.Length && i < offset) {
-				char ch1 = baseInput[i];
-				char ch2 = formattedInput[j];
-				bool ch1IsWs = Char.IsWhiteSpace (ch1);
-				bool ch2IsWs = Char.IsWhiteSpace (ch2);
-				if (ch1 == ch2 || ch1IsWs && ch2IsWs) {
-					i++;
-					j++;
-				} else if (!ch1IsWs && ch2IsWs) {
-					j++;
-				} else if (ch1IsWs && !ch2IsWs) {
-					i++;
-				} else {
-					return -1;
-				}
+			UnregisterAdjustments ();
+			if (HAdjustment != null) {
+				cachedHAdjustment = HAdjustment;
+				HAdjustment.ValueChanged += HAdjustment_ValueChanged;
+			}
+			if (VAdjustment != null) {
+				cachedVAdjustment = VAdjustment;
+				VAdjustment.ValueChanged += VAdjustment_ValueChanged;
 			}
-			return j;
 		}
-		
-		protected override void HAdjustmentValueChanged ()
+
+		void VAdjustment_ValueChanged (object sender, EventArgs e)
+		{
+			CompletionWindowManager.HideWindow ();
+			ParameterInformationWindowManager.HideWindow (null, view);
+		}
+
+		void HAdjustment_ValueChanged (object sender, EventArgs e)
 		{
-			base.HAdjustmentValueChanged ();
 			if (!isInKeyStroke) {
 				CompletionWindowManager.HideWindow ();
-				ParameterInformationWindowManager.HideWindow (view);
+				ParameterInformationWindowManager.HideWindow (null, view);
+			} else {
+				CompletionWindowManager.RepositionWindow ();
+				ParameterInformationWindowManager.RepositionWindow (null, view);
 			}
 		}
 		
-		protected override void VAdjustmentValueChanged ()
-		{
-			base.VAdjustmentValueChanged ();
-			CompletionWindowManager.HideWindow ();
-			ParameterInformationWindowManager.HideWindow (view);
-		}
-		
-		
 #endregion
 		
 #region Key bindings
@@ -967,6 +936,12 @@ namespace MonoDevelop.SourceEditor
 			RunAction (SelectionActions.MoveToDocumentStart);
 		}
 
+		[CommandHandler (MonoDevelop.Ide.Commands.TextEditorCommands.ExpandSelectionToLine)]
+		internal void OnExpandSelectionToLine ()
+		{
+			RunAction (SelectionActions.ExpandSelectionToLine);
+		}
+
 		[CommandHandler (MonoDevelop.Ide.Commands.TextEditorCommands.SelectionMoveToDocumentEnd)]
 		internal void OnSelectionMoveToDocumentEnd ()
 		{
@@ -1012,14 +987,13 @@ namespace MonoDevelop.SourceEditor
 		[CommandHandler (MonoDevelop.Ide.Commands.TextEditorCommands.CompleteStatement)]
 		internal void OnCompleteStatement ()
 		{
-			CodeRefactorer refactorer = IdeApp.Workspace.GetCodeRefactorer (IdeApp.ProjectOperations.CurrentSelectedSolution);
-			DomLocation caretLocation = refactorer.CompleteStatement (ProjectDom, Document.FileName, new DomLocation (Caret.Line, Caret.Column));
-			Caret.Line   = caretLocation.Line;
-			Caret.Column = caretLocation.Column;
+			var doc = IdeApp.Workbench.ActiveDocument;
+			var generator = CodeGenerator.CreateGenerator (doc);
+			if (generator != null) {
+				generator.CompleteStatement (doc);
+			}
 		}
 		
-		
-		
 		[CommandHandler (MonoDevelop.Ide.Commands.TextEditorCommands.DeletePrevWord)]
 		internal void OnDeletePrevWord ()
 		{
@@ -1062,18 +1036,18 @@ namespace MonoDevelop.SourceEditor
 			StartCaretPulseAnimation ();
 		}
 		
-		[CommandHandler (MonoDevelop.SourceEditor.SourceEditorCommands.ToggleCodeFocus)]
-		internal void OnToggleCodeFocus ()
-		{
-			foldMarkerMargin.IsInCodeFocusMode = !foldMarkerMargin.IsInCodeFocusMode;
-		}
-		
 		[CommandHandler (MonoDevelop.Ide.Commands.TextEditorCommands.TransposeCharacters)]
 		internal void TransposeCharacters ()
 		{
 			RunAction (MiscActions.TransposeCharacters);
 		}
-		
+
+		[CommandHandler (MonoDevelop.Ide.Commands.TextEditorCommands.DuplicateLine)]
+		internal void DuplicateLine ()
+		{
+			RunAction (MiscActions.DuplicateLine);
+		}
+
 		[CommandHandler (MonoDevelop.Ide.Commands.TextEditorCommands.RecenterEditor)]
 		internal void RecenterEditor ()
 		{
@@ -1083,11 +1057,8 @@ namespace MonoDevelop.SourceEditor
 		[CommandHandler (MonoDevelop.Ide.Commands.EditCommands.JoinWithNextLine)]
 		internal void JoinLines ()
 		{
-			try {
-				Document.BeginAtomicUndo ();
+			using (var undo = Document.OpenUndoGroup ()) {
 				RunAction (Mono.TextEditor.Vi.ViActions.Join);
-			} finally {
-				Document.EndAtomicUndo ();
 			}
 		}
 		
