@@ -28,19 +28,23 @@
 
 using System;
 using System.Collections;
-using System.Threading;
 using Gtk;
 using Gdk;
 
 using MonoDevelop.Core;
 using MonoDevelop.Core.Execution;
-using MonoDevelop.Core.Gui;
-using MonoDevelop.Ide.Gui;
 using MonoDevelop.Ide.Gui.Pads;
-using MonoDevelop.Projects;
 using MonoDevelop.Components.Commands;
 using MonoDevelop.NUnit.Commands;
 using MonoDevelop.Ide.Gui.Components;
+using MonoDevelop.Ide.Execution;
+using MonoDevelop.Components.Docking;
+using MonoDevelop.Ide;
+using MonoDevelop.Projects;
+using Mono.TextEditor;
+using System.Linq;
+using MonoDevelop.Components;
+using MonoDevelop.Ide.Commands;
 
 namespace MonoDevelop.NUnit
 {
@@ -64,7 +68,6 @@ namespace MonoDevelop.NUnit
 		ListStore regressionStore;
 		TreeView failedTree;
 		ListStore failedStore;
-		Gtk.Tooltips tips = new Gtk.Tooltips ();
 		
 		int TestSummaryPage;
 		int TestRegressionsPage;
@@ -77,7 +80,8 @@ namespace MonoDevelop.NUnit
 		
 		ArrayList testNavigationHistory = new ArrayList ();
 
-		ToolButton buttonRunAll, buttonRun, buttonStop;
+		Button buttonRunAll, buttonStop;
+		
 		public override void Initialize (NodeBuilder[] builders, TreePadOption[] options, string menuPath)
 		{
 			base.Initialize (builders, options, menuPath);
@@ -87,45 +91,23 @@ namespace MonoDevelop.NUnit
 			paned = new VPaned ();
 			
 			VBox vbox = new VBox ();
-			Toolbar topToolbar = new Toolbar ();
-			topToolbar.IconSize = IconSize.Menu;
-			topToolbar.ToolbarStyle = ToolbarStyle.Icons;
-			vbox.PackStart (topToolbar, false, false, 0);
-			
-			Gdk.Pixbuf pixbuf  = Gtk.IconTheme.Default.LoadIcon (Gtk.Stock.GoUp, 16, (Gtk.IconLookupFlags) 0);
-//			Gdk.Pixbuf pixbuf2 = Gtk.IconTheme.Default.LoadIcon (Gtk.Stock.Execute, 16, (Gtk.IconLookupFlags) 0);
-//			pixbuf2.CopyArea (0, 0, 15, 16, pixbuf, 1, 0);
-//			Gtk.Image image = new Gtk.Image (Gtk.Stock.Execute, IconSize.Menu);
-//			Gdk.Pixmap pixbuf, mask;
-//			image.GetPixmap (out pixbuf, out mask);
-//			
-//			pixbuf.DrawDrawable (vbox.Style.BaseGC (StateType.Normal), pixbuf, 5, 5 , 0, 0, image.WidthRequest, image.HeightRequest);
-			
-			buttonRunAll = new ToolButton (new Gtk.Image (pixbuf), null);
-			buttonRunAll.IsImportant = true;
+			DockItemToolbar topToolbar = Window.GetToolbar (PositionType.Top);
+
+			var hbox = new HBox { Spacing = 6 };
+			hbox.PackStart (new ImageView (ImageService.GetIcon ("nunit-run", IconSize.Menu)), false, false, 0);
+			hbox.PackStart (new Label (GettextCatalog.GetString ("Run All")), false, false, 0);
+			buttonRunAll = new Button (hbox);
 			buttonRunAll.Clicked += new EventHandler (OnRunAllClicked);
 			buttonRunAll.Sensitive = true;
-			buttonRunAll.SetTooltip (tips, 
-			                     GettextCatalog.GetString ("Run all tests"), 
-			                     GettextCatalog.GetString ("Run all tests"));
-			topToolbar.Insert (buttonRunAll, -1);
+			buttonRunAll.TooltipText = GettextCatalog.GetString ("Run all tests");
+			topToolbar.Add (buttonRunAll);
 			
-			buttonRun = new ToolButton (new Gtk.Image (Gtk.Stock.Execute, IconSize.Menu), null);
-			buttonRun.IsImportant = true;
-			buttonRun.Clicked += new EventHandler (OnRunClicked);
-			buttonRun.Sensitive = true;
-			buttonRun.SetTooltip (tips, 
-			                     GettextCatalog.GetString ("Run test"), 
-			                     GettextCatalog.GetString ("Run test"));
-			topToolbar.Insert (buttonRun, -1);
-			
-			buttonStop = new ToolButton (Gtk.Stock.Stop);
+			buttonStop = new Button (new Gtk.Image (Ide.Gui.Stock.Stop, IconSize.Menu));
 			buttonStop.Clicked += new EventHandler (OnStopClicked);
 			buttonStop.Sensitive = false;
-			buttonStop.SetTooltip (tips, 
-			                     GettextCatalog.GetString ("Cancel running test"), 
-			                     GettextCatalog.GetString ("Cancel running test"));
-			topToolbar.Insert (buttonStop, -1);
+			buttonStop.TooltipText = GettextCatalog.GetString ("Cancel running test");
+			topToolbar.Add (buttonStop);
+			topToolbar.ShowAll ();
 			
 			vbox.PackEnd (base.Control, true, true, 0);
 			vbox.FocusChain = new Gtk.Widget [] { base.Control };
@@ -165,7 +147,7 @@ namespace MonoDevelop.NUnit
 			VBox boxPaned1 = new VBox ();
 			
 			chart = new TestChart ();
-			chart.ButtonReleaseEvent += new Gtk.ButtonReleaseEventHandler (OnChartPopupMenu);
+			chart.ButtonPressEvent += OnChartButtonPress;
 			chart.SelectionChanged += new EventHandler (OnChartDateChanged);
 			chart.HeightRequest = 50;
 			
@@ -219,7 +201,7 @@ namespace MonoDevelop.NUnit
 			TreeViewColumn col3 = new TreeViewColumn ();
 			col3.Expand = false;
 //			col3.Alignment = 0.5f;
-			col3.Widget = new Gtk.Image (CircleImage.Success);
+			col3.Widget = new ImageView (TestStatusIcon.Success);
 			col3.Widget.Show ();
 			tr = new CellRendererText ();
 //			tr.Xalign = 0.5f;
@@ -230,7 +212,7 @@ namespace MonoDevelop.NUnit
 			TreeViewColumn col4 = new TreeViewColumn ();
 			col4.Expand = false;
 //			col4.Alignment = 0.5f;
-			col4.Widget = new Gtk.Image (CircleImage.Failure);
+			col4.Widget = new ImageView (TestStatusIcon.Failure);
 			col4.Widget.Show ();
 			tr = new CellRendererText ();
 //			tr.Xalign = 0.5f;
@@ -241,7 +223,7 @@ namespace MonoDevelop.NUnit
 			TreeViewColumn col5 = new TreeViewColumn ();
 			col5.Expand = false;
 //			col5.Alignment = 0.5f;
-			col5.Widget = new Gtk.Image (CircleImage.NotRun);
+			col5.Widget = new ImageView (TestStatusIcon.NotRun);
 			col5.Widget.Show ();
 			tr = new CellRendererText ();
 //			tr.Xalign = 0.5f;
@@ -273,14 +255,14 @@ namespace MonoDevelop.NUnit
 			regressionTree = new TreeView ();
 			regressionTree.HeadersVisible = false;
 			regressionTree.RulesHint = true;
-			regressionStore = new ListStore (typeof(object), typeof(string), typeof (Pixbuf));
+			regressionStore = new ListStore (typeof(object), typeof(string), typeof (Xwt.Drawing.Image));
 			
 			CellRendererText trtest2 = new CellRendererText ();
-			CellRendererPixbuf pr = new CellRendererPixbuf ();
+			var pr = new CellRendererImage ();
 			
 			TreeViewColumn col = new TreeViewColumn ();
 			col.PackStart (pr, false);
-			col.AddAttribute (pr, "pixbuf", 2);
+			col.AddAttribute (pr, "image", 2);
 			col.PackStart (trtest2, false);
 			col.AddAttribute (trtest2, "markup", 1);
 			regressionTree.AppendColumn (col);
@@ -298,14 +280,14 @@ namespace MonoDevelop.NUnit
 			failedTree = new TreeView ();
 			failedTree.HeadersVisible = false;
 			failedTree.RulesHint = true;
-			failedStore = new ListStore (typeof(object), typeof(string), typeof (Pixbuf));
+			failedStore = new ListStore (typeof(object), typeof(string), typeof (Xwt.Drawing.Image));
 			
 			trtest2 = new CellRendererText ();
-			pr = new CellRendererPixbuf ();
+			pr = new CellRendererImage ();
 			
 			col = new TreeViewColumn ();
 			col.PackStart (pr, false);
-			col.AddAttribute (pr, "pixbuf", 2);
+			col.AddAttribute (pr, "image", 2);
 			col.PackStart (trtest2, false);
 			col.AddAttribute (trtest2, "markup", 1);
 			failedTree.AppendColumn (col);
@@ -420,8 +402,9 @@ namespace MonoDevelop.NUnit
 		[CommandHandler (TestCommands.RunTestWith)]
 		protected void OnRunTest (object data)
 		{
-			IExecutionMode mode = (IExecutionMode) data;
-			RunSelectedTest (mode.ExecutionHandler);
+			IExecutionHandler h = ExecutionModeCommandService.GetExecutionModeForCommand (data);
+			if (h != null)
+				RunSelectedTest (h);
 		}
 		
 		[CommandUpdateHandler (TestCommands.RunTestWith)]
@@ -429,15 +412,41 @@ namespace MonoDevelop.NUnit
 		{
 			UnitTest test = GetSelectedTest ();
 			if (test != null) {
-				foreach (IExecutionMode mode in Runtime.ProcessService.GetExecutionModes ()) {
-					if (mode != Runtime.ProcessService.DefaultExecutionMode && test.CanRun (mode.ExecutionHandler)) {
-						CommandInfo ci = new CommandInfo (mode.Name);
-						info.Add (ci, mode);
-					}
-				}
+				SolutionEntityItem item = test.OwnerObject as SolutionEntityItem;
+				ExecutionModeCommandService.GenerateExecutionModeCommands (
+				    item,
+				    test.CanRun,
+				    info);
 			}
 		}
 		
+		[CommandHandler (TestCommands.DebugTest)]
+		protected void OnDebugTest (object data)
+		{
+			var debugModeSet = Runtime.ProcessService.GetDebugExecutionMode ();
+			var mode = debugModeSet.ExecutionModes.First (m => m.Id == (string)data);
+			RunSelectedTest (mode.ExecutionHandler);
+		}
+
+		[CommandUpdateHandler (TestCommands.DebugTest)]
+		protected void OnUpdateDebugTest (CommandArrayInfo info)
+		{
+			var debugModeSet = Runtime.ProcessService.GetDebugExecutionMode ();
+			if (debugModeSet == null)
+				return;
+
+			UnitTest test = GetSelectedTest ();
+			if (test == null)
+				return;
+
+			foreach (var mode in debugModeSet.ExecutionModes) {
+				if (test.CanRun (mode.ExecutionHandler))
+					info.Add (GettextCatalog.GetString ("Debug Test ({0})", mode.Name), mode.Id);
+			}
+			if (info.Count == 1)
+				info [0].Text = GettextCatalog.GetString ("Debug Test");
+		}
+
 		public TestPad ()
 		{
 			base.TreeView.CurrentItemActivated += delegate {
@@ -451,11 +460,6 @@ namespace MonoDevelop.NUnit
 				runningTestOperation.Cancel ();
 		}
 		
-		void OnRunClicked (object sender, EventArgs args)
-		{
-			RunSelectedTest (null);
-		}
-			
 		UnitTest GetSelectedTest ()
 		{
 			ITreeNavigator nav = TreeView.GetSelectedNode ();
@@ -463,23 +467,29 @@ namespace MonoDevelop.NUnit
 				return null;
 			return nav.DataItem as UnitTest;
 		}
-		
-		void RunTest (ITreeNavigator nav, IExecutionHandler mode)
+
+		public IAsyncOperation RunTest (UnitTest test, IExecutionHandler mode)
+		{
+			return RunTest (FindTestNode (test), mode, false);
+		}
+
+		IAsyncOperation RunTest (ITreeNavigator nav, IExecutionHandler mode, bool bringToFront = true)
 		{
 			if (nav == null)
-				return;
+				return null;
 			UnitTest test = nav.DataItem as UnitTest;
 			if (test == null)
-				return;
-			TestSession.ResetResult (test.RootTest);
+				return null;
+			NUnitService.ResetResult (test.RootTest);
 			
-			this.buttonRun.Sensitive = false;
 			this.buttonRunAll.Sensitive = false;
 			this.buttonStop.Sensitive = true;
-			
-			IdeApp.Workbench.GetPad<TestPad> ().BringToFront ();
+
+			if (bringToFront)
+				IdeApp.Workbench.GetPad<TestPad> ().BringToFront ();
 			runningTestOperation = testService.RunTest (test, mode);
-			runningTestOperation.Completed += (OperationHandler) DispatchService.GuiDispatch (new OperationHandler (TestSessionCompleted));
+			runningTestOperation.Completed += (OperationHandler) DispatchService.GuiDispatch (new OperationHandler (OnTestSessionCompleted));
+			return runningTestOperation;
 		}
 		
 		void OnRunAllClicked (object sender, EventArgs args)
@@ -492,16 +502,17 @@ namespace MonoDevelop.NUnit
 			RunTest (TreeView.GetSelectedNode (), mode);
 		}
 		
-		void TestSessionCompleted (IAsyncOperation op)
+		void OnTestSessionCompleted (IAsyncOperation op)
 		{
 			if (op.Success)
 				RefreshDetails ();
 			runningTestOperation = null;
-			this.buttonRun.Sensitive = true;
 			this.buttonRunAll.Sensitive = true;
 			this.buttonStop.Sensitive = false;
+
 		}
-		
+
+
 		protected override void OnSelectionChanged (object sender, EventArgs args)
 		{
 			base.OnSelectionChanged (sender, args);
@@ -615,7 +626,7 @@ namespace MonoDevelop.NUnit
 					foreach (UnitTest t in group.Tests) {
 						UnitTestResult res = t.Results.GetLastResult (chart.CurrentDate);
 						if (res != null)
-							detailsStore.AppendValues (t, t.Name, res.TotalSuccess.ToString(), res.TotalFailures.ToString(), res.TotalIgnored.ToString());
+							detailsStore.AppendValues (t, t.Name, res.Passed.ToString (), res.ErrorsAndFailures.ToString (), res.Ignored.ToString());
 						else
 							detailsStore.AppendValues (t, t.Name, "", "", "");
 					}
@@ -625,7 +636,7 @@ namespace MonoDevelop.NUnit
 					UnitTestCollection regs = detailsTest.GetRegressions (chart.ReferenceDate, chart.CurrentDate);
 					if (regs.Count > 0) {
 						foreach (UnitTest t in regs)
-							regressionStore.AppendValues (t, t.Name, CircleImage.Failure);
+							regressionStore.AppendValues (t, t.Name, TestStatusIcon.Failure);
 					} else
 						regressionStore.AppendValues (null, GettextCatalog.GetString ("No regressions found."));
 				}
@@ -634,7 +645,7 @@ namespace MonoDevelop.NUnit
 					UnitTestCollection regs = group.GetFailedTests (chart.CurrentDate);
 					if (regs.Count > 0) {
 						foreach (UnitTest t in regs)
-							failedStore.AppendValues (t, t.Name, CircleImage.Failure);
+							failedStore.AppendValues (t, t.Name, TestStatusIcon.Failure);
 					} else
 						failedStore.AppendValues (null, GettextCatalog.GetString ("No failed tests found."));
 				}
@@ -731,10 +742,11 @@ namespace MonoDevelop.NUnit
 				chart.GoLast ();
 		}
 		
-		void OnChartPopupMenu (object o, Gtk.ButtonReleaseEventArgs args)
+		void OnChartButtonPress (object o, Gtk.ButtonPressEventArgs args)
 		{
-			if (args.Event.Button == 3) {
-				IdeApp.CommandService.ShowContextMenu ("/MonoDevelop/NUnit/ContextMenu/TestChart");
+			if (args.Event.TriggersContextMenu ()) {
+				IdeApp.CommandService.ShowContextMenu (Control, args.Event, "/MonoDevelop/NUnit/ContextMenu/TestChart");
+				args.RetVal = true;
 			}
 		}
 		
