@@ -41,10 +41,12 @@ using MonoDevelop.Ide.Gui;
 using MonoDevelop.Components.Docking;
 using MonoDevelop.Ide;
 using System.Text.RegularExpressions;
+using MonoDevelop.Components;
+using MonoDevelop.Ide.Commands;
 
 namespace MonoDevelop.NUnit
 {
-	class TestResultsPad: IPadContent, ITestProgressMonitor
+	public class TestResultsPad: IPadContent, ITestProgressMonitor
 	{
 		NUnitService testService = NUnitService.Instance;
 		
@@ -74,6 +76,7 @@ namespace MonoDevelop.NUnit
 		ToggleButton buttonSuccess;
 		ToggleButton buttonFailures;
 		ToggleButton buttonIgnored;
+		ToggleButton buttonInconclusive;
 		ToggleButton buttonOutput;
 		
 		bool running;
@@ -91,7 +94,8 @@ namespace MonoDevelop.NUnit
 		
 		bool cancel;
 		
-		public class ResultRecord {
+		public class ResultRecord
+		{
 			public UnitTest Test;
 			public UnitTestResult Result;
 		}
@@ -111,12 +115,12 @@ namespace MonoDevelop.NUnit
 			// Failures tree
 			failuresTreeView = new MonoDevelop.Ide.Gui.Components.PadTreeView ();
 			failuresTreeView.HeadersVisible = false;
-			failuresStore = new TreeStore (typeof(Pixbuf), typeof(string), typeof(object), typeof(string), typeof(int));
-			var pr = new CellRendererPixbuf ();
+			failuresStore = new TreeStore (typeof(Xwt.Drawing.Image), typeof(string), typeof(object), typeof(string), typeof(int));
+			var pr = new CellRendererImage ();
 			CellRendererText tr = new CellRendererText ();
 			TreeViewColumn col = new TreeViewColumn ();
 			col.PackStart (pr, false);
-			col.AddAttribute (pr, "pixbuf", 0);
+			col.AddAttribute (pr, "image", 0);
 			col.PackStart (tr, false);
 			col.AddAttribute (tr, "markup", 1);
 			failuresTreeView.AppendColumn (col);
@@ -124,17 +128,17 @@ namespace MonoDevelop.NUnit
 		
 			var sw = new MonoDevelop.Components.CompactScrolledWindow ();
 			sw.ShadowType = ShadowType.None;
-			sw.Add(failuresTreeView);
+			sw.Add (failuresTreeView);
 			book.Pack1 (sw, true, true);
 			
-			outputView = new TextView();
+			outputView = new MonoDevelop.Ide.Gui.Components.LogView.LogTextView ();
 			outputView.Editable = false;
 			bold = new TextTag ("bold");
 			bold.Weight = Pango.Weight.Bold;
 			outputView.Buffer.TagTable.Add (bold);
 			sw = new MonoDevelop.Components.CompactScrolledWindow ();
 			sw.ShadowType = ShadowType.None;
-			sw.Add(outputView);
+			sw.Add (outputView);
 			book.Pack2 (sw, true, true);
 			outputViewScrolled = sw;
 			
@@ -159,25 +163,34 @@ namespace MonoDevelop.NUnit
 			buttonSuccess = new ToggleButton ();
 			buttonSuccess.Label = GettextCatalog.GetString ("Successful Tests");
 			buttonSuccess.Active = false;
-			buttonSuccess.Image = new Gtk.Image (CircleImage.Success);
+			buttonSuccess.Image = new ImageView (TestStatusIcon.Success);
 			buttonSuccess.Image.Show ();
 			buttonSuccess.Toggled += new EventHandler (OnShowSuccessfulToggled);
 			buttonSuccess.TooltipText = GettextCatalog.GetString ("Show Successful Tests");
 			toolbar.Add (buttonSuccess);
+
+			buttonInconclusive = new ToggleButton ();
+			buttonInconclusive.Label = GettextCatalog.GetString ("Inconclusive Tests");
+			buttonInconclusive.Active = true;
+			buttonInconclusive.Image = new ImageView (TestStatusIcon.Inconclusive);
+			buttonInconclusive.Image.Show ();
+			buttonInconclusive.Toggled += new EventHandler (OnShowInconclusiveToggled);
+			buttonInconclusive.TooltipText = GettextCatalog.GetString ("Show Inconclusive Tests");
+			toolbar.Add (buttonInconclusive);
 			
 			buttonFailures = new ToggleButton ();
 			buttonFailures.Label = GettextCatalog.GetString ("Failed Tests");
 			buttonFailures.Active = true;
-			buttonFailures.Image = new Gtk.Image (CircleImage.Failure);
+			buttonFailures.Image = new ImageView (TestStatusIcon.Failure);
 			buttonFailures.Image.Show ();
 			buttonFailures.Toggled += new EventHandler (OnShowFailuresToggled);
 			buttonFailures.TooltipText = GettextCatalog.GetString ("Show Failed Tests");
 			toolbar.Add (buttonFailures);
-			
+
 			buttonIgnored = new ToggleButton ();
 			buttonIgnored.Label = GettextCatalog.GetString ("Ignored Tests");
 			buttonIgnored.Active = true;
-			buttonIgnored.Image = new Gtk.Image (CircleImage.NotRun);
+			buttonIgnored.Image = new ImageView (TestStatusIcon.NotRun);
 			buttonIgnored.Image.Show ();
 			buttonIgnored.Toggled += new EventHandler (OnShowIgnoredToggled);
 			buttonIgnored.TooltipText = GettextCatalog.GetString ("Show Ignored Tests");
@@ -195,13 +208,13 @@ namespace MonoDevelop.NUnit
 			toolbar.Add (new SeparatorToolItem ());
 			
 			buttonRun = new Button ();
-			buttonRun.Label = GettextCatalog.GetString ("Run Test");
+			buttonRun.Label = GettextCatalog.GetString ("Rerun Tests");
 			buttonRun.Image = new Gtk.Image (Gtk.Stock.Execute, IconSize.Menu);
 			buttonRun.Image.Show ();
 			buttonRun.Sensitive = false;
 			toolbar.Add (buttonRun);
 			
-			buttonStop = new Button (new Gtk.Image (Gtk.Stock.Stop, Gtk.IconSize.Menu));
+			buttonStop = new Button (new Gtk.Image (Ide.Gui.Stock.Stop, Gtk.IconSize.Menu));
 			toolbar.Add (buttonStop);
 			toolbar.ShowAll ();
 			
@@ -243,6 +256,9 @@ namespace MonoDevelop.NUnit
 		
 		public void OnTestSuiteChanged (object sender, EventArgs e)
 		{
+			if (failuresTreeView.IsRealized)
+				failuresTreeView.ScrollToPoint (0, 0);
+
 			results.Clear ();
 			
 			error = null;
@@ -314,6 +330,9 @@ namespace MonoDevelop.NUnit
 			labels.Show ();
 			buttonStop.Sensitive = true;
 			buttonRun.Sensitive = false;
+
+			if (failuresTreeView.IsRealized)
+				failuresTreeView.ScrollToPoint (0, 0);
 			
 			failuresStore.Clear ();
 			outputView.Buffer.Clear ();
@@ -329,7 +348,7 @@ namespace MonoDevelop.NUnit
 		public void AddStartMessage (bool isRunning = true)
 		{
 			if (rootTest != null) {
-				Gdk.Pixbuf infoIcon = failuresTreeView.RenderIcon (Gtk.Stock.DialogInfo, Gtk.IconSize.Menu, "");
+				var infoIcon = ImageService.GetIcon (MonoDevelop.Ide.Gui.Stock.Information, Gtk.IconSize.Menu);
 				string msg = string.Format (isRunning ? GettextCatalog.GetString ("Running tests for <b>{0}</b> configuration <b>{1}</b>") : GettextCatalog.GetString ("Test results for <b>{0}</b> configuration <b>{1}</b>"), rootTest.Name, configuration);
 				startMessageIter = failuresStore.AppendValues (infoIcon, msg, rootTest, null, 0);
 			} else {
@@ -348,9 +367,9 @@ namespace MonoDevelop.NUnit
 		{
 			string msg = GettextCatalog.GetString ("Internal error");
 			if (errorMessage != null)
-				msg += ": " + errorMessage;
+				msg += ": " + Escape (errorMessage);
 
-			Gdk.Pixbuf stock = failuresTreeView.RenderIcon (Gtk.Stock.DialogError, Gtk.IconSize.Menu, "");
+			var stock = ImageService.GetIcon (Ide.Gui.Stock.Error, Gtk.IconSize.Menu);
 			TreeIter testRow = failuresStore.AppendValues (stock, msg, null, null, 0);
 			failuresStore.AppendValues (testRow, null, Escape (error.GetType ().Name + ": " + error.Message), null);
 			TreeIter row = failuresStore.AppendValues (testRow, null, GettextCatalog.GetString ("Stack Trace"), null, null, 0);
@@ -457,13 +476,40 @@ namespace MonoDevelop.NUnit
 				}
 			}
 		}
-		
+
+		[CommandHandler (EditCommands.Copy)]
+		protected void OnCopy ()
+		{
+			UnitTest test = GetSelectedTest ();
+			if (test != null) {
+				var last = test.GetLastResult ();
+				if (last == null)
+					return;
+				var clipboard = Clipboard.Get (Gdk.Atom.Intern ("CLIPBOARD", false));
+				clipboard.Text = last.StackTrace;
+			}
+		}
+
+		[CommandUpdateHandler (EditCommands.Copy)]
+		protected void OnUpdateCopy (CommandInfo info)
+		{
+			UnitTest test = GetSelectedTest ();
+			if (test != null) {
+				var result = test.GetLastResult ();
+				if (result != null) {
+					info.Enabled = !string.IsNullOrEmpty (result.StackTrace);
+					return;
+				}
+			}
+			info.Enabled = false;
+		}
+
 		[CommandHandler (TestCommands.SelectTestInTree)]
 		protected void OnSelectTestInTree ()
 		{
 			Pad pad = IdeApp.Workbench.GetPad<TestPad> ();
 			pad.BringToFront ();
-			TestPad content = (TestPad) pad.Content;
+			TestPad content = (TestPad)pad.Content;
 			content.SelectTest (GetSelectedTest ());
 		}
 		
@@ -516,7 +562,7 @@ namespace MonoDevelop.NUnit
 			if (!failuresTreeView.Selection.GetSelected (out foo, out iter))
 				return null;
 				
-			UnitTest t = (UnitTest) failuresStore.GetValue (iter, 2);
+			UnitTest t = (UnitTest)failuresStore.GetValue (iter, 2);
 			return t;
 		}
 		
@@ -524,12 +570,17 @@ namespace MonoDevelop.NUnit
 		{
 			RefreshList ();
 		}
-
+		
 		void OnShowFailuresToggled (object sender, EventArgs args)
 		{
 			RefreshList ();
 		}
 		
+		void OnShowInconclusiveToggled (object sender, EventArgs args)
+		{
+			RefreshList ();
+		}
+		
 		void OnShowIgnoredToggled (object sender, EventArgs args)
 		{
 			RefreshList ();
@@ -542,6 +593,9 @@ namespace MonoDevelop.NUnit
 		
 		void RefreshList ()
 		{
+			if (failuresTreeView.IsRealized)
+				failuresTreeView.ScrollToPoint (0, 0);
+
 			failuresStore.Clear ();
 			outputView.Buffer.Clear ();
 			outIters.Clear ();
@@ -560,14 +614,14 @@ namespace MonoDevelop.NUnit
 			if (result.IsSuccess) {
 				if (!buttonSuccess.Active)
 					return;
-				TreeIter testRow = failuresStore.AppendValues (CircleImage.Success, Escape (test.FullName), test);
+				TreeIter testRow = failuresStore.AppendValues (TestStatusIcon.Success, Escape (test.FullName), test);
 				failuresTreeView.ScrollToCell (failuresStore.GetPath (testRow), null, false, 0, 0);
 			}
 			if (result.IsFailure) {
 				if (!buttonFailures.Active)
 					return;
 				string file = test.SourceCodeLocation != null ? test.SourceCodeLocation.FileName + ":" + test.SourceCodeLocation.Line : null;
-				TreeIter testRow = failuresStore.AppendValues (CircleImage.Failure, Escape (test.FullName), test, file);
+				TreeIter testRow = failuresStore.AppendValues (TestStatusIcon.Failure, Escape (test.FullName), test, file);
 				bool hasMessage = result.Message != null && result.Message.Length > 0;
 				if (hasMessage)
 					failuresStore.AppendValues (testRow, null, Escape (result.Message), test);
@@ -582,7 +636,15 @@ namespace MonoDevelop.NUnit
 			if (result.IsNotRun) {
 				if (!buttonIgnored.Active)
 					return;
-				TreeIter testRow = failuresStore.AppendValues (CircleImage.NotRun, Escape (test.FullName), test);
+				TreeIter testRow = failuresStore.AppendValues (TestStatusIcon.NotRun, Escape (test.FullName), test);
+				if (result.Message != null)
+					failuresStore.AppendValues (testRow, null, Escape (result.Message), test);
+				failuresTreeView.ScrollToCell (failuresStore.GetPath (testRow), null, false, 0, 0);
+			}
+			if (result.IsInconclusive) {
+				if (!buttonInconclusive.Active)
+					return;
+				TreeIter testRow = failuresStore.AppendValues (TestStatusIcon.Inconclusive, Escape (test.FullName), test);
 				if (result.Message != null)
 					failuresStore.AppendValues (testRow, null, Escape (result.Message), test);
 				failuresTreeView.ScrollToCell (failuresStore.GetPath (testRow), null, false, 0, 0);
@@ -599,7 +661,14 @@ namespace MonoDevelop.NUnit
 				outputView.Buffer.Insert (ref it, result.ConsoleError);
 			outputView.ScrollMarkOnscreen (outputView.Buffer.InsertMark);
 		}
-		
+
+		void ITestProgressMonitor.WriteGlobalLog (string message)
+		{
+			TextIter it = outputView.Buffer.EndIter;
+			outputView.Buffer.Insert (ref it, message);
+			outputView.ScrollMarkOnscreen (outputView.Buffer.InsertMark);
+		}
+
 		string Escape (string s)
 		{
 			return GLib.Markup.EscapeText (s);
@@ -627,6 +696,9 @@ namespace MonoDevelop.NUnit
 				frac = ((double)testsRun / (double)testsToRun);
 			else
 				frac = 1;
+
+			frac = Math.Min (1, Math.Max (0, frac));
+
 			progressBar.Fraction = frac;
 			progressBar.Text = testsRun + " / " + testsToRun;
 		}
@@ -643,7 +715,7 @@ namespace MonoDevelop.NUnit
 				return;
 			cancel = true;
 			Gtk.Application.Invoke (delegate {
-				failuresStore.AppendValues (CircleImage.Failure, GettextCatalog.GetString ("Test execution cancelled."), null);
+				failuresStore.AppendValues (TestStatusIcon.Failure, GettextCatalog.GetString ("Test execution cancelled."), null);
 			});
 			if (CancelRequested != null)
 				CancelRequested ();
@@ -656,36 +728,57 @@ namespace MonoDevelop.NUnit
 		public event TestHandler CancelRequested;
 	}
 	
-	class TestMonitor: GuiSyncObject, ITestProgressMonitor
+	class TestMonitor: ITestProgressMonitor
 	{
-		// TestResultsPad can't be a GuiSyncObject because there are some
-		// object identity issues. If the pad is registered using the
-		// proxy, it will get different hash codes depending on the caller.
-		
 		ITestProgressMonitor monitor;
 		TestResultsPad pad;
 		
-		public TestMonitor (TestResultsPad pad) {
+		public TestMonitor (TestResultsPad pad)
+		{
 			this.pad = pad;
 			this.monitor = pad;
 		}
-		public void InitializeTestRun (UnitTest test) {
-			pad.InitializeTestRun (test);
+		public void InitializeTestRun (UnitTest test)
+		{
+			DispatchService.GuiDispatch (delegate {
+				pad.InitializeTestRun (test);
+			});
 		}
-		public void FinishTestRun () {
-			pad.FinishTestRun ();
+		public void FinishTestRun ()
+		{
+			DispatchService.GuiDispatch (delegate {
+				pad.FinishTestRun ();
+			});
 		}
-		public void Cancel () {
-			pad.Cancel ();
+		public void Cancel ()
+		{
+			DispatchService.GuiDispatch (delegate {
+				pad.Cancel ();
+			});
 		}
-		public void BeginTest (UnitTest test) {
-			monitor.BeginTest (test);
+		public void BeginTest (UnitTest test)
+		{
+			DispatchService.GuiDispatch (delegate {
+				monitor.BeginTest (test);
+			});
 		}
-		public void EndTest (UnitTest test, UnitTestResult result) {
-			monitor.EndTest (test, result);
+		public void EndTest (UnitTest test, UnitTestResult result)
+		{
+			DispatchService.GuiDispatch (delegate {
+				monitor.EndTest (test, result);
+			});
 		}
-		public void ReportRuntimeError (string message, Exception exception) {
-			monitor.ReportRuntimeError (message, exception);
+		public void ReportRuntimeError (string message, Exception exception)
+		{
+			DispatchService.GuiDispatch (delegate {
+				monitor.ReportRuntimeError (message, exception);
+			});
+		}
+		public void WriteGlobalLog (string message)
+		{
+			DispatchService.GuiDispatch (delegate {
+				monitor.WriteGlobalLog (message);
+			});
 		}
 		public bool IsCancelRequested {
 			get { return monitor.IsCancelRequested; }
