@@ -42,6 +42,9 @@ using NUnit.Core;
 using NUnit.Core.Filters;
 using MonoDevelop.NUnit.External;
 using MonoDevelop.Ide;
+using System.Xml.Linq;
+using System.Linq;
+using System.Globalization;
 
 namespace MonoDevelop.NUnit
 {
@@ -52,14 +55,20 @@ namespace MonoDevelop.NUnit
 		TestInfoCache testInfoCache = new TestInfoCache ();
 		bool cacheLoaded;
 		DateTime lastAssemblyTime;
-		
+
 		static Queue<LoadData> loadQueue = new Queue<LoadData> ();
 		static bool loaderRunning;
-		
+
+		public virtual IList<string> UserAssemblyPaths {
+			get {
+				return null;
+			}
+		}
+
 		public NUnitAssemblyTestSuite (string name): base (name)
 		{
 		}
-		
+
 		public NUnitAssemblyTestSuite (string name, SolutionItem ownerSolutionItem): base (name, ownerSolutionItem)
 		{
 		}
@@ -68,7 +77,6 @@ namespace MonoDevelop.NUnit
 		{
 			try {
 				if (TestInfoCachePath != null) {
-					Console.WriteLine ("saving TestInfoCachePath = " + TestInfoCachePath);
 					testInfoCache.Write (TestInfoCachePath);
 				}
 			} catch {
@@ -82,7 +90,30 @@ namespace MonoDevelop.NUnit
 			}
 		}
 
-		
+		public virtual void GetCustomTestRunner (out string assembly, out string type)
+		{
+			assembly = type = null;
+		}
+
+		public virtual void GetCustomConsoleRunner (out string command, out string args)
+		{
+			command = args = null;
+		}
+
+		ProcessExecutionCommand GetCustomConsoleRunnerCommand ()
+		{
+			string file, args;
+
+			GetCustomConsoleRunner (out file, out args);
+			file = file != null ? file.Trim () : null;
+			if (string.IsNullOrEmpty (file))
+				return null;
+
+			var cmd = Runtime.ProcessService.CreateCommand (file);
+			cmd.Arguments = args;
+			return cmd;
+		}
+
 		protected override void OnActiveConfigurationChanged ()
 		{
 			UpdateTests ();
@@ -91,17 +122,10 @@ namespace MonoDevelop.NUnit
 		
 		internal SourceCodeLocation GetSourceCodeLocation (UnitTest test)
 		{
-			if (test is NUnitTestCase) {
-				NUnitTestCase t = (NUnitTestCase) test;
-				return GetSourceCodeLocation (t.ClassName, t.Name);
-			} else if (test is NUnitTestSuite) {
-				NUnitTestSuite t = (NUnitTestSuite) test;
-				return GetSourceCodeLocation (t.ClassName, null);
-			} else
-				return null;
+			return GetSourceCodeLocation (test.FixtureTypeNamespace, test.FixtureTypeName, test.Name);
 		}
 		
-		protected virtual SourceCodeLocation GetSourceCodeLocation (string fullClassName, string methodName)
+		protected virtual SourceCodeLocation GetSourceCodeLocation (string fixtureTypeNamespace, string fixtureTypeName, string methodName)
 		{
 			return null;
 		}
@@ -180,7 +204,7 @@ namespace MonoDevelop.NUnit
 			}
 
 			OnTestStatusChanged ();
-			
+
 			LoadData ld = new LoadData ();
 			ld.Path = AssemblyPath;
 			ld.TestInfoCachePath = cacheLoaded ? null : TestInfoCachePath;
@@ -242,10 +266,15 @@ namespace MonoDevelop.NUnit
 			if (ti.Tests == null)
 				return;
 			foreach (NunitTestInfo test in ti.Tests) {
+				UnitTest newTest;
 				if (test.Tests != null)
-					Tests.Add (new NUnitTestSuite (this, test));
+					newTest = new NUnitTestSuite (this, test);
 				else
-					Tests.Add (new NUnitTestCase (this, test, test.PathName));
+					newTest = new NUnitTestCase (this, test, test.PathName);
+				newTest.FixtureTypeName = test.FixtureTypeName;
+				newTest.FixtureTypeNamespace = test.FixtureTypeNamespace;
+				Tests.Add (newTest);
+
 			}
 			oldList = new UnitTest [Tests.Count];
 			Tests.CopyTo (oldList, 0);
@@ -268,8 +297,7 @@ namespace MonoDevelop.NUnit
 		
 		static void RunAsyncLoadTest ()
 		{
-			while (true)
-			{
+			while (true) {
 				LoadData ld;
 				lock (loadQueue) {
 					if (loadQueue.Count == 0) {
@@ -297,12 +325,12 @@ namespace MonoDevelop.NUnit
 				} catch (Exception ex) {
 					LoggingService.LogError (ex.ToString ());
 				}
-				
+
 				ExternalTestRunner runner = null;
-				
+
 				try {
 					if (File.Exists (ld.Path)) {
-						runner = (ExternalTestRunner) Runtime.ProcessService.CreateExternalProcessObject (typeof(ExternalTestRunner), false);
+						runner = (ExternalTestRunner)Runtime.ProcessService.CreateExternalProcessObject (typeof(ExternalTestRunner), false);
 						ld.Info = runner.GetTestInfo (ld.Path, ld.SupportAssemblies);
 					}
 				} catch (Exception ex) {
@@ -315,7 +343,7 @@ namespace MonoDevelop.NUnit
 							runner.Dispose ();
 					} catch {}
 				}
-				
+
 				try {
 					ld.Callback (ld);
 				} catch {
@@ -327,20 +355,47 @@ namespace MonoDevelop.NUnit
 		{
 			return RunUnitTest (this, "", "", null, testContext);
 		}
-		
+
 		protected override bool OnCanRun (MonoDevelop.Core.Execution.IExecutionHandler executionContext)
 		{
+			var runnerCmd = GetCustomConsoleRunnerCommand ();
+			if (runnerCmd != null) {
+				return executionContext.CanExecute (runnerCmd);
+			}
 			return Runtime.ProcessService.IsValidForRemoteHosting (executionContext);
 		}
-		
+
+		public string[] CollectTests (UnitTestGroup group)
+		{
+			List<string> result = new List<string> ();
+			foreach (var t in group.Tests) {
+				if (t.IsExplicit)
+					continue;
+				if (t is UnitTestGroup) {
+					result.AddRange (CollectTests ((UnitTestGroup)t));
+				} else {
+					result.Add (t.TestId);
+				}
+			}
+			return result.ToArray ();
+		}
+
 		internal UnitTestResult RunUnitTest (UnitTest test, string suiteName, string pathName, string testName, TestContext testContext)
 		{
-			ExternalTestRunner runner = (ExternalTestRunner) Runtime.ProcessService.CreateExternalProcessObject (typeof(ExternalTestRunner), testContext.ExecutionContext);
-			LocalTestMonitor localMonitor = new LocalTestMonitor (testContext, runner, test, suiteName, testName != null);
-			
+			var runnerExe = GetCustomConsoleRunnerCommand ();
+			if (runnerExe != null)
+				return RunWithConsoleRunner (runnerExe, test, suiteName, pathName, testName, testContext);
+
+			ExternalTestRunner runner = (ExternalTestRunner)Runtime.ProcessService.CreateExternalProcessObject (typeof(ExternalTestRunner), testContext.ExecutionContext, UserAssemblyPaths);
+			LocalTestMonitor localMonitor = new LocalTestMonitor (testContext, test, suiteName, testName != null);
+
 			ITestFilter filter = null;
-			if (testName != null) {
-				filter = new TestNameFilter (pathName + "." + testName);
+			if (test != null) {
+				if (test is UnitTestGroup) {
+					filter = new TestNameFilter (CollectTests ((UnitTestGroup)test));
+				} else {
+					filter = new TestNameFilter (test.TestId);
+				}
 			} else {
 				NUnitCategoryOptions categoryOptions = (NUnitCategoryOptions) test.GetOptions (typeof(NUnitCategoryOptions));
 				if (categoryOptions.EnableFilter && categoryOptions.Categories.Count > 0) {
@@ -351,22 +406,26 @@ namespace MonoDevelop.NUnit
 						filter = new NotFilter (filter);
 				}
 			}
-			
+
 			RunData rd = new RunData ();
 			rd.Runner = runner;
 			rd.Test = this;
 			rd.LocalMonitor = localMonitor;
 			testContext.Monitor.CancelRequested += new TestHandler (rd.Cancel);
-			
+
 			UnitTestResult result;
-			
+
 			try {
 				if (string.IsNullOrEmpty (AssemblyPath)) {
 					string msg = GettextCatalog.GetString ("Could not get a valid path to the assembly. There may be a conflict in the project configurations.");
 					throw new Exception (msg);
 				}
 				System.Runtime.Remoting.RemotingServices.Marshal (localMonitor, null, typeof (IRemoteEventListener));
-				result = runner.Run (localMonitor, filter, AssemblyPath, suiteName, new List<string> (SupportAssemblies));
+
+				string testRunnerAssembly, testRunnerType;
+				GetCustomTestRunner (out testRunnerAssembly, out testRunnerType);
+
+				result = runner.Run (localMonitor, filter, AssemblyPath, "", new List<string> (SupportAssemblies), testRunnerType, testRunnerAssembly);
 				if (testName != null)
 					result = localMonitor.SingleTestResult;
 			} catch (Exception ex) {
@@ -376,7 +435,7 @@ namespace MonoDevelop.NUnit
 						RuntimeErrorCleanup (testContext, localMonitor.RunningTest, ex);
 					} else {
 						testContext.Monitor.ReportRuntimeError (null, ex);
-						throw ex;
+						throw;
 					}
 					result = UnitTestResult.CreateFailure (ex);
 				} else {
@@ -401,6 +460,171 @@ namespace MonoDevelop.NUnit
 				t = t.Parent;
 			}
 		}
+
+		UnitTestResult RunWithConsoleRunner (ProcessExecutionCommand cmd, UnitTest test, string suiteName, string pathName, string testName, TestContext testContext)
+		{
+			var outFile = Path.GetTempFileName ();
+			LocalConsole cons = new LocalConsole ();
+
+			try {
+				MonoDevelop.NUnit.External.TcpTestListener tcpListener = null;
+				LocalTestMonitor localMonitor = new LocalTestMonitor (testContext, test, suiteName, testName != null);
+
+				if (!string.IsNullOrEmpty (cmd.Arguments))
+					cmd.Arguments += " ";
+				cmd.Arguments += "\"-xml=" + outFile + "\" " + AssemblyPath;
+
+				bool automaticUpdates = cmd.Command != null && (cmd.Command.Contains ("GuiUnit") || (cmd.Command.Contains ("mdtool.exe") && cmd.Arguments.Contains ("run-md-tests")));
+				if (!string.IsNullOrEmpty(pathName))
+					cmd.Arguments += " -run=" + test.TestId;
+				if (automaticUpdates) {
+					tcpListener = new MonoDevelop.NUnit.External.TcpTestListener (localMonitor, suiteName);
+					cmd.Arguments += " -port=" + tcpListener.Port;
+				}
+
+				// Note that we always dispose the tcp listener as we don't want it listening
+				// forever if the test runner does not try to connect to it
+				using (tcpListener) {
+					var p = testContext.ExecutionContext.Execute (cmd, cons);
+
+					testContext.Monitor.CancelRequested += p.Cancel;
+					if (testContext.Monitor.IsCancelRequested)
+						p.Cancel ();
+					p.WaitForCompleted ();
+					
+					if (new FileInfo (outFile).Length == 0)
+						throw new Exception ("Command failed");
+				}
+
+				// mdtool.exe does not necessarily guarantee we get automatic updates. It just guarantees
+				// that if guiunit is being used then it will give us updates. If you have a regular test
+				// assembly compiled against nunit.framework.dll 
+				if (automaticUpdates && tcpListener.HasReceivedConnection) {
+					if (testName != null)
+						return localMonitor.SingleTestResult;
+					return test.GetLastResult ();
+				}
+
+				XDocument doc = XDocument.Load (outFile);
+
+				if (doc.Root != null) {
+					var root = doc.Root.Elements ("test-suite").FirstOrDefault ();
+					if (root != null) {
+						cons.SetDone ();
+						var ot = cons.Out.ReadToEnd ();
+						var et = cons.Error.ReadToEnd ();
+						testContext.Monitor.WriteGlobalLog (ot);
+						if (!string.IsNullOrEmpty (et)) {
+							testContext.Monitor.WriteGlobalLog ("ERROR:\n");
+							testContext.Monitor.WriteGlobalLog (et);
+						}
+
+						bool macunitStyle = doc.Root.Element ("environment") != null && doc.Root.Element ("environment").Attribute ("macunit-version") != null;
+						var result = ReportXmlResult (localMonitor, root, "", macunitStyle);
+						if (testName != null)
+							result = localMonitor.SingleTestResult;
+						return result;
+					}
+				}
+				throw new Exception ("Test results could not be parsed.");
+			} catch (Exception ex) {
+				cons.SetDone ();
+				var ot = cons.Out.ReadToEnd ();
+				var et = cons.Error.ReadToEnd ();
+				testContext.Monitor.WriteGlobalLog (ot);
+				if (!string.IsNullOrEmpty (et)) {
+					testContext.Monitor.WriteGlobalLog ("ERROR:\n");
+					testContext.Monitor.WriteGlobalLog (et);
+				}
+				testContext.Monitor.ReportRuntimeError ("Test execution failed.\n" + ot + "\n" + et, ex);
+				return UnitTestResult.CreateIgnored ("Test execution failed");
+			} finally {
+				File.Delete (outFile);
+			}
+		}
+
+		UnitTestResult ReportXmlResult (IRemoteEventListener listener, XElement elem, string testPrefix, bool macunitStyle)
+		{
+			UnitTestResult result = new UnitTestResult ();
+			var time = (string)elem.Attribute ("time");
+			if (time != null)
+				result.Time = TimeSpan.FromSeconds (double.Parse (time, CultureInfo.InvariantCulture));
+			result.TestDate = DateTime.Now;
+
+			var reason = elem.Element ("reason");
+			if (reason != null)
+				result.Message = (string) reason;
+
+			var failure = elem.Element ("failure");
+			if (failure != null) {
+				var msg = failure.Element ("message");
+				if (msg != null)
+					result.Message = (string)msg;
+				var stack = failure.Element ("stack-trace");
+				if (stack != null)
+					result.StackTrace = (string)stack;
+			}
+
+			switch ((string)elem.Attribute ("result")) {
+			case "Error":
+			case "Failure":
+				result.Status = ResultStatus.Failure;
+				break;
+			case "Success":
+				result.Status = ResultStatus.Success;
+				break;
+			case "Ignored":
+				result.Status = ResultStatus.Ignored;
+				break;
+			default:
+				result.Status = ResultStatus.Inconclusive;
+				break;
+			}
+
+			if (elem.Name == "test-suite") {
+				// nunitlite does not emit <test-suite type="Namespace" elements so we need to fake
+				// them by deconstructing the full type name and emitting the suite started events manually
+				var names = new List<string> ();
+				if (!macunitStyle || (string)elem.Attribute ("type") == "Assembly")
+					names.Add ("<root>");
+				else
+					names.AddRange (elem.Attribute ("name").Value.Split ('.'));
+
+				for (int i = 0; i < names.Count; i ++)
+					listener.SuiteStarted (testPrefix + string.Join (".", names.Take (i + 1)));
+
+				var name = (string)elem.Attribute ("type") == "Assembly" ? "<root>" : (string) elem.Attribute ("name");
+				var cts = elem.Element ("results");
+				if (cts != null) {
+					foreach (var ct in cts.Elements ()) {
+						var r = ReportXmlResult (listener, ct, name != "<root>" ? testPrefix + name + "." : "", macunitStyle);
+						result.Add (r);
+					}
+				}
+				for (int i = 0; i < names.Count; i ++)
+					listener.SuiteFinished (testPrefix + string.Join (".", names.Take (i + 1)), result);
+			} else {
+				string name = (string)elem.Attribute ("name");
+				switch (result.Status) {
+				case ResultStatus.Success:
+					result.Passed++;
+					break;
+				case ResultStatus.Failure:
+					result.Failures++;
+					break;
+				case ResultStatus.Ignored:
+					result.Ignored++;
+					break;
+				case ResultStatus.Inconclusive:
+					result.Inconclusive++;
+					break;
+				}
+
+				listener.TestStarted (name);
+				listener.TestFinished (name, result);
+			}
+			return result;
+		}
 		
 		protected abstract string AssemblyPath {
 			get;
