@@ -37,7 +37,7 @@ using System.Reflection;
 
 namespace Mono.TextTemplating
 {
-	public class TemplatingEngine : MarshalByRefObject, Microsoft.VisualStudio.TextTemplating.ITextTemplatingEngine
+	public class TemplatingEngine : MarshalByRefObject, ITextTemplatingEngine
 	{
 		public string ProcessTemplate (string content, ITextTemplatingEngineHost host)
 		{
@@ -96,7 +96,7 @@ namespace Mono.TextTemplating
 			using (var sw = new StringWriter ()) {
 				settings.Provider.GenerateCodeFromCompileUnit (ccu, sw, options);
 				return sw.ToString ();
-			};
+			}
 		}
 
 		public CompiledTemplate CompileTemplate (string content, ITextTemplatingEngineHost host)
@@ -133,7 +133,7 @@ namespace Mono.TextTemplating
 				return null;
 			}
 			
-			var results = GenerateCode (host, references, settings, ccu);
+			var results = GenerateCode (references, settings, ccu);
 			if (results.Errors.HasErrors) {
 				host.LogErrors (pt.Errors);
 				host.LogErrors (results.Errors);
@@ -143,20 +143,19 @@ namespace Mono.TextTemplating
 			var templateClassFullName = settings.Namespace + "." + settings.Name;
 			var domain = host.ProvideTemplatingAppDomain (content);
 			if (domain != null) {
-				var type = typeof (CompiledTemplate);
-				var obj = domain.CreateInstanceAndUnwrap (type.Assembly.FullName, type.FullName, false,
-					BindingFlags.CreateInstance, null,
+				var type = typeof(CompiledTemplate);
+				var obj = domain.CreateInstanceFromAndUnwrap (type.Assembly.Location, type.FullName, false,
+					BindingFlags.Default, null,
 					new object[] { host, results, templateClassFullName, settings.Culture, references.ToArray () },
-					null, null, null);
-				return (CompiledTemplate) obj;
-			} else {
-				return new CompiledTemplate (host, results, templateClassFullName, settings.Culture, references.ToArray ());
+					null, null);
+				return (CompiledTemplate)obj;
 			}
+			return new CompiledTemplate (host, results, templateClassFullName, settings.Culture, references.ToArray ());
 		}
 		
-		static CompilerResults GenerateCode (ITextTemplatingEngineHost host, IEnumerable<string> references, TemplateSettings settings, CodeCompileUnit ccu)
+		static CompilerResults GenerateCode (IEnumerable<string> references, TemplateSettings settings, CodeCompileUnit ccu)
 		{
-			var pars = new CompilerParameters () {
+			var pars = new CompilerParameters {
 				GenerateExecutable = false,
 				CompilerOptions = settings.CompilerOptions,
 				IncludeDebugInformation = settings.Debug,
@@ -194,6 +193,8 @@ namespace Mono.TextTemplating
 		public static TemplateSettings GetSettings (ITextTemplatingEngineHost host, ParsedTemplate pt)
 		{
 			var settings = new TemplateSettings ();
+
+			bool relativeLinePragmas = host.GetHostOption ("UseRelativeLinePragmas") as bool? ?? false;
 			
 			foreach (Directive dt in pt.Directives) {
 				switch (dt.Name) {
@@ -217,12 +218,29 @@ namespace Mono.TextTemplating
 					}
 					val = dt.Extract ("hostspecific");
 					if (val != null) {
-						settings.HostSpecific = string.Compare (val, "true", StringComparison.OrdinalIgnoreCase) == 0;
+						if (string.Compare (val, "trueFromBase", StringComparison.OrdinalIgnoreCase) == 0) {
+							settings.HostPropertyOnBase = true;
+							settings.HostSpecific = true;
+						} else {
+							settings.HostSpecific = string.Compare (val, "true", StringComparison.OrdinalIgnoreCase) == 0;
+						}
 					}
 					val = dt.Extract ("CompilerOptions");
 					if (val != null) {
 						settings.CompilerOptions = val;
 					}
+					val = dt.Extract ("relativeLinePragmas");
+					if (val != null) {
+						relativeLinePragmas = string.Compare (val, "true", StringComparison.OrdinalIgnoreCase) == 0;
+					}
+					val = dt.Extract ("linePragmas");
+					if (val != null) {
+						settings.NoLinePragmas = string.Compare (val, "false", StringComparison.OrdinalIgnoreCase) == 0;
+					}
+					val = dt.Extract ("visibility");
+					if (val != null) {
+						settings.InternalVisibility = string.Compare (val, "internal", StringComparison.OrdinalIgnoreCase) == 0;
+					}
 					break;
 					
 				case "assembly":
@@ -256,7 +274,6 @@ namespace Mono.TextTemplating
 					continue;
 					
 				default:
-					bool isParameterProcessor = dt.Name == "parameter";
 					string processorName = dt.Extract ("Processor");
 					if (processorName == null)
 						throw new InvalidOperationException ("Custom directive '" + dt.Name + "' does not specify a processor");
@@ -270,18 +287,26 @@ namespace Mono.TextTemplating
 			//initialize the custom processors
 			foreach (var kv in settings.DirectiveProcessors) {
 				kv.Value.Initialize (host);
-				var hs = kv.Value as IRecognizeHostSpecific;
-				if (hs == null)
+
+				IRecognizeHostSpecific hs;
+				if (settings.HostSpecific || (
+				        !((IDirectiveProcessor)kv.Value).RequiresProcessingRunIsHostSpecific &&
+				        ((hs = kv.Value as IRecognizeHostSpecific) == null || !hs.RequiresProcessingRunIsHostSpecific)))
 					continue;
-				if (hs.RequiresProcessingRunIsHostSpecific && !settings.HostSpecific) {
-					settings.HostSpecific = true;
-					pt.LogWarning ("Directive processor '" + kv.Key + "' requires hostspecific=true, forcing on.");
-				}
-				hs.SetProcessingRunIsHostSpecific (settings.HostSpecific);
+
+				settings.HostSpecific = true;
+				pt.LogWarning ("Directive processor '" + kv.Key + "' requires hostspecific=true, forcing on.");
+			}
+
+			foreach (var kv in settings.DirectiveProcessors) {
+				((IDirectiveProcessor)kv.Value).SetProcessingRunIsHostSpecific (settings.HostSpecific);
+				var hs = kv.Value as IRecognizeHostSpecific;
+				if (hs != null)
+					hs.SetProcessingRunIsHostSpecific (settings.HostSpecific);
 			}
 			
 			if (settings.Name == null)
-				settings.Name = string.Format ("GeneratedTextTransformation{0:x}", new System.Random ().Next ());
+				settings.Name = string.Format ("GeneratedTextTransformation{0:x}", new Random ().Next ());
 			if (settings.Namespace == null)
 				settings.Namespace = typeof (TextTransformation).Namespace;
 			
@@ -292,7 +317,7 @@ namespace Mono.TextTemplating
 			}
 			
 			if (settings.Language == "C#v3.5") {
-				Dictionary<string, string> providerOptions = new Dictionary<string, string> ();
+				var providerOptions = new Dictionary<string, string> ();
 				providerOptions.Add ("CompilerVersion", "v3.5");
 				settings.Provider = new CSharpCodeProvider (providerOptions);
 			}
@@ -304,6 +329,8 @@ namespace Mono.TextTemplating
 				pt.LogError ("A provider could not be found for the language '" + settings.Language + "'");
 				return settings;
 			}
+
+			settings.RelativeLinePragmas = relativeLinePragmas;
 			
 			return settings;
 		}
@@ -364,7 +391,7 @@ namespace Mono.TextTemplating
 		{
 			if (dt.Attributes.Count == 0)
 				return false;
-			StringBuilder sb = new StringBuilder ("Unknown attributes ");
+			var sb = new StringBuilder ("Unknown attributes ");
 			bool first = true;
 			foreach (string key in dt.Attributes.Keys) {
 				if (!first) {
@@ -381,7 +408,7 @@ namespace Mono.TextTemplating
 			return false;
 		}
 		
-		static void ProcessDirectives (ITextTemplatingEngineHost host, string content, ParsedTemplate pt, TemplateSettings settings)
+		static void ProcessDirectives (string content, ParsedTemplate pt, TemplateSettings settings)
 		{
 			foreach (var processor in settings.DirectiveProcessors.Values) {
 				processor.StartProcessingRun (settings.Provider, content, pt.Errors);
@@ -389,10 +416,6 @@ namespace Mono.TextTemplating
 			
 			foreach (var dt in settings.CustomDirectives) {
 				var processor = settings.DirectiveProcessors[dt.ProcessorName];
-				
-				if (processor is RequiresProvidesDirectiveProcessor)
-					throw new NotImplementedException ("RequiresProvidesDirectiveProcessor");
-				
 				processor.ProcessDirective (dt.Directive.Name, dt.Directive.Attributes);
 			}
 			
@@ -410,7 +433,9 @@ namespace Mono.TextTemplating
 		
 		public static CodeCompileUnit GenerateCompileUnit (ITextTemplatingEngineHost host, string content, ParsedTemplate pt, TemplateSettings settings)
 		{
-			ProcessDirectives (host, content, pt, settings);
+			ProcessDirectives (content, pt, settings);
+
+			string baseDirectory = Path.GetDirectoryName (host.TemplateFile);
 
 			//prep the compile unit
 			var ccu = new CodeCompileUnit ();
@@ -423,6 +448,9 @@ namespace Mono.TextTemplating
 			//prep the type
 			var type = new CodeTypeDeclaration (settings.Name);
 			type.IsPartial = true;
+			if (settings.InternalVisibility) {
+				type.TypeAttributes = (type.TypeAttributes & ~TypeAttributes.VisibilityMask) | TypeAttributes.NotPublic;
+			}
 			if (!string.IsNullOrEmpty (settings.Inherits)) {
 				type.BaseTypes.Add (new CodeTypeReference (settings.Inherits));
 			} else if (!settings.IncludePreprocessingHelpers) {
@@ -433,7 +461,7 @@ namespace Mono.TextTemplating
 			namespac.Types.Add (type);
 			
 			//prep the transform method
-			var transformMeth = new CodeMemberMethod () {
+			var transformMeth = new CodeMemberMethod {
 				Name = "TransformText",
 				ReturnType = new CodeTypeReference (typeof (String)),
 				Attributes = MemberAttributes.Public,
@@ -461,7 +489,13 @@ namespace Mono.TextTemplating
 			//build the code from the segments
 			foreach (TemplateSegment seg in pt.Content) {
 				CodeStatement st = null;
-				var location = new CodeLinePragma (seg.StartLocation.FileName ?? host.TemplateFile, seg.StartLocation.Line);
+				CodeLinePragma location = null;
+				if (!settings.NoLinePragmas) {
+					var f = seg.StartLocation.FileName ?? host.TemplateFile;
+					if (settings.RelativeLinePragmas)
+						f = FileUtil.AbsoluteToRelativePath (baseDirectory, f).Replace ('\\', '/');
+					location = new CodeLinePragma (f, seg.StartLocation.Line);
+				}
 				switch (seg.Type) {
 				case SegmentType.Block:
 					if (helperMode)
@@ -478,7 +512,8 @@ namespace Mono.TextTemplating
 					st = new CodeExpressionStatement (new CodeMethodInvokeExpression (writeMeth, new CodePrimitiveExpression (seg.Text)));
 					break;
 				case SegmentType.Helper:
-					type.Members.Add (new CodeSnippetTypeMember (seg.Text) { LinePragma = location });
+					if (!string.IsNullOrEmpty (seg.Text))
+						type.Members.Add (CreateSnippetMember (seg.Text, location));
 					helperMode = true;
 					break;
 				default:
@@ -490,7 +525,9 @@ namespace Mono.TextTemplating
 						//TODO: is there a way to do this for languages that use indentation for blocks, e.g. python?
 						using (var writer = new StringWriter ()) {
 							settings.Provider.GenerateCodeFromStatement (st, writer, null);
-							type.Members.Add (new CodeSnippetTypeMember (writer.ToString ()) { LinePragma = location });
+							var text = writer.ToString ();
+							if (!string.IsNullOrEmpty (text))
+								type.Members.Add (CreateSnippetMember (text, location));
 						}
 					} else {
 						st.LinePragma = location;
@@ -509,16 +546,22 @@ namespace Mono.TextTemplating
 						"ToString")));
 			type.Members.Add (transformMeth);
 			
-			//class code from processors
+			//class code and attributes from processors
 			foreach (var processor in settings.DirectiveProcessors.Values) {
 				string classCode = processor.GetClassCodeForProcessingRun ();
-				if (classCode != null)
-					type.Members.Add (new CodeSnippetTypeMember (classCode));
+				if (!string.IsNullOrEmpty (classCode))
+					type.Members.Add (CreateSnippetMember (classCode));
+				var atts = processor.GetTemplateClassCustomAttributes ();
+				if (atts != null) {
+					if (type.CustomAttributes == null)
+						type.CustomAttributes = new CodeAttributeDeclarationCollection ();
+					type.CustomAttributes.AddRange (atts);
+				}
 			}
 			
 			//generate the Host property if needed
-			if (settings.HostSpecific) {
-				GenerateHostProperty (type, settings);
+			if (settings.HostSpecific && !settings.HostPropertyOnBase) {
+				GenerateHostProperty (type);
 			}
 			
 			GenerateInitializationMethod (type, settings);
@@ -531,8 +574,20 @@ namespace Mono.TextTemplating
 			}
 			return ccu;
 		}
+
+		static CodeSnippetTypeMember CreateSnippetMember (string value, CodeLinePragma location = null)
+		{
+			//HACK: workaround for code generator not indenting first line of member snippet when inserting into class
+			const string indent = "\n        ";
+			if (!char.IsWhiteSpace (value[0]))
+				value = indent + value;
+
+			return new CodeSnippetTypeMember (value) {
+				LinePragma = location
+			};
+		}
 		
-		static void GenerateHostProperty (CodeTypeDeclaration type, TemplateSettings settings)
+		static void GenerateHostProperty (CodeTypeDeclaration type)
 		{
 			var hostField = new CodeMemberField (TypeRef<ITextTemplatingEngineHost> (), "hostValue");
 			hostField.Attributes = (hostField.Attributes & ~MemberAttributes.AccessMask) | MemberAttributes.Private;
@@ -546,7 +601,7 @@ namespace Mono.TextTemplating
 		static void GenerateInitializationMethod (CodeTypeDeclaration type, TemplateSettings settings)
 		{
 			//initialization method
-			var initializeMeth = new CodeMemberMethod () {
+			var initializeMeth = new CodeMemberMethod {
 				Name = "Initialize",
 				ReturnType = new CodeTypeReference (typeof (void), CodeTypeReferenceOptions.GlobalReference),
 				Attributes = MemberAttributes.Family
@@ -583,7 +638,6 @@ namespace Mono.TextTemplating
 		static void GenerateProcessingHelpers (CodeTypeDeclaration type, TemplateSettings settings)
 		{
 			var thisRef = new CodeThisReferenceExpression ();
-			var stringTypeRef = TypeRef<string> ();
 			var sbTypeRef = TypeRef<StringBuilder> ();
 			
 			var sessionField = PrivateField (TypeRef<IDictionary<string,object>> (), "session");
@@ -601,9 +655,9 @@ namespace Mono.TextTemplating
 			type.Members.Add (sessionProp);
 			type.Members.Add (generationEnvironmentProp);
 			
-			AddErrorHelpers (type, settings);
-			AddIndentHelpers (type, settings);
-			AddWriteHelpers (type, settings);
+			AddErrorHelpers (type);
+			AddIndentHelpers (type);
+			AddWriteHelpers (type);
 		}
 		
 		static void AddPropertyGetterInitializationIfFieldIsNull (CodeMemberProperty property, CodeFieldReferenceExpression fieldRef, CodeTypeReference typeRef)
@@ -612,7 +666,7 @@ namespace Mono.TextTemplating
 			property.GetStatements.Insert (0, fieldInit);
 		}
 
-		static CodeConditionStatement FieldInitializationIfNull (CodeFieldReferenceExpression fieldRef, CodeTypeReference typeRef)
+		static CodeConditionStatement FieldInitializationIfNull (CodeExpression fieldRef, CodeTypeReference typeRef)
 		{
 			return new CodeConditionStatement (
 				new CodeBinaryOperatorExpression (fieldRef,
@@ -620,7 +674,7 @@ namespace Mono.TextTemplating
 				new CodeAssignStatement (fieldRef, new CodeObjectCreateExpression (typeRef)));
 		}
 		
-		static void AddErrorHelpers (CodeTypeDeclaration type, TemplateSettings settings)
+		static void AddErrorHelpers (CodeTypeDeclaration type)
 		{
 			var cecTypeRef = TypeRef<CompilerErrorCollection> ();
 			var thisRef = new CodeThisReferenceExpression ();
@@ -638,7 +692,7 @@ namespace Mono.TextTemplating
 			var errorsPropRef = new CodePropertyReferenceExpression (new CodeThisReferenceExpression (), "Errors");
 			
 			var compilerErrorTypeRef = TypeRef<CompilerError> ();
-			var errorMeth = new CodeMemberMethod () {
+			var errorMeth = new CodeMemberMethod {
 				Name = "Error",
 				Attributes = MemberAttributes.Public | MemberAttributes.Final,
 			};
@@ -647,7 +701,7 @@ namespace Mono.TextTemplating
 				new CodeObjectCreateExpression (compilerErrorTypeRef, nullPrim, minusOnePrim, minusOnePrim, nullPrim,
 					new CodeArgumentReferenceExpression ("message"))));
 			
-			var warningMeth = new CodeMemberMethod () {
+			var warningMeth = new CodeMemberMethod {
 				Name = "Warning",
 				Attributes = MemberAttributes.Public | MemberAttributes.Final,
 			};
@@ -666,7 +720,7 @@ namespace Mono.TextTemplating
 			type.Members.Add (errorsProp);
 		}
 		
-		static void AddIndentHelpers (CodeTypeDeclaration type, TemplateSettings settings)
+		static void AddIndentHelpers (CodeTypeDeclaration type)
 		{
 			var stringTypeRef = TypeRef<string> ();
 			var thisRef = new CodeThisReferenceExpression ();
@@ -688,7 +742,7 @@ namespace Mono.TextTemplating
 			currentIndentField.InitExpression = stringEmptyRef;
 			var currentIndentFieldRef = new CodeFieldReferenceExpression (thisRef, currentIndentField.Name);
 			
-			var popIndentMeth = new CodeMemberMethod () {
+			var popIndentMeth = new CodeMemberMethod {
 				Name = "PopIndent",
 				ReturnType = stringTypeRef,
 				Attributes = MemberAttributes.Public | MemberAttributes.Final,
@@ -708,7 +762,7 @@ namespace Mono.TextTemplating
 				new CodeMethodInvokeExpression (currentIndentFieldRef, "Substring", zeroPrim, new CodeVariableReferenceExpression ("lastPos"))));
 			popIndentMeth.Statements.Add (new CodeMethodReturnStatement (new CodeVariableReferenceExpression ("last")));
 			
-			var pushIndentMeth = new CodeMemberMethod () {
+			var pushIndentMeth = new CodeMemberMethod {
 				Name = "PushIndent",
 				Attributes = MemberAttributes.Public | MemberAttributes.Final,
 			};
@@ -718,7 +772,7 @@ namespace Mono.TextTemplating
 			pushIndentMeth.Statements.Add (new CodeAssignStatement (currentIndentFieldRef,
 				new CodeBinaryOperatorExpression (currentIndentFieldRef, CodeBinaryOperatorType.Add, new CodeArgumentReferenceExpression ("indent"))));
 			
-			var clearIndentMeth = new CodeMemberMethod () {
+			var clearIndentMeth = new CodeMemberMethod {
 				Name = "ClearIndent",
 				Attributes = MemberAttributes.Public | MemberAttributes.Final,
 			};
@@ -735,7 +789,7 @@ namespace Mono.TextTemplating
 			type.Members.Add (indentsProp);
 		}
 		
-		static void AddWriteHelpers (CodeTypeDeclaration type, TemplateSettings settings)
+		static void AddWriteHelpers (CodeTypeDeclaration type)
 		{
 			var stringTypeRef = TypeRef<string> ();
 			var thisRef = new CodeThisReferenceExpression ();
@@ -745,20 +799,20 @@ namespace Mono.TextTemplating
 			var textToAppendParam = new CodeParameterDeclarationExpression (stringTypeRef, "textToAppend");
 			var formatParam = new CodeParameterDeclarationExpression (stringTypeRef, "format");
 			var argsParam = new CodeParameterDeclarationExpression (TypeRef<object[]> (), "args");
-			argsParam.CustomAttributes.Add (new CodeAttributeDeclaration (TypeRef<System.ParamArrayAttribute> ()));
+			argsParam.CustomAttributes.Add (new CodeAttributeDeclaration (TypeRef<ParamArrayAttribute> ()));
 			
 			var textToAppendParamRef = new CodeArgumentReferenceExpression ("textToAppend");
 			var formatParamRef = new CodeArgumentReferenceExpression ("format");
 			var argsParamRef = new CodeArgumentReferenceExpression ("args");
 			
-			var writeMeth = new CodeMemberMethod () {
+			var writeMeth = new CodeMemberMethod {
 				Name = "Write",
 				Attributes = MemberAttributes.Public | MemberAttributes.Final,
 			};
 			writeMeth.Parameters.Add (textToAppendParam);
 			writeMeth.Statements.Add (new CodeMethodInvokeExpression (genEnvPropRef, "Append", new CodeArgumentReferenceExpression ("textToAppend")));
 			
-			var writeArgsMeth = new CodeMemberMethod () {
+			var writeArgsMeth = new CodeMemberMethod {
 				Name = "Write",
 				Attributes = MemberAttributes.Public | MemberAttributes.Final,
 			};
@@ -766,7 +820,7 @@ namespace Mono.TextTemplating
 			writeArgsMeth.Parameters.Add (argsParam);
 			writeArgsMeth.Statements.Add (new CodeMethodInvokeExpression (genEnvPropRef, "AppendFormat", formatParamRef, argsParamRef));
 			
-			var writeLineMeth = new CodeMemberMethod () {
+			var writeLineMeth = new CodeMemberMethod {
 				Name = "WriteLine",
 				Attributes = MemberAttributes.Public | MemberAttributes.Final,
 			};
@@ -774,7 +828,7 @@ namespace Mono.TextTemplating
 			writeLineMeth.Statements.Add (new CodeMethodInvokeExpression (genEnvPropRef, "Append", currentIndentFieldRef));
 			writeLineMeth.Statements.Add (new CodeMethodInvokeExpression (genEnvPropRef, "AppendLine", textToAppendParamRef));
 			
-			var writeLineArgsMeth = new CodeMemberMethod () {
+			var writeLineArgsMeth = new CodeMemberMethod {
 				Name = "WriteLine",
 				Attributes = MemberAttributes.Public | MemberAttributes.Final,
 			};
@@ -794,7 +848,7 @@ namespace Mono.TextTemplating
 		{
 			var helperCls = new CodeTypeDeclaration ("ToStringInstanceHelper") {
 				IsClass = true,
-				TypeAttributes = System.Reflection.TypeAttributes.NestedPublic,
+				TypeAttributes = TypeAttributes.NestedPublic,
 			};
 			
 			var formatProviderField = PrivateField (TypeRef<IFormatProvider> (), "formatProvider");
@@ -808,7 +862,7 @@ namespace Mono.TextTemplating
 			helperCls.Members.Add (formatProviderField);
 			helperCls.Members.Add (formatProviderProp);
 			
-			var meth = new CodeMemberMethod () {
+			var meth = new CodeMemberMethod {
 				Name = "ToStringWithCulture",
 				Attributes = MemberAttributes.Public | MemberAttributes.Final,
 				ReturnType = TypeRef<string> (),
@@ -866,7 +920,7 @@ namespace Mono.TextTemplating
 		
 		static CodeMemberProperty GenerateGetterSetterProperty (string propertyName, CodeMemberField field)
 		{
-			var prop = new CodeMemberProperty () {
+			var prop = new CodeMemberProperty {
 				Name = propertyName,
 				Attributes = MemberAttributes.Public | MemberAttributes.Final,
 				Type = field.Type
@@ -879,7 +933,7 @@ namespace Mono.TextTemplating
 		
 		static CodeMemberProperty GenerateGetterProperty (string propertyName, CodeMemberField field)
 		{
-			var prop = new CodeMemberProperty () {
+			var prop = new CodeMemberProperty {
 				Name = propertyName,
 				Attributes = MemberAttributes.Public | MemberAttributes.Final,
 				HasSet = false,
