@@ -25,7 +25,6 @@
 // THE SOFTWARE.
 
 using System;
-using System.IO;
 using Gtk;
 using MonoDevelop.Core;
 using MonoDevelop.Ide.Gui;
@@ -39,7 +38,7 @@ namespace MonoDevelop.VersionControl.Views
 	[System.ComponentModel.ToolboxItem(true)]
 	public partial class LogWidget : Gtk.Bin
 	{
-		MonoDevelop.VersionControl.Revision[] history;
+		Revision[] history;
 		public Revision[] History {
 			get {
 				return history;
@@ -50,17 +49,10 @@ namespace MonoDevelop.VersionControl.Views
 			}
 		}
 		
-		public Toolbar CommandBar {
-			get {
-				return commandBar;
-			}
-		}
-		
-		
 		ListStore logstore = new ListStore (typeof (Revision));
 		FileTreeView treeviewFiles;
 		TreeStore changedpathstore;
-		Gtk.ToolButton revertButton, revertToButton;
+		Gtk.Button revertButton, revertToButton, refreshButton;
 		SearchEntry searchEntry;
 		string currentFilter;
 		
@@ -69,7 +61,7 @@ namespace MonoDevelop.VersionControl.Views
 		CellRendererDiff diffRenderer = new CellRendererDiff ();
 		CellRendererText messageRenderer = new CellRendererText ();
 		CellRendererText textRenderer = new CellRendererText ();
-		CellRendererPixbuf pixRenderer = new CellRendererPixbuf ();
+		CellRendererImage pixRenderer = new CellRendererImage ();
 		
 		bool currentRevisionShortened;
 		
@@ -99,7 +91,7 @@ namespace MonoDevelop.VersionControl.Views
 					double center_x = cell_area.X + Math.Round ((double) (cell_area.Width / 2d));
 					double center_y = cell_area.Y + Math.Round ((double) (cell_area.Height / 2d));
 					cr.Arc (center_x, center_y, 5, 0, 2 * Math.PI);
-					cr.Color = new Cairo.Color (0, 0, 0);
+					cr.SetSourceRGBA (0, 0, 0, 1);
 					cr.Stroke ();
 					if (!FirstNode) {
 						cr.MoveTo (center_x, cell_area.Y - 2);
@@ -122,22 +114,26 @@ namespace MonoDevelop.VersionControl.Views
 			this.info = info;
 			if (info.Document != null)
 				this.preselectFile = info.Item.Path;
+
+			var separator = new HeaderBox ();
+			separator.SetMargins (1, 0, 0, 0);
+			separator.HeightRequest = 4;
+			separator.ShowAll ();
 			
-			revertButton = new Gtk.ToolButton (new Gtk.Image ("vc-revert-command", Gtk.IconSize.Menu), GettextCatalog.GetString ("Revert changes from this revision"));
-			revertButton.IsImportant = true;
-//			revertButton.Sensitive = false;
+			hpaned1 = hpaned1.ReplaceWithWidget (new HPanedThin (), true);
+			vpaned1 = vpaned1.ReplaceWithWidget (new VPanedThin () { HandleWidget = separator }, true);
+
+			revertButton = new DocumentToolButton ("vc-revert-command", GettextCatalog.GetString ("Revert changes from this revision"));
+			revertButton.Sensitive = false;
 			revertButton.Clicked += new EventHandler (RevertRevisionClicked);
-			CommandBar.Insert (revertButton, -1);
-			
-			revertToButton = new Gtk.ToolButton (new Gtk.Image ("vc-revert-command", Gtk.IconSize.Menu), GettextCatalog.GetString ("Revert to this revision"));
-			revertToButton.IsImportant = true;
-//			revertToButton.Sensitive = false;
+
+			revertToButton = new DocumentToolButton ("vc-revert-command", GettextCatalog.GetString ("Revert to this revision"));
+			revertToButton.Sensitive = false;
 			revertToButton.Clicked += new EventHandler (RevertToRevisionClicked);
-			CommandBar.Insert (revertToButton, -1);
-			
-			Gtk.ToolItem item = new Gtk.ToolItem ();
-			Gtk.HBox a = new Gtk.HBox ();
-			item.Add (a);
+
+			refreshButton = new DocumentToolButton (Gtk.Stock.Refresh, GettextCatalog.GetString ("Refresh"));
+			refreshButton.Clicked += new EventHandler (RefreshClicked);
+
 			searchEntry = new SearchEntry ();
 			searchEntry.WidthRequest = 200;
 			searchEntry.ForceFilterButtonVisible = true;
@@ -145,12 +141,7 @@ namespace MonoDevelop.VersionControl.Views
 			searchEntry.Changed += HandleSearchEntryFilterChanged;
 			searchEntry.Ready = true;
 			searchEntry.Show ();
-			a.PackEnd (searchEntry, false, false, 0);
-			CommandBar.Insert (item, -1);
-			((Gtk.Toolbar.ToolbarChild)CommandBar[item]).Expand = true;
-			
-			CommandBar.ShowAll ();
-			
+
 			messageRenderer.Ellipsize = Pango.EllipsizeMode.End;
 			TreeViewColumn colRevMessage = new TreeViewColumn ();
 			colRevMessage.Title = GettextCatalog.GetString ("Message");
@@ -194,20 +185,20 @@ namespace MonoDevelop.VersionControl.Views
 			scrolledwindowFiles.Child = treeviewFiles;
 			scrolledwindowFiles.ShowAll ();
 			
-			changedpathstore = new TreeStore (typeof(Gdk.Pixbuf), typeof (string), // icon/file name
-				typeof(Gdk.Pixbuf), typeof (string), // icon/operation
+			changedpathstore = new TreeStore (typeof(Xwt.Drawing.Image), typeof (string), // icon/file name
+			                                  typeof(Xwt.Drawing.Image), typeof (string), // icon/operation
 				typeof (string), // path
 				typeof (string), // revision path (invisible)
-				typeof (string[]) // diff
+				typeof (string []) // diff
 				);
 			
 			TreeViewColumn colChangedFile = new TreeViewColumn ();
-			var crp = new CellRendererPixbuf ();
+			var crp = new CellRendererImage ();
 			var crt = new CellRendererText ();
 			colChangedFile.Title = GettextCatalog.GetString ("File");
 			colChangedFile.PackStart (crp, false);
 			colChangedFile.PackStart (crt, true);
-			colChangedFile.AddAttribute (crp, "pixbuf", 2);
+			colChangedFile.AddAttribute (crp, "image", 2);
 			colChangedFile.AddAttribute (crt, "text", 3);
 			treeviewFiles.AppendColumn (colChangedFile);
 			
@@ -215,7 +206,7 @@ namespace MonoDevelop.VersionControl.Views
 			colOperation.Title = GettextCatalog.GetString ("Operation");
 			colOperation.PackStart (crp, false);
 			colOperation.PackStart (crt, true);
-			colOperation.AddAttribute (crp, "pixbuf", 0);
+			colOperation.AddAttribute (crp, "image", 0);
 			colOperation.AddAttribute (crt, "text", 1);
 			treeviewFiles.AppendColumn (colOperation);
 			
@@ -235,6 +226,44 @@ namespace MonoDevelop.VersionControl.Views
 			labelAuthor.Text = "";
 			labelDate.Text = "";
 			labelRevision.Text = "";
+
+			vbox2.Remove (scrolledwindow1);
+			HeaderBox tb = new HeaderBox ();
+			tb.Show ();
+			tb.SetMargins (1, 0, 0, 0);
+			tb.ShowTopShadow = true;
+			tb.ShadowSize = 4;
+			tb.SetPadding (8, 8, 8, 8);
+			tb.UseChildBackgroundColor = true;
+			tb.Add (scrolledwindow1);
+			vbox2.PackStart (tb, true, true, 0);
+		}
+
+		protected override void OnRealized ()
+		{
+			base.OnRealized ();
+			var c = Style.Base (StateType.Normal).ToXwtColor ();
+			c.Light *= 0.8;
+			commitBox.ModifyBg (StateType.Normal, c.ToGdkColor ());
+
+			var tcol = new Gdk.Color (255, 251, 242);
+			textviewDetails.ModifyBase (StateType.Normal, tcol);
+			scrolledwindow1.ModifyBase (StateType.Normal, tcol);
+		}
+
+		internal void SetToolbar (DocumentToolbar toolbar)
+		{
+			if (info.Repository.SupportsRevertRevision)
+				toolbar.Add (revertButton);
+
+			toolbar.Add (revertToButton);
+			toolbar.Add (refreshButton);
+
+			Gtk.HBox a = new Gtk.HBox ();
+			a.PackEnd (searchEntry, false, false, 0);
+			toolbar.Add (a, true);
+
+			toolbar.ShowAll ();
 		}
 		
 		bool filtering;
@@ -257,20 +286,29 @@ namespace MonoDevelop.VersionControl.Views
 			scrolledLog.Hide ();
 		}
 		
-		void RevertToRevisionClicked (object src, EventArgs args) {
+		void RevertToRevisionClicked (object src, EventArgs args)
+		{
 			Revision d = SelectedRevision;
 			if (RevertRevisionsCommands.RevertToRevision (info.Repository, info.Item.Path, d, false))
 				VersionControlService.SetCommitComment (info.Item.Path, 
 				  string.Format ("(Revert to revision {0})", d.ToString ()), true);
 		}
 		
-		void RevertRevisionClicked (object src, EventArgs args) {
+		void RevertRevisionClicked (object src, EventArgs args)
+		{
 			Revision d = SelectedRevision;
 			if (RevertRevisionsCommands.RevertRevision (info.Repository, info.Item.Path, d, false))
 				VersionControlService.SetCommitComment (info.Item.Path, 
 				  string.Format ("(Revert revision {0})", d.ToString ()), true);
 		}
 
+		void RefreshClicked (object src, EventArgs args)
+		{
+			ShowLoading ();
+			info.Start (true);
+			revertButton.Sensitive = revertToButton.Sensitive = false;
+		}
+
 		void HandleTreeviewFilesDiffLineActivated (object sender, EventArgs e)
 		{
 			TreePath[] paths = treeviewFiles.Selection.GetSelectedRows ();
@@ -283,6 +321,8 @@ namespace MonoDevelop.VersionControl.Views
 			
 			string fileName = (string)changedpathstore.GetValue (iter, colPath);
 			int line = diffRenderer.GetSelectedLine (paths[0]);
+			if (line == -1)
+				line = 1;
 			var doc = IdeApp.Workbench.OpenDocument (fileName, line, 0, OpenDocumentOptions.Default | OpenDocumentOptions.OnlyInternalViewer);
 			int i = 1;
 			foreach (var content in doc.Window.SubViewContents) {
@@ -301,23 +341,25 @@ namespace MonoDevelop.VersionControl.Views
 				i++;
 			}
 		}
-		
+
+		const int colOperation = 4;
 		const int colPath = 5;
 		const int colDiff = 6;
 		
 		void HandleTreeviewFilesTestExpandRow (object o, TestExpandRowArgs args)
 		{
-			string[] diff = changedpathstore.GetValue (args.Iter, colDiff) as string[];
-			if (diff != null) {
-				return;
-			}
 			TreeIter iter;
 			if (changedpathstore.IterChildren (out iter, args.Iter)) {
-				string path = (string)changedpathstore .GetValue (args.Iter, colPath);
+				string[] diff = changedpathstore.GetValue (iter, colDiff) as string[];
+				if (diff != null)
+					return;
+
+				string path = (string)changedpathstore.GetValue (args.Iter, colPath);
+
 				changedpathstore.SetValue (iter, colDiff, new string[] { GettextCatalog.GetString ("Loading data...") });
 				var rev = SelectedRevision;
 				ThreadPool.QueueUserWorkItem (delegate {
-					string text;
+					string text = "";
 					try {
 						text = info.Repository.GetTextAtRevision (path, rev);
 					} catch (Exception e) {
@@ -338,25 +380,42 @@ namespace MonoDevelop.VersionControl.Views
 						return;
 					}
 					string[] lines;
-					var changedDocument = new Mono.TextEditor.TextDocument (text);
-					if (prevRev == null) {
-						lines = new string[changedDocument.LineCount];
-						for (int i = 0; i < changedDocument.LineCount; i++) {
-							lines[i] = "+ " + changedDocument.GetLineText (i + 1).TrimEnd ('\r','\n');
-						}
-						
+					// Indicator that the file was binary
+					if (text == null) {
+						lines = new [] { " Binary files differ" };
 					} else {
-						string prevRevisionText = "";
-						try {
-							prevRevisionText = info.Repository.GetTextAtRevision (path, prevRev);
-						} catch (Exception e) {
-							// The file did not exist at this point in time, so just treat it as empty
+						var changedDocument = new Mono.TextEditor.TextDocument (text);
+						if (prevRev == null) {
+							lines = new string [changedDocument.LineCount];
+							for (int i = 0; i < changedDocument.LineCount; i++) {
+								lines[i] = "+ " + changedDocument.GetLineText (i + 1).TrimEnd ('\r','\n');
+							}
+						} else {
+							string prevRevisionText = "";
+							try {
+								prevRevisionText = info.Repository.GetTextAtRevision (path, prevRev);
+							} catch (Exception e) {
+								Application.Invoke (delegate {
+									LoggingService.LogError ("Error while getting revision text", e);
+									MessageService.ShowError ("Error while getting revision text.", "The file may not be part of the working copy.");
+								});
+								return;
+							}
+
+							if (String.IsNullOrEmpty (text)) {
+								if (!String.IsNullOrEmpty (prevRevisionText)) {
+									lines = new string [changedDocument.LineCount];
+									for (int i = 0; i < changedDocument.LineCount; i++) {
+										lines [i] = "- " + changedDocument.GetLineText (i + 1).TrimEnd ('\r','\n');
+									}
+								}
+							}
+
+							var originalDocument = new Mono.TextEditor.TextDocument (prevRevisionText);
+							originalDocument.FileName = "Revision " + prevRev;
+							changedDocument.FileName = "Revision " + rev;
+							lines = Mono.TextEditor.Utils.Diff.GetDiffString (originalDocument, changedDocument).Split ('\n');
 						}
-						
-						var originalDocument = new Mono.TextEditor.TextDocument (prevRevisionText);
-						originalDocument.FileName = "Revision " + prevRev.ToString ();
-						changedDocument.FileName = "Revision " + rev.ToString ();
-						lines = Mono.TextEditor.Utils.Diff.GetDiffString (originalDocument, changedDocument).Split ('\n');
 					}
 					Application.Invoke (delegate {
 						changedpathstore.SetValue (iter, colDiff, lines);
@@ -376,7 +435,7 @@ namespace MonoDevelop.VersionControl.Views
 			TreeIter iter;
 			if (!treeviewFiles.Selection.GetSelected (out iter))
 				return;
-			string path = (string)changedpathstore.GetValue (iter, 5);
+			string path = (string)changedpathstore.GetValue (iter, colPath);
 			ThreadPool.QueueUserWorkItem (delegate {
 				string text = info.Repository.GetTextAtRevision (path, rev);
 				string prevRevision = text; // info.Repository.GetTextAtRevision (path, rev.GetPrevious ());
@@ -407,9 +466,9 @@ namespace MonoDevelop.VersionControl.Views
 			var rev = (Revision)model.GetValue (iter, 0);
 			string day;
 			var age = rev.Time - DateTime.Now;
-			if (age.TotalDays == 0) {
+			if (age.Days == 0) {
 				day = GettextCatalog.GetString ("Today");
-			} else if (age.TotalDays == 1) {
+			} else if (age.Days == 1) {
 				day = GettextCatalog.GetString ("Yesterday");
 			} else {
 				day = rev.Time.ToShortDateString ();
@@ -454,8 +513,8 @@ namespace MonoDevelop.VersionControl.Views
 			string author = rev.Author;
 			if (string.IsNullOrEmpty (author))
 				return;
-			int idx = author.IndexOf ("<");
-			if (idx >= 0 && idx < author.IndexOf (">"))
+			int idx = author.IndexOf ("<", StringComparison.Ordinal);
+			if (idx >= 0 && idx < author.IndexOf (">", StringComparison.Ordinal))
 				author = author.Substring (0, idx).Trim ();
 			if (string.IsNullOrEmpty (currentFilter))
 				renderer.Text = author;
@@ -465,19 +524,18 @@ namespace MonoDevelop.VersionControl.Views
 		
 		void AuthorIconFunc (Gtk.TreeViewColumn tree_column, Gtk.CellRenderer cell, Gtk.TreeModel model, Gtk.TreeIter iter)
 		{
-			CellRendererPixbuf renderer = (CellRendererPixbuf)cell;
+			CellRendererImage renderer = (CellRendererImage)cell;
 			var rev = (Revision)model.GetValue (iter, 0);
 			if (string.IsNullOrEmpty (rev.Email))
 				return;
 			ImageLoader img = ImageService.GetUserIcon (rev.Email, 16);
-			if (img.LoadOperation.IsCompleted)
-				renderer.Pixbuf = img.Pixbuf;
-			else {
-				img.LoadOperation.Completed += delegate {
-					Gtk.Application.Invoke (delegate {
-						if (logstore.IterIsValid (iter))
-							model.EmitRowChanged (model.GetPath (iter), iter);
-					});
+
+			renderer.Image = img.Image;
+			if (img.Downloading) {
+				img.Completed += (sender, e) => {
+					renderer.Image = img.Image;
+					if (logstore.IterIsValid (iter))
+						model.EmitRowChanged (model.GetPath (iter), iter);
 				};
 			}
 		}
@@ -497,7 +555,7 @@ namespace MonoDevelop.VersionControl.Views
 			CellRendererDiff rc = (CellRendererDiff)cell;
 			string[] lines = (string[])changedpathstore.GetValue (iter, colDiff);
 			if (lines == null)
-				lines = new string[] { (string)changedpathstore.GetValue (iter, 4) };
+				lines = new string[] { (string)changedpathstore.GetValue (iter, colOperation) };
 			rc.InitCell (treeviewFiles, changedpathstore.IterDepth (iter) != 0, lines, changedpathstore.GetPath (iter));
 		}
 		
@@ -540,28 +598,30 @@ namespace MonoDevelop.VersionControl.Views
 			
 			if (d == null)
 				return;
+
+			revertButton.Sensitive = revertToButton.Sensitive = true;
 			Gtk.TreeIter selectIter = Gtk.TreeIter.Zero;
 			bool select = false;
 			foreach (RevisionPath rp in info.Repository.GetRevisionChanges (d)) {
-				Gdk.Pixbuf actionIcon;
+				Xwt.Drawing.Image actionIcon;
 				string action = null;
 				if (rp.Action == RevisionAction.Add) {
 					action = GettextCatalog.GetString ("Add");
-					actionIcon = ImageService.GetPixbuf (Gtk.Stock.Add, Gtk.IconSize.Menu);
+					actionIcon = ImageService.GetIcon (Gtk.Stock.Add, Gtk.IconSize.Menu);
 				} else if (rp.Action == RevisionAction.Delete) {
 					action = GettextCatalog.GetString ("Delete");
-					actionIcon = ImageService.GetPixbuf (Gtk.Stock.Remove, Gtk.IconSize.Menu);
+					actionIcon = ImageService.GetIcon (Gtk.Stock.Remove, Gtk.IconSize.Menu);
 				} else if (rp.Action == RevisionAction.Modify) {
 					action = GettextCatalog.GetString ("Modify");
-					actionIcon = ImageService.GetPixbuf ("gtk-edit", Gtk.IconSize.Menu);
+					actionIcon = ImageService.GetIcon ("gtk-edit", Gtk.IconSize.Menu);
 				} else if (rp.Action == RevisionAction.Replace) {
 					action = GettextCatalog.GetString ("Replace");
-					actionIcon = ImageService.GetPixbuf ("gtk-edit", Gtk.IconSize.Menu);
+					actionIcon = ImageService.GetIcon ("gtk-edit", Gtk.IconSize.Menu);
 				} else {
 					action = rp.ActionDescription;
-					actionIcon = ImageService.GetPixbuf (MonoDevelop.Ide.Gui.Stock.Empty, Gtk.IconSize.Menu);
+					actionIcon = ImageService.GetIcon (MonoDevelop.Ide.Gui.Stock.Empty, Gtk.IconSize.Menu);
 				}
-				Gdk.Pixbuf fileIcon = DesktopService.GetPixbufForFile (rp.Path, Gtk.IconSize.Menu);
+				Xwt.Drawing.Image fileIcon = DesktopService.GetIconForFile (rp.Path, Gtk.IconSize.Menu);
 				var iter = changedpathstore.AppendValues (actionIcon, action, fileIcon, System.IO.Path.GetFileName (rp.Path), System.IO.Path.GetDirectoryName (rp.Path), rp.Path, null);
 				changedpathstore.AppendValues (iter, null, null, null, null, null, rp.Path, null);
 				if (rp.Path == preselectFile) {
@@ -588,8 +648,10 @@ namespace MonoDevelop.VersionControl.Views
 			labelRevision.Text = GettextCatalog.GetString ("revision: {0}", rev);
 			textviewDetails.Buffer.Text = d.Message;
 			
-			if (select)
+			if (select) {
 				treeviewFiles.Selection.SelectIter (selectIter);
+				treeviewFiles.ExpandRow (treeviewFiles.Model.GetPath (selectIter), true);
+			}
 		}
 		
 		void UpdateHistory ()
@@ -654,5 +716,17 @@ namespace MonoDevelop.VersionControl.Views
 				currentRevisionShortened = false;
 			}
 		}
+
+		internal string DiffText {
+			get {
+				TreeIter iter;
+				if (treeviewFiles.Selection.GetSelected (out iter)) {
+					string [] items = changedpathstore.GetValue (iter, colDiff) as string [];
+					if (items != null)
+						return String.Join (Environment.NewLine, items);
+				}
+				return null;
+			}
+		}
 	}
 }
