@@ -21,22 +21,22 @@ namespace MonoDevelop.VersionControl
 {
 	public class VersionControlService
 	{
-		static Gdk.Pixbuf overlay_modified;
-		static Gdk.Pixbuf overlay_removed;
-		static Gdk.Pixbuf overlay_conflicted;
-		static Gdk.Pixbuf overlay_added;
-		internal static Gdk.Pixbuf overlay_controled;
-		static Gdk.Pixbuf overlay_unversioned;
-		static Gdk.Pixbuf overlay_protected;
-		static Gdk.Pixbuf overlay_locked;
-		static Gdk.Pixbuf overlay_unlocked;
-//		static Gdk.Pixbuf overlay_normal;
+		static Xwt.Drawing.Image overlay_modified;
+		static Xwt.Drawing.Image overlay_removed;
+		static Xwt.Drawing.Image overlay_conflicted;
+		static Xwt.Drawing.Image overlay_added;
+		internal static Xwt.Drawing.Image overlay_controled;
+		static Xwt.Drawing.Image overlay_unversioned;
+		static Xwt.Drawing.Image overlay_protected;
+		static Xwt.Drawing.Image overlay_locked;
+		static Xwt.Drawing.Image overlay_unlocked;
+        static Xwt.Drawing.Image overlay_ignored;
 
-		static Gdk.Pixbuf icon_modified;
-		static Gdk.Pixbuf icon_removed;
-		static Gdk.Pixbuf icon_conflicted;
-		static Gdk.Pixbuf icon_added;
-		internal static Gdk.Pixbuf icon_controled;
+		static Xwt.Drawing.Image icon_modified;
+		static Xwt.Drawing.Image icon_removed;
+		static Xwt.Drawing.Image icon_conflicted;
+		static Xwt.Drawing.Image icon_added;
+		internal static Xwt.Drawing.Image icon_controled;
 		
 		static Hashtable comments;
 		static object commentsLock = new object ();
@@ -53,39 +53,44 @@ namespace MonoDevelop.VersionControl
 		
 		static VersionControlService ()
 		{
-			try {
-				overlay_modified = Gdk.Pixbuf.LoadFromResource("overlay_modified.png");
-				overlay_removed = Gdk.Pixbuf.LoadFromResource("overlay_removed.png");
-				overlay_conflicted = Gdk.Pixbuf.LoadFromResource("overlay_conflicted.png");
-				overlay_added = Gdk.Pixbuf.LoadFromResource("overlay_added.png");
-				overlay_controled = Gdk.Pixbuf.LoadFromResource("overlay_controled.png");
-				overlay_unversioned = Gdk.Pixbuf.LoadFromResource("overlay_unversioned.png");
-				overlay_protected = Gdk.Pixbuf.LoadFromResource("overlay_lock_required.png");
-				overlay_unlocked = Gdk.Pixbuf.LoadFromResource("overlay_unlocked.png");
-				overlay_locked = Gdk.Pixbuf.LoadFromResource("overlay_locked.png");
-	//			overlay_normal = Gdk.Pixbuf.LoadFromResource("overlay_normal.png");
-			
-				icon_modified = ImageService.GetPixbuf ("gtk-edit", Gtk.IconSize.Menu);
-				icon_removed = ImageService.GetPixbuf (Gtk.Stock.Remove, Gtk.IconSize.Menu);
-				icon_conflicted = ImageService.GetPixbuf (Gtk.Stock.DialogWarning, Gtk.IconSize.Menu);
-				icon_added = ImageService.GetPixbuf (Gtk.Stock.Add, Gtk.IconSize.Menu);
-				icon_controled = Gdk.Pixbuf.LoadFromResource("overlay_controled.png");
-			} catch (Exception e) {
-				LoggingService.LogError ("Error while loading icons.", e);
-			}
-			IdeApp.Workspace.FileAddedToProject += OnFileAdded;
-			//IdeApp.Workspace.FileChangedInProject += OnFileChanged;
-			//IdeApp.Workspace.FileRemovedFromProject += OnFileRemoved;
-			//IdeApp.Workspace.FileRenamedInProject += OnFileRenamed;
-			
-			IdeApp.Workspace.ItemAddedToSolution += OnEntryAdded;
-			IdeApp.Exiting += delegate {
-				DelayedSaveComments (null);
+			IdeApp.Initialized += delegate {
+				try {
+					overlay_modified = Xwt.Drawing.Image.FromResource("modified-overlay-light-16.png");
+					overlay_removed = Xwt.Drawing.Image.FromResource("removed-overlay-light-16.png");
+					overlay_conflicted = Xwt.Drawing.Image.FromResource("conflict-overlay-light-16.png");
+					overlay_added = Xwt.Drawing.Image.FromResource("added-overlay-light-16.png");
+					overlay_controled = Xwt.Drawing.Image.FromResource("versioned-overlay-light-16.png");
+					overlay_unversioned = Xwt.Drawing.Image.FromResource("unversioned-overlay-light-16.png");
+					overlay_protected = Xwt.Drawing.Image.FromResource("lock-required-overlay-light-16.png");
+					overlay_unlocked = Xwt.Drawing.Image.FromResource("unlocked-overlay-light-16.png");
+					overlay_locked = Xwt.Drawing.Image.FromResource("locked-overlay-light-16.png");
+					overlay_ignored = Xwt.Drawing.Image.FromResource("ignored-overlay-light-16.png");
+
+					icon_modified = ImageService.GetIcon ("vc-file-modified", Gtk.IconSize.Menu);
+					icon_removed = ImageService.GetIcon ("vc-file-removed", Gtk.IconSize.Menu);
+					icon_conflicted = ImageService.GetIcon ("vc-file-conflicted", Gtk.IconSize.Menu);
+					icon_added = ImageService.GetIcon ("vc-file-added", Gtk.IconSize.Menu);
+					icon_controled = Xwt.Drawing.Image.FromResource("versioned-overlay-light-16.png");
+				} catch (Exception e) {
+					LoggingService.LogError ("Error while loading icons.", e);
+				}
+
+				IdeApp.Workspace.SolutionLoaded += (sender, e) => SessionSolutionDisabled |= IsSolutionDisabled (e.Solution);
+
+				IdeApp.Workspace.FileAddedToProject += OnFileAdded;
+				//IdeApp.Workspace.FileChangedInProject += OnFileChanged;
+				//IdeApp.Workspace.FileRemovedFromProject += OnFileRemoved;
+				//IdeApp.Workspace.FileRenamedInProject += OnFileRenamed;
+
+				IdeApp.Workspace.ItemAddedToSolution += OnEntryAdded;
+				IdeApp.Exiting += delegate {
+					DelayedSaveComments (null);
+				};
 			};
-			
+
 			AddinManager.AddExtensionNodeHandler ("/MonoDevelop/VersionControl/VersionControlSystems", OnExtensionChanged);
 		}
-		
+
 		static void OnExtensionChanged (object s, ExtensionNodeEventArgs args)
 		{
 			VersionControlSystem vcs = (VersionControlSystem) args.ExtensionObject;
@@ -107,8 +112,11 @@ namespace MonoDevelop.VersionControl
 			}
 		}
 		
-		public static Gdk.Pixbuf LoadOverlayIconForStatus(VersionStatus status)
+		public static Xwt.Drawing.Image LoadOverlayIconForStatus(VersionStatus status)
 		{
+			if ((status & VersionStatus.Ignored) != 0)
+				return overlay_ignored;
+
 			if ((status & VersionStatus.Versioned) == 0)
 				return overlay_unversioned;
 			
@@ -138,7 +146,7 @@ namespace MonoDevelop.VersionControl
 			return null;
 		}
 		
-		public static Gdk.Pixbuf LoadIconForStatus (VersionStatus status)
+		public static Xwt.Drawing.Image LoadIconForStatus (VersionStatus status)
 		{
 			switch (status & VersionStatus.LocalChangesMask) {
 				case VersionStatus.Modified:
@@ -175,9 +183,13 @@ namespace MonoDevelop.VersionControl
 			}
 			return String.Empty;
 		}
-		
+
+		internal static Dictionary<Repository, InternalRepositoryReference> referenceCache = new Dictionary<Repository, InternalRepositoryReference> ();
 		public static Repository GetRepository (IWorkspaceObject entry)
 		{
+			if (IsGloballyDisabled)
+				return null;
+
 			InternalRepositoryReference repoRef = (InternalRepositoryReference) entry.ExtendedProperties [typeof(InternalRepositoryReference)];
 			if (repoRef != null)
 				return repoRef.Repo;
@@ -186,14 +198,17 @@ namespace MonoDevelop.VersionControl
 			InternalRepositoryReference rref = null;
 			if (repo != null) {
 				repo.AddRef ();
-				rref = new InternalRepositoryReference (repo);
+				if (!referenceCache.TryGetValue (repo, out rref)) {
+					rref = new InternalRepositoryReference (repo);
+					referenceCache [repo] = rref;
+				}
 			}
 			entry.ExtendedProperties [typeof(InternalRepositoryReference)] = rref;
 			
 			return repo;
 		}
 		
-		static Repository GetRepositoryReference (string path, string id)
+		public static Repository GetRepositoryReference (string path, string id)
 		{
 			foreach (VersionControlSystem vcs in GetVersionControlSystems ()) {
 				Repository repo = vcs.GetRepositoryReference (path, id);
@@ -209,7 +224,7 @@ namespace MonoDevelop.VersionControl
 		{
 			lock (commentsLock) {
 				Hashtable doc = GetCommitComments ();
-				if (comment == null || comment.Length == 0) {
+				if (String.IsNullOrEmpty (comment)) {
 					if (doc.ContainsKey (file)) {
 						doc.Remove (file);
 						if (save) SaveComments ();
@@ -330,7 +345,7 @@ namespace MonoDevelop.VersionControl
 						return;
 					}
 				
-					FileService.EnsureDirectoryExists (file.ParentDirectory);
+					Directory.CreateDirectory (file.ParentDirectory);
 					stream = new FileStream (file, FileMode.Create, FileAccess.Write);
 					BinaryFormatter formatter = new BinaryFormatter ();
 					formatter.Serialize (stream, comments);
@@ -344,40 +359,57 @@ namespace MonoDevelop.VersionControl
 			}
 		}
 		
-		internal static bool NotifyPrepareCommit (Repository repo, ChangeSet changeSet)
+		internal static void NotifyPrepareCommit (Repository repo, ChangeSet changeSet)
 		{
+			if (!DispatchService.IsGuiThread) {
+				Gtk.Application.Invoke (delegate {
+					NotifyPrepareCommit (repo, changeSet);
+				});
+				return;
+			}
+
 			if (PrepareCommit != null) {
 				try {
 					PrepareCommit (null, new CommitEventArgs (repo, changeSet, false));
 				} catch (Exception ex) {
 					MessageService.ShowException (ex);
-					return false;
 				}
 			}
-			return true;
 		}
 		
-		internal static bool NotifyBeforeCommit (Repository repo, ChangeSet changeSet)
+		internal static void NotifyBeforeCommit (Repository repo, ChangeSet changeSet)
 		{
+			if (!DispatchService.IsGuiThread) {
+				Gtk.Application.Invoke (delegate {
+					NotifyBeforeCommit (repo, changeSet);
+				});
+				return;
+			}
+
 			if (BeginCommit != null) {
 				try {
 					BeginCommit (null, new CommitEventArgs (repo, changeSet, false));
 				} catch (Exception ex) {
 					MessageService.ShowException (ex);
-					return false;
 				}
 			}
-			return true;
 		}
 		
-		internal static bool NotifyAfterCommit (Repository repo, ChangeSet changeSet, bool success)
+		internal static void NotifyAfterCommit (Repository repo, ChangeSet changeSet, bool success)
 		{
+			if (!DispatchService.IsGuiThread) {
+				Gtk.Application.Invoke (delegate {
+					NotifyAfterCommit (repo, changeSet, success);
+				});
+				return;
+			}
+
 			if (EndCommit != null) {
 				try {
 					EndCommit (null, new CommitEventArgs (repo, changeSet, success));
 				} catch (Exception ex) {
 					MessageService.ShowException (ex);
-					return false;
+					return;
 				}
 			}
 			if (success) {
@@ -385,7 +417,6 @@ namespace MonoDevelop.VersionControl
 					SetCommitComment (it.LocalPath, null, false);
 				SaveComments ();
 			}
-			return true;
 		}
 
 		public static void NotifyFileStatusChanged (IEnumerable<VersionControlItem> items) 
@@ -423,7 +454,7 @@ namespace MonoDevelop.VersionControl
 					Repository repo = GetRepository (repoFiles.Key);
 					if (repo == null)
 						continue;
-					var versionInfos = repo.GetVersionInfo (repoFiles.Select (f => f.ProjectFile.FilePath));
+					var versionInfos = repo.GetVersionInfo (repoFiles.Select (f => f.ProjectFile.FilePath), VersionInfoQueryFlags.IgnoreCache);
 					FilePath[] paths = versionInfos.Where (i => i.CanAdd).Select (i => i.LocalPath).ToArray ();
 					if (paths.Length > 0) {
 						if (monitor == null)
@@ -463,11 +494,11 @@ namespace MonoDevelop.VersionControl
 			}
 		}
 */
-		static void SolutionItemAddFiles (string rootPath, SolutionItem entry, ArrayList files)
+		static void SolutionItemAddFiles (string rootPath, SolutionItem entry, HashSet<string> files)
 		{
 			if (entry is SolutionEntityItem) {
-				string file = ((SolutionEntityItem)entry).FileName;
-				SolutionItemAddFile (rootPath, files, file);
+				foreach (var file in ((SolutionEntityItem)entry).GetItemFiles (false))
+					SolutionItemAddFile (rootPath, files, file);
 			}
 			
 			if (entry is Project) {
@@ -481,26 +512,24 @@ namespace MonoDevelop.VersionControl
 			}
 		}
 		
-		static void SolutionItemAddFile (string rootPath, ArrayList files, string file)
+		static void SolutionItemAddFile (string rootPath, HashSet<string> files, string file)
 		{
-			if (!file.StartsWith (rootPath + Path.DirectorySeparatorChar))
+			if (!file.StartsWith (rootPath + Path.DirectorySeparatorChar, StringComparison.Ordinal))
 			    return;
 			if (!File.Exists (file))
 				return;
-			if (files.Contains (file))
-				return;
-			string dir = Path.GetDirectoryName (file);
-			while (dir != rootPath) {
-				if (files.Contains (dir))
-					break;
-				files.Add (dir);
-				dir = Path.GetDirectoryName (dir);
+			if (files.Add (file)) {
+				string dir = Path.GetDirectoryName (file);
+				while (dir != rootPath && files.Add (dir))
+					dir = Path.GetDirectoryName (dir);
 			}
-			files.Add (file);
 		}
 		
 		static void OnEntryAdded (object o, SolutionItemEventArgs args)
 		{
+			if (args is SolutionItemChangeEventArgs && ((SolutionItemChangeEventArgs) args).Reloading)
+				return;
+
 			// handles addition of solutions and projects
 			SolutionItem parent = (SolutionItem) args.SolutionItem.ParentFolder;
 			
@@ -525,32 +554,43 @@ namespace MonoDevelop.VersionControl
 			// While we /could/ call repo.Add with `recursive = true', we don't
 			// necessarily want to add files under the project/solution directory
 			// that may not be a part of this project/solution.
-			
-			ArrayList files = new ArrayList ();
-			
-			files.Add (path);
+
+			var files = new HashSet<string> { path };
 			SolutionItemAddFiles (path, entry, files);
-			files.Sort ();
 			
 			using (IProgressMonitor monitor = GetStatusMonitor ()) {
-				string[] paths = (string[]) files.ToArray (typeof (string));
-				
-				foreach (string p in paths) {
-					if (repo.GetVersionInfo (p).CanAdd)
-						repo.Add (p, false, monitor);
+				var status = repo.GetDirectoryVersionInfo (path, false, true);
+				foreach (var v in status) {
+					if (!v.IsVersioned && files.Contains (v.LocalPath))
+						repo.Add (v.LocalPath, false, monitor);
 				}
 			}
-			
+
+			if (entry is SolutionFolder && files.Count == 1)
+				return;
+
 			NotifyFileStatusChanged (new FileUpdateEventArgs (repo, parent.BaseDirectory, true));
 		}
 		
 		public static IProgressMonitor GetProgressMonitor (string operation)
 		{
+			return GetProgressMonitor (operation, VersionControlOperationType.Other);
+		}
+		
+		public static IProgressMonitor GetProgressMonitor (string operation, VersionControlOperationType op)
+		{
+			IconId icon;
+			switch (op) {
+			case VersionControlOperationType.Pull: icon = Stock.StatusDownload; break;
+			case VersionControlOperationType.Push: icon = Stock.StatusUpload; break;
+			default: icon = "md-version-control"; break;
+			}
+
 			IProgressMonitor monitor = IdeApp.Workbench.ProgressMonitors.GetOutputProgressMonitor ("MonoDevelop.VersionControlOutput", "Version Control", "md-version-control", false, true);
 			Pad outPad = IdeApp.Workbench.ProgressMonitors.GetPadForMonitor (monitor);
 			
 			AggregatedProgressMonitor mon = new AggregatedProgressMonitor (monitor);
-			mon.AddSlaveMonitor (IdeApp.Workbench.ProgressMonitors.GetStatusProgressMonitor (operation, "md-version-control", true, true, false, outPad));
+			mon.AddSlaveMonitor (IdeApp.Workbench.ProgressMonitors.GetStatusProgressMonitor (operation, icon, true, true, false, outPad));
 			return mon;
 		}
 		
@@ -581,6 +621,35 @@ namespace MonoDevelop.VersionControl
 		{
 			GetConfiguration ().Repositories.Remove (repo);
 		}
+
+		public static bool IsGloballyDisabled {
+			get { return ConfigurationGlobalDisabled || SessionSolutionDisabled; }
+			set { ConfigurationGlobalDisabled = value; }
+		}
+
+		public static bool IsSolutionDisabled (Solution it)
+		{
+			return it.UserProperties.HasValue ("VersionControlDisabled");
+		}
+
+		public static void SetSolutionDisabled (Solution it, bool value)
+		{
+			if (value)
+				it.UserProperties.SetValue ("VersionControlDisabled", true);
+			else
+				it.UserProperties.RemoveValue ("VersionControlDisabled");
+			SessionSolutionDisabled = value;
+		}
+
+		internal static bool ConfigurationGlobalDisabled {
+			get { return GetConfiguration ().Disabled; }
+			set { GetConfiguration ().Disabled = value; }
+		}
+
+		internal static bool SessionSolutionDisabled {
+			get;
+			private set;
+		}
 		
 		static public IEnumerable<Repository> GetRepositories ()
 		{
@@ -591,12 +660,12 @@ namespace MonoDevelop.VersionControl
 		{
 			if (configuration == null) {
 				if (File.Exists (ConfigFile)) {
-					XmlDataSerializer ser = new XmlDataSerializer (dataContext);
-					XmlTextReader reader = new XmlTextReader (new StreamReader (ConfigFile));
 					try {
-						configuration = (VersionControlConfiguration) ser.Deserialize (reader, typeof (VersionControlConfiguration));
-					} finally {
-						reader.Close ();
+						XmlDataSerializer ser = new XmlDataSerializer (dataContext);
+						using (var reader = File.OpenText (ConfigFile))
+							configuration = (VersionControlConfiguration) ser.Deserialize (reader, typeof (VersionControlConfiguration));
+					} catch {
+						((FilePath) ConfigFile).Delete ();
 					}
 				}
 				if (configuration == null)
@@ -609,12 +678,9 @@ namespace MonoDevelop.VersionControl
 		{
 			if (configuration != null) {
 				XmlDataSerializer ser = new XmlDataSerializer (dataContext);
-				XmlTextWriter tw = new XmlTextWriter (new StreamWriter (ConfigFile));
-				tw.Formatting = Formatting.Indented;
-				try {
+				using (var tw = new XmlTextWriter (File.CreateText (ConfigFile))) {
+					tw.Formatting = Formatting.Indented;
 					ser.Serialize (tw, configuration, typeof (VersionControlConfiguration));
-				} finally {
-					tw.Close ();
 				}
 			}
 		}
@@ -631,12 +697,14 @@ namespace MonoDevelop.VersionControl
 		
 		public static bool CheckVersionControlInstalled ()
 		{
+			if (IsGloballyDisabled)
+				return false;
+
 			foreach (VersionControlSystem vcs in GetVersionControlSystems ()) {
 				if (vcs.IsInstalled)
 					return true;
 			}
-			
-			MessageService.ShowError (GettextCatalog.GetString ("There isn't any supported version control system installed. You may need to install additional add-ins or packages."));
+
 			return false;
 		}
 		
@@ -734,13 +802,21 @@ namespace MonoDevelop.VersionControl
 
 		public Repository Repo {
 			get {
-				return this.repo;
+				return repo;
 			}
 		}
 		
 		public void Dispose ()
 		{
+			VersionControlService.referenceCache.Remove (repo);
 			repo.Unref ();
 		}
 	}
+
+	public enum VersionControlOperationType
+	{
+		Pull,
+		Push,
+		Other
+	}
 }
