@@ -28,10 +28,9 @@
 using System;
 using System.Linq;
 using System.Collections.Generic;
-using System.Collections.ObjectModel;
 using System.Text;
 using System.Xml;
-using System.Diagnostics;
+using System.IO;
 
 namespace Mono.TextEditor.Highlighting
 {
@@ -60,24 +59,29 @@ namespace Mono.TextEditor.Highlighting
 		}
 
 		void HandleTextReplaced (object sender, DocumentChangeEventArgs e)
+		{
+			if (doc == null || doc.SuppressHighlightUpdate || doc.CurrentAtomicUndoOperationType == OperationType.Format)
+				return;
+			SyntaxModeService.StartUpdate (doc, this, e.Offset, e.Offset + e.InsertionLength);
+		}
+
+		public void UpdateDocumentHighlighting ()
 		{
 			if (doc == null || doc.SuppressHighlightUpdate)
 				return;
-			Mono.TextEditor.Highlighting.SyntaxModeService.StartUpdate (doc, this, e.Offset, e.Offset + e.InsertionLength);
+			SyntaxModeService.StartUpdate (doc, this, 0, doc.TextLength);
 		}
 
 		void HandleTextSet (object sender, EventArgs e)
 		{
-			if (doc == null || doc.SuppressHighlightUpdate)
-				return;
-			Mono.TextEditor.Highlighting.SyntaxModeService.StartUpdate (doc, this, 0, doc.TextLength);
+			UpdateDocumentHighlighting ();
 		}
 		
 		public event EventHandler DocumentSet;
 		
 		protected virtual void OnDocumentSet (EventArgs e)
 		{
-			EventHandler handler = this.DocumentSet;
+			var handler = DocumentSet;
 			if (handler != null)
 				handler (this, e);
 		}	
@@ -96,11 +100,11 @@ namespace Mono.TextEditor.Highlighting
 			}
 		}
 		
-		protected SyntaxMode () : base (null)
+		protected SyntaxMode ()
 		{
-			DefaultColor = "text";
+			DefaultColor = "Plain Text";
 			Name = "<root>";
-			this.Delimiter = "&()<>{}[]~!%^*-+=|\\#/:;\"' ,\t.?";
+			SetDelimiter ("&()<>{}[]~!%^*-+=|\\#/:;\"' ,\t.?");
 		}
 		
 		public SyntaxMode (TextDocument doc) : this ()
@@ -133,29 +137,27 @@ namespace Mono.TextEditor.Highlighting
 			}
 			if (result != null) {
 				// crop to begin
-				if (result.Offset != offset) {
-					while (result != null && result.EndOffset < offset)
+				if (result.Offset < offset) {
+					while (result != null && result.EndOffset < offset) {
 						result = result.Next;
+					}
 					if (result != null) {
 						int endOffset = result.EndOffset;
 						result.Offset = offset;
 						result.Length = endOffset - offset;
 					}
 				}
-
-				if (result != null && offset + length != chunkParser.lineOffset + line.Length) {
-					// crop to end
-					Chunk cur = result;
-					while (cur != null && cur.EndOffset < offset + length) {
-						cur = cur.Next;
-					}
-					if (cur != null) {
-						cur.Length = offset + length - cur.Offset;
-						cur.Next = null;
-					}
-				}
 			}
 			while (result != null) {
+				// crop to end
+				if (result.EndOffset >= offset + length) {
+					result.Length = offset + length - result.Offset;
+					result.Next = null;
+					if (result.Length < 0) {
+						result.Length = 0;
+						yield break;
+					}
+				}
 				yield return result;
 				result = result.Next;
 			}
@@ -165,6 +167,10 @@ namespace Mono.TextEditor.Highlighting
 		{
 			return string.Format ("#{0:X2}{1:X2}{2:X2}", color.Red >> 8, color.Green >> 8, color.Blue >> 8);
 		}
+		public static string ColorToPangoMarkup (Cairo.Color color)
+		{
+			return ColorToPangoMarkup ((Gdk.Color)((HslColor)color));
+		}
 
 		public static int GetIndentLength (TextDocument doc, int offset, int length, bool skipFirstLine)
 		{
@@ -194,10 +200,11 @@ namespace Mono.TextEditor.Highlighting
 
 		public class SpanParser
 		{
-			protected SyntaxMode mode;
+			protected readonly SyntaxMode mode;
 			protected CloneableStack<Span> spanStack;
 			protected Stack<Rule> ruleStack;
-			protected TextDocument doc;
+			protected readonly TextDocument doc;
+
 			int maxEnd;
 
 			public Rule CurRule {
@@ -215,7 +222,7 @@ namespace Mono.TextEditor.Highlighting
 					return spanStack;
 				}
 				set {
-					this.spanStack = value;
+					spanStack = value;
 				}
 			}
 
@@ -224,8 +231,8 @@ namespace Mono.TextEditor.Highlighting
 					return ruleStack;
 				}
 				set {
-					this.ruleStack = value;
-					this.CurRule = ruleStack.Peek ();
+					ruleStack = value;
+					CurRule = ruleStack.Peek ();
 				}
 			}
 
@@ -236,8 +243,8 @@ namespace Mono.TextEditor.Highlighting
 					return result;
 				Rule rule = mode;
 				result.Push (mode);
-				foreach (Span span in this.spanStack.Reverse ()) {
-					Rule tmp = rule.GetRule (doc, span.Rule) ?? this.CurRule;
+				foreach (Span span in spanStack.Reverse ()) {
+					Rule tmp = rule.GetRule (doc, span.Rule) ?? CurRule;
 					result.Push (tmp);
 					rule = tmp;
 				}
@@ -249,6 +256,8 @@ namespace Mono.TextEditor.Highlighting
 				if (mode == null)
 					throw new ArgumentNullException ("mode");
 				this.doc  = mode.Document;
+				if (this.doc == null)
+					throw new ArgumentException ("Syntax mode isn't bound to any document.", "mode");
 				this.mode = mode;
 				this.SpanStack = spanStack;
 				this.CurRule = mode;
@@ -275,24 +284,27 @@ namespace Mono.TextEditor.Highlighting
 			{
 				spanStack.Push (span);
 				ruleStack.Push (rule);
-				this.CurRule = rule;
-				this.CurSpan = span;
+				CurRule = rule;
+				CurSpan = span;
 			}
 
-			public void PopSpan ()
+			public Span PopSpan ()
 			{
-				if (spanStack.Count > 0)
-					spanStack.Pop ();
+				Span result = null;
+				if (spanStack.Count > 0) {
+					result = spanStack.Pop ();
+				}
 				if (ruleStack.Count > 1)
 					ruleStack.Pop ();
-				this.CurRule = ruleStack.Peek ();
-				this.CurSpan = spanStack.Count > 0 ? spanStack.Peek () : null;
+				CurRule = ruleStack.Peek ();
+				CurSpan = spanStack.Count > 0 ? spanStack.Peek () : null;
+				return result;
 			}
 
 			public Rule GetRule (Span span)
 			{
 				if (string.IsNullOrEmpty (span.Rule))
-					return new Rule (mode);
+					return new Rule ();
 				return CurRule.GetRule (doc, span.Rule);
 			}
 
@@ -318,10 +330,9 @@ namespace Mono.TextEditor.Highlighting
 			}
 		
 						
-			protected virtual void ScanSpan (ref int i)
+			protected virtual bool ScanSpan (ref int i)
 			{
 				int textOffset = i - StartOffset;
-				
 				for (int j = 0; j < CurRule.Spans.Length; j++) {
 					Span span = CurRule.Spans [j];
 
@@ -334,32 +345,38 @@ namespace Mono.TextEditor.Highlighting
 					} 
 
 					RegexMatch match = span.Begin.TryMatch (CurText, textOffset);
-					if (!match.Success)
+					if (!match.Success) {
 						continue;
+					}
 					// scan for span exit which cancels the span.
 					if ((span.ExitFlags & SpanExitFlags.CancelSpan) == SpanExitFlags.CancelSpan && span.Exit != null) {
 						bool foundEnd = false;
 						for (int k = i + match.Length; k < CurText.Length; k++) {
 							if (span.Exit.TryMatch (CurText, k - StartOffset).Success)
-								return;
+								return false;
 							if (span.End.TryMatch (CurText, k - StartOffset).Success) {
 								foundEnd = true;
 								break;
 							}
 						}
 						if (!foundEnd)
-							return;
+							return false;
 					}
 					
 					bool mismatch = false;
 					if ((span.BeginFlags & SpanBeginFlags.FirstNonWs) == SpanBeginFlags.FirstNonWs)
 						mismatch = CurText.Take (i).Any (ch => !char.IsWhiteSpace (ch));
+					if ((span.BeginFlags & SpanBeginFlags.NewWord) == SpanBeginFlags.NewWord) {
+						if (textOffset - 1 > 0 && textOffset - 1 < CurText.Length)
+							mismatch = !char.IsWhiteSpace (CurText [textOffset - 1]);
+					}
 					if (mismatch)
 						continue;
 					FoundSpanBegin (span, i, match.Length);
-					i += match.Length - 1;
-					return;
+					i += System.Math.Max (0, match.Length - 1);
+					return true;
 				}
+				return false;
 			}
 
 			protected virtual bool ScanSpanEnd (Span cur, ref int i)
@@ -370,7 +387,7 @@ namespace Mono.TextEditor.Highlighting
 					RegexMatch match = cur.End.TryMatch (CurText, textOffset);
 					if (match.Success) {
 						FoundSpanEnd (cur, i, match.Length);
-						i += match.Length - 1;
+						i += System.Math.Max (0, match.Length - 1);
 						return true;
 					}
 				}
@@ -379,7 +396,7 @@ namespace Mono.TextEditor.Highlighting
 					RegexMatch match = cur.Exit.TryMatch (CurText, textOffset);
 					if (match.Success) {
 						FoundSpanExit (cur, i, match.Length);
-						i += match.Length - 1;
+						i += System.Math.Max (0, match.Length - 1);
 						return true;
 					}
 				}
@@ -407,33 +424,29 @@ namespace Mono.TextEditor.Highlighting
 					Span cur = CurSpan;
 					if (cur != null) {
 						if (cur.Escape != null) {
-							bool mismatch = false;
-							for (int j = 0; j < cur.Escape.Length; j++) {
-								if (textIndex + j >= CurText.Length || CurText[textIndex + j] != cur.Escape[j]) {
-									mismatch = true;
-									break;
-								}
-							}
-							if (!mismatch) {
-								i += cur.Escape.Length;
-								if (cur.Escape.Length > 1)
-									i--;
+							RegexMatch match = cur.Escape.TryMatch(CurText, textIndex);
+							if (match.Success) {
+								int j = i + match.Length - 1;
+								ParseChar (ref i, CurText [textIndex]);
+								i = j;
 								continue;
 							}
 						}
 						if (ScanSpanEnd (cur, ref i))
 							continue;
+
+					}
+					if (!ScanSpan (ref i)) {
+						if (i < doc.TextLength)
+							ParseChar (ref i, CurText [textIndex]);
 					}
-					ScanSpan (ref i);
-					if (i < doc.TextLength)
-						ParseChar (ref i, CurText[textIndex]);
 				}
 			}
 		}
 
 		public class ChunkParser
 		{
-			readonly string defaultStyle = "text";
+			readonly string defaultStyle = "Plain Text";
 			protected SpanParser spanParser;
 			protected TextDocument doc;
 			protected DocumentLine line;
@@ -455,7 +468,7 @@ namespace Mono.TextEditor.Highlighting
 					throw new ArgumentNullException ("line");
 			}
 
-			Chunk startChunk = null;
+			Chunk startChunk;
 			Chunk endChunk;
 
 			protected virtual void AddRealChunk (Chunk chunk)
@@ -515,7 +528,7 @@ namespace Mono.TextEditor.Highlighting
 					return result;
 				}
 				string rule = spanParser.SpanStack.Peek ().Rule;
-				if (!string.IsNullOrEmpty (rule) && rule.StartsWith ("mode:"))
+				if (!string.IsNullOrEmpty (rule) && rule.StartsWith ("mode:", StringComparison.Ordinal))
 					return spanParser.CurRule.DefaultColor ?? defaultStyle;
 				return spanParser.SpanStack.Peek ().Color;
 			}
@@ -549,11 +562,11 @@ namespace Mono.TextEditor.Highlighting
 
 				curChunk.Offset = offset;
 				curChunk.Length = length;
-				curChunk.Style = span.TagColor ?? GetChunkStyle (span);
+				curChunk.Style = span.BeginTagColor ?? GetChunkStyle (span);
 				curChunk.SpanStack.Push (span);
 				AddChunk (ref curChunk, 0, curChunk.Style);
 				foreach (SemanticRule semanticRule in spanRule.SemanticRules) {
-					semanticRule.Analyze (this.doc, line, curChunk, offset, line.EndOffsetIncludingDelimiter);
+					semanticRule.Analyze (doc, line, curChunk, offset, line.EndOffsetIncludingDelimiter);
 				}
 			}
 
@@ -573,43 +586,58 @@ namespace Mono.TextEditor.Highlighting
 
 				curChunk.Offset = offset;
 				curChunk.Length = length;
-				curChunk.Style  = span.TagColor ?? GetChunkStyle (span);
+				curChunk.Style  = span.EndTagColor ?? GetChunkStyle (span);
 				AddChunk (ref curChunk, 0, defaultStyle);
 				spanParser.PopSpan ();
 			}
 			protected StringBuilder wordbuilder = new StringBuilder ();
-			bool inWord = false;
+			bool inWord;
 			public void ParseChar (ref int i, char ch)
 			{
-				int textOffset = i - spanParser.StartOffset;
-
 				Rule cur = spanParser.CurRule;
-				bool isWordPart = cur.Delimiter.IndexOf (ch) < 0;
-
+				bool isWordPart = cur.GetDelimiter (mode).IndexOf (ch) < 0;
 				if (inWord && !isWordPart || !inWord && isWordPart)
 					AddChunk (ref curChunk, 0, curChunk.Style = GetStyle (curChunk) ?? GetSpanStyle ());
 
 				inWord = isWordPart;
 
-				if (cur.HasMatches && i - curChunk.Offset == 0) {
+				if (cur.HasMatches && (i - curChunk.Offset == 0 || string.IsNullOrEmpty (cur.GetDelimiter (mode)))) {
 					Match foundMatch = null;
-					int   foundMatchLength = 0;
+					var   foundMatchLength = new int[0];
+					int textOffset = i - spanParser.StartOffset;
 					foreach (Match ruleMatch in cur.Matches) {
-						int matchLength = ruleMatch.TryMatch (spanParser.CurText, textOffset);
-						if (foundMatchLength < matchLength) {
+						var matchLength = ruleMatch.TryMatch (spanParser.CurText, textOffset);
+						if (foundMatchLength.Length < matchLength.Length) {
 							foundMatch = ruleMatch;
 							foundMatchLength = matchLength;
 						}
 					}
 					if (foundMatch != null) {
-						AddChunk (ref curChunk, foundMatchLength, GetChunkStyleColor (foundMatch.Color));
-						i += foundMatchLength - 1;
-						curChunk.Length = i - curChunk.Offset + 1;
-						return;
+						if (foundMatch.IsGroupMatch) {
+							for (int j = 1; j < foundMatchLength.Length; j++) {
+								var len = foundMatchLength [j];
+								if (len > 0) {
+									AddChunk (ref curChunk, len, GetChunkStyleColor (foundMatch.Groups [j - 1]));
+									i += len - 1;
+									curChunk.Length = 0;
+								}
+							}
+							return;
+						}
+						if (foundMatchLength[0] > 0) {
+							AddChunk (ref curChunk, foundMatchLength[0], GetChunkStyleColor (foundMatch.Color));
+							i += foundMatchLength[0] - 1;
+							curChunk.Length = 0;
+							return;
+						}
 					}
 				}
 				wordbuilder.Append (ch);
 				curChunk.Length = i - curChunk.Offset + 1;
+				if (!isWordPart) {
+					AddChunk (ref curChunk, 0, curChunk.Style = GetStyle (curChunk) ?? GetSpanStyle ());
+				}
+
 			}
 
 			protected virtual string GetStyle (Chunk chunk)
@@ -629,6 +657,7 @@ namespace Mono.TextEditor.Highlighting
 				length = System.Math.Min (doc.TextLength - offset, length);
 				curChunk = new Chunk (offset, 0, GetSpanStyle ());
 				spanParser.ParseSpans (offset, length);
+				curChunk.SpanStack = spanParser.SpanStack;
 				curChunk.Length = offset + length - curChunk.Offset;
 				if (curChunk.Length > 0) {
 					curChunk.Style = GetStyle (curChunk) ?? GetSpanStyle ();
@@ -638,13 +667,13 @@ namespace Mono.TextEditor.Highlighting
 			}
 		}
 
-		public override Rule GetRule (TextDocument doc, string name)
+		public override Rule GetRule (TextDocument document, string name)
 		{
 			if (name == null || name == "<root>") {
 				return this;
 			}
-			if (name.StartsWith ("mode:"))
-				return SyntaxModeService.GetSyntaxMode (doc, name.Substring ("mode:".Length));
+			if (name.StartsWith ("mode:", StringComparison.Ordinal))
+				return SyntaxModeService.GetSyntaxMode (document, name.Substring ("mode:".Length));
 
 			foreach (Rule rule in rules) {
 				if (rule.Name == name)
@@ -698,12 +727,13 @@ namespace Mono.TextEditor.Highlighting
 
 		public const string MimeTypesAttribute = "mimeTypes";
 
-		public static SyntaxMode Read (XmlReader reader)
+		public static SyntaxMode Read (Stream stream)
 		{
+			var reader = XmlReader.Create (stream);
 			var result = new SyntaxMode (null);
-			List<Match> matches = new List<Match> ();
-			List<Span> spanList = new List<Span> ();
-			List<Marker> prevMarkerList = new List<Marker> ();
+			var matches = new List<Match> ();
+			var spanList = new List<Span> ();
+			var prevMarkerList = new List<Marker> ();
 			XmlReadHelper.ReadList (reader, Node, delegate () {
 				if (reader == null)
 					return true;
@@ -775,12 +805,13 @@ namespace Mono.TextEditor.Highlighting
 		
 		Rule DeepCopy (TextDocument doc, SyntaxMode mode, Rule rule)
 		{
-			Rule newRule = new Rule (mode);
+			var newRule = new Rule ();
 			newRule.spans = new Span[rule.Spans.Length];
 			for (int i = 0; i < rule.Spans.Length; i++) {
 				newRule.spans [i] = rule.Spans [i].Clone ();
 			}
-			newRule.Delimiter = rule.Delimiter;
+			if (rule.HasDelimiter)
+				newRule.SetDelimiter (rule.GetDelimiter (null));
 			newRule.IgnoreCase = rule.IgnoreCase;
 			newRule.Name = rule.Name;
 			newRule.DefaultColor = rule.DefaultColor;
@@ -815,3 +846,4 @@ namespace Mono.TextEditor.Highlighting
 		}
 	}
 }
+
