@@ -33,10 +33,12 @@ using Mono.TextEditor.Utils;
 using System.Linq;
 using System.ComponentModel;
 using ICSharpCode.NRefactory.Editor;
+using System.Threading.Tasks;
+using System.Threading;
 
 namespace Mono.TextEditor
 {
-	public class TextDocument : AbstractAnnotatable, IBuffer, ICSharpCode.NRefactory.Editor.IDocument
+	public class TextDocument : ICSharpCode.NRefactory.AbstractAnnotatable, ICSharpCode.NRefactory.Editor.IDocument
 	{
 		readonly IBuffer buffer;
 		readonly ILineSplitter splitter;
@@ -56,17 +58,34 @@ namespace Mono.TextEditor
 			}
 			set {
 				if (mimeType != value) {
-					mimeType = value;
-					SyntaxMode = SyntaxModeService.GetSyntaxMode (this);
+					lock (this) {
+						mimeType = value;
+						SyntaxMode = SyntaxModeService.GetSyntaxMode (this, value);
+					}
 				}
 			}
 		}
 		
+		string fileName;
 		public string FileName {
-			get;
-			set;
+			get {
+				return fileName;
+			}
+			set {
+				fileName = value;
+				OnFileNameChanged (EventArgs.Empty);
+			}
 		}	
-		
+
+		public event EventHandler FileNameChanged;
+
+		protected virtual void OnFileNameChanged (EventArgs e)
+		{
+			EventHandler handler = this.FileNameChanged;
+			if (handler != null)
+				handler (this, e);
+		}
+
 		public bool HeightChanged {
 			get;
 			set;
@@ -106,6 +125,15 @@ namespace Mono.TextEditor
 			get;
 			set;
 		}
+
+		public bool HasLineEndingMismatchOnTextSet {
+			get {
+				return splitter.LineEndingMismatch;
+			}
+			set {
+				splitter.LineEndingMismatch = value;
+			}
+		}
 		
 		protected TextDocument (IBuffer buffer,ILineSplitter splitter)
 		{
@@ -114,6 +142,8 @@ namespace Mono.TextEditor
 			splitter.LineChanged += SplitterLineSegmentTreeLineChanged;
 			splitter.LineRemoved += HandleSplitterLineSegmentTreeLineRemoved;
 			foldSegmentTree.tree.NodeRemoved += HandleFoldSegmentTreetreeNodeRemoved; 
+			textSegmentMarkerTree.InstallListener (this);
+			this.diffTracker.SetTrackDocument (this); 
 		}
 
 		void HandleFoldSegmentTreetreeNodeRemoved (object sender, RedBlackTree<FoldSegment>.RedBlackTreeNodeEventArgs e)
@@ -140,14 +170,6 @@ namespace Mono.TextEditor
 			};
 		}
 
-		~TextDocument ()
-		{
-			if (foldSegmentWorker != null) {
-				foldSegmentWorker.Dispose ();
-				foldSegmentWorker = null;
-			}
-		}
-
 		void SplitterLineSegmentTreeLineChanged (object sender, LineEventArgs e)
 		{
 			if (LineChanged != null)
@@ -165,16 +187,19 @@ namespace Mono.TextEditor
 		}
 
 		public bool SuppressHighlightUpdate { get; set; }
-		
+		internal DocumentLine longestLineAtTextSet;
+
 		public string Text {
 			get {
 				return buffer.Text;
 			}
 			set {
 				var args = new DocumentChangeEventArgs (0, Text, value);
+				textSegmentMarkerTree.Clear ();
 				OnTextReplacing (args);
 				buffer.Text = value;
-				splitter.Initalize (value);
+				extendingTextMarkers = new List<TextLineMarker> ();
+				splitter.Initalize (value, out longestLineAtTextSet);
 				ClearFoldSegments ();
 				OnTextReplaced (args);
 				versionProvider = new TextSourceVersionProvider ();
@@ -184,9 +209,9 @@ namespace Mono.TextEditor
 			}
 		}
 
-		public void Insert (int offset, string text)
+		public void Insert (int offset, string text, ICSharpCode.NRefactory.Editor.AnchorMovementType anchorMovementType = AnchorMovementType.Default)
 		{
-			Replace (offset, 0, text);
+			Replace (offset, 0, text, anchorMovementType);
 		}
 		
 		public void Remove (int offset, int count)
@@ -200,6 +225,11 @@ namespace Mono.TextEditor
 		}
 
 		public void Replace (int offset, int count, string value)
+		{
+			Replace (offset, count, value, AnchorMovementType.Default);
+		}
+
+		public void Replace (int offset, int count, string value, ICSharpCode.NRefactory.Editor.AnchorMovementType anchorMovementType = AnchorMovementType.Default)
 		{
 			if (offset < 0)
 				throw new ArgumentOutOfRangeException ("offset", "must be > 0, was: " + offset);
@@ -209,11 +239,10 @@ namespace Mono.TextEditor
 				throw new ArgumentOutOfRangeException ("count", "must be > 0, was: " + count);
 
 			InterruptFoldWorker ();
-			
 			//int oldLineCount = LineCount;
-			var args = new DocumentChangeEventArgs (offset, count > 0 ? GetTextAt (offset, count) : "", value);
+			var args = new DocumentChangeEventArgs (offset, count > 0 ? GetTextAt (offset, count) : "", value, anchorMovementType);
 			OnTextReplacing (args);
-			value = args.InsertedText;
+			value = args.InsertedText.Text;
 			UndoOperation operation = null;
 			if (!isInUndo) {
 				operation = new UndoOperation (args);
@@ -232,10 +261,6 @@ namespace Mono.TextEditor
 			splitter.TextReplaced (this, args);
 			versionProvider.AppendChange (args);
 			OnTextReplaced (args);
-			
-			UpdateUndoStackOnReplace (args);
-			if (operation != null)
-				operation.Setup (this, args);
 		}
 		
 		public string GetTextBetween (int startOffset, int endOffset)
@@ -308,6 +333,10 @@ namespace Mono.TextEditor
 		
 		public char GetCharAt (int offset)
 		{
+			if (offset < 0)
+				throw new ArgumentException ("offset < 0");
+			if (offset >= TextLength)
+				throw new ArgumentException ("offset >= TextLength");
 			return buffer.GetCharAt (offset);
 		}
 
@@ -460,7 +489,8 @@ namespace Mono.TextEditor
 			if (lineNr < DocumentLocation.MinLine)
 				return DocumentLocation.Empty;
 			DocumentLine line = GetLine (lineNr);
-			return new DocumentLocation (lineNr, System.Math.Min (line.LengthIncludingDelimiter, offset - line.Offset) + 1);
+			var col = System.Math.Max (1, System.Math.Min (line.LengthIncludingDelimiter, offset - line.Offset) + 1);
+			return new DocumentLocation (lineNr, col);
 		}
 
 		public string GetLineIndent (int lineNumber)
@@ -498,8 +528,7 @@ namespace Mono.TextEditor
 		public class UndoOperation
 		{
 			DocumentChangeEventArgs args;
-			int startOffset, length;
-			
+
 			public virtual DocumentChangeEventArgs Args {
 				get {
 					return args;
@@ -515,49 +544,23 @@ namespace Mono.TextEditor
 			{
 			}
 
-			internal virtual bool ChangedLine (int offset, int lastLineOffset, int endOffset)
-			{
-				if (Args == null)
-					return false;
-				return startOffset <= lastLineOffset && offset <= startOffset + length 
-						|| lastLineOffset <= startOffset && startOffset <= endOffset
-						;
-					;
-			}
-			
 			public UndoOperation (DocumentChangeEventArgs args)
 			{
 				this.args = args;
 			}
-			
-			
-			internal void Setup (TextDocument doc, DocumentChangeEventArgs args)
+
+			public virtual void Undo (TextDocument doc, bool fireEvent = true)
 			{
-				if (args != null) {
-					startOffset = args.Offset;
-					length = args.InsertionLength;
-				}
+				doc.Replace (args.Offset, args.InsertionLength, args.RemovedText.Text);
+				if (fireEvent)
+					OnUndoDone ();
 			}
 			
-			internal virtual void InformTextReplace (DocumentChangeEventArgs args)
+			public virtual void Redo (TextDocument doc, bool fireEvent = true)
 			{
-				if (args.Offset < startOffset) {
-					startOffset = System.Math.Max (startOffset + args.ChangeDelta, args.Offset);
-				} else if (args.Offset < startOffset + length) {
-					length = System.Math.Max (length + args.ChangeDelta, startOffset - args.Offset);
-				}
-			}
-			
-			public virtual void Undo (TextDocument doc)
-			{
-				doc.Replace (args.Offset, args.InsertionLength, args.RemovedText);
-				OnUndoDone ();
-			}
-			
-			public virtual void Redo (TextDocument doc)
-			{
-				doc.Replace (args.Offset, args.RemovalLength, args.InsertedText);
-				OnRedoDone ();
+				doc.Replace (args.Offset, args.RemovalLength, args.InsertedText.Text);
+				if (fireEvent)
+					OnRedoDone ();
 			}
 			
 			protected virtual void OnUndoDone ()
@@ -577,7 +580,14 @@ namespace Mono.TextEditor
 		
 		class AtomicUndoOperation : UndoOperation
 		{
+			OperationType operationType;
 			protected List<UndoOperation> operations = new List<UndoOperation> ();
+
+			public OperationType OperationType {
+				get {
+					return operationType;
+				}
+			}
 			
 			public List<UndoOperation> Operations {
 				get {
@@ -590,22 +600,13 @@ namespace Mono.TextEditor
 					return null;
 				}
 			}
-			
-			
-			internal override void InformTextReplace (DocumentChangeEventArgs args)
+
+			public AtomicUndoOperation (OperationType operationType = OperationType.Undefined)
 			{
-				operations.ForEach (o => o.InformTextReplace (args));
+				this.operationType = operationType;
 			}
-			
-			internal override bool ChangedLine (int lineOffset, int lastLineOffset, int lineEndOffset)
-			{
-				foreach (var op in Operations) {
-					if (op.ChangedLine (lineOffset, lastLineOffset, lineEndOffset))
-						return true;
-				}
-				return false;
-			}
-			
+		
+
 			public void Insert (int index, UndoOperation operation)
 			{
 				operations.Insert (index, operation);
@@ -616,22 +617,28 @@ namespace Mono.TextEditor
 				operations.Add (operation);
 			}
 			
-			public override void Undo (TextDocument doc)
+			public override void Undo (TextDocument doc, bool fireEvent = true)
 			{
+				doc.currentAtomicUndoOperationType.Push (operationType);
 				for (int i = operations.Count - 1; i >= 0; i--) {
-					operations [i].Undo (doc);
+					operations [i].Undo (doc, false);
 					doc.OnUndone (new UndoOperationEventArgs (operations[i]));
 				}
-				OnUndoDone ();
+				doc.currentAtomicUndoOperationType.Pop (); 
+				if (fireEvent)
+					OnUndoDone ();
 			}
 			
-			public override void Redo (TextDocument doc)
+			public override void Redo (TextDocument doc, bool fireEvent = true)
 			{
+				doc.currentAtomicUndoOperationType.Push (operationType);
 				foreach (UndoOperation operation in this.operations) {
-					operation.Redo (doc);
+					operation.Redo (doc, false);
 					doc.OnRedone (new UndoOperationEventArgs (operation));
 				}
-				OnRedoDone ();
+				doc.currentAtomicUndoOperationType.Pop (); 
+				if (fireEvent)
+					OnRedoDone ();
 			}
 		}
 		
@@ -647,7 +654,7 @@ namespace Mono.TextEditor
 					isClosed = value;
 				}
 			}
-			
+
 			public override DocumentChangeEventArgs Args {
 				get {
 					return operations.Count > 0 ? operations [operations.Count - 1].Args : null;
@@ -659,20 +666,33 @@ namespace Mono.TextEditor
 		Stack<UndoOperation> undoStack = new Stack<UndoOperation> ();
 		Stack<UndoOperation> redoStack = new Stack<UndoOperation> ();
 		AtomicUndoOperation currentAtomicOperation = null;
-		
-		// The undo stack needs to be updated on replace, because the text editor
-		// draws a replace operation marker at the left margin.
-		public void UpdateUndoStackOnReplace (DocumentChangeEventArgs args)
-		{
-			foreach (UndoOperation op in undoStack) {
-				op.InformTextReplace (args);
+
+		internal int UndoBeginOffset {
+			get {
+				if (undoStack.Count == 0)
+					return -1;
+				var op = undoStack.Peek ();
+				while (op is AtomicUndoOperation)
+					op = ((AtomicUndoOperation)op).Operations.FirstOrDefault ();
+				if (op == null)
+					return -1;
+				return ((UndoOperation)op).Args.Offset;
 			}
-			// since we're only displaying the undo stack it's not required to update the redo stack
-//			foreach (UndoOperation op in redoStack) {
-//				op.InformTextReplace (args);
-//			}
 		}
-		
+
+		internal int RedoBeginOffset {
+			get {
+				if (redoStack.Count == 0)
+					return -1;
+				var op = redoStack.Peek ();
+				while (op is AtomicUndoOperation)
+					op = ((AtomicUndoOperation)op).Operations.FirstOrDefault ();
+				if (op == null)
+					return -1;
+				return ((UndoOperation)op).Args.Offset;
+			}
+		}
+
 		public bool CanUndo {
 			get {
 				return this.undoStack.Count > 0 || currentAtomicOperation != null;
@@ -702,27 +722,21 @@ namespace Mono.TextEditor
 			Dirty,
 			Changed
 		}
+
+		public DiffTracker diffTracker = new DiffTracker ();
+
+		public DiffTracker DiffTracker {
+			get {
+				return diffTracker;
+			}
+			set {
+				diffTracker = value;
+			}
+		}
 		
 		public LineState GetLineState (DocumentLine line)
 		{
-			if (line == null)
-				return LineState.Unchanged;
-				
-			int lineOffset = line.Offset;
-			int lastLineEnd = line.Offset - line.DelimiterLength;
-			int lineEndOffset = lineOffset + line.LengthIncludingDelimiter;
-			foreach (UndoOperation op in undoStack) {
-				if (op.ChangedLine (lineOffset, lastLineEnd, lineEndOffset)) {
-					if (savePoint != null) {
-						foreach (UndoOperation savedUndo in savePoint) {
-							if (op == savedUndo)
-								return LineState.Changed;
-						}
-					}
-					return LineState.Dirty;
-				}
-			}
-			return LineState.Unchanged;
+			return diffTracker.GetLineState (line);
 		}
 		
 		
@@ -736,6 +750,7 @@ namespace Mono.TextEditor
 				((KeyboardStackUndo)undoStack.Peek ()).IsClosed = true;
 			savePoint = undoStack.ToArray ();
 			this.CommitUpdateAll ();
+			DiffTracker.SetBaseDocument (CreateDocumentSnapshot ());
 		}
 		
 		public void OptimizeTypedUndo ()
@@ -754,7 +769,7 @@ namespace Mono.TextEditor
 				undoStack.Push (keyUndo);
 				keyUndo = new KeyboardStackUndo ();
 			}
-			if (keyUndo.Args != null && keyUndo.Args.Offset + 1 != top.Args.Offset) {
+			if (keyUndo.Args != null && keyUndo.Args.Offset + 1 != top.Args.Offset || !char.IsLetterOrDigit (top.Args.InsertedText.GetCharAt (0))) {
 				keyUndo.IsClosed = true;
 				undoStack.Push (keyUndo);
 				keyUndo = new KeyboardStackUndo ();
@@ -795,15 +810,23 @@ namespace Mono.TextEditor
 				return;
 			OnBeforeUndoOperation (EventArgs.Empty);
 			isInUndo = true;
-			UndoOperation operation = undoStack.Pop ();
+			var operation = undoStack.Pop ();
 			redoStack.Push (operation);
 			operation.Undo (this);
 			isInUndo = false;
 			OnUndone (new UndoOperationEventArgs (operation));
-			this.RequestUpdate (new UpdateAll ());
-			this.CommitDocumentUpdate ();
 		}
-		
+
+		public void RollbackTo (ICSharpCode.NRefactory.Editor.ITextSourceVersion version)
+		{
+			var steps = Version.CompareAge (version);
+			if (steps < 0)
+				throw new InvalidOperationException ("Invalid version");
+			while (steps-- > 0) {
+				undoStack.Pop ().Undo (this);
+			}
+		}
+
 		internal protected virtual void OnUndone (UndoOperationEventArgs e)
 		{
 			EventHandler<UndoOperationEventArgs> handler = this.Undone;
@@ -838,8 +861,6 @@ namespace Mono.TextEditor
 			operation.Redo (this);
 			isInUndo = false;
 			OnRedone (new UndoOperationEventArgs (operation));
-			this.RequestUpdate (new UpdateAll ());
-			this.CommitDocumentUpdate ();
 		}
 		
 		internal protected virtual void OnRedone (UndoOperationEventArgs e)
@@ -850,7 +871,8 @@ namespace Mono.TextEditor
 		}
 		
 		public event EventHandler<UndoOperationEventArgs> Redone;
-		
+		 
+		Stack<OperationType> currentAtomicUndoOperationType =  new Stack<OperationType> ();
 		int atomicUndoLevel;
 
 		public bool IsInAtomicUndo {
@@ -858,17 +880,23 @@ namespace Mono.TextEditor
 				return atomicUndoLevel > 0;
 			}
 		}
+
+		public OperationType CurrentAtomicUndoOperationType {
+			get {
+				return currentAtomicUndoOperationType.Count > 0 ?  currentAtomicUndoOperationType.Peek () : OperationType.Undefined;
+			}
+		}
 		
 		class UndoGroup : IDisposable
 		{
 			TextDocument doc;
 			
-			public UndoGroup (TextDocument doc)
+			public UndoGroup (TextDocument doc, OperationType operationType)
 			{
 				if (doc == null)
 					throw new ArgumentNullException ("doc");
+				doc.BeginAtomicUndo (operationType);
 				this.doc = doc;
-				doc.BeginAtomicUndo ();
 			}
 
 			public void Dispose ()
@@ -882,18 +910,24 @@ namespace Mono.TextEditor
 		
 		public IDisposable OpenUndoGroup()
 		{
-			return new UndoGroup (this);
+			return OpenUndoGroup(OperationType.Undefined);
 		}
-		
-		internal void BeginAtomicUndo ()
+
+		public IDisposable OpenUndoGroup(OperationType operationType)
 		{
+			return new UndoGroup (this, operationType);
+		}
+
+		internal void BeginAtomicUndo (OperationType operationType = OperationType.Undefined)
+		{
+			currentAtomicUndoOperationType.Push (operationType);
 			if (atomicUndoLevel == 0) {
 				if (this.syntaxMode != null && !SuppressHighlightUpdate)
 					Mono.TextEditor.Highlighting.SyntaxModeService.WaitUpdate (this);
 			}
 			if (currentAtomicOperation == null) {
 				Debug.Assert (atomicUndoLevel == 0); 
-				currentAtomicOperation = new AtomicUndoOperation ();
+				currentAtomicOperation = new AtomicUndoOperation (operationType);
 				OnBeginUndo ();
 			}
 			atomicUndoLevel++;
@@ -920,6 +954,7 @@ namespace Mono.TextEditor
 				}
 				currentAtomicOperation = null;
 			}
+			currentAtomicUndoOperationType.Pop ();
 		}
 		
 		protected virtual void OnBeginUndo ()
@@ -979,36 +1014,47 @@ namespace Mono.TextEditor
 		}
 		
 		readonly object syncObject = new object();
-		BackgroundWorker foldSegmentWorker = null;
-		BackgroundWorker FoldSegmentWorker {
-			get {
-				if (foldSegmentWorker == null) {
-					foldSegmentWorker = new BackgroundWorker ();
-					foldSegmentWorker.WorkerSupportsCancellation = true;
-					foldSegmentWorker.DoWork += UpdateFoldSegmentWorker;
-				}
-				return foldSegmentWorker;
-			}
-		}
-		
-		public void UpdateFoldSegments (List<FoldSegment> newSegments)
-		{
-			UpdateFoldSegments (newSegments, true);
-		}
-		
 
-		public void UpdateFoldSegments (List<FoldSegment> newSegments, bool runInThread)
+		CancellationTokenSource foldSegmentSrc;
+		Task foldSegmentTask;
+
+		public void UpdateFoldSegments (List<FoldSegment> newSegments, bool startTask = false, bool useApplicationInvoke = false, CancellationToken masterToken = default(CancellationToken))
 		{
 			if (newSegments == null) {
 				return;
 			}
-			
+
 			InterruptFoldWorker ();
-			if (!runInThread) {
-				UpdateFoldSegmentWorker (null, new DoWorkEventArgs (newSegments));
+			bool update;
+			if (!startTask) {
+				var newFoldedSegments = UpdateFoldSegmentWorker (newSegments, out update);
+				if (useApplicationInvoke) {
+					Gtk.Application.Invoke (delegate {
+						foldedSegments = newFoldedSegments;
+						InformFoldTreeUpdated ();
+					});
+				} else {
+					foldedSegments = newFoldedSegments;
+					InformFoldTreeUpdated ();
+				}
 				return;
 			}
-			FoldSegmentWorker.RunWorkerAsync (newSegments);
+			foldSegmentSrc = new CancellationTokenSource ();
+			masterToken.Register (InterruptFoldWorker); 
+			var token = foldSegmentSrc.Token;
+			foldSegmentTask = Task.Factory.StartNew (delegate {
+				var segments = UpdateFoldSegmentWorker (newSegments, out update, token);
+				if (token.IsCancellationRequested)
+					return;
+				Gtk.Application.Invoke (delegate {
+					if (token.IsCancellationRequested)
+						return;
+					foldedSegments = segments;
+					InformFoldTreeUpdated ();
+					if (update)
+						CommitUpdateAll ();
+				});
+			}, token);
 		}
 		
 		void RemoveFolding (FoldSegment folding)
@@ -1023,18 +1069,18 @@ namespace Mono.TextEditor
 		/// Updates the fold segments in a background worker thread. Don't call this method outside of a background worker.
 		/// Use UpdateFoldSegments instead.
 		/// </summary>
-		public void UpdateFoldSegmentWorker (object sender, DoWorkEventArgs e)
+		HashSet<FoldSegment> UpdateFoldSegmentWorker (List<FoldSegment> newSegments, out bool update, CancellationToken token = default(CancellationToken))
 		{
-			var worker = sender as BackgroundWorker;
-			var newSegments = (List<FoldSegment>)e.Argument;
 			var oldSegments = new List<FoldSegment> (FoldSegments);
 			int oldIndex = 0;
 			bool foldedSegmentAdded = false;
 			newSegments.Sort ();
 			var newFoldedSegments = new HashSet<FoldSegment> ();
 			foreach (FoldSegment newFoldSegment in newSegments) {
-				if (worker != null && worker.CancellationPending)
-					return;
+				if (token.IsCancellationRequested) {
+					update = false;
+					return null;
+				}
 				int offset = newFoldSegment.Offset;
 				while (oldIndex < oldSegments.Count && offset > oldSegments [oldIndex].Offset) {
 					RemoveFolding (oldSegments [oldIndex]);
@@ -1073,35 +1119,33 @@ namespace Mono.TextEditor
 				}
 			}
 			while (oldIndex < oldSegments.Count) {
+				if (token.IsCancellationRequested) {
+					update = false;
+					return null;
+				}
 				RemoveFolding (oldSegments [oldIndex]);
 				oldIndex++;
 			}
 			bool countChanged = foldedSegments.Count != newFoldedSegments.Count;
-			if (worker != null) {
-				Gtk.Application.Invoke (delegate {
-					foldedSegments = newFoldedSegments;
-					InformFoldTreeUpdated ();
-					if (foldedSegmentAdded || countChanged)
-						CommitUpdateAll ();
-				});
-			} else {
-				foldedSegments = newFoldedSegments;
-				InformFoldTreeUpdated ();
-			}
+			update = foldedSegmentAdded || countChanged;
+			return newFoldedSegments;
 		}
 		
 		public void WaitForFoldUpdateFinished ()
 		{
-			while (FoldSegmentWorker.IsBusy)
-				System.Threading.Thread.Sleep (10);
+			if (foldSegmentTask != null) {
+				foldSegmentTask.Wait (5000);
+				foldSegmentTask = null;
+			}
 		}
 		
 		void InterruptFoldWorker ()
 		{
-			if (!FoldSegmentWorker.IsBusy)
+			if (foldSegmentSrc == null)
 				return;
-			FoldSegmentWorker.CancelAsync ();
+			foldSegmentSrc.Cancel ();
 			WaitForFoldUpdateFinished ();
+			foldSegmentSrc = null;
 		}
 		
 		public void ClearFoldSegments ()
@@ -1159,7 +1203,7 @@ namespace Mono.TextEditor
 		
 		public int GetLineCount (FoldSegment segment)
 		{
-			return OffsetToLineNumber(segment.EndLine.Offset) - OffsetToLineNumber(segment.StartLine.Offset);
+			return segment.EndLine.LineNumber - segment.StartLine.LineNumber;
 		}
 		
 		public void EnsureOffsetIsUnfolded (int offset)
@@ -1203,7 +1247,9 @@ namespace Mono.TextEditor
 		
 		public event EventHandler<FoldSegmentEventArgs> Folded;
 		#endregion
-		
+
+		#region Text line markers
+
 		public event EventHandler<TextMarkerEvent> MarkerAdded;
 		protected virtual void OnMarkerAdded (TextMarkerEvent e)
 		{
@@ -1221,28 +1267,28 @@ namespace Mono.TextEditor
 		}
 
 		
-		List<TextMarker> extendingTextMarkers = new List<TextMarker> ();
+		List<TextLineMarker> extendingTextMarkers = new List<TextLineMarker> ();
 		public IEnumerable<DocumentLine> LinesWithExtendingTextMarkers {
 			get {
 				return from marker in extendingTextMarkers where marker.LineSegment != null select marker.LineSegment;
 			}
 		}
 		
-		public void AddMarker (int lineNumber, TextMarker marker)
+		public void AddMarker (int lineNumber, TextLineMarker marker)
 		{
 			AddMarker (this.GetLine (lineNumber), marker);
 		}
 		
-		public void AddMarker (DocumentLine line, TextMarker marker)
+		public void AddMarker (DocumentLine line, TextLineMarker marker)
 		{
 			AddMarker (line, marker, true);
 		}
-		
-		public void AddMarker (DocumentLine line, TextMarker marker, bool commitUpdate)
+
+		public void AddMarker (DocumentLine line, TextLineMarker marker, bool commitUpdate)
 		{
 			if (line == null || marker == null)
 				return;
-			if (marker is IExtendingTextMarker) {
+			if (marker is IExtendingTextLineMarker) {
 				lock (extendingTextMarkers) {
 					HeightChanged = true;
 					extendingTextMarkers.Add (marker);
@@ -1255,26 +1301,26 @@ namespace Mono.TextEditor
 				this.CommitLineUpdate (line);
 		}
 		
-		static int CompareMarkers (TextMarker left, TextMarker right)
+		static int CompareMarkers (TextLineMarker left, TextLineMarker right)
 		{
 			if (left.LineSegment == null || right.LineSegment == null)
 				return 0;
 			return left.LineSegment.Offset.CompareTo (right.LineSegment.Offset);
 		}
 		
-		public void RemoveMarker (TextMarker marker)
+		public void RemoveMarker (TextLineMarker marker)
 		{
 			RemoveMarker (marker, true);
 		}
 		
-		public void RemoveMarker (TextMarker marker, bool updateLine)
+		public void RemoveMarker (TextLineMarker marker, bool updateLine)
 		{
 			if (marker == null)
 				return;
 			var line = marker.LineSegment;
 			if (line == null)
 				return;
-			if (marker is IExtendingTextMarker) {
+			if (marker is IExtendingTextLineMarker) {
 				lock (extendingTextMarkers) {
 					HeightChanged = true;
 					extendingTextMarkers.Remove (marker);
@@ -1304,10 +1350,10 @@ namespace Mono.TextEditor
 		{
 			if (line == null || type == null)
 				return;
-			if (typeof(IExtendingTextMarker).IsAssignableFrom (type)) {
+			if (typeof(IExtendingTextLineMarker).IsAssignableFrom (type)) {
 				lock (extendingTextMarkers) {
 					HeightChanged = true;
-					foreach (TextMarker marker in line.Markers.Where (marker => marker is IExtendingTextMarker)) {
+					foreach (TextLineMarker marker in line.Markers.Where (marker => marker is IExtendingTextLineMarker)) {
 						extendingTextMarkers.Remove (marker);
 					}
 				}
@@ -1316,16 +1362,64 @@ namespace Mono.TextEditor
 			if (updateLine)
 				this.CommitLineUpdate (line);
 		}
+
+		#endregion
+
+		#region Text segment markers
+
+		SegmentTree<TextSegmentMarker> textSegmentMarkerTree = new SegmentTree<TextSegmentMarker> (); 
+
+		public IEnumerable<TextSegmentMarker> GetTextSegmentMarkersAt (DocumentLine line)
+		{
+			return textSegmentMarkerTree.GetSegmentsOverlapping (line.Segment);
+		}
+
+		public IEnumerable<TextSegmentMarker> GetTextSegmentMarkersAt (TextSegment segment)
+		{
+			return textSegmentMarkerTree.GetSegmentsOverlapping (segment);
+		}
+
+		public IEnumerable<TextSegmentMarker> GetTextSegmentMarkersAt (int offset)
+		{
+			return textSegmentMarkerTree.GetSegmentsAt (offset);
+		}
 		
+
+		public void AddMarker (TextSegmentMarker marker)
+		{
+			textSegmentMarkerTree.Add (marker);
+			var startLine = OffsetToLineNumber (marker.Offset);
+			var endLine = OffsetToLineNumber (marker.EndOffset);
+			CommitMultipleLineUpdate (startLine, endLine);
+		}
+
+		/// <summary>
+		/// Removes a marker from the document.
+		/// </summary>
+		/// <returns><c>true</c>, if marker was removed, <c>false</c> otherwise.</returns>
+		/// <param name="marker">Marker.</param>
+		public bool RemoveMarker (TextSegmentMarker marker)
+		{
+			bool wasRemoved = textSegmentMarkerTree.Remove (marker);
+			if (wasRemoved) {
+				var startLine = OffsetToLineNumber (marker.Offset);
+				var endLine = OffsetToLineNumber (marker.EndOffset);
+				CommitMultipleLineUpdate (startLine, endLine);
+			}
+			return wasRemoved;
+		}
+
+		#endregion
+
 		void HandleSplitterLineSegmentTreeLineRemoved (object sender, LineEventArgs e)
 		{
-			foreach (TextMarker marker in e.Line.Markers) {
-				if (marker is IExtendingTextMarker) {
+			foreach (TextLineMarker marker in e.Line.Markers) {
+				if (marker is IExtendingTextLineMarker) {
 					lock (extendingTextMarkers) {
 						HeightChanged = true;
 						extendingTextMarkers.Remove (marker);
 					}
-					UnRegisterVirtualTextMarker ((IExtendingTextMarker)marker);
+					UnRegisterVirtualTextMarker ((IExtendingTextLineMarker)marker);
 				}
 			}
 		}
@@ -1391,7 +1485,7 @@ namespace Mono.TextEditor
 		
 		public void CommitLineUpdate (DocumentLine line)
 		{
-			CommitLineUpdate (this.OffsetToLineNumber (line.Offset));
+			CommitLineUpdate (line.LineNumber);
 		}
 
 		public void CommitUpdateAll ()
@@ -1419,17 +1513,11 @@ namespace Mono.TextEditor
 		}
 		
 		public static bool IsWordSeparator (char ch)
-
 		{
-
-			return Char.IsWhiteSpace (ch) || (Char.IsPunctuation (ch) && ch != '_');
-
+			return !(char.IsLetterOrDigit (ch) || ch == '_');
 		}
 
-		
-
 		public bool IsWholeWordAt (int offset, int length)
-
 		{
 			return (offset == 0 || IsWordSeparator (GetCharAt (offset - 1))) &&
 				   (offset + length == TextLength || IsWordSeparator (GetCharAt (offset + length)));
@@ -1540,15 +1628,15 @@ namespace Mono.TextEditor
 			}
 		}
 		
-		Dictionary<int, IExtendingTextMarker> virtualTextMarkers = new Dictionary<int, IExtendingTextMarker> ();
-		public void RegisterVirtualTextMarker (int lineNumber, IExtendingTextMarker marker)
+		Dictionary<int, IExtendingTextLineMarker> virtualTextMarkers = new Dictionary<int, IExtendingTextLineMarker> ();
+		public void RegisterVirtualTextMarker (int lineNumber, IExtendingTextLineMarker marker)
 		{
 			virtualTextMarkers[lineNumber] = marker;
 		}
 		
-		public IExtendingTextMarker GetExtendingTextMarker (int lineNumber)
+		public IExtendingTextLineMarker GetExtendingTextMarker (int lineNumber)
 		{
-			IExtendingTextMarker result;
+			IExtendingTextLineMarker result;
 			if (virtualTextMarkers.TryGetValue (lineNumber, out result))
 				return result;
 			return null;
@@ -1560,7 +1648,7 @@ namespace Mono.TextEditor
 		/// <param name='marker'>
 		/// marker.
 		/// </param>
-		public void UnRegisterVirtualTextMarker (IExtendingTextMarker marker)
+		public void UnRegisterVirtualTextMarker (IExtendingTextLineMarker marker)
 		{
 			var keys = new List<int> (from pair in virtualTextMarkers where pair.Value == marker select pair.Key);
 			keys.ForEach (key => { virtualTextMarkers.Remove (key); CommitLineUpdate (key); });
@@ -1695,14 +1783,29 @@ namespace Mono.TextEditor
 			return OffsetToLocation (offset);
 		}
 
+		void ICSharpCode.NRefactory.Editor.IDocument.Insert (int offset, ITextSource text)
+		{
+			Insert (offset, text.Text);
+		}
+
+		void ICSharpCode.NRefactory.Editor.IDocument.Replace (int offset, int count, ITextSource text)
+		{
+			Replace (offset, count, text.Text);
+		}
+
 		void ICSharpCode.NRefactory.Editor.IDocument.Insert (int offset, string text)
 		{
 			Insert (offset, text);
 		}
 
-		public void Insert (int offset, string text, ICSharpCode.NRefactory.Editor.AnchorMovementType defaultAnchorMovementType)
+		void ICSharpCode.NRefactory.Editor.IDocument.Insert (int offset, ITextSource text, AnchorMovementType anchorMovementType)
 		{
-			Insert (offset, text);
+			Insert (offset, text.Text, anchorMovementType);
+		}
+
+		void ICSharpCode.NRefactory.Editor.IDocument.Insert (int offset, string text, AnchorMovementType anchorMovementType)
+		{
+			Insert (offset, text, anchorMovementType);
 		}
 
 		void ICSharpCode.NRefactory.Editor.IDocument.StartUndoableAction ()
@@ -1729,6 +1832,16 @@ namespace Mono.TextEditor
 		#endregion
 
 		#region ITextSource implementation
+		void ICSharpCode.NRefactory.Editor.ITextSource.WriteTextTo (System.IO.TextWriter writer)
+		{
+			throw new NotImplementedException ();
+		}
+
+		void ICSharpCode.NRefactory.Editor.ITextSource.WriteTextTo (System.IO.TextWriter writer, int offset, int length)
+		{
+			throw new NotImplementedException ();
+		}
+
 		ICSharpCode.NRefactory.Editor.ITextSource ICSharpCode.NRefactory.Editor.ITextSource.CreateSnapshot ()
 		{
 			throw new NotImplementedException ();
@@ -1788,10 +1901,18 @@ namespace Mono.TextEditor
 			}
 		}
 
+		public TextDocument CreateDocumentSnapshot ()
+		{
+			return new SnapshotDocument (Text, Version);
+		}
+
 		ICSharpCode.NRefactory.Editor.IDocument ICSharpCode.NRefactory.Editor.IDocument.CreateDocumentSnapshot ()
 		{
 			return new SnapshotDocument (Text, Version);
 		}
+
+
+
 		#endregion
 	}
 	
