@@ -31,6 +31,7 @@ using System.IO;
 using System.Diagnostics;
 using Mono.TextEditor.Highlighting;
 using ICSharpCode.NRefactory.Editor;
+using Xwt.Drawing;
 
 namespace Mono.TextEditor
 {
@@ -41,9 +42,9 @@ namespace Mono.TextEditor
 	
 	public class TextEditorData : IDisposable
 	{
-		ITextEditorOptions options;
-		readonly TextDocument document; 
-		Caret      caret;
+		ITextEditorOptions    options;
+		TextDocument document; 
+		readonly Caret        caret;
 		
 		static Adjustment emptyAdjustment = new Adjustment (0, 0, 0, 0, 0, 0);
 		
@@ -78,8 +79,21 @@ namespace Mono.TextEditor
 				currentMode.AddedToEditor (this);
 				if (oldMode != null)
 					oldMode.RemovedFromEditor (this);
+				OnEditModeChanged (new EditModeChangedEventArgs (oldMode, currentMode));
 			}
 		}
+
+		protected virtual void OnEditModeChanged (EditModeChangedEventArgs e)
+		{
+			var handler = EditModeChanged;
+			if (handler != null)
+				handler (this, e);
+		}
+
+		/// <summary>
+		/// Occurs when the edit mode changed.
+		/// </summary>
+		public event EventHandler<EditModeChangedEventArgs> EditModeChanged;
 		
 		public TextEditor Parent {
 			get;
@@ -88,13 +102,13 @@ namespace Mono.TextEditor
 		
 		public string FileName {
 			get {
-				return Document.FileName;
+				return Document != null ? Document.FileName : null;
 			}
 		}
 		
 		public string MimeType {
 			get {
-				return Document.MimeType;
+				return Document != null ? Document.MimeType : null;
 			}
 		}
 
@@ -103,7 +117,63 @@ namespace Mono.TextEditor
 			protected set;
 		}
 
+		ISelectionSurroundingProvider selectionSurroundingProvider = new DefaultSelectionSurroundingProvider ();
+		public ISelectionSurroundingProvider SelectionSurroundingProvider {
+			get {
+				return selectionSurroundingProvider;
+			}
+			set {
+				if (value == null)
+					throw new ArgumentNullException ("surrounding provider needs to be != null");
+				selectionSurroundingProvider = value;
+			}
+		}
+
+		bool? customTabsToSpaces;
+		public bool TabsToSpaces {
+			get {
+				return customTabsToSpaces.HasValue ? customTabsToSpaces.Value : options.TabsToSpaces;
+			}
+			set {
+				customTabsToSpaces = value;
+			}
+		}
+		#region Tooltip providers
+		internal List<TooltipProvider> tooltipProviders = new List<TooltipProvider> ();
+		public IEnumerable<TooltipProvider> TooltipProviders {
+			get { return tooltipProviders; }
+		}
 		
+		/// <summary>
+		/// If set the tooltips wont show up.
+		/// </summary>
+		public bool SuppressTooltips {
+			get;
+			set;
+		}
+
+		public void ClearTooltipProviders ()
+		{
+			foreach (var tp in tooltipProviders) {
+				var disposableProvider = tp as IDisposable;
+				if (disposableProvider == null)
+					continue;
+				disposableProvider.Dispose ();
+			}
+			tooltipProviders.Clear ();
+		}
+		
+		public void AddTooltipProvider (TooltipProvider provider)
+		{
+			tooltipProviders.Add (provider);
+		}
+		
+		public void RemoveTooltipProvider (TooltipProvider provider)
+		{
+			tooltipProviders.Remove (provider);
+		}
+		#endregion
+
 		public TextEditorData () : this (new TextDocument ())
 		{
 		}
@@ -133,6 +203,7 @@ namespace Mono.TextEditor
 
 			HeightTree = new HeightTree (this);
 			HeightTree.Rebuild ();
+			IndentationTracker = new DefaultIndentationTracker (document);
 		}
 
 		void HandleFoldTreeUpdated (object sender, EventArgs e)
@@ -142,9 +213,13 @@ namespace Mono.TextEditor
 
 		void HandleDocTextSet (object sender, EventArgs e)
 		{
-			caret.SetDocument (document);
+			if (vadjustment != null)
+				vadjustment.Value = vadjustment.Lower;
+			if (hadjustment != null)
+				hadjustment.Value = hadjustment.Lower;
 			HeightTree.Rebuild ();
 			ClearSelection ();
+			caret.SetDocument (document);
 		}
 
 		public double GetLineHeight (DocumentLine line)
@@ -175,7 +250,7 @@ namespace Mono.TextEditor
 
 		void HandleTextReplaced (object sender, DocumentChangeEventArgs e)
 		{
-			caret.UpdateCaretPosition ();
+			caret.UpdateCaretPosition (e);
 		}
 
 
@@ -216,16 +291,55 @@ namespace Mono.TextEditor
 		ColorScheme colorStyle;
 		public ColorScheme ColorStyle {
 			get {
-				return colorStyle ?? ColorScheme.Empty;
+				return colorStyle ?? SyntaxModeService.DefaultColorStyle;
 			}
 			set {
 				colorStyle = value;
 			}
 		}
+
+		string ConvertToPangoMarkup (string str, bool replaceTabs = true)
+		{
+			if (str == null)
+				throw new ArgumentNullException ("str");
+			var result = new StringBuilder ();
+			foreach (char ch in str) {
+				switch (ch) {
+				case '&':
+					result.Append ("&amp;");
+					break;
+				case '<':
+					result.Append ("&lt;");
+					break;
+				case '>':
+					result.Append ("&gt;");
+					break;
+				case '\t':
+					if (replaceTabs) {
+						result.Append (new string (' ', options.TabSize));
+					} else {
+						result.Append ('\t');
+					}
+					break;
+				default:
+					result.Append (ch);
+					break;
+				}
+			}
+			return result.ToString ();
+		}
 		
 		public string GetMarkup (int offset, int length, bool removeIndent, bool useColors = true, bool replaceTabs = true)
 		{
 			ISyntaxMode mode = Document.SyntaxMode;
+			var style = ColorStyle;
+
+			if (style == null) {
+				var str = Document.GetTextAt (offset, length);
+				if (removeIndent)
+					str = str.TrimStart (' ', '\t');
+				return ConvertToPangoMarkup (str, replaceTabs);
+			}
 
 			int indentLength = SyntaxMode.GetIndentLength (Document, offset, length, false);
 			int curOffset = offset;
@@ -234,17 +348,17 @@ namespace Mono.TextEditor
 			while (curOffset < offset + length && curOffset < Document.TextLength) {
 				DocumentLine line = Document.GetLineByOffset (curOffset);
 				int toOffset = System.Math.Min (line.Offset + line.Length, offset + length);
-				Stack<ChunkStyle> styleStack = new Stack<ChunkStyle> ();
-				foreach (var chunk in mode.GetChunks (ColorStyle, line, curOffset, toOffset - curOffset)) {
+				var styleStack = new Stack<ChunkStyle> ();
 
-					ChunkStyle chunkStyle = ColorStyle.GetChunkStyle (chunk);
-					bool setBold = chunkStyle.Bold && (styleStack.Count == 0 || !styleStack.Peek ().Bold) ||
-							!chunkStyle.Bold && (styleStack.Count == 0 || styleStack.Peek ().Bold);
-					bool setItalic = chunkStyle.Italic && (styleStack.Count == 0 || !styleStack.Peek ().Italic) ||
-							!chunkStyle.Italic && (styleStack.Count == 0 || styleStack.Peek ().Italic);
+				foreach (var chunk in mode.GetChunks (style, line, curOffset, toOffset - curOffset)) {
+					var chunkStyle = style.GetChunkStyle (chunk);
+					bool setBold = (styleStack.Count > 0 && styleStack.Peek ().FontWeight != chunkStyle.FontWeight) || 
+						chunkStyle.FontWeight != FontWeight.Normal;
+					bool setItalic = (styleStack.Count > 0 && styleStack.Peek ().FontStyle != chunkStyle.FontStyle) || 
+						chunkStyle.FontStyle != FontStyle.Normal;
 					bool setUnderline = chunkStyle.Underline && (styleStack.Count == 0 || !styleStack.Peek ().Underline) ||
 							!chunkStyle.Underline && (styleStack.Count == 0 || styleStack.Peek ().Underline);
-					bool setColor = styleStack.Count == 0 || TextViewMargin.GetPixel (styleStack.Peek ().Color) != TextViewMargin.GetPixel (chunkStyle.Color);
+					bool setColor = styleStack.Count == 0 || TextViewMargin.GetPixel (styleStack.Peek ().Foreground) != TextViewMargin.GetPixel (chunkStyle.Foreground);
 					if (setColor || setBold || setItalic || setUnderline) {
 						if (styleStack.Count > 0) {
 							result.Append ("</span>");
@@ -253,43 +367,19 @@ namespace Mono.TextEditor
 						result.Append ("<span");
 						if (useColors) {
 							result.Append (" foreground=\"");
-							result.Append (SyntaxMode.ColorToPangoMarkup (chunkStyle.Color));
+							result.Append (SyntaxMode.ColorToPangoMarkup (chunkStyle.Foreground));
 							result.Append ("\"");
 						}
-						if (chunkStyle.Bold)
-							result.Append (" weight=\"bold\"");
-						if (chunkStyle.Italic)
-							result.Append (" style=\"italic\"");
+						if (chunkStyle.FontWeight != Xwt.Drawing.FontWeight.Normal)
+							result.Append (" weight=\"" + chunkStyle.FontWeight + "\"");
+						if (chunkStyle.FontStyle != Xwt.Drawing.FontStyle.Normal)
+							result.Append (" style=\"" + chunkStyle.FontStyle + "\"");
 						if (chunkStyle.Underline)
 							result.Append (" underline=\"single\"");
 						result.Append (">");
 						styleStack.Push (chunkStyle);
 					}
-
-					for (int i = 0; i < chunk.Length && chunk.Offset + i < Document.TextLength; i++) {
-						char ch = Document.GetCharAt (chunk.Offset + i);
-						switch (ch) {
-						case '&':
-							result.Append ("&amp;");
-							break;
-						case '<':
-							result.Append ("&lt;");
-							break;
-						case '>':
-							result.Append ("&gt;");
-							break;
-						case '\t':
-							if (replaceTabs) {
-								result.Append (new string (' ', options.TabSize));
-							} else {
-								result.Append ('\t');
-							}
-							break;
-						default:
-							result.Append (ch);
-							break;
-						}
-					}
+					result.Append (ConvertToPangoMarkup (Document.GetTextBetween (chunk.Offset, System.Math.Min (chunk.EndOffset, Document.TextLength)), replaceTabs));
 				}
 				while (styleStack.Count > 0) {
 					result.Append ("</span>");
@@ -335,8 +425,8 @@ namespace Mono.TextEditor
 			if (string.IsNullOrEmpty (str))
 				return "";
 			StringBuilder sb = new StringBuilder ();
-			bool convertTabs = Options.TabsToSpaces;
-			
+			bool convertTabs = TabsToSpaces;
+			var tabSize = Options.TabSize;
 			for (int i = 0; i < str.Length; i++) {
 				char ch = str [i];
 				switch (ch) {
@@ -345,7 +435,7 @@ namespace Mono.TextEditor
 					break;
 				case '\t':
 					if (convertTabs) {
-						int tabWidth = TextViewMargin.GetNextTabstop (this, loc.Column) - loc.Column;
+						int tabWidth = TextViewMargin.GetNextTabstop (this, loc.Column, tabSize) - loc.Column;
 						sb.Append (new string (' ', tabWidth));
 						loc = new DocumentLocation (loc.Line, loc.Column + tabWidth);
 					} else 
@@ -385,6 +475,11 @@ namespace Mono.TextEditor
 			if (String.IsNullOrEmpty (text))
 				return;
 			using (var undo = OpenUndoGroup ()) {
+				DeleteSelectedText (IsSomethingSelected ? MainSelection.SelectionMode != SelectionMode.Block : true);
+				// Needs to be called after delete text, delete text handles virtual caret postitions itself,
+				// but afterwards the virtual position may need to be restored.
+				EnsureCaretIsNotVirtual ();
+
 				if (IsSomethingSelected && MainSelection.SelectionMode == SelectionMode.Block) {
 					var visualInsertLocation = LogicalToVisualLocation (MainSelection.Anchor);
 					var selection = MainSelection;
@@ -404,10 +499,12 @@ namespace Mono.TextEditor
 						}
 						Insert (lineSegment.Offset + insertOffset, textToInsert);
 					}
+					var visualColumn = GetLine (Caret.Location.Line).GetVisualColumn (this, Caret.Column);
 					MainSelection = new Selection (
-								new DocumentLocation (selection.Anchor.Line, Caret.Column),
-								new DocumentLocation (selection.Lead.Line, Caret.Column),
-								Mono.TextEditor.SelectionMode.Block);
+						new DocumentLocation (selection.Anchor.Line, GetLine (selection.Anchor.Line).GetLogicalColumn (this, visualColumn)),
+						new DocumentLocation (selection.Lead.Line, GetLine (selection.Lead.Line).GetLogicalColumn (this, visualColumn)),
+						SelectionMode.Block
+					);
 					Caret.PreserveSelection = false;
 					Document.CommitMultipleLineUpdate (selection.MinLine, selection.MaxLine);
 				} else {
@@ -419,26 +516,33 @@ namespace Mono.TextEditor
 
 		void DetachDocument ()
 		{
+			if (document == null)
+				return;
 			document.BeginUndo -= OnBeginUndo;
 			document.EndUndo -= OnEndUndo;
 
 			document.Undone -= DocumentHandleUndone;
 			document.Redone -= DocumentHandleRedone;
 			document.LineChanged -= HandleDocLineChanged;
-			
+			document.TextReplaced -= HandleTextReplaced;
+
 			document.TextSet -= HandleDocTextSet;
 			document.Folded -= HandleTextEditorDataDocumentFolded;
 			document.FoldTreeUpdated -= HandleFoldTreeUpdated;
+			document = null;
 		}
 
 		public void Dispose ()
 		{
 			if (IsDisposed)
 				return;
+			document.WaitForFoldUpdateFinished ();
 			IsDisposed = true;
 			options = options.Kill ();
 			HeightTree.Dispose ();
 			DetachDocument ();
+			ClearTooltipProviders ();
+			tooltipProviders = null;
 		}
 
 		/// <summary>
@@ -505,17 +609,7 @@ namespace Mono.TextEditor
 		{
 			return Options.WordFindStrategy.FindCurrentWordStart (Document, offset);
 		}
-		
-		public delegate void PasteCallback (int insertionOffset, string text, int insertedChars);
-		
-		public event PasteCallback Paste;
-		
-		public void PasteText (int insertionOffset, string text, int insertedChars)
-		{
-			if (Paste != null)
-				Paste (insertionOffset, text, insertedChars);
-		}
-		
+
 		#region undo/redo handling
 		DocumentLocation savedCaretPos;
 		Selection savedSelection;
@@ -524,7 +618,7 @@ namespace Mono.TextEditor
 		void OnBeginUndo (object sender, EventArgs args)
 		{
 			savedCaretPos  = Caret.Location;
-			savedSelection = Selection.Clone (MainSelection);
+			savedSelection = MainSelection;
 		}
 
 		void OnEndUndo (object sender, TextDocument.UndoOperationEventArgs e)
@@ -565,7 +659,7 @@ namespace Mono.TextEditor
 				undoSelection = selection;
 				
 				redoCaretPos  = editor.Caret.Location;
-				redoSelection = Mono.TextEditor.Selection.Clone (editor.MainSelection);
+				redoSelection = editor.MainSelection;
 			}
 			
 			public void UndoState ()
@@ -585,7 +679,7 @@ namespace Mono.TextEditor
 		#region Selection management
 		public bool IsSomethingSelected {
 			get {
-				return MainSelection != null && MainSelection.Anchor != MainSelection.Lead; 
+				return !MainSelection.IsEmpty && MainSelection.Anchor != MainSelection.Lead; 
 			}
 		}
 		
@@ -628,39 +722,31 @@ namespace Mono.TextEditor
 		
 		public SelectionMode SelectionMode {
 			get {
-				return MainSelection != null ? MainSelection.SelectionMode : SelectionMode.Normal;
+				return !MainSelection.IsEmpty ? MainSelection.SelectionMode : SelectionMode.Normal;
 			}
 			set {
-				if (MainSelection != null)
-					MainSelection.SelectionMode = value;
+				if (MainSelection.IsEmpty)
+					return;
+				MainSelection = MainSelection.WithSelectionMode (value);
 			}
 		}
 		
-		Selection mainSelection = null;
+		Selection mainSelection = Selection.Empty;
 		public Selection MainSelection {
 			get {
 				return mainSelection;
 			}
 			set {
-				if (mainSelection == null && value == null)
+				if (mainSelection.IsEmpty && value.IsEmpty)
 					return;
-				if (mainSelection == null && value != null || mainSelection != null && value == null || !mainSelection.Equals (value)) {
+				if (mainSelection.IsEmpty && !value.IsEmpty || !mainSelection.IsEmpty && value.IsEmpty || !mainSelection.Equals (value)) {
 					OnSelectionChanging (EventArgs.Empty);
-					if (mainSelection != null)
-						mainSelection.Changed -= HandleMainSelectionChanged;
 					mainSelection = value;
-					if (mainSelection != null) 
-						mainSelection.Changed += HandleMainSelectionChanged;
 					OnSelectionChanged (EventArgs.Empty);
 				}
 			}
 		}
 
-		void HandleMainSelectionChanged (object sender, EventArgs e)
-		{
-			OnSelectionChanged (EventArgs.Empty);
-		}
-		
 		public IEnumerable<Selection> Selections {
 			get {
 				yield return MainSelection;
@@ -680,18 +766,20 @@ namespace Mono.TextEditor
 //		}
 		public int SelectionAnchor {
 			get {
-				if (MainSelection == null)
+				if (MainSelection.IsEmpty)
 					return -1;
 				return MainSelection.GetAnchorOffset (this);
 			}
 			set {
 				DocumentLocation location = Document.OffsetToLocation (value);
-				if (mainSelection == null) {
+				if (mainSelection.IsEmpty) {
 					MainSelection = new Selection (location, location);
 				} else {
-					if (MainSelection.Lead == location)
-						MainSelection.Lead = MainSelection.Anchor;
-					MainSelection.Anchor = location;
+					if (MainSelection.Lead == location) {
+						MainSelection = MainSelection.WithLead (MainSelection.Anchor);
+					} else {
+						MainSelection = MainSelection.WithAnchor (location);
+					}
 				}
 			}
 		}
@@ -701,30 +789,29 @@ namespace Mono.TextEditor
 		/// </summary>
 		public TextSegment SelectionRange {
 			get {
-				return MainSelection != null ? MainSelection.GetSelectionRange (this) : new TextSegment (Caret.Offset, 0);
+				return !MainSelection.IsEmpty ? MainSelection.GetSelectionRange (this) : new TextSegment (Caret.Offset, 0);
 			}
 			set {
 				if (SelectionRange != value) {
 					OnSelectionChanging (EventArgs.Empty);
 					if (value.IsEmpty) {
-						MainSelection = null;
+						MainSelection = Selection.Empty;
 					} else {
 						DocumentLocation loc1 = document.OffsetToLocation (value.Offset);
 						DocumentLocation loc2 = document.OffsetToLocation (value.EndOffset);
-						if (MainSelection == null) {
+						if (MainSelection.IsEmpty) {
 							MainSelection = new Selection (loc1, loc2);
 						} else {
 							if (MainSelection.Anchor == loc1) {
-								MainSelection.Lead = loc2;
+								MainSelection = MainSelection.WithLead (loc2);
 							} else if (MainSelection.Anchor == loc2) {
-								MainSelection.Lead = loc1;
+								MainSelection = MainSelection.WithLead (loc1);
 							} else {
 								MainSelection = new Selection (loc1, loc2);
 							}
 						}
 						
 					}
-					OnSelectionChanged (EventArgs.Empty);
 				}
 			}
 		}
@@ -766,15 +853,16 @@ namespace Mono.TextEditor
 		{
 			if (!IsSomethingSelected)
 				return;
-			MainSelection = null;
-			OnSelectionChanged (EventArgs.Empty);
+			MainSelection = Selection.Empty;
 		}
 		
 		public void ExtendSelectionTo (DocumentLocation location)
 		{
-			if (MainSelection == null)
+			if (MainSelection.IsEmpty) {
 				MainSelection = new Selection (location, location);
-			MainSelection.Lead = location;
+			} else {
+				MainSelection = MainSelection.WithLead (location);
+			}
 		}
 		
 		public void SetSelection (int anchorOffset, int leadOffset)
@@ -807,16 +895,18 @@ namespace Mono.TextEditor
 
 		internal void DeleteSelection (Selection selection)
 		{
-			if (selection == null)
-				throw new ArgumentNullException ("selection");
+			if (selection.IsEmpty)
+				throw new ArgumentNullException ("selection was empty.");
 			switch (selection.SelectionMode) {
 			case SelectionMode.Normal:
 				var segment = selection.GetSelectionRange (this);
-				if (Caret.Offset > segment.Offset)
-					Caret.Offset -= System.Math.Min (segment.Length, Caret.Offset - segment.Offset);
 				int len = System.Math.Min (segment.Length, Document.TextLength - segment.Offset);
+				var loc = selection.Anchor < selection.Lead ? selection.Anchor : selection.Lead;
+				caret.Location = loc;
+				EnsureCaretIsNotVirtual ();
 				if (len > 0)
 					Remove (segment.Offset, len);
+				caret.Location = loc;
 				break;
 			case SelectionMode.Block:
 				DocumentLocation visStart = LogicalToVisualLocation (selection.Anchor);
@@ -837,9 +927,10 @@ namespace Mono.TextEditor
 						Caret.Column = col1 + 1;
 				}
 				int column = System.Math.Min (selection.Anchor.Column, selection.Lead.Column);
-				selection.Anchor = new DocumentLocation (selection.Anchor.Line, column);
-				selection.Lead = new DocumentLocation (selection.Lead.Line, column);
-				Caret.Column = column;
+				MainSelection = selection.WithRange (
+					new DocumentLocation (selection.Anchor.Line, column),
+					new DocumentLocation (selection.Lead.Line, column)
+				);
 				Caret.PreserveSelection = preserve;
 				break;
 			}
@@ -859,6 +950,8 @@ namespace Mono.TextEditor
 			using (var undo = OpenUndoGroup ()) {
 				EnsureCaretIsNotVirtual ();
 				foreach (Selection selection in Selections) {
+					EnsureIsNotVirtual (selection.Anchor);
+					EnsureIsNotVirtual (selection.Lead);
 					var segment = selection.GetSelectionRange (this);
 					needUpdate |= Document.OffsetToLineNumber (segment.Offset) != Document.OffsetToLineNumber (segment.EndOffset);
 					DeleteSelection (selection);
@@ -942,6 +1035,9 @@ namespace Mono.TextEditor
 		
 		public SearchResult FindNext (bool setSelection)
 		{
+			if (SearchEngine.SearchRequest == null || string.IsNullOrEmpty (SearchEngine.SearchRequest.SearchPattern))
+				return null;
+
 			int startOffset = Caret.Offset;
 			if (IsSomethingSelected && IsMatchAt (startOffset)) {
 				startOffset = MainSelection.GetLeadOffset (this);
@@ -958,6 +1054,8 @@ namespace Mono.TextEditor
 		
 		public SearchResult FindPrevious (bool setSelection)
 		{
+			if (SearchEngine.SearchRequest == null || string.IsNullOrEmpty (SearchEngine.SearchRequest.SearchPattern))
+				return null;
 			int startOffset = Caret.Offset - SearchEngine.SearchRequest.SearchPattern.Length;
 			if (IsSomethingSelected && IsMatchAt (MainSelection.GetAnchorOffset (this))) 
 				startOffset = MainSelection.GetAnchorOffset (this);
@@ -996,24 +1094,7 @@ namespace Mono.TextEditor
 		
 		public int SearchReplaceAll (string withPattern)
 		{
-			int result = 0;
-			using (var undo = OpenUndoGroup ()) {
-				int offset = 0;
-				if (!SearchRequest.SearchRegion.IsInvalid)
-					offset = SearchRequest.SearchRegion.Offset;
-				SearchResult searchResult; 
-				while (true) {
-					searchResult = SearchForward (offset);
-					if (searchResult == null || searchResult.SearchWrapped)
-						break;
-					searchEngine.Replace (searchResult, withPattern);
-					offset = searchResult.Offset + withPattern.Length;
-					result++;
-				}
-				if (result > 0)
-					ClearSelection ();
-			}
-			return result;
+			return searchEngine.ReplaceAll (withPattern);
 		}
 		#endregion
 		
@@ -1072,28 +1153,47 @@ namespace Mono.TextEditor
 		/// </summary>
 		public int EnsureCaretIsNotVirtual ()
 		{
-			Debug.Assert (document.IsInAtomicUndo);
-			DocumentLine line = Document.GetLine (Caret.Line);
-			if (line == null)
+			return EnsureIsNotVirtual (Caret.Location);
+		}
+
+		public bool IsCaretInVirtualLocation {
+			get {
+				DocumentLine documentLine = Document.GetLine (Caret.Line);
+				if (documentLine == null)
+					return true;
+				return Caret.Column > documentLine.Length + 1;
+			}
+		}
+
+		int EnsureIsNotVirtual (DocumentLocation loc)
+		{
+			return EnsureIsNotVirtual (loc.Line, loc.Column);
+		}
+
+		int EnsureIsNotVirtual (int line, int column)
+		{
+			DocumentLine documentLine = Document.GetLine (line);
+			if (documentLine == null)
 				return 0;
-			if (Caret.Column > line.Length + 1) {
+			if (column > documentLine.Length + 1) {
 				string virtualSpace;
-				if (HasIndentationTracker && line.Length == 0) {
-					virtualSpace = GetIndentationString (Caret.Location);
+				if (HasIndentationTracker && documentLine.Length == 0) {
+					virtualSpace = GetIndentationString (line, column);
 				} else {
-					virtualSpace = new string (' ', Caret.Column - 1 - line.Length);
+					virtualSpace = new string (' ', column - 1 - documentLine.Length);
 				}
 				var oldPreserve = Caret.PreserveSelection;
 				Caret.PreserveSelection = true;
-				Insert (Caret.Offset, virtualSpace);
+				Insert (documentLine.Offset, virtualSpace);
 				Caret.PreserveSelection = oldPreserve;
-			
+				
 				// No need to reposition the caret, because it's already at the correct position
 				// The only difference is that the position is not virtual anymore.
 				return virtualSpace.Length;
 			}
 			return 0;
 		}
+
 		#endregion
 		
 		public Stream OpenStream ()
@@ -1123,6 +1223,52 @@ namespace Mono.TextEditor
 		}
 		public event EventHandler RecenterEditor;
 
+		#region Text Paste
+		/// <summary>
+		/// Gets or sets the text paste handler.
+		/// </summary>
+		public ITextPasteHandler TextPasteHandler {
+			get;
+			set;
+		}
+
+		public int PasteText (int offset, string text, byte[] copyData, ref IDisposable undoGroup)
+		{
+			if (TextPasteHandler != null) {
+				string newText;
+				try {
+					newText = TextPasteHandler.FormatPlainText (offset, text, copyData);
+				} catch (Exception e) {
+					Console.WriteLine ("Text paste handler exception:" + e);
+					newText = text;
+				}
+				if (newText != text) {
+					var inserted = Insert (offset, text);
+					if (options.GenerateFormattingUndoStep) {
+						undoGroup.Dispose ();
+						undoGroup = OpenUndoGroup ();
+					}
+					var result = Replace (offset, inserted, newText);
+					if (Paste != null)
+						Paste (offset, text, result);
+					return result;
+				}
+			}
+			var insertedChars = Insert (offset, text);
+			if (options.GenerateFormattingUndoStep) {
+				undoGroup.Dispose ();
+				undoGroup = OpenUndoGroup ();
+			}
+			if (Paste != null)
+				Paste (offset, text, insertedChars);
+			return insertedChars;
+		}
+
+		public delegate void PasteCallback (int insertionOffset, string text, int insertedChars);
+		
+		public event PasteCallback Paste;
+		#endregion
+
 		#region Document delegation
 		public int Length {
 			get {
@@ -1208,7 +1354,7 @@ namespace Mono.TextEditor
 		
 		public int LineCount {
 			get {
-				return Document.LineCount;
+				return Document != null ? Document.LineCount : 0;
 			}
 		}
 		
@@ -1256,15 +1402,15 @@ namespace Mono.TextEditor
 		{
 			return Document.OpenUndoGroup ();
 		}
+
+		public IDisposable OpenUndoGroup(OperationType operationType)
+		{
+			return Document.OpenUndoGroup (operationType);
+		}
 		#endregion
 		
 		#region Parent functions
-		public bool HasFocus {
-			get {
-				return Parent != null ? Parent.HasFocus : false;
-			}
-		}
-		
+
 		public void ScrollToCaret ()
 		{
 			if (Parent != null)
@@ -1381,8 +1527,8 @@ namespace Mono.TextEditor
 
 		void HandleTextEditorDataDocumentFolded (object sender, FoldSegmentEventArgs e)
 		{
-			int start = OffsetToLineNumber (e.FoldSegment.StartLine.Offset);
-			int end = OffsetToLineNumber (e.FoldSegment.EndLine.Offset);
+			int start = e.FoldSegment.StartLine.LineNumber;
+			int end = e.FoldSegment.EndLine.LineNumber;
 			
 			if (e.FoldSegment.IsFolded) {
 				if (e.FoldSegment.Marker != null)
