@@ -50,7 +50,7 @@ namespace MonoDevelop.Core.Serialization
 		
 		protected override void Initialize ()
 		{
-			DataItemAttribute atd = (DataItemAttribute) Context.AttributeProvider.GetCustomAttribute (ValueType, typeof(DataItemAttribute), false);
+			IDataItemAttribute atd = (IDataItemAttribute) Context.AttributeProvider.GetCustomAttribute (ValueType, typeof(IDataItemAttribute), false);
 			if (atd != null) {
 				if (!string.IsNullOrEmpty (atd.Name)) {
 					Name = atd.Name;
@@ -316,7 +316,7 @@ namespace MonoDevelop.Core.Serialization
 				object val = prop.GetValue (obj);
 				if (val == null)
 					continue;
-				if (!serCtx.Serializer.IncludeDefaultValues && val.Equals (prop.DefaultValue))
+				if (!serCtx.IsDefaultValueSerializationForced (prop) && val.Equals (prop.DefaultValue))
 					continue;
 				
 				DataCollection col = itemCol;
@@ -557,14 +557,21 @@ namespace MonoDevelop.Core.Serialization
 					}
 				}
 			}
-			if (mapData != null) {
-				isFallbackType = true;
-				return Context.GetConfigurationDataType ((Type)mapData);
-			}
+			
+			// Try a direct type name query
 			DataType dt = Context.GetConfigurationDataType (name);
 			if (dt != null && ValueType.IsAssignableFrom (dt.ValueType))
 				return dt;
 			
+			// The type could not be found. Use the fallback type if available.
+			
+			// Properties can specify a specific fallback type. It is provided in the custom map data.
+			if (mapData != null) {
+				isFallbackType = true;
+				return Context.GetConfigurationDataType ((Type)mapData);
+			}
+			
+			// A type can have a default fallback type.
 			if (fallbackType != null) {
 				isFallbackType = true;
 				return Context.GetConfigurationDataType (fallbackType);
