@@ -33,10 +33,13 @@ using System.IO;
 using System.Text;
 
 using Mono.Addins;
+using Mono.Unix.Native;
 using MonoDevelop.Core.FileSystem;
 using System.Collections.Generic;
 using System.Threading;
 using System.Linq;
+using System.Threading.Tasks;
+using System.Net;
 
 namespace MonoDevelop.Core
 {
@@ -47,11 +50,11 @@ namespace MonoDevelop.Core
 		static FileServiceErrorHandler errorHandler;
 		
 		static FileSystemExtension fileSystemChain;
-		static FileSystemExtension defaultExtension = Platform.IsWindows ? new DefaultFileSystemExtension () : new UnixFileSystemExtension () ;
+		static readonly FileSystemExtension defaultExtension = Platform.IsWindows ? new DefaultFileSystemExtension () : new UnixFileSystemExtension () ;
 		
-		static EventQueue eventQueue = new EventQueue ();
+		static readonly EventQueue eventQueue = new EventQueue ();
 		
-		static string applicationRootPath = Path.Combine (PropertyService.EntryAssemblyPath, "..");
+		static readonly string applicationRootPath = Path.Combine (PropertyService.EntryAssemblyPath, "..");
 		public static string ApplicationRootPath {
 			get {
 				return applicationRootPath;
@@ -64,6 +67,7 @@ namespace MonoDevelop.Core
 				if (args.PathChanged (addinFileSystemExtensionPath))
 					UpdateExtensions ();
 			};
+			UpdateExtensions ();
 		}
 		
 		static void UpdateExtensions ()
@@ -73,7 +77,7 @@ namespace MonoDevelop.Core
 				return;
 			}
 			
-			FileSystemExtension[] extensions = (FileSystemExtension[]) AddinManager.GetExtensionObjects (addinFileSystemExtensionPath, typeof(FileSystemExtension));
+			var extensions = AddinManager.GetExtensionObjects (addinFileSystemExtensionPath, typeof(FileSystemExtension)).Cast<FileSystemExtension> ().ToArray ();
 			for (int n=0; n<extensions.Length - 1; n++) {
 				extensions [n].Next = extensions [n + 1];
 			}
@@ -235,24 +239,59 @@ namespace MonoDevelop.Core
 				srcExt.DeleteDirectory (srcPath);
 			}
 		}
-		
-		public static bool RequestFileEdit (string fileName)
+
+		/// <summary>
+		/// Requests permission for modifying a file
+		/// </summary>
+		/// <param name="fileName">The file to be modified</param>
+		/// <param name="throwIfFails">If set to false, it will catch the exception that would've been thrown.</param>
+		/// <remarks>This method must be called before trying to write any file. It throws an exception if permission is not granted.</remarks>
+		public static bool RequestFileEdit (FilePath fileName, bool throwIfFails = true)
 		{
 			Debug.Assert (!String.IsNullOrEmpty (fileName));
-			return GetFileSystemForPath (fileName, false).RequestFileEdit (fileName);
+			return RequestFileEdit (new FilePath[] { fileName }, throwIfFails);
 		}
-		
+
+		/// <summary>
+		/// Requests permission for modifying a set of files
+		/// </summary>
+		/// <param name="fileNames">Files</param>
+		/// <remarks>This method must be called before trying to write any file. It throws an exception if permission is not granted.</remarks>
+		public static bool RequestFileEdit (IEnumerable<FilePath> fileNames, bool throwIfFails = true)
+		{
+			try {
+				foreach (var fg in fileNames.GroupBy (f => GetFileSystemForPath (f, false)))
+					fg.Key.RequestFileEdit (fg);
+				return true;
+			} catch (Exception ex) {
+				if (throwIfFails)
+					throw;
+				LoggingService.LogError ("File can't be written", ex);
+				return false;
+			}
+		}
+
 		public static void NotifyFileChanged (FilePath fileName)
 		{
-			NotifyFilesChanged (new FilePath[] { fileName });
+			NotifyFileChanged (fileName, false);
+		}
+
+		public static void NotifyFileChanged (FilePath fileName, bool autoReload)
+		{
+			NotifyFilesChanged (new FilePath[] { fileName }, autoReload);
 		}
 		
 		public static void NotifyFilesChanged (IEnumerable<FilePath> files)
+		{
+			NotifyFilesChanged (files, false);
+		}
+
+		public static void NotifyFilesChanged (IEnumerable<FilePath> files, bool autoReload)
 		{
 			try {
 				foreach (var fsFiles in files.GroupBy (f => GetFileSystemForPath (f, false)))
 					fsFiles.Key.NotifyFilesChanged (fsFiles);
-				OnFileChanged (new FileEventArgs (files, false));
+				OnFileChanged (new FileEventArgs (files, false, autoReload));
 			} catch (Exception ex) {
 				LoggingService.LogError ("File change notification failed", ex);
 			}
@@ -275,8 +314,6 @@ namespace MonoDevelop.Core
 		internal static FileSystemExtension GetFileSystemForPath (string path, bool isDirectory)
 		{
 			Debug.Assert (!String.IsNullOrEmpty (path));
-			if (fileSystemChain == null)
-				UpdateExtensions ();
 			FileSystemExtension nx = fileSystemChain;
 			while (nx != null && !nx.CanHandlePath (path, isDirectory))
 				nx = nx.Next;
@@ -330,8 +367,8 @@ namespace MonoDevelop.Core
 			if (!Path.IsPathRooted (absPath) || string.IsNullOrEmpty (baseDirectoryPath))
 				return absPath;
 				
-			absPath = Path.GetFullPath (absPath);
-			baseDirectoryPath = Path.GetFullPath (baseDirectoryPath.TrimEnd (Path.DirectorySeparatorChar));
+			absPath = GetFullPath (absPath);
+			baseDirectoryPath = GetFullPath (baseDirectoryPath).TrimEnd (Path.DirectorySeparatorChar);
 			
 			fixed (char* bPtr = baseDirectoryPath, aPtr = absPath) {
 				var bEnd = bPtr + baseDirectoryPath.Length;
@@ -354,18 +391,27 @@ namespace MonoDevelop.Core
 					a++;
 					b++;
 					if (b >= bEnd) {
-						indx++;
-						lastStartA = a + 1;
-						lastStartB = b;
+						if (a >= aEnd || IsSeparator (*a)) {
+							indx++;
+							lastStartA = a + 1;
+							lastStartB = b;
+						}
 						break;
 					}
 				}
-				
 				if (indx == 0) 
 					return absPath;
 				
 				if (lastStartA >= aEnd)
 					return ".";
+
+				// handle case a: some/path b: some/path/deeper...
+				if (a >= aEnd) {
+					if (IsSeparator (*b)) {
+						lastStartA = a + 1;
+						lastStartB = b;
+					}
+				}
 				
 				// look how many levels to go up into the base path
 				int goUpCount = 0;
@@ -374,7 +420,6 @@ namespace MonoDevelop.Core
 						goUpCount++;
 					lastStartB++;
 				}
-				
 				var size = goUpCount * 2 + goUpCount + aEnd - lastStartA;
 				var result = new char [size];
 				fixed (char* rPtr = result) {
@@ -418,16 +463,24 @@ namespace MonoDevelop.Core
 		
 		public static bool IsDirectory (string filename)
 		{
-			return Directory.Exists (filename) && (File.GetAttributes (filename) & FileAttributes.Directory) != 0;
+			return Directory.Exists (filename);
 		}
 		
 		public static string GetFullPath (string path)
 		{
 			if (path == null)
 				throw new ArgumentNullException ("path");
-			// Note: It's not required for Path.GetFullPath (path) that path exists.
-			return Path.GetFullPath (path); 
+			if (!Platform.IsWindows || path.IndexOf ('*') == -1)
+				return Path.GetFullPath (path);
+			else {
+				// On Windows, GetFullPath doesn't work if the path contains wildcards.
+				path = path.Replace ("*", wildcardMarker);
+				path = Path.GetFullPath (path);
+				return path.Replace (wildcardMarker, "*");
+			}
 		}
+
+		static readonly string wildcardMarker = "_" + Guid.NewGuid ().ToString () + "_";
 		
 		public static string CreateTempDirectory ()
 		{
@@ -455,6 +508,12 @@ namespace MonoDevelop.Core
 		// Atomic rename of a file. It does not fire events.
 		public static void SystemRename (string sourceFile, string destFile)
 		{
+			if (string.IsNullOrEmpty (sourceFile))
+				throw new ArgumentException ("sourceFile");
+
+			if (string.IsNullOrEmpty (destFile))
+				throw new ArgumentException ("destFile");
+
 			//FIXME: use the atomic System.IO.File.Replace on NTFS
 			if (Platform.IsWindows) {
 				string wtmp = null;
@@ -488,34 +547,73 @@ namespace MonoDevelop.Core
 				}
 			}
 			else {
-				Mono.Unix.Native.Syscall.rename (sourceFile, destFile);
+				if (Syscall.rename (sourceFile, destFile) != 0) {
+					switch (Stdlib.GetLastError ()) {
+					case Errno.EACCES:
+					case Errno.EPERM:
+						throw new UnauthorizedAccessException ();
+					case Errno.EINVAL:
+						throw new InvalidOperationException ();
+					case Errno.ENOTDIR:
+						throw new DirectoryNotFoundException ();
+					case Errno.ENOENT:
+						throw new FileNotFoundException ();
+					case Errno.ENAMETOOLONG:
+						throw new PathTooLongException ();
+					default:
+						throw new IOException ();
+					}
+				}
 			}
 		}
+
+		/// <summary>
+		/// Renames a directory
+		/// </summary>
+		/// <param name="sourceDir">Source directory</param>
+		/// <param name="destDir">Destination directory</param>
+		/// <remarks>
+		/// It works like Directory.Move, but it supports changing the case of a directory name in case-insensitive file systems
+		/// </remarks>
+		public static void SystemDirectoryRename (string sourceDir, string destDir)
+		{
+			if (Directory.Exists (destDir) && string.Equals (Path.GetFullPath (sourceDir), Path.GetFullPath (destDir), StringComparison.CurrentCultureIgnoreCase)) {
+				// If the destination directory exists but we can't find it with the provided name casing, then it means we are just changing the case
+				var existingDir = Directory.GetDirectories (Path.GetDirectoryName (destDir), Path.GetFileName (destDir)).FirstOrDefault ();
+				if (existingDir == null || (Path.GetFileName (existingDir) == Path.GetFileName (sourceDir))) {
+					var temp = destDir + ".renaming";
+					int n = 0;
+					while (Directory.Exists (temp) || File.Exists (temp))
+						temp = destDir + ".renaming_" + (n++);
+					Directory.Move (sourceDir, temp);
+					try {
+						Directory.Move (temp, destDir);
+					} catch {
+						Directory.Move (temp, sourceDir);
+					}
+					return;
+				}
+			}
+			Directory.Move (sourceDir, destDir);
+		}
 		
 		/// <summary>
 		/// Removes the directory if it's empty.
 		/// </summary>
 		public static void RemoveDirectoryIfEmpty (string directory)
 		{
-			if (Directory.Exists (directory) && !Directory.EnumerateFiles (directory).Any ())
+			// HACK: we should use EnumerateFiles but it's broken in some Mono releases
+			// https://bugzilla.xamarin.com/show_bug.cgi?id=2975
+			if (Directory.Exists (directory) && !Directory.GetFiles (directory).Any ())
 				Directory.Delete (directory);
 		}
 		
-		/// <summary>
-		/// Creates a directory if it does not already exist.
-		/// </summary>
-		public static void EnsureDirectoryExists (string directory)
-		{
-			if (!Directory.Exists (directory))
-				Directory.CreateDirectory (directory);
-		}
-		
 		/// <summary>
 		/// Makes the path separators native.
 		/// </summary>
 		public static string MakePathSeparatorsNative (string path)
 		{
-			if (path == null || path.Length == 0)
+			if (string.IsNullOrEmpty (path))
 				return path;
 			char c = Path.DirectorySeparatorChar == '\\'? '/' : '\\'; 
 			return path.Replace (c, Path.DirectorySeparatorChar);
@@ -586,6 +684,75 @@ namespace MonoDevelop.Core
 			Counters.FileChangeNotifications++;
 			eventQueue.RaiseEvent (() => FileChanged, null, args);
 		}
+
+		public static Task<bool> UpdateDownloadedCacheFile (string url, string cacheFile,
+			Func<Stream,bool> validateDownload = null, CancellationToken ct = default (CancellationToken))
+		{
+			return WebRequestHelper.GetResponseAsync (
+				() => (HttpWebRequest)WebRequest.Create (url),
+				r => {
+					//check to see if the online file has been modified since it was last downloaded
+					var localNewsXml = new FileInfo (cacheFile);
+					if (localNewsXml.Exists)
+						r.IfModifiedSince = localNewsXml.LastWriteTime;
+				},
+				ct
+			).ContinueWith (t => {
+				bool deleteTempFile = true;
+				var tempFile = cacheFile + ".temp";
+
+				try {
+					ct.ThrowIfCancellationRequested ();
+
+					if (t.IsFaulted) {
+						var wex = t.Exception.Flatten ().InnerException as WebException;
+						if (wex != null) {
+							var resp = wex.Response as HttpWebResponse;
+							if (resp != null && resp.StatusCode == HttpStatusCode.NotModified)
+								return false;
+						}
+					}
+
+					//TODO: limit this size in case open wifi hotspots provide junk data
+					var response = t.Result;
+					if (response.StatusCode == HttpStatusCode.OK) {
+						using (var fs = File.Create (tempFile))
+								response.GetResponseStream ().CopyTo (fs, 2048);
+					}
+
+					//check the document is valid, might get bad ones from wifi hotspots etc
+					if (validateDownload != null) {
+						ct.ThrowIfCancellationRequested ();
+
+						using (var f = File.OpenRead (tempFile)) {
+							bool validated;
+							try {
+								validated = validateDownload (f);
+							} catch (Exception ex) {
+								throw new Exception ("Failed to validate downloaded file", ex);
+							}
+							if (!validated) {
+								throw new Exception ("Failed to validate downloaded file");
+							}
+						}
+					}
+
+					ct.ThrowIfCancellationRequested ();
+
+					SystemRename (tempFile, cacheFile);
+					deleteTempFile = false;
+					return true;
+				} finally {
+					if (deleteTempFile) {
+						try {
+							File.Delete (tempFile);
+						} catch (Exception ex) {
+							LoggingService.LogError ("Failed to delete temp download file", ex);
+						}
+					}
+				}
+			}, ct);
+		}
 	}
 	
 	class EventQueue
@@ -662,8 +829,11 @@ namespace MonoDevelop.Core
 					return;
 				}
 			}
-			if (del != null)
-				del.DynamicInvoke (thisObj, args);
+			if (del != null) {
+				Runtime.MainSynchronizationContext.Post (delegate {
+					del.DynamicInvoke (thisObj, args);
+				}, null);
+			}
 		}
 	}
 	
