@@ -31,41 +31,52 @@ using System.IO;
 
 using Mono.Unix;
 using System.Globalization;
+using System.Runtime.InteropServices;
+using System.Threading;
 
 namespace MonoDevelop.Core
 {
 	public static class GettextCatalog
 	{
+		static Thread mainThread;
+
+		[DllImport ("kernel32.dll", SetLastError = true)]
+		static extern int SetThreadUILanguage (int LangId);
+
 		static GettextCatalog ()
 		{
+			mainThread = Thread.CurrentThread;
+
 			//variable can be used to override where Gettext looks for the catalogues
-			string catalog = System.Environment.GetEnvironmentVariable ("MONODEVELOP_LOCALE_PATH");
+			string catalog = Environment.GetEnvironmentVariable ("MONODEVELOP_LOCALE_PATH");
 
 			// Set the user defined language
 			string lang = PropertyService.Get ("MonoDevelop.Ide.UserInterfaceLanguage", "");
 			if (!string.IsNullOrEmpty (lang)) {
-				if (PropertyService.IsWindows) {
+				if (Platform.IsWindows) {
 					lang = lang.Replace("_", "-");
 					CultureInfo ci = CultureInfo.GetCultureInfo(lang);
 					if (ci.IsNeutralCulture) {
-						// We need a neutral culture
+						// We need a non-neutral culture
 						foreach (CultureInfo c in CultureInfo.GetCultures (CultureTypes.AllCultures & ~CultureTypes.NeutralCultures))
 							if (c.Parent != null && c.Parent.Name == ci.Name) {
 								ci = c;
 								break;
 							}
 					}
-					if (!ci.IsNeutralCulture)
-						System.Threading.Thread.CurrentThread.CurrentCulture = ci;
+					if (!ci.IsNeutralCulture) {
+						SetThreadUILanguage (ci.LCID);
+						mainThread.CurrentUICulture = ci;
+					}
 				}
 				else
 					Environment.SetEnvironmentVariable ("LANGUAGE", lang);
 			}
 			
-			if (string.IsNullOrEmpty (catalog)) {
+			if (string.IsNullOrEmpty (catalog) || !Directory.Exists (catalog)) {
 				string location = System.Reflection.Assembly.GetExecutingAssembly ().Location;
 				location = Path.GetDirectoryName (location);
-				if (PropertyService.IsWindows) {
+				if (Platform.IsWindows) {
 					// On windows, load the catalog from a child dir
 					catalog = Path.Combine (location, "locale");
 				}
@@ -73,6 +84,8 @@ namespace MonoDevelop.Core
 					// MD is located at $prefix/lib/monodevelop/bin
 					// adding "../../.." should give us $prefix
 					string prefix = Path.Combine (Path.Combine (Path.Combine (location, ".."), ".."), "..");
+					if (Platform.IsMac)
+						prefix = Path.Combine (prefix, "..", "MacOS");
 					//normalise it
 					prefix = Path.GetFullPath (prefix);
 					//catalogue is installed to "$prefix/share/locale" by default
@@ -86,6 +99,10 @@ namespace MonoDevelop.Core
 				Console.WriteLine (ex);
 			}
 		}
+
+		public static CultureInfo UICulture {
+			get { return mainThread.CurrentUICulture; }
+		}
 		
 		#region GetString
 		
