@@ -26,33 +26,41 @@
 //
 
 using System;
+using System.Linq;
 using System.Diagnostics;
 using System.Collections.Generic;
 using System.IO;
-using System.Xml;
 using System.Globalization;
 using System.Runtime.Serialization.Formatters.Binary;
 using Mono.Addins;
+using MonoDevelop.Core.ProgressMonitoring;
 using MonoDevelop.Projects;
 using MonoDevelop.Projects.Extensions;
 using MonoDevelop.Core.Serialization;
 using MonoDevelop.Core;
 using MonoDevelop.Core.Assemblies;
+using Cecil = Mono.Cecil;
+using System.Threading;
 
 namespace MonoDevelop.Projects.Formats.MSBuild
 {
 	public static class MSBuildProjectService
 	{
 		const string ItemTypesExtensionPath = "/MonoDevelop/ProjectModel/MSBuildItemTypes";
-		public const string GenericItemGuid = "{9344bdbb-3e7f-41fc-a0dd-8665d75ee146}";
+		public const string GenericItemGuid = "{9344BDBB-3E7F-41FC-A0DD-8665D75EE146}";
 		public const string FolderTypeGuid = "{2150E333-8FDC-42A3-9474-1A3956D46DE8}";
-
-		public const string DefaultFormat = "MSBuild08";
+		
+		//NOTE: default toolsversion should match the default format.
+		// remember to update the builder process' app.config too
+		public const string DefaultFormat = "MSBuild12";
 		
 		static DataContext dataContext;
 		
-		static Dictionary<TargetRuntime,RemoteBuildEngine> builders = new Dictionary<TargetRuntime, RemoteBuildEngine> ();
+		static IMSBuildGlobalPropertyProvider[] globalPropertyProviders;
+		static Dictionary<string,RemoteBuildEngine> builders = new Dictionary<string, RemoteBuildEngine> ();
 		static GenericItemTypeNode genericItemTypeNode = new GenericItemTypeNode ();
+
+		internal static bool ShutDown { get; private set; }
 		
 		public static DataContext DataContext {
 			get {
@@ -62,6 +70,8 @@ namespace MonoDevelop.Projects.Formats.MSBuild
 					foreach (ItemMember prop in MSBuildProjectHandler.ExtendedMSBuildProperties) {
 						ItemProperty iprop = new ItemProperty (prop.Name, prop.Type);
 						iprop.IsExternal = prop.IsExternal;
+						if (prop.CustomAttributes != null)
+							iprop.CustomAttributes = prop.CustomAttributes;
 						dataContext.RegisterProperty (prop.DeclaringType, iprop);
 					}
 				}
@@ -74,23 +84,62 @@ namespace MonoDevelop.Projects.Formats.MSBuild
 			Services.ProjectService.DataContextChanged += delegate {
 				dataContext = null;
 			};
+
+			PropertyService.PropertyChanged += HandlePropertyChanged;
+			DefaultMSBuildVerbosity = PropertyService.Get ("MonoDevelop.Ide.MSBuildVerbosity", MSBuildVerbosity.Normal);
+
+			Runtime.ShuttingDown += (sender, e) => ShutDown = true;
+
+			const string gppPath = "/MonoDevelop/ProjectModel/MSBuildGlobalPropertyProviders";
+			globalPropertyProviders = AddinManager.GetExtensionObjects<IMSBuildGlobalPropertyProvider> (gppPath);
+			foreach (var gpp in globalPropertyProviders) {
+				gpp.GlobalPropertiesChanged += HandleGlobalPropertyProviderChanged;
+			}
 		}
+
+		static void HandleGlobalPropertyProviderChanged (object sender, EventArgs e)
+		{
+			lock (builders) {
+				var gpp = (IMSBuildGlobalPropertyProvider) sender;
+				foreach (var builder in builders.Values)
+					builder.SetGlobalProperties (gpp.GetGlobalProperties ());
+			}
+		}
+
+		static void HandlePropertyChanged (object sender, PropertyChangedEventArgs e)
+		{
+			if (e.Key == "MonoDevelop.Ide.MSBuildVerbosity") {
+				DefaultMSBuildVerbosity = (MSBuildVerbosity) e.NewValue;
+			}
+		}
+
+		internal static MSBuildVerbosity DefaultMSBuildVerbosity { get; private set; }
 		
-		public static SolutionEntityItem LoadItem (IProgressMonitor monitor, string fileName, string typeGuid, string itemGuid)
+		public static SolutionEntityItem LoadItem (IProgressMonitor monitor, string fileName, MSBuildFileFormat expectedFormat, string typeGuid, string itemGuid)
 		{
 			foreach (ItemTypeNode node in GetItemTypeNodes ()) {
 				if (node.CanHandleFile (fileName, typeGuid))
-					return node.LoadSolutionItem (monitor, fileName, itemGuid);
+					return node.LoadSolutionItem (monitor, fileName, expectedFormat, itemGuid);
 			}
 			
 			if (string.IsNullOrEmpty (typeGuid) && IsProjectSubtypeFile (fileName)) {
 				typeGuid = LoadProjectTypeGuids (fileName);
 				foreach (ItemTypeNode node in GetItemTypeNodes ()) {
 					if (node.CanHandleFile (fileName, typeGuid))
-						return node.LoadSolutionItem (monitor, fileName, itemGuid);
+						return node.LoadSolutionItem (monitor, fileName, expectedFormat, itemGuid);
 				}
 			}
 
+			// If it is a known unsupported project, load it as UnknownProject
+			var projectInfo = MSBuildProjectService.GetUnknownProjectTypeInfo (typeGuid != null ? new [] { typeGuid } : new string[0], fileName);
+			if (projectInfo != null && projectInfo.LoadFiles) {
+				if (typeGuid == null)
+					typeGuid = projectInfo.Guid;
+				var h = new MSBuildProjectHandler (typeGuid, "", itemGuid);
+				h.SetUnsupportedType (projectInfo);
+				return h.Load (monitor, fileName, expectedFormat, "", null);
+			}
+
 			return null;
 		}
 		
@@ -144,27 +193,58 @@ namespace MonoDevelop.Projects.Formats.MSBuild
 			}
 		}
 		
+		public static bool SupportsProjectType (string projectFile)
+		{
+			if (!string.IsNullOrWhiteSpace (projectFile)) {
+				// If we have a project file, try to load it.
+				try {
+					using (var monitor = new ConsoleProgressMonitor ()) {
+						return MSBuildProjectService.LoadItem (monitor, projectFile, null, null, null) != null;
+					}
+				} catch {
+					return false;
+				}
+			}
+
+			return false;
+		}
+
+		public static void CheckHandlerUsesMSBuildEngine (SolutionItem item, out bool useByDefault, out bool require)
+		{
+			var handler = item.ItemHandler as MSBuildProjectHandler;
+			if (handler == null) {
+				useByDefault = require = false;
+				return;
+			}
+			useByDefault = handler.UseMSBuildEngineByDefault;
+			require = handler.RequireMSBuildEngine;
+		}
+
 		internal static DotNetProjectSubtypeNode GetDotNetProjectSubtype (string typeGuids)
 		{
-			if (!string.IsNullOrEmpty (typeGuids)) {
-				Type ptype = null;
-				DotNetProjectSubtypeNode foundNode = null;
-				foreach (string guid in typeGuids.Split (';')) {
-					string tguid = guid.Trim ();
-					foreach (DotNetProjectSubtypeNode st in GetItemSubtypeNodes ()) {
-						if (st.SupportsType (tguid)) {
-							if (ptype == null || ptype.IsAssignableFrom (st.Type)) {
-								ptype = st.Type;
-								foundNode = st;
-							}
+			if (!string.IsNullOrEmpty (typeGuids))
+				return GetDotNetProjectSubtype (typeGuids.Split (';').Select (t => t.Trim ()));
+			else
+				return null;
+		}
+
+		internal static DotNetProjectSubtypeNode GetDotNetProjectSubtype (IEnumerable<string> typeGuids)
+		{
+			Type ptype = null;
+			DotNetProjectSubtypeNode foundNode = null;
+			foreach (string guid in typeGuids) {
+				foreach (DotNetProjectSubtypeNode st in GetItemSubtypeNodes ()) {
+					if (st.SupportsType (guid)) {
+						if (ptype == null || ptype.IsAssignableFrom (st.Type)) {
+							ptype = st.Type;
+							foundNode = st;
 						}
 					}
 				}
-				return foundNode;
 			}
-			return null;
+			return foundNode;
 		}
-		
+
 		static IEnumerable<ItemTypeNode> GetItemTypeNodes ()
 		{
 			foreach (ExtensionNode node in AddinManager.GetExtensionNodes (ItemTypesExtensionPath)) {
@@ -174,7 +254,7 @@ namespace MonoDevelop.Projects.Formats.MSBuild
 			yield return genericItemTypeNode;
 		}
 		
-		static IEnumerable<DotNetProjectSubtypeNode> GetItemSubtypeNodes ()
+		internal static IEnumerable<DotNetProjectSubtypeNode> GetItemSubtypeNodes ()
 		{
 			foreach (ExtensionNode node in AddinManager.GetExtensionNodes (ItemTypesExtensionPath)) {
 				if (node is DotNetProjectSubtypeNode)
@@ -182,22 +262,22 @@ namespace MonoDevelop.Projects.Formats.MSBuild
 			}
 		}
 		
-		internal static ItemTypeNode FindHandlerForFile (FilePath file)
+		internal static bool CanReadFile (FilePath file)
 		{
 			foreach (ItemTypeNode node in GetItemTypeNodes ()) {
 				if (node.CanHandleFile (file, null)) {
-					return node;
+					return true;
 				}
 			}
 			if (IsProjectSubtypeFile (file)) {
 				string typeGuids = LoadProjectTypeGuids (file);
 				foreach (ItemTypeNode node in GetItemTypeNodes ()) {
 					if (node.CanHandleFile (file, typeGuids)) {
-						return node;
+						return true;
 					}
 				}
 			}
-			return null;
+			return GetUnknownProjectTypeInfo (new string[0], file) != null;
 		}
 		
 		internal static string GetExtensionForItem (SolutionEntityItem item)
@@ -224,7 +304,7 @@ namespace MonoDevelop.Projects.Formats.MSBuild
 			return false;
 		}
 		
-		static char[] specialCharacters = new char [] {'%', '$', '@', '(', ')', '\'', ';', '?', '*' };
+		static char[] specialCharacters = new char [] {'%', '$', '@', '(', ')', '\'', ';', '?' };
 		
 		public static string EscapeString (string str)
 		{
@@ -235,6 +315,17 @@ namespace MonoDevelop.Projects.Formats.MSBuild
 			}
 			return str;
 		}
+
+		public static string UnescapePath (string path)
+		{
+			if (string.IsNullOrEmpty (path))
+				return path;
+			
+			if (!Platform.IsWindows)
+				path = path.Replace ("\\", "/");
+			
+			return UnscapeString (path);
+		}
 		
 		public static string UnscapeString (string str)
 		{
@@ -250,12 +341,10 @@ namespace MonoDevelop.Projects.Formats.MSBuild
 		
 		public static string ToMSBuildPath (string baseDirectory, string absPath)
 		{
-			absPath = EscapeString (absPath);
 			if (baseDirectory != null) {
-				absPath = FileService.NormalizeRelativePath (FileService.AbsoluteToRelativePath (
-				         baseDirectory, absPath));
+				absPath = FileService.NormalizeRelativePath (FileService.AbsoluteToRelativePath (baseDirectory, absPath));
 			}
-			return absPath.Replace ('/', '\\');
+			return EscapeString (absPath).Replace ('/', '\\');
 		}
 		
 		internal static string ToMSBuildPathRelative (string baseDirectory, string absPath)
@@ -271,7 +360,7 @@ namespace MonoDevelop.Projects.Formats.MSBuild
 			return file.ToRelative (basePath);
 		}
 
-		internal static string FromMSBuildPath (string basePath, string relPath)
+		public static string FromMSBuildPath (string basePath, string relPath)
 		{
 			string res;
 			FromMSBuildPath (basePath, relPath, out res);
@@ -294,70 +383,79 @@ namespace MonoDevelop.Projects.Formats.MSBuild
 			if (string.IsNullOrEmpty (relPath))
 				return false;
 			
-			string path = relPath;
-			if (!PropertyService.IsWindows)
-				path = path.Replace ("\\", "/");
-			
-			path = UnscapeString (path);
+			string path = UnescapePath (relPath);
 
 			if (char.IsLetter (path [0]) && path.Length > 1 && path[1] == ':') {
-				if (PropertyService.IsWindows) {
+				if (Platform.IsWindows) {
 					resultPath = path; // Return the escaped value
 					return true;
 				} else
 					return false;
 			}
 			
-			if (basePath != null)
-				path = Path.Combine (basePath, path);
+			bool isRooted = Path.IsPathRooted (path);
 			
+			if (!isRooted && basePath != null) {
+				path = Path.Combine (basePath, path);
+				isRooted = Path.IsPathRooted (path);
+			}
+			
+			// Return relative paths as-is, we can't do anything else with them
+			if (!isRooted) {
+				resultPath = FileService.NormalizeRelativePath (path);
+				return true;
+			}
+			
+			// If we're on Windows, don't need to fix file casing.
+			if (Platform.IsWindows) {
+				resultPath = FileService.GetFullPath (path);
+				return true;
+			}
+			
+			// If the path exists with the exact casing specified, then we're done
 			if (System.IO.File.Exists (path) || System.IO.Directory.Exists (path)){
 				resultPath = Path.GetFullPath (path);
 				return true;
 			}
-				
-			if (Path.IsPathRooted (path) && !PropertyService.IsWindows) {
-					
-				// Windows paths are case-insensitive. When mapping an absolute path
-				// we can try to find the correct case for the path.
-				
-				string[] names = path.Substring (1).Split ('/');
-				string part = "/";
-				
-				for (int n=0; n<names.Length; n++) {
-					string[] entries;
+			
+			// Not on Windows, file doesn't exist. That could mean the project was brought from Windows
+			// and the filename case in the project doesn't match the file on disk, because Windows is
+			// case-insensitive. Since we have an absolute path, search the directory for the file with
+			// the correct case.
+			string[] names = path.Substring (1).Split ('/');
+			string part = "/";
+			
+			for (int n=0; n<names.Length; n++) {
+				string[] entries;
 
-					if (names [n] == ".."){
-						if (part == "/")
-							return false; // Can go further back. It's not an existing file
-						part = Path.GetFullPath (part + "/..");
-						continue;
-					}
-					
-					entries = Directory.GetFileSystemEntries (part);
-					
-					string fpath = null;
-					foreach (string e in entries) {
-						if (string.Compare (Path.GetFileName (e), names[n], true) == 0) {
-							fpath = e;
-							break;
-						}
-					}
-					if (fpath == null) {
-						// Part of the path does not exist. Can't do any more checking.
-						part = Path.GetFullPath (part);
-						for (; n < names.Length; n++)
-							part += "/" + names[n];
-						resultPath = part;
-						return true;
-					}
-
-					part = fpath;
+				if (names [n] == ".."){
+					if (part == "/")
+						return false; // Can go further back. It's not an existing file
+					part = Path.GetFullPath (part + "/..");
+					continue;
 				}
-				resultPath = Path.GetFullPath (part);
-			} else {
-				resultPath = Path.GetFullPath (path);
+				
+				entries = Directory.GetFileSystemEntries (part);
+				
+				string fpath = null;
+				foreach (string e in entries) {
+					if (string.Compare (Path.GetFileName (e), names [n], StringComparison.OrdinalIgnoreCase) == 0) {
+						fpath = e;
+						break;
+					}
+				}
+				if (fpath == null) {
+					// Part of the path does not exist. Can't do any more checking.
+					part = Path.GetFullPath (part);
+					for (; n < names.Length; n++)
+						part += "/" + names[n];
+					resultPath = part;
+					return true;
+				}
+
+				part = fpath;
 			}
+			resultPath = Path.GetFullPath (part);
 			return true;
 		}
 
@@ -401,99 +499,136 @@ namespace MonoDevelop.Projects.Formats.MSBuild
 			return true;
 		}
 		
-		static RemoteBuildEngine currentBuildEngine;
-		
-		public static RemoteProjectBuilder GetProjectBuilder (TargetRuntime runtime, TargetFramework fx, string file)
+		internal static RemoteProjectBuilder GetProjectBuilder (TargetRuntime runtime, string minToolsVersion, string file, string solutionFile)
 		{
 			lock (builders) {
-				string binDir = runtime.GetMSBuildBinPath (fx);
-				
+				//attempt to use 12.0 builder first if available
+				string toolsVersion = "12.0";
+				string binDir = runtime.GetMSBuildBinPath ("12.0");
+				if (binDir == null) {
+					//fall back to 4.0, we know it's always available
+					toolsVersion = "4.0";
+				}
+
+				//check the ToolsVersion we found can handle the project
+				Version tv, mtv;
+				if (Version.TryParse (toolsVersion, out tv) && Version.TryParse (minToolsVersion, out mtv) && tv < mtv) {
+					string error = null;
+					if (runtime is MsNetTargetRuntime && minToolsVersion == "12.0")
+						error = "MSBuild 2013 is not installed. Please download and install it from " +
+						"http://www.microsoft.com/en-us/download/details.aspx?id=40760";
+					throw new InvalidOperationException (error ?? string.Format (
+						"Runtime '{0}' does not have MSBuild '{1}' ToolsVersion installed",
+						runtime.Id, toolsVersion)
+					);
+				}
+
+				//one builder per solution
+				string builderKey = runtime.Id + " # " + solutionFile;
 				RemoteBuildEngine builder;
-				if (builders.TryGetValue (runtime, out builder)) {
+				if (builders.TryGetValue (builderKey, out builder)) {
 					builder.ReferenceCount++;
-					return new RemoteProjectBuilder (file, binDir, builder);
+					return new RemoteProjectBuilder (file, builder);
+				}
+
+				//always start the remote process explicitly, even if it's using the current runtime and fx
+				//else it won't pick up the assembly redirects from the builder exe
+				var exe = GetExeLocation (runtime, toolsVersion);
+				MonoDevelop.Core.Execution.RemotingService.RegisterRemotingChannel ();
+				var pinfo = new ProcessStartInfo (exe) {
+					UseShellExecute = false,
+					CreateNoWindow = true,
+					RedirectStandardError = true,
+					RedirectStandardInput = true,
+				};
+				runtime.GetToolsExecutionEnvironment ().MergeTo (pinfo);
+				
+				Process p = null;
+				try {
+					p = runtime.ExecuteAssembly (pinfo);
+					p.StandardInput.WriteLine (Process.GetCurrentProcess ().Id.ToString ());
+					string responseKey = "[MonoDevelop]";
+					string sref;
+					while (true) {
+						sref = p.StandardError.ReadLine ();
+						if (sref.StartsWith (responseKey, StringComparison.Ordinal)) {
+							sref = sref.Substring (responseKey.Length);
+							break;
+						}
+					}
+					byte[] data = Convert.FromBase64String (sref);
+					MemoryStream ms = new MemoryStream (data);
+					BinaryFormatter bf = new BinaryFormatter ();
+					var engine = (IBuildEngine)bf.Deserialize (ms);
+					engine.SetCulture (GettextCatalog.UICulture);
+					engine.SetGlobalProperties (GetCoreGlobalProperties (solutionFile));
+					foreach (var gpp in globalPropertyProviders)
+						builder.SetGlobalProperties (gpp.GetGlobalProperties ());
+					builder = new RemoteBuildEngine (p, engine);
+				} catch {
+					if (p != null) {
+						try {
+							p.Kill ();
+						} catch { }
+					}
+					throw;
 				}
 				
-				if (runtime.IsRunning) {
-					if (currentBuildEngine == null)
-						currentBuildEngine = new RemoteBuildEngine (null, new BuildEngine ());
-					return new RemoteProjectBuilder (file, binDir, currentBuildEngine);
-				}
-				else {
-					MonoDevelop.Core.Execution.RemotingService.RegisterRemotingChannel ();
-					string exe = typeof(ProjectBuilder).Assembly.Location;
-					ProcessStartInfo pinfo = new ProcessStartInfo (exe);
-					runtime.GetToolsExecutionEnvironment (fx).MergeTo (pinfo);
-					pinfo.UseShellExecute = false;
-					pinfo.RedirectStandardError = true;
-					pinfo.RedirectStandardInput = true;
-					
-					Process p = null;
-					try {
-						p = runtime.ExecuteAssembly (pinfo, fx);
-						p.StandardInput.WriteLine (Process.GetCurrentProcess ().Id.ToString ());
-						string sref = p.StandardError.ReadLine ();
-						byte[] data = Convert.FromBase64String (sref);
-						MemoryStream ms = new MemoryStream (data);
-						BinaryFormatter bf = new BinaryFormatter ();
-						builder = new RemoteBuildEngine (p, (IBuildEngine) bf.Deserialize (ms));
-					} catch {
-						if (p != null) {
-							try {
-								p.Kill ();
-							} catch { }
-						}
-						throw;
-					}
-				}
-				builders [runtime] = builder;
+				builders [builderKey] = builder;
 				builder.ReferenceCount = 1;
-				return new RemoteProjectBuilder (file, binDir, builder);
+				builder.Disconnected += delegate {
+					lock (builders)
+						builders.Remove (builderKey);
+				};
+				return new RemoteProjectBuilder (file, builder);
 			}
 		}
 
+		static IDictionary<string,string> GetCoreGlobalProperties (string slnFile)
+		{
+			var dictionary = new Dictionary<string,string> ();
+
+			//this causes build targets to behave how they should inside an IDE, instead of in a command-line process
+			dictionary.Add ("BuildingInsideVisualStudio", "true");
+
+			//we don't have host compilers in MD, and this is set to true by some of the MS targets
+			//which causes it to always run the CoreCompile task if BuildingInsideVisualStudio is also
+			//true, because the VS in-process compiler would take care of the deps tracking
+			dictionary.Add ("UseHostCompilerIfAvailable", "false" );
+
+			if (string.IsNullOrEmpty (slnFile))
+				return dictionary;
+
+			dictionary.Add ("SolutionPath", Path.GetFullPath (slnFile));
+			dictionary.Add ("SolutionName", Path.GetFileNameWithoutExtension (slnFile));
+			dictionary.Add ("SolutionFilename", Path.GetFileName (slnFile));
+			dictionary.Add ("SolutionDir", Path.GetDirectoryName (slnFile) + Path.DirectorySeparatorChar);
+
+			return dictionary;;
+		}
+		
+		static string GetExeLocation (TargetRuntime runtime, string toolsVersion)
+		{
+			FilePath sourceExe = typeof(MSBuildProjectService).Assembly.Location;
+
+			if ((runtime is MsNetTargetRuntime) && int.Parse (toolsVersion.Split ('.')[0]) >= 4)
+				toolsVersion = "dotnet." + toolsVersion;
+
+			var exe = sourceExe.ParentDirectory.Combine ("MSBuild", toolsVersion, "MonoDevelop.Projects.Formats.MSBuild.exe");
+			if (File.Exists (exe))
+				return exe;
+			
+			throw new InvalidOperationException ("Unsupported MSBuild ToolsVersion '" + toolsVersion + "'");
+		}
+
 		internal static void ReleaseProjectBuilder (RemoteBuildEngine engine)
 		{
 			lock (builders) {
-				if (engine.ReferenceCount > 0) {
-					if (--engine.ReferenceCount == 0) {
-						engine.ReleaseTime = DateTime.Now.AddSeconds (3);
-						ScheduleProjectBuilderCleanup (engine.ReleaseTime.AddMilliseconds (500));
-					}
-				}
-			}
-		}
-		
-		static DateTime nextCleanup = DateTime.MinValue;
-		
-		static void ScheduleProjectBuilderCleanup (DateTime cleanupTime)
-		{
-			lock (builders) {
-				if (cleanupTime < nextCleanup)
+				if (--engine.ReferenceCount != 0)
 					return;
-				nextCleanup = cleanupTime;
-				System.Threading.ThreadPool.QueueUserWorkItem (delegate {
-					DateTime tnow = DateTime.Now;
-					while (tnow < nextCleanup) {
-						System.Threading.Thread.Sleep ((int)(nextCleanup - tnow).TotalMilliseconds);
-						CleanProjectBuilders ();
-						tnow = DateTime.Now;
-					}
-				});
-			}
-		}
-		
-		static void CleanProjectBuilders ()
-		{
-			lock (builders) {
-				DateTime tnow = DateTime.Now;
-				foreach (var val in new Dictionary<TargetRuntime,RemoteBuildEngine> (builders)) {
-					if (val.Value.ReferenceCount == 0 && val.Value.ReleaseTime <= tnow) {
-						builders.Remove (val.Key);
-						val.Value.Dispose ();
-					}
-				}
+				builders.Remove (builders.First (kvp => kvp.Value == engine).Key);
 			}
+			engine.Dispose ();
 		}
 
 		static Dictionary<string, string> cultureNamesTable;
@@ -512,14 +647,27 @@ namespace MonoDevelop.Projects.Formats.MSBuild
 		static string LoadProjectTypeGuids (string fileName)
 		{
 			MSBuildProject project = new MSBuildProject ();
-			project.LoadXml (File.ReadAllText (fileName));
+			project.Load (fileName);
 			
-			MSBuildPropertyGroup globalGroup = project.GetGlobalPropertyGroup ();
+			MSBuildPropertySet globalGroup = project.GetGlobalPropertyGroup ();
 			if (globalGroup == null)
 				return null;
 
 			return globalGroup.GetPropertyValue ("ProjectTypeGuids");
 		}
+
+		internal static UnknownProjectTypeNode GetUnknownProjectTypeInfo (string[] guids, string fileName = null)
+		{
+			var ext = fileName != null ? Path.GetExtension (fileName).TrimStart ('.') : null;
+			var nodes = AddinManager.GetExtensionNodes<UnknownProjectTypeNode> ("/MonoDevelop/ProjectModel/UnknownMSBuildProjectTypes")
+				.Where (p => guids.Any (p.MatchesGuid) || (ext != null && p.Extension == ext)).ToList ();
+			return nodes.FirstOrDefault (n => !n.IsSolvable) ?? nodes.FirstOrDefault (n => n.IsSolvable);
+		}
+
+		public static MSBuildProjectHandler GetHandler (Project project)
+		{
+			return (MSBuildProjectHandler) project.GetItemHandler ();
+		}
 	}
 	
 	class MSBuildDataContext: DataContext
@@ -528,6 +676,8 @@ namespace MonoDevelop.Projects.Formats.MSBuild
 		{
 			if (type == typeof(bool))
 				return new MSBuildBoolDataType ();
+			else if (type == typeof(bool?))
+				return new MSBuildNullableBoolDataType ();
 			else
 				return base.CreateConfigurationDataType (type);
 		}
@@ -541,7 +691,7 @@ namespace MonoDevelop.Projects.Formats.MSBuild
 		
 		internal protected override DataNode OnSerialize (SerializationContext serCtx, object mapData, object value)
 		{
-			return new DataValue (Name, (bool)value ? "true" : "false");
+			return new MSBuildBoolDataValue (Name, (bool) value);
 		}
 		
 		internal protected override object OnDeserialize (SerializationContext serCtx, object mapData, DataNode data)
@@ -549,6 +699,48 @@ namespace MonoDevelop.Projects.Formats.MSBuild
 			return String.Equals (((DataValue)data).Value, "true", StringComparison.OrdinalIgnoreCase);
 		}
 	}
+
+	class MSBuildBoolDataValue : DataValue
+	{
+		public MSBuildBoolDataValue (string name, bool value)
+			: base (name, value ? "True" : "False")
+		{
+			RawValue = value;
+		}
+
+		public bool RawValue { get; private set; }
+	}
+
+	class MSBuildNullableBoolDataType: PrimitiveDataType
+	{
+		public MSBuildNullableBoolDataType (): base (typeof(bool))
+		{
+		}
+
+		internal protected override DataNode OnSerialize (SerializationContext serCtx, object mapData, object value)
+		{
+			return new MSBuildNullableBoolDataValue (Name, (bool?) value);
+		}
+
+		internal protected override object OnDeserialize (SerializationContext serCtx, object mapData, DataNode data)
+		{
+			var d = (DataValue)data;
+			if (string.IsNullOrEmpty (d.Value))
+				return (bool?) null;
+			return (bool?) String.Equals (d.Value, "true", StringComparison.OrdinalIgnoreCase);
+		}
+	}
+
+	class MSBuildNullableBoolDataValue : DataValue
+	{
+		public MSBuildNullableBoolDataValue (string name, bool? value)
+			: base (name, value.HasValue? (value.Value? "True" : "False") : null)
+		{
+			RawValue = value;
+		}
+
+		public bool? RawValue { get; private set; }
+	}
 	
 	public class MSBuildResourceHandler: IResourceHandler
 	{
@@ -593,10 +785,10 @@ namespace MonoDevelop.Projects.Formats.MSBuild
 			return true;
 		}
 		
-		public override SolutionEntityItem LoadSolutionItem (IProgressMonitor monitor, string fileName, string itemGuid)
+		public override SolutionEntityItem LoadSolutionItem (IProgressMonitor monitor, string fileName, MSBuildFileFormat expectedFormat, string itemGuid)
 		{
 			MSBuildProjectHandler handler = new MSBuildProjectHandler (Guid, Import, itemGuid);
-			return handler.Load (monitor, fileName, null, null);
+			return handler.Load (monitor, fileName, expectedFormat, null, null);
 		}
 	}
 }
