@@ -34,17 +34,17 @@ using System.Globalization;
 using System.IO;
 using System.Text;
 using System.Text.RegularExpressions;
-using System.Xml;
 
 using MonoDevelop.Projects;
 using MonoDevelop.Core.Serialization;
 using MonoDevelop.Projects.Extensions;
 using MonoDevelop.Core;
-using MonoDevelop.Core.ProgressMonitoring;
+using System.Reflection;
+using System.Linq;
 
 namespace MonoDevelop.Projects.Formats.MSBuild
 {
-	internal class SlnFileFormat
+	public class SlnFileFormat
 	{
 		public string GetValidFormatName (object obj, string fileName, MSBuildFileFormat format)
 		{
@@ -53,10 +53,10 @@ namespace MonoDevelop.Projects.Formats.MSBuild
 		
 		public bool CanReadFile (string file, MSBuildFileFormat format)
 		{
-			if (String.Compare (Path.GetExtension (file), ".sln", true) == 0) {
+			if (String.Compare (Path.GetExtension (file), ".sln", StringComparison.OrdinalIgnoreCase) == 0) {
 				string tmp;
 				string version = GetSlnFileVersion (file, out tmp);
-				return version == format.SlnVersion;
+				return format.SupportsSlnVersion (version);
 			}
 			return false;
 		}
@@ -122,9 +122,14 @@ namespace MonoDevelop.Projects.Formats.MSBuild
 				slnData.UpdateVersion (format);
 
 				sw.WriteLine ();
+
 				//Write Header
 				sw.WriteLine ("Microsoft Visual Studio Solution File, Format Version " + slnData.VersionString);
 				sw.WriteLine (slnData.HeaderComment);
+				if (slnData.VisualStudioVersion != null)
+					sw.WriteLine ("VisualStudioVersion = {0}", slnData.VisualStudioVersion);
+				if (slnData.MinimumVisualStudioVersion != null)
+					sw.WriteLine ("MinimumVisualStudioVersion = {0}", slnData.MinimumVisualStudioVersion);
 
 				//Write the projects
 				monitor.BeginTask (GettextCatalog.GetString ("Saving projects"), 1);
@@ -241,6 +246,16 @@ namespace MonoDevelop.Projects.Formats.MSBuild
 						WriteDataItem (writer, data);
 						writer.WriteLine ("\tEndProjectSection");
 					}
+					if (item.ItemDependencies.Count > 0 || handler.UnresolvedProjectDependencies != null) {
+						writer.WriteLine ("\tProjectSection(ProjectDependencies) = postProject");
+						foreach (var dep in item.ItemDependencies)
+							writer.WriteLine ("\t\t{0} = {0}", dep.ItemId);
+						if (handler.UnresolvedProjectDependencies != null) {
+							foreach (var dep in handler.UnresolvedProjectDependencies)
+								writer.WriteLine ("\t\t{0} = {0}", dep);
+						}
+						writer.WriteLine ("\tEndProjectSection");
+					}
 				} else if (ce is SolutionFolder) {
 					//Solution
 					SlnData slnData = GetSlnData (ce);
@@ -302,14 +317,31 @@ namespace MonoDevelop.Projects.Formats.MSBuild
 
 				foreach (SolutionConfigurationEntry cce in cc.Configurations) {
 					SolutionEntityItem p = cce.Item;
-					list.Add (String.Format (
-						"\t\t{0}.{1}.ActiveCfg = {2}", p.ItemId, ToSlnConfigurationId (cc), ToSlnConfigurationId (cce.ItemConfiguration)));
+
+					// Don't save configurations for shared projects
+					if (!p.SupportsConfigurations ())
+						continue;
+					
+                    // <ProjectGuid>...</ProjectGuid> in some Visual Studio generated F# project files 
+                    // are missing "{"..."}" in their guid. This is not generally a problem since it
+                    // is a valid GUID format. However the solution file format requires that these are present. 
+                    string itemGuid = p.ItemId;
+                    if (!itemGuid.StartsWith("{") && !itemGuid.EndsWith("}")) 
+                        itemGuid = "{" + itemGuid + "}";
+
+                    list.Add (String.Format (
+                        "\t\t{0}.{1}.ActiveCfg = {2}", itemGuid, ToSlnConfigurationId (cc), ToSlnConfigurationId (cce.ItemConfiguration)));
 
 					if (cce.Build)
 						list.Add (String.Format (
-							"\t\t{0}.{1}.Build.0 = {2}", p.ItemId, ToSlnConfigurationId (cc), ToSlnConfigurationId (cce.ItemConfiguration)));
+                            "\t\t{0}.{1}.Build.0 = {2}", itemGuid, ToSlnConfigurationId (cc), ToSlnConfigurationId (cce.ItemConfiguration)));
+					
+					if (cce.Deploy)
+						list.Add (String.Format (
+							"\t\t{0}.{1}.Deploy.0 = {2}", itemGuid, ToSlnConfigurationId (cc), ToSlnConfigurationId (cce.ItemConfiguration)));
 				}
 			}
+			
 		}
 
 		void WriteNestedProjects (SolutionFolder folder, SolutionFolder root, StreamWriter writer)
@@ -322,7 +354,7 @@ namespace MonoDevelop.Projects.Formats.MSBuild
 		{
 			// Find a project section of type MonoDevelopProperties
 			int start, end;
-			if (!FindSection (lines, "MonoDevelopProperties", out start, out end))
+			if (!FindSection (lines, "MonoDevelopProperties", true, out start, out end))
 				return null;
 			
 			// Deserialize the object
@@ -333,13 +365,33 @@ namespace MonoDevelop.Projects.Formats.MSBuild
 			return it;
 		}
 		
+		List<string> ReadSolutionItemDependencies (List<string> lines)
+		{
+			// Find a project section of type MonoDevelopProperties
+			int start, end;
+			if (!FindSection (lines, "ProjectDependencies", false, out start, out end))
+				return null;
+
+			var ids = new List<string> ();
+			for (int n=start + 1; n < end; n++) {
+				string line = lines [n];
+				int i = line.IndexOf ('=');
+				if (i != -1)
+					ids.Add (line.Substring (0, i).Trim ());
+			}
+
+			// Remove the lines, since they have already been preocessed
+			lines.RemoveRange (start, end - start + 1);
+			return ids;
+		}
+
 		List<string> ReadFolderFiles (List<string> lines)
 		{
 			// Find a solution item section of type SolutionItems
 
 			List<string> list = new List<string> ();
 			int start, end;
-			if (!FindSection (lines, "SolutionItems", out start, out end))
+			if (!FindSection (lines, "SolutionItems", true, out start, out end))
 				return list;
 			
 			for (int n=start + 1; n < end; n++) {
@@ -357,14 +409,17 @@ namespace MonoDevelop.Projects.Formats.MSBuild
 			return list;
 		}
 		
-		bool FindSection (List<string> lines, string name, out int start, out int end)
+		bool FindSection (List<string> lines, string name, bool preProject, out int start, out int end)
 		{
 			start = -1;
 			end = -1;
+
+			string prePost = preProject ? "preProject" : "postProject";
+			var sectionLine = "ProjectSection(" + name + ")=" + prePost;
 			
 			for (int n=0; n<lines.Count && start == -1; n++) {
 				string line = lines [n].Replace ("\t","").Replace (" ", "");
-				if (line == "ProjectSection(" + name + ")=preProject")
+				if (line == sectionLine)
 					start = n;
 			}
 			if (start == -1)
@@ -584,6 +639,9 @@ namespace MonoDevelop.Projects.Formats.MSBuild
 				ProjectExtensionUtil.BeginLoadOperation ();
 				sol = new Solution ();
 				monitor.BeginTask (string.Format (GettextCatalog.GetString ("Loading solution: {0}"), fileName), 1);
+				var projectLoadMonitor = monitor as IProjectLoadProgressMonitor;
+				if (projectLoadMonitor != null)
+					projectLoadMonitor.CurrentSolution = sol;
 				LoadSolution (sol, fileName, format, monitor);
 			} catch (Exception ex) {
 				monitor.ReportError (GettextCatalog.GetString ("Could not load solution: {0}", fileName), ex);
@@ -633,12 +691,12 @@ namespace MonoDevelop.Projects.Formats.MSBuild
 				while (reader.Peek () >= 0) {
 					s = GetNextLine (reader, lines).Trim ();
 
-					if (String.Compare (s, "Global", true) == 0) {
+					if (String.Compare (s, "Global", StringComparison.OrdinalIgnoreCase) == 0) {
 						ParseGlobal (reader, lines, globals);
 						continue;
 					}
 
-					if (s.StartsWith ("Project")) {
+					if (s.StartsWith ("Project", StringComparison.Ordinal)) {
 						Section sec = new Section ();
 						projectSections.Add (sec);
 
@@ -649,6 +707,22 @@ namespace MonoDevelop.Projects.Formats.MSBuild
 
 						continue;
 					}
+
+					if (s.StartsWith ("VisualStudioVersion = ", StringComparison.Ordinal)) {
+						Version v;
+						if (Version.TryParse (s.Substring ("VisualStudioVersion = ".Length), out v))
+							data.VisualStudioVersion = v;
+						else
+							monitor.Log.WriteLine ("Ignoring unparseable VisualStudioVersion value in sln file");
+					}
+
+					if (s.StartsWith ("MinimumVisualStudioVersion = ", StringComparison.Ordinal)) {
+						Version v;
+						if (Version.TryParse (s.Substring ("MinimumVisualStudioVersion = ".Length), out v))
+							data.MinimumVisualStudioVersion = v;
+						else
+							monitor.Log.WriteLine ("Ignoring unparseable MinimumVisualStudioVersion value in sln file");
+					}
 				}
 			}
 
@@ -682,7 +756,8 @@ namespace MonoDevelop.Projects.Formats.MSBuild
 				string projTypeGuid = match.Groups [1].Value.ToUpper ();
 				string projectName = match.Groups [2].Value;
 				string projectPath = match.Groups [3].Value;
-				string projectGuid = match.Groups [4].Value;
+				string projectGuid = match.Groups [4].Value.ToUpper ();
+				List<string> projLines;
 
 				if (projTypeGuid == MSBuildProjectService.FolderTypeGuid) {
 					//Solution folder
@@ -691,7 +766,7 @@ namespace MonoDevelop.Projects.Formats.MSBuild
 					MSBuildProjectService.InitializeItemHandler (sfolder);
 					MSBuildProjectService.SetId (sfolder, projectGuid);
 
-					List<string> projLines = lines.GetRange (sec.Start + 1, sec.Count - 2);
+					projLines = lines.GetRange (sec.Start + 1, sec.Count - 2);
 					DeserializeSolutionItem (sol, sfolder, projLines);
 					
 					foreach (string f in ReadFolderFiles (projLines))
@@ -729,40 +804,91 @@ namespace MonoDevelop.Projects.Formats.MSBuild
 				SolutionEntityItem item = null;
 				
 				try {
-					item = ProjectExtensionUtil.LoadSolutionItem (monitor, projectPath, delegate {
-						return MSBuildProjectService.LoadItem (monitor, projectPath, projTypeGuid, projectGuid);
-					});
-					
-					if (item == null) {
-						LoggingService.LogWarning (GettextCatalog.GetString (
-							"Unknown project type guid '{0}' on line #{1}. Ignoring.",
-							projTypeGuid,
-							sec.Start + 1));
-						monitor.ReportWarning (GettextCatalog.GetString (
-							"{0}({1}): Unsupported or unrecognized project : '{2}'.", 
-							fileName, sec.Start + 1, projectPath));
-						continue;
+					if (sol.IsSolutionItemEnabled (projectPath)) {
+						item = ProjectExtensionUtil.LoadSolutionItem (monitor, projectPath, delegate {
+							return MSBuildProjectService.LoadItem (monitor, projectPath, format, projTypeGuid, projectGuid);
+						});
+						
+						if (item == null) {
+							throw new UnknownSolutionItemTypeException (projTypeGuid);
+						}
+					} else {
+						var uitem = new UnloadedSolutionItem () {
+							FileName = projectPath
+						};
+						var h = new MSBuildHandler (projTypeGuid, projectGuid) {
+							Item = uitem,
+						};
+						uitem.SetItemHandler (h);
+						item = uitem;
 					}
-
-					MSBuildProjectHandler handler = (MSBuildProjectHandler) item.ItemHandler;
-					List<string> projLines = lines.GetRange (sec.Start + 1, sec.Count - 2);
-					DataItem it = GetSolutionItemData (projLines);
-					handler.SlnProjectContent = projLines.ToArray ();
-					handler.ReadSlnData (it);
 					
 				} catch (Exception e) {
-					LoggingService.LogError (GettextCatalog.GetString (
-								"Error while trying to load the project {0}. Exception : {1}",
-								projectPath, e.ToString ()));
-					monitor.ReportWarning (GettextCatalog.GetString (
-						"Error while trying to load the project '{0}': {1}", projectPath, e.Message));
+					// If we get a TargetInvocationException from using Activator.CreateInstance we
+					// need to unwrap the real exception
+					while (e is TargetInvocationException)
+						e = ((TargetInvocationException) e).InnerException;
+					
+					bool loadAsProject = false;
 
-					UnknownSolutionItem uitem = new UnknownSolutionItem ();
-					uitem.FileName = projectPath;
-					uitem.LoadError = e.Message;
+					if (e is UnknownSolutionItemTypeException) {
+						var name = ((UnknownSolutionItemTypeException)e).TypeName;
+
+						var relPath = new FilePath (path).ToRelative (sol.BaseDirectory);
+						if (!string.IsNullOrEmpty (name)) {
+							var guids = name.Split (';');
+							var projectInfo = MSBuildProjectService.GetUnknownProjectTypeInfo (guids, fileName);
+							if (projectInfo != null) {
+								loadAsProject = projectInfo.LoadFiles;
+								LoggingService.LogWarning (string.Format ("Could not load {0} project '{1}'. {2}", projectInfo.Name, relPath, projectInfo.GetInstructions ()));
+								monitor.ReportWarning (GettextCatalog.GetString ("Could not load {0} project '{1}'. {2}", projectInfo.Name, relPath, projectInfo.GetInstructions ()));
+							} else {
+								LoggingService.LogWarning (string.Format ("Could not load project '{0}' with unknown item type '{1}'", relPath, name));
+								monitor.ReportWarning (GettextCatalog.GetString ("Could not load project '{0}' with unknown item type '{1}'", relPath, name));
+							}
+						} else {
+							LoggingService.LogWarning (string.Format ("Could not load project '{0}' with unknown item type", relPath));
+							monitor.ReportWarning (GettextCatalog.GetString ("Could not load project '{0}' with unknown item type", relPath));
+						}
+
+					} else if (e is UserException) {
+						var ex = (UserException) e;
+						LoggingService.LogError ("{0}: {1}", ex.Message, ex.Details);
+						monitor.ReportError (string.Format ("{0}{1}{1}{2}", ex.Message, Environment.NewLine, ex.Details), null);
+					} else {
+						LoggingService.LogError (string.Format ("Error while trying to load the project {0}", projectPath), e);
+						monitor.ReportWarning (GettextCatalog.GetString (
+							"Error while trying to load the project '{0}': {1}", projectPath, e.Message));
+					}
+
+					SolutionEntityItem uitem;
+					if (loadAsProject) {
+						uitem = new UnknownProject () {
+							FileName = projectPath,
+							LoadError = e.Message,
+						};
+					} else {
+						uitem = new UnknownSolutionItem () {
+							FileName = projectPath,
+							LoadError = e.Message,
+						};
+					}
+
+					var h = new MSBuildHandler (projTypeGuid, projectGuid) {
+						Item = uitem,
+					};
+					uitem.SetItemHandler (h);
 					item = uitem;
 				}
-				
+
+				MSBuildHandler handler = (MSBuildHandler) item.ItemHandler;
+				projLines = lines.GetRange (sec.Start + 1, sec.Count - 2);
+				DataItem it = GetSolutionItemData (projLines);
+
+				handler.UnresolvedProjectDependencies = ReadSolutionItemDependencies (projLines);
+				handler.SlnProjectContent = projLines.ToArray ();
+				handler.ReadSlnData (it);
+
 				if (!items.ContainsKey (projectGuid)) {
 					items.Add (projectGuid, item);
 					sortedList.Add (item);
@@ -778,6 +904,22 @@ namespace MonoDevelop.Projects.Formats.MSBuild
 				globals.Remove ("NestedProjects");
 			}
 
+			// Resolve project dependencies
+			foreach (var it in items.Values.OfType<SolutionEntityItem> ()) {
+				MSBuildHandler handler = (MSBuildHandler) it.ItemHandler;
+				if (handler.UnresolvedProjectDependencies != null) {
+					foreach (var id in handler.UnresolvedProjectDependencies.ToArray ()) {
+						SolutionItem dep;
+						if (items.TryGetValue (id, out dep) && dep is SolutionEntityItem) {
+							handler.UnresolvedProjectDependencies.Remove (id);
+							it.ItemDependencies.Add ((SolutionEntityItem)dep);
+						}
+					}
+					if (handler.UnresolvedProjectDependencies.Count == 0)
+						handler.UnresolvedProjectDependencies = null;
+				}
+			}
+
 			//Add top level folders and projects to the main folder
 			foreach (SolutionItem ce in sortedList) {
 				if (ce.ParentFolder == null)
@@ -823,6 +965,17 @@ namespace MonoDevelop.Projects.Formats.MSBuild
 
 			data.GlobalExtra = globalLines;
 			monitor.EndTask ();
+
+			// When reloading a project, keep the solution data and item id
+			sol.SolutionItemAdded += delegate(object sender, SolutionItemChangeEventArgs e) {
+				if (e.Reloading) {
+					ItemSlnData.TransferData (e.ReplacedItem, e.SolutionItem);
+					var ih = e.SolutionItem.ItemHandler as MSBuildHandler;
+					if (ih != null)
+						ih.ItemId = e.ReplacedItem.ItemId;
+				}
+			};
+
 			return folder;
 		}
 
@@ -856,7 +1009,6 @@ namespace MonoDevelop.Projects.Formats.MSBuild
 			if (sec == null || String.Compare (sec.Val, "postSolution", true) != 0)
 				return;
 
-			List<SolutionConfigurationEntry> noBuildList = new List<SolutionConfigurationEntry> ();
 			Dictionary<string, SolutionConfigurationEntry> cache = new Dictionary<string, SolutionConfigurationEntry> ();
 			Dictionary<string, string> ignoredProjects = new Dictionary<string, string> ();
 			SlnData slnData = GetSlnData (sln.RootFolder);
@@ -871,6 +1023,7 @@ namespace MonoDevelop.Projects.Formats.MSBuild
 				//Format:
 				// {projectGuid}.SolutionConfigName|SolutionPlatform.ActiveCfg = ProjConfigName|ProjPlatform
 				// {projectGuid}.SolutionConfigName|SolutionPlatform.Build.0 = ProjConfigName|ProjPlatform
+				// {projectGuid}.SolutionConfigName|SolutionPlatform.Deploy.0 = ProjConfigName|ProjPlatform
 
 				string [] parts = s.Split (new char [] {'='}, 2);
 				if (parts.Length < 2) {
@@ -888,8 +1041,11 @@ namespace MonoDevelop.Projects.Formats.MSBuild
 				} else if (left.EndsWith (".Build.0")) {
 					action = "Build.0";
 					left = left.Substring (0, left.Length - 8);
+				} else if (left.EndsWith (".Deploy.0")) {
+					action = "Deploy.0";
+					left = left.Substring (0, left.Length - 9);
 				} else { 
-					LoggingService.LogWarning (GettextCatalog.GetString ("{0} ({1}) : Unknown action. Only ActiveCfg & Build.0 supported.",
+					LoggingService.LogWarning (GettextCatalog.GetString ("{0} ({1}) : Unknown action. Only ActiveCfg, Build.0 and Deploy.0 supported.",
 						sln.FileName, lineNum + 1));
 					continue;
 				}
@@ -901,7 +1057,7 @@ namespace MonoDevelop.Projects.Formats.MSBuild
 					continue;
 				}
 
-				string projGuid = t [0];
+				string projGuid = t [0].ToUpper ();
 				string slnConfig = t [1];
 
 				if (!slnData.ItemsByGuid.ContainsKey (projGuid)) {
@@ -916,13 +1072,14 @@ namespace MonoDevelop.Projects.Formats.MSBuild
 				}
 
 				SolutionEntityItem item;
-				if (slnData.ItemsByGuid.TryGetValue (projGuid, out item)) {
+				if (slnData.ItemsByGuid.TryGetValue (projGuid, out item) && item.SupportsConfigurations ()) {
 					string key = projGuid + "." + slnConfig;
 					SolutionConfigurationEntry combineConfigEntry = null;
 					if (cache.ContainsKey (key)) {
 						combineConfigEntry = cache [key];
 					} else {
 						combineConfigEntry = GetConfigEntry (sln, item, slnConfig);
+						combineConfigEntry.Build = false; // Not buildable by default. Build will be enabled if a Build.0 entry is found
 						cache [key] = combineConfigEntry;
 					}
 	
@@ -937,21 +1094,16 @@ namespace MonoDevelop.Projects.Formats.MSBuild
 					 */
 					if (action == "ActiveCfg") {
 						combineConfigEntry.ItemConfiguration = FromSlnConfigurationId (projConfig);
-						noBuildList.Add (combineConfigEntry);
 					} else if (action == "Build.0") {
-						noBuildList.Remove (combineConfigEntry);
+						combineConfigEntry.Build = true;
+					} else if (action == "Deploy.0") {
+						combineConfigEntry.Deploy = true;
 					}
 				}
 				extras.RemoveAt (extras.Count - 1);
 			}
 
 			slnData.SectionExtras ["ProjectConfigurationPlatforms"] = extras;
-
-			foreach (SolutionConfigurationEntry e in noBuildList) {
-				//Mark (build=false) of all projects for which 
-				//ActiveCfg was found but no Build.0
-				e.Build = false;
-			}
 		}
 
 		/* Gets the CombineConfigurationEntry corresponding to the @entry in its parentCombine's 
@@ -1027,7 +1179,9 @@ namespace MonoDevelop.Projects.Formats.MSBuild
 				return;
 
 			for (int i = 0; i < sec.Count - 2; i ++) {
+				// Guids should be upper case for VS compatibility
 				KeyValuePair<string, string> pair = SplitKeyValue (lines [i + sec.Start + 1].Trim ());
+				pair = new KeyValuePair<string, string> (pair.Key.ToUpper (), pair.Value.ToUpper ());
 
 				SolutionItem folderItem;
 				SolutionItem item;
