@@ -49,9 +49,8 @@ namespace MonoDevelop.Projects
 {
 	[DataInclude(typeof(DotNetProjectConfiguration))]
 	[ProjectModelDataItem ("AbstractDotNetProject")]
-	public abstract class DotNetProject : Project, IAssemblyProject
+	public abstract class DotNetProject : Project, IAssemblyProject, IDotNetFileContainer
 	{
-
 		bool usePartialTypes = true;
 		ProjectParameters languageParameters;
 
@@ -141,8 +140,7 @@ namespace MonoDevelop.Projects
 				Configurations.Add (configRelease);
 			}
 
-			if ((projectOptions != null) && (projectOptions.Attributes["TargetFrameworkVersion"] != null))
-				newProjectTargetFrameworkId = TargetFrameworkMoniker.Parse (projectOptions.Attributes["TargetFrameworkVersion"].Value);
+			targetFramework = GetTargetFrameworkForNewProject (projectOptions, GetDefaultTargetFrameworkId ());
 
 			string binPath;
 			if (projectCreateInfo != null) {
@@ -164,8 +162,31 @@ namespace MonoDevelop.Projects
 			}
 		}
 
-		public override string ProjectType {
-			get { return "DotNet"; }
+		TargetFramework GetTargetFrameworkForNewProject (XmlElement projectOptions, TargetFrameworkMoniker defaultMoniker)
+		{
+			if (projectOptions == null)
+				return Runtime.SystemAssemblyService.GetTargetFramework (defaultMoniker);
+
+			var att = projectOptions.Attributes ["TargetFrameworkVersion"];
+			if (att == null) {
+				att = projectOptions.Attributes ["TargetFramework"];
+				if (att == null)
+					return Runtime.SystemAssemblyService.GetTargetFramework (defaultMoniker);
+			}
+
+			var moniker = TargetFrameworkMoniker.Parse (att.Value);
+
+			//if the string did not include a framework identifier, use the project's default
+			var netID = TargetFrameworkMoniker.ID_NET_FRAMEWORK;
+			if (moniker.Identifier == netID && !att.Value.StartsWith (netID, StringComparison.Ordinal))
+				moniker = new TargetFrameworkMoniker (defaultMoniker.Identifier, moniker.Version, moniker.Profile);
+
+			return Runtime.SystemAssemblyService.GetTargetFramework (moniker);
+		}
+
+		public override IEnumerable<string> GetProjectTypes ()
+		{
+			yield return "DotNet";
 		}
 
 		private string languageName;
@@ -194,6 +215,30 @@ namespace MonoDevelop.Projects
 			get { return projectReferences; }
 		}
 
+		/// <summary>
+		/// Checks the status of references. To be called when referenced files may have been deleted or created.
+		/// </summary>
+		public void RefreshReferenceStatus ()
+		{
+			for (int n = 0; n < References.Count; n++) {
+				var cp = References [n].GetRefreshedReference ();
+				if (cp != null)
+					References [n] = cp;
+			}
+		}
+
+		public virtual bool CanReferenceProject (DotNetProject targetProject, out string reason)
+		{
+			if (!TargetFramework.CanReferenceAssembliesTargetingFramework (targetProject.TargetFramework)) {
+				reason = GettextCatalog.GetString ("Incompatible target framework: {0}", targetProject.TargetFramework.Id);
+				return false;
+			}
+
+			reason = null;
+
+			return true;
+		}
+
 		public IDotNetLanguageBinding LanguageBinding {
 			get {
 				if (languageBinding == null) {
@@ -277,15 +322,12 @@ namespace MonoDevelop.Projects
 			}
 		}
 		
-		//only used for initializing new projects
-		TargetFrameworkMoniker newProjectTargetFrameworkId;
-		
 		TargetFramework targetFramework;
 
 		public TargetFramework TargetFramework {
 			get {
 				if (targetFramework == null) {
-					var id = newProjectTargetFrameworkId ?? GetDefaultTargetFrameworkId ();
+					var id = GetDefaultTargetFrameworkId ();
 					targetFramework = Runtime.SystemAssemblyService.GetTargetFramework (id);
 				}
 				return targetFramework;
@@ -412,12 +454,17 @@ namespace MonoDevelop.Projects
 			get { return new string[] { "AnyCPU" }; }
 		}
 
-		void CheckReferenceChange (string updatedFile)
+		void CheckReferenceChange (FilePath updatedFile)
 		{
-			foreach (ProjectReference pr in References) {
-				if (pr.ReferenceType == ReferenceType.Assembly) {
-					if (updatedFile == Path.GetFullPath (pr.Reference))
+			for (int n=0; n<References.Count; n++) {
+				ProjectReference pr = References [n];
+				if (pr.ReferenceType == ReferenceType.Assembly && DefaultConfiguration != null) {
+					if (pr.GetReferencedFileNames (DefaultConfiguration.Selector).Any (f => f == updatedFile))
 						pr.NotifyStatusChanged ();
+				} else if (pr.HintPath == updatedFile) {
+					var nr = pr.GetRefreshedReference ();
+					if (nr != null)
+						References [n] = nr;
 				}
 			}
 		}
@@ -470,7 +517,7 @@ namespace MonoDevelop.Projects
 			// Generated satellite resource files
 			
 			FilePath outputDir = conf.OutputDirectory;
-			string satelliteAsmName = Path.GetFileNameWithoutExtension (conf.OutputAssembly) + ".resources.dll";
+			string satelliteAsmName = Path.GetFileNameWithoutExtension (conf.CompiledOutputName) + ".resources.dll";
 			
 			HashSet<string> cultures = new HashSet<string> ();
 			foreach (ProjectFile finfo in Files) {
@@ -533,37 +580,26 @@ namespace MonoDevelop.Projects
 						LoggingService.LogWarning ("Project '{0}' referenced from '{1}' could not be found", projectReference.Reference, this.Name);
 						continue;
 					}
-
-					string refOutput = p.GetOutputFileName (configuration);
-					if (string.IsNullOrEmpty (refOutput)) {
-						LoggingService.LogWarning ("Project '{0}' referenced from '{1}' has an empty output filename", p.Name, this.Name);
-						continue;
-					}
-
-					list.Add (refOutput);
-
+					DotNetProjectConfiguration conf = p.GetConfiguration (configuration) as DotNetProjectConfiguration;
 					//VS COMPAT: recursively copy references's "local copy" files
 					//but only copy the "copy to output" files from the immediate references
-					if (processedProjects.Add (p))
-						foreach (var f in p.GetSupportFileList (configuration))
-							list.Add (f.Src, f.CopyOnlyIfNewer, f.Target);
+					if (processedProjects.Add (p) || supportReferDistance == 1) {
+						foreach (var v in p.GetOutputFiles (configuration))
+							list.Add (v, true, v.CanonicalPath.ToString ().Substring (conf.OutputDirectory.CanonicalPath.ToString ().Length + 1));
 
-					DotNetProjectConfiguration refConfig = p.GetConfiguration (configuration) as DotNetProjectConfiguration;
-
-					if (refConfig != null && refConfig.DebugMode) {
-						string mdbFile = TargetRuntime.GetAssemblyDebugInfoFile (refOutput);
-						if (File.Exists (mdbFile)) {
-							list.Add (mdbFile);
-						}
+						foreach (var v in p.GetSupportFileList (configuration))
+							list.Add (v.Src, v.CopyOnlyIfNewer, v.Target);
 					}
 				}
 				else if (projectReference.ReferenceType == ReferenceType.Assembly) {
 					// VS COMPAT: Copy the assembly, but also all other assemblies referenced by it
 					// that are located in the same folder
-					foreach (string file in GetAssemblyRefsRec (projectReference.Reference, new HashSet<string> ())) {
+					var visitedAssemblies = new HashSet<string> ();
+					var referencedFiles = projectReference.GetReferencedFileNames (configuration);
+					foreach (string file in referencedFiles.SelectMany (ar => GetAssemblyRefsRec (ar, visitedAssemblies))) {
 						// Indirectly referenced assemblies are only copied if a newer copy doesn't exist. This avoids overwritting directly referenced assemblies
 						// by indirectly referenced stale copies of the same assembly. See bug #655566.
-						bool copyIfNewer = file != projectReference.Reference;
+						bool copyIfNewer = !referencedFiles.Contains (file);
 						list.Add (file, copyIfNewer);
 						if (File.Exists (file + ".config"))
 							list.Add (file + ".config", copyIfNewer);
@@ -572,7 +608,7 @@ namespace MonoDevelop.Projects
 							list.Add (mdbFile, copyIfNewer);
 					}
 				}
-				else if (projectReference.ReferenceType == ReferenceType.Custom) {
+				else {
 					foreach (string refFile in projectReference.GetReferencedFileNames (configuration))
 						list.Add (refFile);
 				}
@@ -649,14 +685,9 @@ namespace MonoDevelop.Projects
 			}
 			
 			yield return fileName;
-			Mono.Cecil.AssemblyDefinition adef;
-			try {
-				adef = Mono.Cecil.AssemblyDefinition.ReadAssembly (fileName);
-			} catch {
-				yield break;
-			}
-			foreach (Mono.Cecil.AssemblyNameReference aref in adef.MainModule.AssemblyReferences) {
-				string asmFile = Path.Combine (Path.GetDirectoryName (fileName), aref.Name);
+
+			foreach (var reference in SystemAssemblyService.GetAssemblyReferences (fileName)) {
+				string asmFile = Path.Combine (Path.GetDirectoryName (fileName), reference);
 				foreach (string refa in GetAssemblyRefsRec (asmFile, visited))
 					yield return refa;
 			}
@@ -676,7 +707,7 @@ namespace MonoDevelop.Projects
 
 		public override IEnumerable<SolutionItem> GetReferencedItems (ConfigurationSelector configuration)
 		{
-			List<SolutionItem> items = new List<SolutionItem> ();
+			List<SolutionItem> items = new List<SolutionItem> (base.GetReferencedItems (configuration));
 			if (ParentSolution == null)
 				return items;
 
@@ -712,7 +743,12 @@ namespace MonoDevelop.Projects
 		/// When set to true, it will include assemblies generated by referenced project. When set to false,
 		/// it will only include package and direct assembly references.
 		/// </param>
-		public virtual IEnumerable<string> GetReferencedAssemblies (ConfigurationSelector configuration, bool includeProjectReferences)
+		public IEnumerable<string> GetReferencedAssemblies (ConfigurationSelector configuration, bool includeProjectReferences)
+		{
+			return Services.ProjectService.GetExtensionChain (this).GetReferencedAssemblies (this, configuration, includeProjectReferences);
+		}
+
+		internal protected virtual IEnumerable<string> OnGetReferencedAssemblies (ConfigurationSelector configuration, bool includeProjectReferences)
 		{
 			IAssemblyReferenceHandler handler = this.ItemHandler as IAssemblyReferenceHandler;
 			if (handler != null) {
@@ -733,6 +769,22 @@ namespace MonoDevelop.Projects
 					}
 				}
 			}
+
+			var config = (DotNetProjectConfiguration)GetConfiguration (configuration);
+			bool noStdLib = false;
+			if (config != null) {
+				var parameters = config.CompilationParameters as DotNetConfigurationParameters;
+				if (parameters != null) {
+					noStdLib = parameters.NoStdLib;
+				}
+			}
+
+			// System.Core is an implicit reference
+			if (!noStdLib) {
+				var sa = AssemblyContext.GetAssemblies (TargetFramework).FirstOrDefault (a => a.Name == "System.Core" && a.Package.IsFrameworkPackage);
+				if (sa != null)
+					yield return sa.Location;
+			}
 		}
 
 		protected internal override void OnSave (IProgressMonitor monitor)
@@ -817,10 +869,11 @@ namespace MonoDevelop.Projects
 					break;
 				}
 			}
-			
+
+			var config = (DotNetProjectConfiguration) GetConfiguration (configuration);
 			return Files.Any (file => file.BuildAction == BuildAction.EmbeddedResource
 					&& String.Compare (Path.GetExtension (file.FilePath), ".resx", StringComparison.OrdinalIgnoreCase) == 0
-					&& MD1DotNetProjectHandler.IsResgenRequired (file.FilePath));
+					&& MD1DotNetProjectHandler.IsResgenRequired (file.FilePath, config.IntermediateOutputDirectory.Combine (file.ResourceId)));
 		}
 		
 		protected internal override DateTime OnGetLastBuildTime (ConfigurationSelector configuration)
@@ -850,7 +903,7 @@ namespace MonoDevelop.Projects
 				return null;
 			//return all projects in the sln in case some are loaded dynamically
 			//FIXME: should we do this for the whole workspace?
-			return ParentSolution.GetAllProjects ().OfType<DotNetProject> ()
+			return ParentSolution.RootFolder.GetAllBuildableEntries (configuration).OfType<DotNetProject> ()
 				.Select (d => (string) d.GetOutputFileName (configuration))
 				.Where (d => !string.IsNullOrEmpty (d)).ToList ();
 		}
@@ -872,6 +925,8 @@ namespace MonoDevelop.Projects
 			if (config == null)
 				return false;
 			ExecutionCommand cmd = CreateExecutionCommand (configuration, config);
+			if (context.ExecutionTarget != null)
+				cmd.Target = context.ExecutionTarget;
 
 			return (compileTarget == CompileTarget.Exe || compileTarget == CompileTarget.WinExe) && context.ExecutionHandler.CanExecute (cmd);
 		}
@@ -880,9 +935,12 @@ namespace MonoDevelop.Projects
 		{
 			List<FilePath> col = base.OnGetItemFiles (includeReferencedFiles);
 			if (includeReferencedFiles) {
-				foreach (ProjectReference pref in References)
-					if (pref.ReferenceType == ReferenceType.Assembly)
-						col.Add (pref.Reference);
+				foreach (ProjectReference pref in References) {
+					if (pref.ReferenceType == ReferenceType.Assembly) {
+						foreach (var f in pref.GetReferencedFileNames (DefaultConfiguration.Selector))
+							col.Add (f);
+					}
+				}
 				foreach (DotNetProjectConfiguration c in Configurations) {
 					if (c.SignAssembly)
 						col.Add (c.AssemblyKeyFile);
@@ -898,20 +956,29 @@ namespace MonoDevelop.Projects
 				return false;
 			return LanguageBinding.IsSourceCodeFile (fileName);
 		}
-		
+
 		/// <summary>
 		/// Gets the default namespace for the file, according to the naming policy.
 		/// </summary>
 		/// <remarks>Always returns a valid namespace, even if the fileName is null.</remarks>
 		public virtual string GetDefaultNamespace (string fileName)
 		{
-			DotNetNamingPolicy pol = Policies.Get<DotNetNamingPolicy> ();
+			return GetDefaultNamespace (this, DefaultNamespace, fileName);
+		}
+
+		/// <summary>
+		/// Gets the default namespace for the file, according to the naming policy.
+		/// </summary>
+		/// <remarks>Always returns a valid namespace, even if the fileName is null.</remarks>
+		internal static string GetDefaultNamespace (Project project, string defaultNamespace, string fileName)
+		{
+			DotNetNamingPolicy pol = project.Policies.Get<DotNetNamingPolicy> ();
 
 			string root = null;
 			string dirNamespc = null;
-			string defaultNmspc = !string.IsNullOrEmpty (DefaultNamespace)
-				? DefaultNamespace
-				: SanitisePotentialNamespace (Name) ?? "Application";
+			string defaultNmspc = !string.IsNullOrEmpty (defaultNamespace)
+				? defaultNamespace
+				: SanitisePotentialNamespace (project.Name) ?? "Application";
 			
 			if (string.IsNullOrEmpty (fileName)) {
 				return defaultNmspc;
@@ -920,7 +987,7 @@ namespace MonoDevelop.Projects
 			string dirname = Path.GetDirectoryName (fileName);
 			string relativeDirname = null;
 			if (!String.IsNullOrEmpty (dirname)) {
-				relativeDirname = GetRelativeChildPath (dirname);
+				relativeDirname = project.GetRelativeChildPath (dirname);
 				if (string.IsNullOrEmpty (relativeDirname) || relativeDirname.StartsWith (".."))
 					relativeDirname = null;
 			}
@@ -932,7 +999,8 @@ namespace MonoDevelop.Projects
 						root = defaultNmspc;
 						goto case DirectoryNamespaceAssociation.Flat;
 					case DirectoryNamespaceAssociation.Flat:
-						dirNamespc = SanitisePotentialNamespace (relativeDirname);
+						//use the last component only
+						dirNamespc = SanitisePotentialNamespace (Path.GetFileName (relativeDirname));
 						break;
 
 					case DirectoryNamespaceAssociation.PrefixedHierarchical:
@@ -955,7 +1023,7 @@ namespace MonoDevelop.Projects
 			return defaultNmspc;
 		}
 
-		string GetHierarchicalNamespace (string relativePath)
+		static string GetHierarchicalNamespace (string relativePath)
 		{
 			StringBuilder sb = new StringBuilder (relativePath);
 			for (int i = 0; i < sb.Length; i++) {
@@ -965,7 +1033,7 @@ namespace MonoDevelop.Projects
 			return sb.ToString ();
 		}
 
-		string SanitisePotentialNamespace (string potential)
+		static string SanitisePotentialNamespace (string potential)
 		{
 			StringBuilder sb = new StringBuilder ();
 			foreach (char c in potential) {
@@ -993,12 +1061,17 @@ namespace MonoDevelop.Projects
 		{
 			foreach (ProjectReference pref in References) {
 				if (pref.ReferenceType == ReferenceType.Package) {
-					string newRef = AssemblyContext.GetAssemblyNameForVersion (pref.Reference, pref.Package != null ? pref.Package.Name : null, this.TargetFramework);
-					if (newRef == null) {
+					// package name is only relevant if it's not a framework package
+					var pkg = pref.Package;
+					string packageName = pkg != null && !pkg.IsFrameworkPackage? pkg.Name : null;
+					// find the version of the assembly that's valid for the new framework
+					var newAsm = AssemblyContext.GetAssemblyForVersion (pref.Reference, packageName, TargetFramework);
+					// if it changed, clear assembly resolution caches and update reference
+					if (newAsm == null) {
 						pref.ResetReference ();
-					} else if (newRef != pref.Reference) {
-						pref.Reference = newRef;
-					} else if (!pref.IsValid) {
+					} else if (newAsm.FullName != pref.Reference) {
+						pref.Reference = newAsm.FullName;
+					} else if (!pref.IsValid || newAsm.Package != pref.Package) {
 						pref.ResetReference ();
 					}
 				}
@@ -1059,7 +1132,7 @@ namespace MonoDevelop.Projects
 			if (oldHandler.GetType () != newHandler.GetType ()) {
 				// If the file format has a default resource handler different from the one
 				// choosen for this project, then all resource ids must be converted
-				foreach (ProjectFile file in Files) {
+				foreach (ProjectFile file in Files.Where (f => f.BuildAction == BuildAction.EmbeddedResource)) {
 					if (file.Subtype == Subtype.Directory)
 						continue;
 					string oldId = file.GetResourceId (oldHandler);
@@ -1098,14 +1171,12 @@ namespace MonoDevelop.Projects
 
 		internal void NotifyReferenceRemovedFromProject (ProjectReference reference)
 		{
-			SetNeedsBuilding (true);
 			NotifyModified ("References");
 			OnReferenceRemovedFromProject (new ProjectReferenceEventArgs (this, reference));
 		}
 
 		internal void NotifyReferenceAddedToProject (ProjectReference reference)
 		{
-			SetNeedsBuilding (true);
 			NotifyModified ("References");
 			OnReferenceAddedToProject (new ProjectReferenceEventArgs (this, reference));
 		}
@@ -1148,6 +1219,8 @@ namespace MonoDevelop.Projects
 			try {
 				try {
 					ExecutionCommand executionCommand = CreateExecutionCommand (configuration, dotNetProjectConfig);
+					if (context.ExecutionTarget != null)
+						executionCommand.Target = context.ExecutionTarget;
 
 					if (!context.ExecutionHandler.CanExecute (executionCommand)) {
 						monitor.ReportError (GettextCatalog.GetString ("Can not execute \"{0}\". The selected execution mode is not supported for .NET projects.", dotNetProjectConfig.CompiledOutputName), null);
@@ -1164,8 +1237,37 @@ namespace MonoDevelop.Projects
 					aggregatedOperationMonitor.Dispose ();
 				}
 			} catch (Exception ex) {
+				LoggingService.LogError (string.Format ("Cannot execute \"{0}\"", dotNetProjectConfig.CompiledOutputName), ex);
 				monitor.ReportError (GettextCatalog.GetString ("Cannot execute \"{0}\"", dotNetProjectConfig.CompiledOutputName), ex);
 			}
 		}
+
+		public void AddImportIfMissing (string name, string condition)
+		{
+			importsAdded.Add (new DotNetProjectImport (name, condition));
+		}
+
+		public void RemoveImport (string name)
+		{
+			importsRemoved.Add (new DotNetProjectImport (name));
+		}
+
+		List <DotNetProjectImport> importsAdded = new List<DotNetProjectImport> ();
+
+		internal IList<DotNetProjectImport> ImportsAdded {
+			get { return importsAdded; }
+		}
+
+		List <DotNetProjectImport> importsRemoved = new List<DotNetProjectImport> ();
+
+		internal IList<DotNetProjectImport> ImportsRemoved {
+			get { return importsRemoved; }
+		}
+
+		public void ImportsSaved ()
+		{
+			importsAdded.Clear ();
+			importsRemoved.Clear ();
+		}
 	}
 }
