@@ -41,13 +41,6 @@ using MonoDevelop.Projects;
 
 namespace MonoDevelop.Projects
 {
-	public enum NewFileSearch
-	{
-		None,
-		OnLoad,
-		OnLoadAutoInsert
-	}
-	
 	/// <summary>
 	/// A project
 	/// </summary>
@@ -60,9 +53,8 @@ namespace MonoDevelop.Projects
 	public abstract class Project : SolutionEntityItem
 	{
 		string[] buildActions;
-		bool isDirty;
 
-		public Project ()
+		protected Project ()
 		{
 			FileService.FileChanged += OnFileChanged;
 			files = new ProjectFileCollection ();
@@ -104,16 +96,22 @@ namespace MonoDevelop.Projects
 			get { return files; }
 		}
 		private ProjectFileCollection files;
-		
-		
-		[ItemProperty("newfilesearch", DefaultValue = NewFileSearch.None)]
-		protected NewFileSearch newFileSearch = NewFileSearch.None;
-		public NewFileSearch NewFileSearch {
-			get { return newFileSearch; }
 
+		[ProjectPathItemProperty ("BaseIntermediateOutputPath")]
+		FilePath baseIntermediateOutputPath;
+
+		public virtual FilePath BaseIntermediateOutputPath {
+			get {
+				if (!baseIntermediateOutputPath.IsNullOrEmpty)
+					return baseIntermediateOutputPath;
+				return BaseDirectory.Combine ("obj");
+			}
 			set {
-				newFileSearch = value;
-				NotifyModified ("NewFileSearch");
+				if (value.IsNullOrEmpty)
+					value = FilePath.Null;
+				if (baseIntermediateOutputPath == value)
+					return;
+				NotifyModified ("BaseIntermediateOutputPath");
 			}
 		}
 
@@ -123,10 +121,16 @@ namespace MonoDevelop.Projects
 		/// <value>
 		/// The type of the project.
 		/// </value>
-		public abstract string ProjectType {
-			get;
+		[Obsolete ("Use GetProjectTypes")]
+		public virtual string ProjectType {
+			get { return GetProjectTypes ().First (); }
 		}
 
+		/// <summary>
+		/// Gets the project type and its base types.
+		/// </summary>
+		public abstract IEnumerable<string> GetProjectTypes ();
+
 		/// <summary>
 		/// Gets or sets the icon of the project.
 		/// </summary>
@@ -388,10 +392,10 @@ namespace MonoDevelop.Projects
 		
 		//HACK: the build code is structured such that support file copying is in here instead of the item handler
 		//so in order to avoid doing them twice when using the msbuild engine, we special-case them
-		bool UsingMSBuildEngine ()
+		bool UsingMSBuildEngine (ConfigurationSelector sel)
 		{
 			var msbuildHandler = ItemHandler as MonoDevelop.Projects.Formats.MSBuild.MSBuildProjectHandler;
-			return msbuildHandler != null && msbuildHandler.UseXbuild;
+			return msbuildHandler != null && msbuildHandler.UseMSBuildEngineForItem (this, sel);
 		}
 
 		protected override BuildResult OnBuild (IProgressMonitor monitor, ConfigurationSelector configuration)
@@ -406,10 +410,8 @@ namespace MonoDevelop.Projects
 			
 			StringParserService.Properties["Project"] = Name;
 			
-			if (UsingMSBuildEngine ()) {
-				var r = DoBuild (monitor, configuration);
-				isDirty = false;
-				return r;
+			if (UsingMSBuildEngine (configuration)) {
+				return DoBuild (monitor, configuration);
 			}
 			
 			string outputDir = conf.OutputDirectory;
@@ -429,8 +431,6 @@ namespace MonoDevelop.Projects
 			
 			BuildResult res = DoBuild (monitor, configuration);
 
-			isDirty = false;
-
 			if (res != null) {
 				string errorString = GettextCatalog.GetPluralString ("{0} error", "{0} errors", res.ErrorCount, res.ErrorCount);
 				string warningString = GettextCatalog.GetPluralString ("{0} warning", "{0} warnings", res.WarningCount, res.WarningCount);
@@ -630,15 +630,13 @@ namespace MonoDevelop.Projects
 
 		protected override void OnClean (IProgressMonitor monitor, ConfigurationSelector configuration)
 		{
-			SetDirty ();
-			
 			ProjectConfiguration config = GetConfiguration (configuration) as ProjectConfiguration;
 			if (config == null) {
-				monitor.ReportError (GettextCatalog.GetString ("Configuration '{0}' not found in project '{1}'", config.Id, Name), null);
+				monitor.ReportError (GettextCatalog.GetString ("Configuration '{0}' not found in project '{1}'", configuration, Name), null);
 				return;
 			}
 			
-			if (UsingMSBuildEngine ()) {
+			if (UsingMSBuildEngine (configuration)) {
 				DoClean (monitor, config.Selector);
 				return;
 			}
@@ -665,18 +663,6 @@ namespace MonoDevelop.Projects
 			ItemHandler.RunTarget (monitor, "Clean", configuration);
 		}
 
-		void GetBuildableReferencedItems (List<SolutionItem> referenced, SolutionItem item, ConfigurationSelector configuration)
-		{
-			if (referenced.Contains (item))
-				return;
-
-			if (item.NeedsBuilding (configuration))
-				referenced.Add (item);
-
-			foreach (SolutionItem ritem in item.GetReferencedItems (configuration))
-				GetBuildableReferencedItems (referenced, ritem, configuration);
-		}
-
 		protected internal override void OnExecute (IProgressMonitor monitor, ExecutionContext context, ConfigurationSelector configuration)
 		{
 			ProjectConfiguration config = GetConfiguration (configuration) as ProjectConfiguration;
@@ -719,22 +705,7 @@ namespace MonoDevelop.Projects
 
 		protected internal override bool OnGetNeedsBuilding (ConfigurationSelector configuration)
 		{
-			if (!isDirty) {
-				if (CheckNeedsBuild (configuration))
-					SetDirty ();
-			}
-			return isDirty;
-		}
-
-		protected internal override void OnSetNeedsBuilding (bool value, ConfigurationSelector configuration)
-		{
-			isDirty = value;
-		}
-
-		void SetDirty ()
-		{
-			if (!Loading)
-				isDirty = true;
+			return CheckNeedsBuild (configuration);
 		}
 
 		/// <summary>
@@ -764,7 +735,7 @@ namespace MonoDevelop.Projects
 			}
 
 			foreach (SolutionItem pref in GetReferencedItems (configuration)) {
-				if (pref.GetLastBuildTime (configuration) > tim || pref.NeedsBuilding (configuration))
+				if (pref.GetLastBuildTime (configuration) > tim)
 					return true;
 			}
 
@@ -796,7 +767,6 @@ namespace MonoDevelop.Projects
 			foreach (FileEventInfo fi in e) {
 				ProjectFile file = GetProjectFile (fi.FileName);
 				if (file != null) {
-					SetDirty ();
 					try {
 						NotifyFileChangedInProject (file);
 					} catch {
@@ -837,13 +807,15 @@ namespace MonoDevelop.Projects
 			OnFileChangedInProject (new ProjectFileEventArgs (this, file));
 		}
 
-		internal void NotifyFilePropertyChangedInProject (ProjectFile file)
+		internal void NotifyFilePropertyChangedInProject (ProjectFile file, string property)
 		{
 			NotifyModified ("Files");
-			OnFilePropertyChangedInProject (new ProjectFileEventArgs (this, file));
+			OnFilePropertyChangedInProject (new ProjectFileEventArgs (this, file, property));
 		}
 
-		List<ProjectFile> unresolvedDeps;
+		// A collection of files that depend on other files for which the dependencies
+		// have not yet been resolved.
+		UnresolvedFileCollection unresolvedDeps;
 
 		void NotifyFileRemovedFromProject (IEnumerable<ProjectFile> objs)
 		{
@@ -856,8 +828,7 @@ namespace MonoDevelop.Projects
 				file.SetProject (null);
 				args.Add (new ProjectFileEventInfo (this, file));
 				if (DependencyResolutionEnabled) {
-					if (unresolvedDeps.Contains (file))
-						unresolvedDeps.Remove (file);
+					unresolvedDeps.Remove (file);
 					foreach (ProjectFile f in file.DependentChildren) {
 						f.DependsOnFile = null;
 						if (!string.IsNullOrEmpty (f.DependsOn))
@@ -866,7 +837,6 @@ namespace MonoDevelop.Projects
 					file.DependsOnFile = null;
 				}
 			}
-			SetDirty ();
 			NotifyModified ("Files");
 			OnFileRemovedFromProject (args);
 		}
@@ -886,11 +856,16 @@ namespace MonoDevelop.Projects
 				ResolveDependencies (file);
 			}
 
-			SetDirty ();
 			NotifyModified ("Files");
 			OnFileAddedToProject (args);
 		}
 
+		internal void UpdateDependency (ProjectFile file, FilePath oldPath)
+		{
+			unresolvedDeps.Remove (file, oldPath);
+			ResolveDependencies (file);
+		}
+
 		internal void ResolveDependencies (ProjectFile file)
 		{
 			if (!DependencyResolutionEnabled)
@@ -900,11 +875,13 @@ namespace MonoDevelop.Projects
 				unresolvedDeps.Add (file);
 
 			List<ProjectFile> resolved = null;
-			foreach (ProjectFile unres in unresolvedDeps) {
+			foreach (ProjectFile unres in unresolvedDeps.GetUnresolvedFilesForPath (file.FilePath)) {
 				if (string.IsNullOrEmpty (unres.DependsOn)) {
+					if (resolved == null)
+						resolved = new List<ProjectFile> ();
 					resolved.Add (unres);
 				}
-				if (unres.ResolveParent ()) {
+				if (unres.ResolveParent (file)) {
 					if (resolved == null)
 						resolved = new List<ProjectFile> ();
 					resolved.Add (unres);
@@ -922,7 +899,7 @@ namespace MonoDevelop.Projects
 				if (value) {
 					if (unresolvedDeps != null)
 						return;
-					unresolvedDeps = new List<ProjectFile> ();
+					unresolvedDeps = new UnresolvedFileCollection ();
 					foreach (ProjectFile file in files)
 						ResolveDependencies (file);
 				} else {
@@ -933,7 +910,6 @@ namespace MonoDevelop.Projects
 
 		internal void NotifyFileRenamedInProject (ProjectFileRenamedEventArgs args)
 		{
-			SetDirty ();
 			NotifyModified ("Files");
 			OnFileRenamedInProject (args);
 		}
@@ -1017,30 +993,6 @@ namespace MonoDevelop.Projects
 		public event ProjectFileRenamedEventHandler FileRenamedInProject;
 	}
 
-	public class UnknownProject : Project
-	{
-		public override string ProjectType {
-			get { return ""; }
-		}
-
-		public override SolutionItemConfiguration CreateConfiguration (string name)
-		{
-			return null;
-		}
-		
-		internal protected override bool OnGetCanExecute (ExecutionContext context, ConfigurationSelector configuration)
-		{
-			return false;
-		}
-		
-		protected override BuildResult OnBuild (IProgressMonitor monitor, ConfigurationSelector configuration)
-		{
-			BuildResult res = new BuildResult ();
-			res.AddError ("Unknown project type");
-			return res;
-		}
-	}
-
 	public delegate void ProjectEventHandler (Object sender, ProjectEventArgs e);
 	public class ProjectEventArgs : EventArgs
 	{
@@ -1054,4 +1006,73 @@ namespace MonoDevelop.Projects
 			get { return project; }
 		}
 	}
+
+	class UnresolvedFileCollection
+	{
+		// Holds a dictionary of files that depend on other files, and for which the dependency
+		// has not yet been resolved. The key of the dictionary is the path to a parent
+		// file to be resolved, and the value can be a ProjectFile object or a List<ProjectFile>
+		// (This may happen if several files depend on the same parent file)
+		Dictionary<FilePath,object> unresolvedDeps = new Dictionary<FilePath, object> ();
+
+		public void Remove (ProjectFile file)
+		{
+			Remove (file, null);
+		}
+
+		public void Remove (ProjectFile file, FilePath dependencyPath)
+		{
+			if (dependencyPath.IsNullOrEmpty) {
+				if (string.IsNullOrEmpty (file.DependsOn))
+					return;
+				dependencyPath = file.DependencyPath;
+			}
+
+			object depFile;
+			if (unresolvedDeps.TryGetValue (dependencyPath, out depFile)) {
+				if ((depFile is ProjectFile) && ((ProjectFile)depFile == file))
+					unresolvedDeps.Remove (dependencyPath);
+				else if (depFile is List<ProjectFile>) {
+					var list = (List<ProjectFile>) depFile;
+					list.Remove (file);
+					if (list.Count == 1)
+						unresolvedDeps [dependencyPath] = list[0];
+				}
+			}
+		}
+
+		public void Add (ProjectFile file)
+		{
+			object depFile;
+			if (unresolvedDeps.TryGetValue (file.DependencyPath, out depFile)) {
+				if (depFile is ProjectFile) {
+					if ((ProjectFile)depFile != file) {
+						var list = new List<ProjectFile> ();
+						list.Add ((ProjectFile)depFile);
+						list.Add (file);
+						unresolvedDeps [file.DependencyPath] = list;
+					}
+				}
+				else if (depFile is List<ProjectFile>) {
+					var list = (List<ProjectFile>) depFile;
+					if (!list.Contains (file))
+						list.Add (file);
+				}
+			} else
+				unresolvedDeps [file.DependencyPath] = file;
+		}
+
+		public IEnumerable<ProjectFile> GetUnresolvedFilesForPath (FilePath filePath)
+		{
+			object depFile;
+			if (unresolvedDeps.TryGetValue (filePath, out depFile)) {
+				if (depFile is ProjectFile)
+					yield return (ProjectFile) depFile;
+				else {
+					foreach (var f in (List<ProjectFile>) depFile)
+						yield return f;
+				}
+			}
+		}
+	}
 }
