@@ -51,7 +51,9 @@ namespace MonoDevelop.Projects
 		internal object MemoryProbe = Counters.ItemsInMemory.CreateMemoryProbe ();
 			
 		ProjectItemCollection items;
-		
+		ProjectItemCollection wildcardItems;
+		ItemCollection<SolutionEntityItem> dependencies = new ItemCollection<SolutionEntityItem> ();
+
 		SolutionItemEventArgs thisItemArgs;
 		
 		FileStatusTracker<SolutionItemEventArgs> fileStatusTracker;
@@ -74,6 +76,7 @@ namespace MonoDevelop.Projects
 		public SolutionEntityItem ()
 		{
 			items = new ProjectItemCollection (this);
+			wildcardItems = new ProjectItemCollection (this);
 			thisItemArgs = new SolutionItemEventArgs (this);
 			configurations = new SolutionItemConfigurationCollection (this);
 			configurations.ConfigurationAdded += new ConfigurationEventHandler (OnConfigurationAddedToCollection);
@@ -89,13 +92,14 @@ namespace MonoDevelop.Projects
 			
 			Counters.ItemsLoaded--;
 			
-			foreach (var item in items) {
+			foreach (var item in items.Concat (wildcardItems)) {
 				IDisposable disp = item as IDisposable;
 				if (disp != null)
 					disp.Dispose ();
 			}
 			
 			// items = null;
+			// wildcardItems = null;
 			// thisItemArgs = null;
 			// fileStatusTracker = null;
 			// fileFormat = null;
@@ -105,6 +109,35 @@ namespace MonoDevelop.Projects
 			base.Dispose ();
 		}
 
+		protected override void OnBoundToSolution ()
+		{
+			base.OnBoundToSolution ();
+			ParentSolution.SolutionItemRemoved += HandleSolutionItemRemoved;
+			ParentSolution.SolutionItemAdded += HandleSolutionItemAdded;
+		}
+
+		protected override void OnUnboundFromSolution ()
+		{
+			base.OnUnboundFromSolution ();
+			ParentSolution.SolutionItemAdded -= HandleSolutionItemAdded;
+			ParentSolution.SolutionItemRemoved -= HandleSolutionItemRemoved;
+		}
+
+		void HandleSolutionItemAdded (object sender, SolutionItemChangeEventArgs e)
+		{
+			if (e.Reloading && dependencies.Count > 0 && (e.SolutionItem is SolutionEntityItem) && (e.ReplacedItem is SolutionEntityItem)) {
+				int i = dependencies.IndexOf ((SolutionEntityItem)e.ReplacedItem);
+				if (i != -1)
+					dependencies [i] = (SolutionEntityItem) e.SolutionItem;
+			}
+		}
+
+		void HandleSolutionItemRemoved (object sender, SolutionItemChangeEventArgs e)
+		{
+			if (!e.Reloading && (e.SolutionItem is SolutionEntityItem))
+				dependencies.Remove ((SolutionEntityItem)e.SolutionItem);
+		}
+
 		
 		internal override void SetItemHandler (ISolutionItemHandler handler)
 		{
@@ -184,7 +217,15 @@ namespace MonoDevelop.Projects
 				NotifyModified ("FileName");
 			}
 		}
-		
+
+		public bool Enabled {
+			get { return ParentSolution != null ? ParentSolution.IsSolutionItemEnabled (FileName) : true; }
+			set { 
+				if (ParentSolution != null)
+					ParentSolution.SetSolutionItemEnabled (FileName, value);
+			}
+		}
+
 		public FileFormat FileFormat {
 			get {
 				if (ParentSolution != null) {
@@ -205,11 +246,33 @@ namespace MonoDevelop.Projects
 				NotifyModified ("FileFormat");
 			}
 		}
+
+		public virtual bool SupportsConfigurations ()
+		{
+			return SupportsBuild ();
+		}
 		
 		public ProjectItemCollection Items {
 			get { return items; }
 		}
+
+		internal ProjectItemCollection WildcardItems {
+			get { return wildcardItems; }
+		}
 		
+		/// <summary>
+		/// Projects that need to be built before building this one
+		/// </summary>
+		/// <value>The dependencies.</value>
+		public ItemCollection<SolutionEntityItem> ItemDependencies {
+			get { return dependencies; }
+		}
+
+		public override IEnumerable<SolutionItem> GetReferencedItems (ConfigurationSelector configuration)
+		{
+			return base.GetReferencedItems (configuration).Concat (dependencies);
+		}
+
 		void IWorkspaceFileObject.ConvertToFormat (FileFormat format, bool convertChildren)
 		{
 			this.FileFormat = format;
@@ -306,10 +369,10 @@ namespace MonoDevelop.Projects
 			ItemHandler.Save (monitor);
 		}
 		
+		[Obsolete ("This method will be removed in future releases")]
 		public void SetNeedsBuilding (bool value)
 		{
-			foreach (string conf in GetConfigurations ())
-				SetNeedsBuilding (value, new ItemConfigurationSelector (conf));
+			// Nothing to be done
 		}
 
 		public FilePath GetAbsoluteChildPath (FilePath relPath)
