@@ -66,6 +66,11 @@ namespace MonoDevelop.Projects
 				return items;
 			}
 		}
+
+		internal SolutionFolderItemCollection GetItemsWithoutCreating ()
+		{
+			return items;
+		}
 		
 		[ItemProperty]
 		[ProjectPathItemProperty ("File", Scope="*")]
@@ -194,19 +199,6 @@ namespace MonoDevelop.Projects
 			SetItemHandler (new DummySolutionFolderHandler (this));
 		}
 
-		
-		protected internal override bool OnGetNeedsBuilding (ConfigurationSelector configuration)
-		{
-			foreach (SolutionItem item in Items)
-				if (item.NeedsBuilding (configuration)) return true;
-			return false;
-		}
-		
-		protected internal override void OnSetNeedsBuilding (bool value, ConfigurationSelector configuration)
-		{
-			// Ignore
-		}
-		
 		public override void Dispose()
 		{
 			if (items != null) {
@@ -229,14 +221,37 @@ namespace MonoDevelop.Projects
 				
 				SolutionEntityItem newItem;
 				try {
-					newItem = Services.ProjectService.ReadSolutionItem (monitor, item.FileName);
+					if (ParentSolution.IsSolutionItemEnabled (item.FileName))
+						newItem = Services.ProjectService.ReadSolutionItem (monitor, item.FileName);
+					else {
+						UnknownSolutionItem e = new UnloadedSolutionItem () {
+							FileName = item.FileName
+						};
+						var ch = item.GetItemHandler () as MonoDevelop.Projects.Formats.MSBuild.MSBuildHandler;
+						if (ch != null) {
+							var h = new MonoDevelop.Projects.Formats.MSBuild.MSBuildHandler (ch.TypeGuid, ch.ItemId) {
+								Item = e,
+							};
+							e.SetItemHandler (h);
+						}
+						newItem = e;
+					}
 				} catch (Exception ex) {
 					UnknownSolutionItem e = new UnknownSolutionItem ();
 					e.LoadError = ex.Message;
 					e.FileName = item.FileName;
 					newItem = e;
 				}
-				
+
+				if (!Items.Contains (item)) {
+					// The old item is gone, which probably means it has already been reloaded (BXC20615), or maybe removed.
+					// In this case, there isn't anything else we can do
+					newItem.Dispose ();
+
+					// Find the replacement if it exists
+					return Items.OfType<SolutionEntityItem> ().FirstOrDefault (it => it.FileName == item.FileName);
+				}
+
 				// Replace in the file list
 				Items.Replace (item, newItem);
 				
@@ -244,8 +259,8 @@ namespace MonoDevelop.Projects
 				ConnectChildEntryEvents (newItem);
 	
 				NotifyModified ("Items");
-				OnItemRemoved (new SolutionItemChangeEventArgs (item, ParentSolution, true), true);
-				OnItemAdded (new SolutionItemChangeEventArgs (newItem, ParentSolution, true), true);
+				OnItemRemoved (new SolutionItemChangeEventArgs (item, ParentSolution, true) { ReplacedItem = item } , true);
+				OnItemAdded (new SolutionItemChangeEventArgs (newItem, ParentSolution, true) { ReplacedItem = item }, true);
 				
 				item.Dispose ();
 				return newItem;
@@ -324,7 +339,7 @@ namespace MonoDevelop.Projects
 			Items.Add (item);
 			
 			SolutionEntityItem eitem = item as SolutionEntityItem;
-			if (eitem != null && createSolutionConfigurations) {
+			if (eitem != null && createSolutionConfigurations && eitem.SupportsBuild ()) {
 				// Create new solution configurations for item configurations
 				foreach (ItemConfiguration iconf in eitem.Configurations) {
 					bool found = false;
@@ -336,7 +351,9 @@ namespace MonoDevelop.Projects
 					}
 					if (!found) {
 						SolutionConfiguration sconf = new SolutionConfiguration (iconf.Id);
-						sconf.AddItem (eitem);
+						// Add all items to the new configuration
+						foreach (var it in ParentSolution.GetAllSolutionItems<SolutionEntityItem> ())
+							sconf.AddItem (it);
 						ParentSolution.Configurations.Add (sconf);
 					}
 				}
@@ -468,7 +485,7 @@ namespace MonoDevelop.Projects
 			foreach (SolutionItem item in Items) {
 				if (item is SolutionFolder)
 					((SolutionFolder)item).GetAllBuildableEntries (list, configuration, includeExternalReferences);
-				else if ((item is SolutionEntityItem) && conf.BuildEnabledForItem ((SolutionEntityItem) item))
+				else if ((item is SolutionEntityItem) && conf.BuildEnabledForItem ((SolutionEntityItem) item) && item.SupportsBuild ())
 					GetAllBuildableReferences (list, item, configuration, includeExternalReferences);
 			}
 		}
@@ -483,17 +500,43 @@ namespace MonoDevelop.Projects
 					GetAllBuildableReferences (list, it, configuration, includeExternalReferences);
 			}
 		}
-		
+
+		[Obsolete("Use GetProjectsContainingFile() (plural) instead")]
 		public Project GetProjectContainingFile (string fileName) 
 		{
 			ReadOnlyCollection<Project> projects = GetAllProjects ();
 			foreach (Project projectEntry in projects) {
-				if (projectEntry.IsFileInProject(fileName)) {
+				if (projectEntry.FileName == fileName || projectEntry.IsFileInProject(fileName)) {
 					return projectEntry;
 				}
 			}
 			return null;
 		}
+
+		public IEnumerable<Project> GetProjectsContainingFile (string fileName)
+		{
+			ReadOnlyCollection<Project> projects = GetAllProjects ();
+
+			Project mainProject = null;
+			var projectsWithLinks = new List<Project>();
+			foreach (Project projectEntry in projects) {
+				if (projectEntry.FileName == fileName || projectEntry.IsFileInProject(fileName)) {
+					var projectPath = Path.GetDirectoryName (projectEntry.FileName);
+					if (fileName.StartsWith (projectPath)) {
+						mainProject = projectEntry;
+					} else {
+						projectsWithLinks.Add (projectEntry);
+					}
+				}
+			}
+
+			if (mainProject != null) {
+				yield return mainProject;
+			}
+			foreach (var project in projectsWithLinks) {
+				yield return project;
+			}
+		}
 		
 		public SolutionEntityItem FindSolutionItem (string fileName)
 		{
@@ -619,7 +662,7 @@ namespace MonoDevelop.Projects
 			}
 			
 			try {
-				List<SolutionItem> toBuild = new List<SolutionItem> (allProjects.Where (p => p.NeedsBuilding (configuration)));
+				List<SolutionItem> toBuild = new List<SolutionItem> (allProjects);
 				
 				monitor.BeginTask (GettextCatalog.GetString ("Building Solution: {0} ({1})", Name, configuration.ToString ()), toBuild.Count);
 
@@ -708,8 +751,11 @@ namespace MonoDevelop.Projects
 		{
 			foreach (Project projectEntry in GetAllProjects()) {
 				foreach (ProjectFile fInfo in projectEntry.Files) {
-					if (fInfo.FilePath == oldName)
+					if (fInfo.FilePath == oldName) {
+						if (fInfo.BuildAction == projectEntry.GetDefaultBuildAction (oldName))
+							fInfo.BuildAction = projectEntry.GetDefaultBuildAction (newName);
 						fInfo.Name = newName;
+					}
 				}
 			}
 		}
@@ -988,9 +1034,18 @@ namespace MonoDevelop.Projects
 			get { return false; }
 		}
 		
+		public void OnModified (string hint)
+		{
+		}
+
 		public void Dispose ()
 		{
 		}
+
+		public object GetService (Type t)
+		{
+			return null;
+		}
 	}
 	
 	public class SolutionFolderFileCollection: System.Collections.ObjectModel.Collection<FilePath>
