@@ -39,6 +39,8 @@ using MonoDevelop.Core;
 using MonoDevelop.Projects;
 using MonoDevelop.Core.Serialization;
 using MonoDevelop.Core.StringParsing;
+using MonoDevelop.Core.Execution;
+using MonoDevelop.Projects.Extensions;
 
 namespace MonoDevelop.Projects
 {
@@ -108,6 +110,10 @@ namespace MonoDevelop.Projects
 				if (oldName != Name)
 					OnNameChanged (new WorkspaceItemRenamedEventArgs (this, oldName, Name));
 				NotifyModified ();
+
+				// Load the user properties after the file name has been set.
+				if (Loading)
+					LoadUserProperties ();
 			}
 		}
 
@@ -157,8 +163,6 @@ namespace MonoDevelop.Projects
 		
 		public virtual object GetService (Type t)
 		{
-			if (t.IsInstanceOfType (this))
-				return this;
 			return Services.ProjectService.GetExtensionChain (this).GetService (this, t);
 		}
 		
@@ -213,10 +217,16 @@ namespace MonoDevelop.Projects
 			return list.AsReadOnly ();
 		}
 
+		[Obsolete("Use GetProjectsContainingFile() (plural) instead")]
 		public virtual Project GetProjectContainingFile (FilePath fileName)
 		{
 			return null;
 		}
+
+		public virtual IEnumerable<Project> GetProjectsContainingFile (FilePath fileName)
+		{
+			yield break;
+		}
 		
 		public virtual ReadOnlyCollection<string> GetConfigurations ()
 		{
@@ -243,6 +253,11 @@ namespace MonoDevelop.Projects
 			return Services.ProjectService.GetExtensionChain (this).RunTarget (monitor, this, target, configuration);
 		}
 		
+		public bool SupportsBuild ()
+		{
+			return SupportsTarget (ProjectService.BuildTarget);
+		}
+
 		public void Clean (IProgressMonitor monitor, string configuration)
 		{
 			Clean (monitor, (SolutionConfigurationSelector) configuration);
@@ -253,6 +268,16 @@ namespace MonoDevelop.Projects
 			Services.ProjectService.GetExtensionChain (this).RunTarget (monitor, this, ProjectService.CleanTarget, configuration);
 		}
 		
+		public bool SupportsTarget (string target)
+		{
+			return Services.ProjectService.GetExtensionChain (this).SupportsTarget (this, target);
+		}
+
+		public bool SupportsExecute ()
+		{
+			return Services.ProjectService.GetExtensionChain (this).SupportsExecute (this);
+		}
+
 		public BuildResult Build (IProgressMonitor monitor, string configuration)
 		{
 			return InternalBuild (monitor, (SolutionConfigurationSelector) configuration);
@@ -282,31 +307,42 @@ namespace MonoDevelop.Projects
 		{
 			return Services.ProjectService.GetExtensionChain (this).CanExecute (this, context, configuration);
 		}
-		
+
+		public IEnumerable<ExecutionTarget> GetExecutionTargets (string configuration)
+		{
+			return GetExecutionTargets ((SolutionConfigurationSelector) configuration);
+		}
+
+		public IEnumerable<ExecutionTarget> GetExecutionTargets (ConfigurationSelector configuration)
+		{
+			return Services.ProjectService.GetExtensionChain (this).GetExecutionTargets (this, configuration);
+		}
+
+		[Obsolete ("This method will be removed in future releases")]
 		public bool NeedsBuilding (string configuration)
 		{
-			return NeedsBuilding ((SolutionConfigurationSelector) configuration);
+			return true;
 		}
 		
+		[Obsolete ("This method will be removed in future releases")]
 		public bool NeedsBuilding (ConfigurationSelector configuration)
 		{
-			return Services.ProjectService.GetExtensionChain (this).GetNeedsBuilding (this, configuration);
+			return true;
 		}
 		
+		[Obsolete ("This method will be removed in future releases")]
 		public void SetNeedsBuilding (bool value)
 		{
-			foreach (string conf in GetConfigurations ())
-				SetNeedsBuilding (value, new SolutionConfigurationSelector (conf));
 		}
 		
+		[Obsolete ("This method will be removed in future releases")]
 		public void SetNeedsBuilding (bool needsBuilding, string configuration)
 		{
-			SetNeedsBuilding (needsBuilding, (SolutionConfigurationSelector) configuration);
 		}
 		
+		[Obsolete ("This method will be removed in future releases")]
 		public void SetNeedsBuilding (bool needsBuilding, ConfigurationSelector configuration)
 		{
-			Services.ProjectService.GetExtensionChain (this).SetNeedsBuilding (this, needsBuilding, configuration);
 		}
 		
 		public virtual FileFormat FileFormat {
@@ -391,6 +427,16 @@ namespace MonoDevelop.Projects
 			return null;
 		}
 		
+		internal protected virtual bool OnGetSupportsTarget (string target)
+		{
+			return true;
+		}
+
+		internal protected virtual bool OnGetSupportsExecute ()
+		{
+			return true;
+		}
+
 		protected virtual void OnClean (IProgressMonitor monitor, ConfigurationSelector configuration)
 		{
 		}
@@ -408,10 +454,15 @@ namespace MonoDevelop.Projects
 		{
 			return true;
 		}
-		
+
+		internal protected virtual IEnumerable<ExecutionTarget> OnGetExecutionTargets (ConfigurationSelector configuration)
+		{
+			yield break;
+		}
+
 		internal protected virtual bool OnGetNeedsBuilding (ConfigurationSelector configuration)
 		{
-			return false;
+			return true;
 		}
 		
 		internal protected virtual void OnSetNeedsBuilding (bool val, ConfigurationSelector configuration)
@@ -437,7 +488,6 @@ namespace MonoDevelop.Projects
 		
 		protected virtual void OnEndLoad ()
 		{
-			LoadUserProperties ();
 		}
 		
 		public virtual void LoadUserProperties ()
@@ -536,11 +586,17 @@ namespace MonoDevelop.Projects
 		
 		protected virtual void OnNameChanged (WorkspaceItemRenamedEventArgs e)
 		{
+			fileStatusTracker.ResetLoadTimes ();
 			NotifyModified ();
 			if (NameChanged != null)
 				NameChanged (this, e);
 		}
-		
+
+		internal protected virtual object OnGetService (Type t)
+		{
+			return null;
+		}
+
 		protected void NotifyModified ()
 		{
 			OnModified (new WorkspaceItemEventArgs (this));
@@ -634,12 +690,12 @@ namespace MonoDevelop.Projects
 			}
 			set {
 //				needsReload = value;
-				reloadCheckTime.Clear ();
-				foreach (FilePath file in item.GetItemFiles (false)) {
-					if (value)
+				if (value) {
+					reloadCheckTime.Clear ();
+					foreach (FilePath file in item.GetItemFiles (false))
 						reloadCheckTime [file] = DateTime.MinValue;
-					else
-						reloadCheckTime [file] = GetLastWriteTime (file);
+				} else {
+					ResetLoadTimes ();
 				}
 			}
 		}
