@@ -38,10 +38,11 @@ using MonoDevelop.Components.Commands;
 using MonoDevelop.Core.Collections;
 using MonoDevelop.Ide.Gui.Components;
 using System.Linq;
+using MonoDevelop.Components;
 
 namespace MonoDevelop.Ide.Gui.Pads.ProjectPad
 {
-	public class ProjectFileNodeBuilder: TypeNodeBuilder
+	class ProjectFileNodeBuilder: TypeNodeBuilder
 	{
 		public override Type NodeDataType {
 			get { return typeof(ProjectFile); }
@@ -56,65 +57,90 @@ namespace MonoDevelop.Ide.Gui.Pads.ProjectPad
 			var file = (ProjectFile) dataObject;
 			return file.Link.IsNullOrEmpty ? file.FilePath.FileName : file.Link.FileName;
 		}
-		
-		public override void GetNodeAttributes (ITreeNavigator treeNavigator, object dataObject, ref NodeAttributes attributes)
+
+		public override void GetNodeAttributes (ITreeNavigator parentNode, object dataObject, ref NodeAttributes attributes)
 		{
-			ProjectFile file = (ProjectFile) dataObject;
-			if (file.DependsOnFile != null) {
-				attributes = NodeAttributes.None;
-			} else {
-				attributes |= NodeAttributes.AllowRename;
+			var file = (ProjectFile) dataObject;
+
+			if ((file.Flags & ProjectItemFlags.Hidden) != 0) {
+				attributes |= NodeAttributes.Hidden;
+				return;
 			}
-			if (!file.Visible && !treeNavigator.Options ["ShowAllFiles"])
+
+			attributes |= NodeAttributes.AllowRename;
+
+			if (!file.Visible && !parentNode.Options ["ShowAllFiles"])
 				attributes |= NodeAttributes.Hidden;
 		}
 		
-		public override void BuildNode (ITreeBuilder treeBuilder, object dataObject, ref string label, ref Gdk.Pixbuf icon, ref Gdk.Pixbuf closedIcon)
+		public override void BuildNode (ITreeBuilder treeBuilder, object dataObject, NodeInfo nodeInfo)
 		{
 			ProjectFile file = (ProjectFile) dataObject;
 
-			label = file.Link.IsNullOrEmpty ? file.FilePath.FileName : file.Link.FileName;
+			nodeInfo.Label = GLib.Markup.EscapeText (file.Link.IsNullOrEmpty ? file.FilePath.FileName : file.Link.FileName);
 			if (!File.Exists (file.FilePath)) {
-				label = "<span foreground='red'>" + label + "</span>";
+				nodeInfo.Label = "<span foreground='red'>" + nodeInfo.Label + "</span>";
 			}
 			
-			icon = DesktopService.GetPixbufForFile (file.FilePath, Gtk.IconSize.Menu);
+			nodeInfo.Icon = DesktopService.GetIconForFile (file.FilePath, Gtk.IconSize.Menu);
 			
-			if (file.IsLink && icon != null) {
-				icon = icon.Copy ();
-				using (Gdk.Pixbuf overlay = Gdk.Pixbuf.LoadFromResource ("Icons.16x16.LinkOverlay.png")) {
-					overlay.Composite (icon,
-						0,  0,
-						icon.Width, icon.Width,
-						0, 0,
-						1, 1, Gdk.InterpType.Bilinear, 255); 
+			if (file.IsLink && nodeInfo.Icon != null) {
+				var overlay = ImageService.GetIcon ("md-link-overlay").WithSize (Xwt.IconSize.Small);
+				var cached = Context.GetComposedIcon (nodeInfo.Icon, overlay);
+				if (cached != null)
+					nodeInfo.Icon = cached;
+				else {
+					var ib = new Xwt.Drawing.ImageBuilder (nodeInfo.Icon.Width, nodeInfo.Icon.Height);
+					ib.Context.DrawImage (nodeInfo.Icon, 0, 0);
+					ib.Context.DrawImage (overlay, 0, 0);
+					var res = ib.ToVectorImage ();
+					ib.Dispose ();
+					Context.CacheComposedIcon (nodeInfo.Icon, overlay, res);
+					nodeInfo.Icon = res;
 				}
 			}
 		}
 		
 		public override object GetParentObject (object dataObject)
 		{
-			ProjectFile file = (ProjectFile) dataObject;
-			string dir = Path.GetDirectoryName (file.FilePath);
+			var file = (ProjectFile) dataObject;
+			var dir = !file.IsLink ? file.FilePath.ParentDirectory : file.Project.BaseDirectory.Combine (file.ProjectVirtualPath).ParentDirectory;
 			
 			if (!string.IsNullOrEmpty (file.DependsOn)) {
-				ProjectFile groupUnder = file.Project.Files.GetFile (Path.Combine (dir, file.DependsOn));
+				ProjectFile groupUnder = file.Project.Files.GetFile (file.FilePath.ParentDirectory.Combine (file.DependsOn));
 				if (groupUnder != null)
 					return groupUnder;
 			}
 			
 			if (dir == file.Project.BaseDirectory)
 				return file.Project;
-			else
-			    return new ProjectFolder (dir, file.Project, null);
+
+			return new ProjectFolder (dir, file.Project, null);
 		}
 		
 		public override int CompareObjects (ITreeNavigator thisNode, ITreeNavigator otherNode)
 		{
 			if (otherNode.DataItem is ProjectFolder)
 				return 1;
-			else
+
+			if (!(thisNode.DataItem is ProjectFile))
 				return DefaultSort;
+			if (!(otherNode.DataItem is ProjectFile))
+				return DefaultSort;
+
+			string name1 = thisNode.NodeName;
+			string name2 = otherNode.NodeName;
+
+			//Compare filenames without extension
+			string path1 = Path.GetFileNameWithoutExtension (name1);
+			string path2 = Path.GetFileNameWithoutExtension (name2);
+			int cmp = string.Compare (path1, path2, StringComparison.CurrentCultureIgnoreCase);
+			if (cmp != 0)
+				return cmp;
+			//Compare extensions
+			string ext1 = Path.GetExtension (name1);
+			string ext2 = Path.GetExtension (name2);
+			return string.Compare (ext1, ext2, StringComparison.CurrentCultureIgnoreCase);
 		}
 		
 		public override bool HasChildNodes (ITreeBuilder builder, object dataObject)
@@ -131,19 +157,25 @@ namespace MonoDevelop.Ide.Gui.Pads.ProjectPad
 				foreach (ProjectFile pf in file.DependentChildren)
 					treeBuilder.AddChild (pf);
 		}
-
-
 	}
 	
-	public class ProjectFileNodeCommandHandler: NodeCommandHandler
+	class ProjectFileNodeCommandHandler: NodeCommandHandler
 	{
+		public override void OnRenameStarting (ref int selectionStart, ref int selectionLength)
+		{
+			string name = CurrentNode.NodeName;
+			selectionStart = 0;
+			selectionLength = Path.GetFileNameWithoutExtension(name).Length;
+		}
+
 		public override void RenameItem (string newName)
 		{
-			ProjectFile file = (ProjectFile) CurrentNode.DataItem;
+			ProjectFile newProjectFile = null;
+			var file = (ProjectFile) CurrentNode.DataItem;
 			
-			FilePath oldPath, newPath, newLink = FilePath.Null, oldLink = FilePath.Null;
+			FilePath oldPath, newPath, newLink = FilePath.Null;
 			if (file.IsLink) {
-				oldLink = file.ProjectVirtualPath;
+				var oldLink = file.ProjectVirtualPath;
 				newLink = oldLink.ParentDirectory.Combine (newName);
 				oldPath = file.Project.BaseDirectory.Combine (oldLink);
 				newPath = file.Project.BaseDirectory.Combine (newLink);
@@ -152,78 +184,133 @@ namespace MonoDevelop.Ide.Gui.Pads.ProjectPad
 				newPath = oldPath.ParentDirectory.Combine (newName);	
 			}
 			
-			if (oldPath != newPath) {
-				try {
-					if (!FileService.IsValidPath (newPath)) {
-						MessageService.ShowWarning (GettextCatalog.GetString ("The name you have chosen contains illegal characters. Please choose a different name."));
-					} else if (File.Exists (newPath) || Directory.Exists (newPath) ||
-					           (file.Project != null && file.Project.Files.GetFileWithVirtualPath (newPath.ToRelative (file.Project.BaseDirectory)) != null)) {
-						MessageService.ShowWarning (GettextCatalog.GetString ("File or directory name is already in use. Please choose a different one."));
-					} else {
-						if (file.IsLink) {
-							file.Link = newLink;
-						} else {
-							FileService.RenameFile (oldPath, newName);
-						}
-						if (file.Project != null)
-							IdeApp.ProjectOperations.Save (file.Project);
-					}
-				} catch (System.ArgumentException) { // new file name with wildcard (*, ?) characters in it
+			try {
+				if (file.Project != null)
+					newProjectFile = file.Project.Files.GetFileWithVirtualPath (newPath.ToRelative (file.Project.BaseDirectory));
+
+				if (!FileService.IsValidPath (newPath)) {
 					MessageService.ShowWarning (GettextCatalog.GetString ("The name you have chosen contains illegal characters. Please choose a different name."));
-				} catch (System.IO.IOException ex) {
-					MessageService.ShowException (ex, GettextCatalog.GetString ("There was an error renaming the file."));
+				} else if (newProjectFile != null && newProjectFile != file) {
+					// If there is already a file under the newPath which is *different*, then throw an exception
+					MessageService.ShowWarning (GettextCatalog.GetString ("File or directory name is already in use. Please choose a different one."));
+				} else {
+					if (file.IsLink) {
+						file.Link = newLink;
+					} else {
+						// This could throw an exception if we try to replace another file during the rename.
+						FileService.RenameFile (oldPath, newName);
+					}
+					if (file.Project != null)
+						IdeApp.ProjectOperations.Save (file.Project);
 				}
+			} catch (ArgumentException) { // new file name with wildcard (*, ?) characters in it
+				MessageService.ShowWarning (GettextCatalog.GetString ("The name you have chosen contains illegal characters. Please choose a different name."));
+			} catch (IOException ex) {
+				MessageService.ShowException (ex, GettextCatalog.GetString ("There was an error renaming the file."));
 			}
 		}
 		
 		public override void ActivateItem ()
 		{
 			ProjectFile file = (ProjectFile) CurrentNode.DataItem;
-			IdeApp.Workbench.OpenDocument (file.FilePath);
+			IdeApp.Workbench.OpenDocument (file.FilePath, file.Project);
+		}
+		
+		public override void ActivateMultipleItems ()
+		{
+			ProjectFile file;
+			for (int i = 0; i < CurrentNodes.Length; i++) {
+				// Only bring the last file to the front
+				file = (ProjectFile) CurrentNodes [i].DataItem;
+				IdeApp.Workbench.OpenDocument (file.FilePath, file.Project, i == CurrentNodes.Length - 1);
+			}
 		}
 		
 		public override DragOperation CanDragNode ()
 		{
-			if (((ProjectFile) CurrentNode.DataItem).DependsOnFile == null)
-				return DragOperation.Copy | DragOperation.Move;
-			else
-				return DragOperation.None;
+			return DragOperation.Copy | DragOperation.Move;
 		}
 		
 		public override bool CanDropNode (object dataObject, DragOperation operation)
 		{
-			return (dataObject is SolutionItem && ((ProjectFile) CurrentNode.DataItem).DependsOnFile == null);
+			var target = (ProjectFile) CurrentNode.DataItem;
+			var pf = dataObject as ProjectFile;
+
+			return pf != null && pf != target && !pf.HasChildren && target.DependsOn == null;
 		}
-		
-		public override void OnNodeDrop (object dataObject, DragOperation operation)
+
+		void Drop (ProjectFile pf, DragOperation operation, HashSet<SolutionEntityItem> projectsToSave)
 		{
+			var target = (ProjectFile) CurrentNode.DataItem;
+			var targetDirectory = target.FilePath.ParentDirectory;
+
+			// file dependencies only work if they are in the same physical folder
+			if (pf.FilePath.ParentDirectory != targetDirectory) {
+				var targetPath = targetDirectory.Combine (pf.FilePath.FileName);
+
+				// if copying to the same directory, make a copy with a different name
+				if (targetPath == pf.FilePath)
+					targetPath = ProjectOperations.GetTargetCopyName (targetPath, false);
+
+				if (File.Exists (targetPath))
+					if (!MessageService.Confirm (GettextCatalog.GetString ("The file '{0}' already exists. Do you want to overwrite it?", targetPath.FileName), AlertButton.OverwriteFile))
+						return;
+
+				// unlink the project file from its current parent
+				pf.DependsOn = null;
+
+				projectsToSave.Add (pf.Project);
+
+				bool move = operation == DragOperation.Move;
+				var opText = move ? GettextCatalog.GetString ("Moving file...") : GettextCatalog.GetString ("Copying file...");
+
+				using (var monitor = IdeApp.Workbench.ProgressMonitors.GetStatusProgressMonitor (opText, Stock.StatusWorking, true))
+					IdeApp.ProjectOperations.TransferFiles (monitor, pf.Project, pf.FilePath, target.Project, targetPath, move, true);
+
+				pf = target.Project.Files.GetFile (targetPath);
+			}
+
+			// the dropped project file now depends on the file it was just dropped onto
+			pf.DependsOn = target.FilePath.FileName;
+			projectsToSave.Add (pf.Project);
+		}
+
+		public override void OnMultipleNodeDrop (object[] dataObjects, DragOperation operation)
+		{
+			var projectsToSave = new HashSet<SolutionEntityItem> ();
+
+			foreach (var dataObject in dataObjects)
+				Drop ((ProjectFile) dataObject, operation, projectsToSave);
+
+			IdeApp.ProjectOperations.Save (projectsToSave);
 		}
 		
 		public override bool CanDeleteItem ()
 		{
-			return ((ProjectFile) CurrentNode.DataItem).DependsOnFile == null;
+			return true;
 		}
 		
 		[CommandHandler (EditCommands.Delete)]
 		[AllowMultiSelection]
 		public override void DeleteMultipleItems ()
 		{
+			var projects = new Set<SolutionEntityItem> ();
+			var files = new List<ProjectFile> ();
 			bool hasChildren = false;
-			List<ProjectFile> files = new List<ProjectFile> ();
-			Set<SolutionEntityItem> projects = new Set<SolutionEntityItem> ();
-			foreach (ITreeNavigator node in CurrentNodes) {
-				ProjectFile pf = (ProjectFile) node.DataItem;
+
+			foreach (var node in CurrentNodes) {
+				var pf = (ProjectFile) node.DataItem;
 				projects.Add (pf.Project);
 				if (pf.HasChildren)
 					hasChildren = true;
 				files.Add (pf);
 			}
-			
-			AlertButton removeFromProject = new AlertButton (GettextCatalog.GetString ("_Remove from Project"), Gtk.Stock.Remove);
-			
+
+			var removeFromProject = new AlertButton (GettextCatalog.GetString ("_Remove from Project"), Gtk.Stock.Remove);
 			string question, secondaryText;
 			
-			secondaryText = GettextCatalog.GetString ("The Delete option permanently removes the file from your hard disk. Click Remove from Project if you only want to remove it from your current solution.");
+			secondaryText = GettextCatalog.GetString ("The Delete option permanently removes the file from your hard disk. " +
+				"Click Remove from Project if you only want to remove it from your current solution.");
 			
 			if (hasChildren) {
 				if (files.Count == 1)
@@ -240,28 +327,28 @@ namespace MonoDevelop.Ide.Gui.Pads.ProjectPad
 				else
 					question = GettextCatalog.GetString ("Are you sure you want to remove the selected files from the project?");
 			}
-			
-			AlertButton result = MessageService.AskQuestion (question, secondaryText,
-			                                                 AlertButton.Delete, AlertButton.Cancel, removeFromProject);
+
+			var result = MessageService.AskQuestion (question, secondaryText, AlertButton.Delete, AlertButton.Cancel, removeFromProject);
 			if (result != removeFromProject && result != AlertButton.Delete) 
 				return;
-			   
-			foreach (ProjectFile file in files) {
-				Project project = file.Project;
+
+			foreach (var file in files) {
+				var project = file.Project;
 				var inFolder = project.Files.GetFilesInVirtualPath (file.ProjectVirtualPath.ParentDirectory).ToList ();
-				if (inFolder.Count == 1 && inFolder [0] == file) {
+
+				if (inFolder.Count == 1 && inFolder [0] == file && project.Files.GetFileWithVirtualPath (file.ProjectVirtualPath.ParentDirectory) == null) {
 					// This is the last project file in the folder. Make sure we keep
 					// a reference to the folder, so it is not deleted from the tree.
-					ProjectFile folderFile = new ProjectFile (project.BaseDirectory.Combine (file.ProjectVirtualPath.ParentDirectory));
+					var folderFile = new ProjectFile (project.BaseDirectory.Combine (file.ProjectVirtualPath.ParentDirectory));
 					folderFile.Subtype = Subtype.Directory;
 					project.Files.Add (folderFile);
 				}
-				
+
 				if (file.HasChildren) {
-					foreach (ProjectFile f in file.DependentChildren) {
-						project.Files.Remove (f);
+					foreach (var child in file.DependentChildren) {
+						project.Files.Remove (child);
 						if (result == AlertButton.Delete)
-							FileService.DeleteFile (f.Name);
+							FileService.DeleteFile (child.Name);
 					}
 				}
 			
@@ -291,9 +378,28 @@ namespace MonoDevelop.Ide.Gui.Pads.ProjectPad
 		[CommandUpdateHandler (ViewCommands.OpenWithList)]
 		public void OnOpenWithUpdate (CommandArrayInfo info)
 		{
-			ProjectFile finfo = (ProjectFile) CurrentNode.DataItem;
+			var pf = (ProjectFile) CurrentNode.DataItem;
+			PopulateOpenWithViewers (info, pf.Project, pf.FilePath);
+		}
+		
+		internal static void PopulateOpenWithViewers (CommandArrayInfo info, Project project, string filePath)
+		{
+			var viewers = DisplayBindingService.GetFileViewers (filePath, project).ToList ();
+			
+			//show the default viewer first
+			var def = viewers.FirstOrDefault (v => v.CanUseAsDefault) ?? viewers.FirstOrDefault (v => v.IsExternal);
+			if (def != null) {
+				CommandInfo ci = info.Add (def.Title, def);
+				ci.Description = GettextCatalog.GetString ("Open with '{0}'", def.Title);
+				if (viewers.Count > 1)
+					info.AddSeparator ();
+			}
+			
+			//then the builtins, followed by externals
 			FileViewer prev = null; 
-			foreach (FileViewer fv in IdeApp.Workbench.GetFileViewers (finfo.Name)) {
+			foreach (FileViewer fv in viewers) {
+				if (def != null && fv.Equals (def))
+					continue;
 				if (prev != null && fv.IsExternal != prev.IsExternal)
 					info.AddSeparator ();
 				CommandInfo ci = info.Add (fv.Title, fv);
@@ -320,10 +426,11 @@ namespace MonoDevelop.Ide.Gui.Pads.ProjectPad
 		[CommandUpdateHandler (FileCommands.SetBuildAction)]
 		public void OnSetBuildActionUpdate (CommandArrayInfo info)
 		{
-			Set<string> toggledActions = new Set<string> ();
+			var toggledActions = new Set<string> ();
 			Project proj = null;
-			foreach (ITreeNavigator node in CurrentNodes) {
-				ProjectFile finfo = (ProjectFile) node.DataItem;
+
+			foreach (var node in CurrentNodes) {
+				var finfo = (ProjectFile) node.DataItem;
 				
 				//disallow multi-slect on more than one project, since available build actions may differ
 				if (proj == null && finfo.Project != null) {
@@ -334,12 +441,15 @@ namespace MonoDevelop.Ide.Gui.Pads.ProjectPad
 				}
 				toggledActions.Add (finfo.BuildAction);
 			}
+
+			if (proj == null)
+				return;
 			
 			foreach (string action in proj.GetBuildActions ()) {
 				if (action == "--") {
 					info.AddSeparator ();
 				} else {
-					CommandInfo ci = info.Add (BuildAction.Translate (action), action);
+					CommandInfo ci = info.Add (action, action);
 					ci.Checked = toggledActions.Contains (action);
 					if (ci.Checked)
 						ci.CheckedInconsistent = toggledActions.Count > 1;
@@ -351,13 +461,7 @@ namespace MonoDevelop.Ide.Gui.Pads.ProjectPad
 		[AllowMultiSelection]
 		public void OnShowProperties ()
 		{
-			foreach (Pad pad in IdeApp.Workbench.Pads) {
-				if (pad.Id == "MonoDevelop.DesignerSupport.PropertyPad") {
-					pad.Visible = true;
-					pad.BringToFront (true);
-					return;
-				}
-			}
+			IdeApp.Workbench.Pads.PropertyPad.BringToFront (true);
 		}
 		
 		//NOTE: This command is slightly odd, as it operates on a tri-state value, 
