@@ -39,10 +39,11 @@ using MonoDevelop.Components.Commands;
 using MonoDevelop.Ide.Gui.Components;
 using System.Collections.Generic;
 using System.Linq;
+using MonoDevelop.Components;
 
 namespace MonoDevelop.Ide.Gui.Pads.ProjectPad
 {
-	public class SystemFileNodeBuilder: TypeNodeBuilder
+	class SystemFileNodeBuilder: TypeNodeBuilder
 	{
 		public override Type NodeDataType {
 			get { return typeof(SystemFile); }
@@ -62,21 +63,21 @@ namespace MonoDevelop.Ide.Gui.Pads.ProjectPad
 			attributes |= NodeAttributes.AllowRename;
 		}
 		
-		public override void BuildNode (ITreeBuilder treeBuilder, object dataObject, ref string label, ref Gdk.Pixbuf icon, ref Gdk.Pixbuf closedIcon)
+		public override void BuildNode (ITreeBuilder treeBuilder, object dataObject, NodeInfo nodeInfo)
 		{
 			SystemFile file = (SystemFile) dataObject;
-			label = file.Name;
+			nodeInfo.Label = GLib.Markup.EscapeText (file.Name);
 			
-			icon = DesktopService.GetPixbufForFile (file.Path, Gtk.IconSize.Menu);
+			nodeInfo.Icon = DesktopService.GetIconForFile (file.Path, Gtk.IconSize.Menu);
 			
 			if (file.ShowTransparent) {
-				Gdk.Pixbuf gicon = Context.GetComposedIcon (icon, "fade");
+				var gicon = Context.GetComposedIcon (nodeInfo.Icon, "fade");
 				if (gicon == null) {
-					gicon = ImageService.MakeTransparent (icon, 0.5);
-					Context.CacheComposedIcon (icon, "fade", gicon);
+					gicon = nodeInfo.Icon.WithAlpha (0.5);
+					Context.CacheComposedIcon (nodeInfo.Icon, "fade", gicon);
 				}
-				icon = gicon;
-				label = "<span foreground='dimgrey'>" + label + "</span>";
+				nodeInfo.Icon = gicon;
+				nodeInfo.Label = "<span foreground='dimgrey'>" + nodeInfo.Label + "</span>";
 			}
 		}
 		
@@ -89,7 +90,7 @@ namespace MonoDevelop.Ide.Gui.Pads.ProjectPad
 		}
 	}
 	
-	public class SystemFileNodeCommandHandler: NodeCommandHandler
+	class SystemFileNodeCommandHandler: NodeCommandHandler
 	{
 		public override void RenameItem (string newName)
 		{
