@@ -47,9 +47,11 @@ using MonoDevelop.Ide.Extensions;
 using System.Linq;
 using System.Threading;
 using MonoDevelop.Ide.TypeSystem;
+using ICSharpCode.NRefactory;
 using ICSharpCode.NRefactory.TypeSystem;
 using ICSharpCode.NRefactory.TypeSystem.Implementation;
 using System.Text;
+using System.Collections.ObjectModel;
 
 namespace MonoDevelop.Ide.Gui
 {
@@ -59,7 +61,10 @@ namespace MonoDevelop.Ide.Gui
 		
 		IWorkbenchWindow window;
 		TextEditorExtension editorExtension;
-		
+		ParsedDocument parsedDocument;
+		IProjectContent singleFileContext;
+		Mono.TextEditor.ITextEditorDataProvider provider = null;
+
 		const int ParseDelay = 600;
 
 		public IWorkbenchWindow Window {
@@ -77,19 +82,21 @@ namespace MonoDevelop.Ide.Gui
  		
 		public T GetContent<T> () where T : class
 		{
+			if (window == null)
+				return null;
 			//check whether the ViewContent can return the type directly
-			T ret = Window.ActiveViewContent.GetContent<T> ();
+			T ret = Window.ActiveViewContent.GetContent (typeof(T)) as T;
 			if (ret != null)
 				return ret;
 			
 			//check the primary viewcontent
 			//not sure if this is the right thing to do, but things depend on this behaviour
 			if (Window.ViewContent != Window.ActiveViewContent) {
-				ret = Window.ViewContent.GetContent<T> ();
+				ret = Window.ViewContent.GetContent (typeof(T)) as T;
 				if (ret != null)
 					return ret;
 			}
-			
+
 			//no, so look through the TexteditorExtensions as well
 			TextEditorExtension nextExtension = editorExtension;
 			while (nextExtension != null) {
@@ -103,14 +110,14 @@ namespace MonoDevelop.Ide.Gui
 		public IEnumerable<T> GetContents<T> () where T : class
 		{
 			//check whether the ViewContent can return the type directly
-			T ret = Window.ActiveViewContent.GetContent<T> ();
+			T ret = (T) Window.ActiveViewContent.GetContent (typeof(T));
 			if (ret != null)
 				yield return ret;
 			
 			//check the primary viewcontent
 			//not sure if this is the right thing to do, but things depend on this behaviour
 			if (Window.ViewContent != Window.ActiveViewContent) {
-				ret = Window.ViewContent.GetContent<T> ();
+				ret = (T) Window.ViewContent.GetContent (typeof(T));
 				if (ret != null)
 					yield return ret;
 			}
@@ -123,7 +130,19 @@ namespace MonoDevelop.Ide.Gui
 				nextExtension = nextExtension.Next as TextEditorExtension;
 			}
 		}
-		
+		static Document ()
+		{
+			if (IdeApp.Workbench != null) {
+				IdeApp.Workbench.ActiveDocumentChanged += delegate {
+					// reparse on document switch to update the current file with changes done in other files.
+					var doc = IdeApp.Workbench.ActiveDocument;
+					if (doc == null || doc.Editor == null)
+						return;
+					doc.StartReparseThread ();
+				};
+			}
+		}
+
 		public Document (IWorkbenchWindow window)
 		{
 			Counters.OpenDocuments++;
@@ -135,6 +154,7 @@ namespace MonoDevelop.Ide.Gui
 				IdeApp.Workspace.ItemRemovedFromSolution += OnEntryRemoved;
 			if (window.ViewContent.Project != null)
 				window.ViewContent.Project.Modified += HandleProjectModified;
+			window.ViewsChanged += HandleViewsChanged;
 		}
 
 /*		void UpdateRegisteredDom (object sender, ProjectDomEventArgs e)
@@ -148,7 +168,7 @@ namespace MonoDevelop.Ide.Gui
 
 		public FilePath FileName {
 			get {
-				if (!Window.ViewContent.IsFile)
+				if (Window == null || !Window.ViewContent.IsFile)
 					return null;
 				return Window.ViewContent.IsUntitled ? Window.ViewContent.UntitledName : Window.ViewContent.ContentName;
 			}
@@ -164,11 +184,11 @@ namespace MonoDevelop.Ide.Gui
 		}
 		
 		public bool HasProject {
-			get { return Window.ViewContent.Project != null; }
+			get { return Window != null ? Window.ViewContent.Project != null : false; }
 		}
 		
 		public Project Project {
-			get { return Window.ViewContent.Project; }
+			get { return Window != null ? Window.ViewContent.Project : null; }
 /*			set { 
 				Window.ViewContent.Project = value; 
 				if (value != null)
@@ -178,21 +198,40 @@ namespace MonoDevelop.Ide.Gui
 				StartReparseThread ();
 			}*/
 		}
-		
-		IProjectContent singleFileContext;
-		public  virtual IProjectContent ProjectContent {
+
+		public bool IsCompileableInProject {
+			get {
+				var project = Project;
+				if (project == null)
+					return false;
+				var solution = project.ParentSolution;
+
+				if (solution != null && IdeApp.Workspace != null) {
+					var config = IdeApp.Workspace.ActiveConfiguration;
+					if (config != null) {
+						var sc = solution.GetConfiguration (config);
+						if (sc != null && !sc.BuildEnabledForItem (project))
+							return false;
+					}
+				}
+
+				var pf = project.GetProjectFile (FileName);
+				return pf != null && pf.BuildAction != BuildAction.None;
+			}
+		}
+
+		public  IProjectContent ProjectContent {
 			get {
 				return Project != null ? TypeSystemService.GetProjectContext (Project) : GetProjectContext ();
 			}
 		}
 		
-		public  virtual ICompilation Compilation {
+		public virtual ICompilation Compilation {
 			get {
 				return Project != null ? TypeSystemService.GetCompilation (Project) : GetProjectContext ().CreateCompilation ();
 			}
 		}
 		
-		ParsedDocument parsedDocument;
 		public virtual ParsedDocument ParsedDocument {
 			get {
 				return parsedDocument;
@@ -208,14 +247,59 @@ namespace MonoDevelop.Ide.Gui
 			window.SelectWindow ();
 		}
 		
-		public IBaseViewContent ActiveView {
-			get { return window.ActiveViewContent; }
+		public DocumentView ActiveView {
+			get {
+				LoadViews (true);
+				return WrapView (window.ActiveViewContent);
+			}
 		}
 		
-		public IViewContent PrimaryView {
-			get { return window.ViewContent; }
+		public DocumentView PrimaryView {
+			get {
+				LoadViews (true);
+				return WrapView (window.ViewContent);
+			}
 		}
-		
+
+		public ReadOnlyCollection<DocumentView> Views {
+			get {
+				LoadViews (true);
+				if (viewsRO == null)
+					viewsRO = new ReadOnlyCollection<DocumentView> (views);
+				return viewsRO;
+			}
+		}
+
+		ReadOnlyCollection<DocumentView> viewsRO;
+		List<DocumentView> views = new List<DocumentView> ();
+
+		void HandleViewsChanged (object sender, EventArgs e)
+		{
+			LoadViews (false);
+		}
+
+		void LoadViews (bool force)
+		{
+			if (!force && views == null)
+				return;
+			var newList = new List<DocumentView> ();
+			newList.Add (WrapView (window.ViewContent));
+			foreach (var v in window.SubViewContents)
+				newList.Add (WrapView (v));
+			views = newList;
+			viewsRO = null;
+		}
+
+		DocumentView WrapView (IBaseViewContent content)
+		{
+			if (content == null)
+				return null;
+			if (views != null)
+				return views.FirstOrDefault (v => v.BaseContent == content) ?? new DocumentView (this, content);
+			else
+				return new DocumentView (this, content);
+		}
+
 		public string Name {
 			get {
 				IViewContent view = Window.ViewContent;
@@ -223,19 +307,14 @@ namespace MonoDevelop.Ide.Gui
 			}
 		}
 
-		Mono.TextEditor.TextEditorData data = null;
 		public Mono.TextEditor.TextEditorData Editor {
 			get {
-				if (data == null) {
-					Mono.TextEditor.ITextEditorDataProvider view = GetContent <Mono.TextEditor.ITextEditorDataProvider> ();
-					if (view != null) {
-						data = view.GetTextEditorData ();
-						if (data.IsDisposed)
-							return null;
-						data.Document.Tag = this;
-					}
+				if (provider == null) {
+					provider = GetContent <Mono.TextEditor.ITextEditorDataProvider> ();
+					if (provider == null)
+						return null;
 				}
-				return data;
+				return provider.GetTextEditorData ();
 			}
 		}
 
@@ -256,7 +335,7 @@ namespace MonoDevelop.Ide.Gui
 			}
 		}
 		
-		public virtual void Save ()
+		public void Save ()
 		{
 			// suspend type service "check all file loop" since we have already a parsed document.
 			// Or at least one that updates "soon".
@@ -273,8 +352,11 @@ namespace MonoDevelop.Ide.Gui
 				if (Window.ViewContent.ContentName == null) {
 					SaveAs ();
 				} else {
-					if (!FileService.RequestFileEdit (Window.ViewContent.ContentName))
-						MessageService.ShowMessage (GettextCatalog.GetString ("The file could not be saved. Write permission has not been granted."));
+					try {
+						FileService.RequestFileEdit (Window.ViewContent.ContentName, true);
+					} catch (Exception ex) {
+						MessageService.ShowException (ex, GettextCatalog.GetString ("The file could not be saved."));
+					}
 					
 					FileAttributes attr = FileAttributes.ReadOnly | FileAttributes.Directory | FileAttributes.Offline | FileAttributes.System;
 	
@@ -297,7 +379,13 @@ namespace MonoDevelop.Ide.Gui
 				// Note that the parsed document may be overwritten by a background thread to a more recent one.
 				var doc = parsedDocument;
 				if (doc != null && doc.ParsedFile != null) {
-					doc.ParsedFile.LastWriteTime = DateTime.Now;
+					string fileName = Window.ViewContent.ContentName;
+					try {
+						doc.ParsedFile.LastWriteTime = File.GetLastWriteTimeUtc (fileName);
+					} catch (Exception e) {
+						doc.ParsedFile.LastWriteTime = DateTime.UtcNow;
+						LoggingService.LogWarning ("Exception while getting the write time from " + fileName, e); 
+					}
 				}
 				TypeSystemService.TrackFileChanges = true;
 			}
@@ -312,7 +400,8 @@ namespace MonoDevelop.Ide.Gui
 		{
 			if (Window.ViewContent.IsViewOnly || !Window.ViewContent.IsFile)
 				return;
-			
+
+
 			Encoding encoding = null;
 			
 			IEncodedTextContent tbuffer = GetContent <IEncodedTextContent> ();
@@ -328,7 +417,6 @@ namespace MonoDevelop.Ide.Gui
 					Encoding = encoding,
 					ShowEncodingSelector = (tbuffer != null),
 				};
-				
 				if (Window.ViewContent.IsUntitled)
 					dlg.InitialFileName = Window.ViewContent.UntitledName;
 				else {
@@ -360,7 +448,7 @@ namespace MonoDevelop.Ide.Gui
 				else
 					Window.ViewContent.Save (filename + "~");
 			}
-			
+			TypeSystemService.RemoveSkippedfile (FileName);
 			// do actual save
 			if (tbuffer != null && encoding != null)
 				tbuffer.Save (filename, encoding);
@@ -374,7 +462,7 @@ namespace MonoDevelop.Ide.Gui
 			UpdateParseDocument ();
 		}
 		
-		public virtual bool IsBuildTarget
+		public bool IsBuildTarget
 		{
 			get
 			{
@@ -387,17 +475,17 @@ namespace MonoDevelop.Ide.Gui
 			}
 		}
 		
-		public virtual IAsyncOperation Build ()
+		public IAsyncOperation Build ()
 		{
 			return IdeApp.ProjectOperations.BuildFile (Window.ViewContent.ContentName);
 		}
 		
-		public virtual IAsyncOperation Rebuild ()
+		public IAsyncOperation Rebuild ()
 		{
 			return Build ();
 		}
 		
-		public virtual void Clean ()
+		public void Clean ()
 		{
 		}
 		
@@ -406,28 +494,29 @@ namespace MonoDevelop.Ide.Gui
 			return Run (Runtime.ProcessService.DefaultExecutionHandler);
 		}
 
-		public virtual IAsyncOperation Run (IExecutionHandler handler)
+		public IAsyncOperation Run (IExecutionHandler handler)
 		{
 			return IdeApp.ProjectOperations.ExecuteFile (Window.ViewContent.ContentName, handler);
 		}
 
-		public virtual bool CanRun ()
+		public bool CanRun ()
 		{
 			return CanRun (Runtime.ProcessService.DefaultExecutionHandler);
 		}
 		
-		public virtual bool CanRun (IExecutionHandler handler)
+		public bool CanRun (IExecutionHandler handler)
 		{
 			return IsBuildTarget && Window.ViewContent.ContentName != null && IdeApp.ProjectOperations.CanExecuteFile (Window.ViewContent.ContentName, handler);
 		}
 		
 		public bool Close ()
 		{
-			return Window.CloseWindow (false, true, 0);
+			return ((SdiWorkspaceWindow)Window).CloseWindow (false, true);
 		}
 		
-		protected virtual void OnSaved (EventArgs args)
+		void OnSaved (EventArgs args)
 		{
+			IdeApp.Workbench.SaveFileStatus ();
 			if (Saved != null)
 				Saved (this, args);
 		}
@@ -447,33 +536,15 @@ namespace MonoDevelop.Ide.Gui
 //			TypeSystemService.DomRegistered -= UpdateRegisteredDom;
 			CancelParseTimeout ();
 			ClearTasks ();
-			
-			string currentParseFile = FileName;
-			Project curentParseProject = Project;
-			
-			if (window is SdiWorkspaceWindow)
-				((SdiWorkspaceWindow)window).DetachFromPathedDocument ();
-			window.Closed -= OnClosed;
-			window.ActiveViewContentChanged -= OnActiveViewContentChanged;
-			if (IdeApp.Workspace != null)
-				IdeApp.Workspace.ItemRemovedFromSolution -= OnEntryRemoved;
-			
+			TypeSystemService.RemoveSkippedfile (FileName);
+
+
 			try {
 				OnClosed (a);
 			} catch (Exception ex) {
 				LoggingService.LogError ("Exception while calling OnClosed.", ex);
 			}
-			
-			while (editorExtension != null) {
-				try {
-					editorExtension.Dispose ();
-				} catch (Exception ex) {
-					LoggingService.LogError ("Exception while disposing extension:" + editorExtension, ex);
-				}
-				editorExtension = editorExtension.Next as TextEditorExtension;
-			}
-			editorExtension = null;
-			
+
 			// Parse the file when the document is closed. In this way if the document
 			// is closed without saving the changes, the saved compilation unit
 			// information will be restored
@@ -487,10 +558,32 @@ namespace MonoDevelop.Ide.Gui
 				dom = null;
 			}*/
 			
-			parsedDocument = null;
-			data = null;
 			Counters.OpenDocuments--;
 		}
+
+		internal void DisposeDocument ()
+		{
+			DetachExtensionChain ();
+			RemoveAnnotations (typeof(System.Object));
+			if (window is SdiWorkspaceWindow)
+				((SdiWorkspaceWindow)window).DetachFromPathedDocument ();
+			window.Closed -= OnClosed;
+			window.ActiveViewContentChanged -= OnActiveViewContentChanged;
+			if (IdeApp.Workspace != null)
+				IdeApp.Workspace.ItemRemovedFromSolution -= OnEntryRemoved;
+
+			// Unsubscribe project events
+			if (window.ViewContent.Project != null)
+				window.ViewContent.Project.Modified -= HandleProjectModified;
+			window.ViewsChanged += HandleViewsChanged;
+			window = null;
+
+			parsedDocument = null;
+			singleFileContext = null;
+			provider = null;
+			views = null;
+			viewsRO = null;
+		}
 #region document tasks
 		object lockObj = new object ();
 		
@@ -523,20 +616,68 @@ namespace MonoDevelop.Ide.Gui
 			OnViewChanged (args);
 		}
 		
-		protected virtual void OnClosed (EventArgs args)
+		void OnClosed (EventArgs args)
 		{
 			if (Closed != null)
 				Closed (this, args);
 		}
 		
-		protected virtual void OnViewChanged (EventArgs args)
+		void OnViewChanged (EventArgs args)
 		{
 			if (ViewChanged != null)
 				ViewChanged (this, args);
 		}
 		
 		bool wasEdited;
-		
+
+		void InitializeExtensionChain ()
+		{
+			DetachExtensionChain ();
+			var editor = GetContent<IExtensibleTextEditor> ();
+
+			ExtensionNodeList extensions = window.ExtensionContext.GetExtensionNodes ("/MonoDevelop/Ide/TextEditorExtensions", typeof(TextEditorExtensionNode));
+			editorExtension = null;
+			TextEditorExtension last = null;
+			var mimetypeChain = DesktopService.GetMimeTypeInheritanceChainForFile (FileName).ToArray ();
+			foreach (TextEditorExtensionNode extNode in extensions) {
+				if (!extNode.Supports (FileName, mimetypeChain))
+					continue;
+				TextEditorExtension ext;
+				try {
+					ext = (TextEditorExtension)extNode.CreateInstance ();
+				} catch (Exception e) {
+					LoggingService.LogError ("Error while creating text editor extension :" + extNode.Id + "(" + extNode.Type +")", e); 
+					continue;
+				}
+				if (ext.ExtendsEditor (this, editor)) {
+					if (last != null) {
+						ext.Next = last.Next;
+						last.Next = ext;
+						last = ext;
+					} else {
+						editorExtension = last = ext;
+						last.Next = editor.AttachExtension (editorExtension);
+					}
+					ext.Initialize (this);
+				}
+			}
+			if (window is SdiWorkspaceWindow)
+				((SdiWorkspaceWindow)window).AttachToPathedDocument (GetContent<MonoDevelop.Ide.Gui.Content.IPathedDocument> ());
+		}
+
+		void DetachExtensionChain ()
+		{
+			while (editorExtension != null) {
+				try {
+					editorExtension.Dispose ();
+				} catch (Exception ex) {
+					LoggingService.LogError ("Exception while disposing extension:" + editorExtension, ex);
+				}
+				editorExtension = editorExtension.Next as TextEditorExtension;
+			}
+			editorExtension = null;
+		}
+
 		void InitializeEditor (IExtensibleTextEditor editor)
 		{
 			Editor.Document.TextReplaced += (o, a) => {
@@ -560,32 +701,8 @@ namespace MonoDevelop.Ide.Gui
 			};
 			Editor.Document.Undone += (o, a) => StartReparseThread ();
 			Editor.Document.Redone += (o, a) => StartReparseThread ();
-			
-			// If the new document is a text editor, attach the extensions
-			
-			ExtensionNodeList extensions = window.ExtensionContext.GetExtensionNodes ("/MonoDevelop/Ide/TextEditorExtensions", typeof(TextEditorExtensionNode));
-			editorExtension = null;
-			TextEditorExtension last = null;
-			foreach (TextEditorExtensionNode extNode in extensions) {
-				if (!extNode.Supports (FileName))
-					continue;
-				
-				TextEditorExtension ext = (TextEditorExtension)extNode.CreateInstance ();
-				if (ext.ExtendsEditor (this, editor)) {
-					if (editorExtension == null)
-						editorExtension = ext;
-					
-					if (last != null)
-						last.Next = ext;
-					
-					last = ext;
-					ext.Initialize (this);
-				}
-			}
-			
-			if (editorExtension != null)
-				last.Next = editor.AttachExtension (editorExtension);
-			
+
+			InitializeExtensionChain ();
 		}
 		
 		internal void OnDocumentAttached ()
@@ -593,13 +710,10 @@ namespace MonoDevelop.Ide.Gui
 			IExtensibleTextEditor editor = GetContent<IExtensibleTextEditor> ();
 			if (editor != null) {
 				InitializeEditor (editor);
-				RunWhenLoaded (() => ReparseDocument ());
+				RunWhenLoaded (delegate { ListenToProjectLoad (Project); });
 			}
 			
 			window.Document = this;
-			
-			if (window is SdiWorkspaceWindow)
-				((SdiWorkspaceWindow)window).AttachToPathedDocument (GetContent<MonoDevelop.Ide.Gui.Content.IPathedDocument> ());
 		}
 		
 		/// <summary>
@@ -617,36 +731,58 @@ namespace MonoDevelop.Ide.Gui
 			}
 			e.Document.RunWhenLoaded (action);
 		}
-		
+
+		public void AttachToProject (Project project)
+		{
+			SetProject (project);
+		}
+
+		TypeSystemService.ProjectContentWrapper currentWrapper;
 		internal void SetProject (Project project)
 		{
-			if (Window.ViewContent.Project == project)
+			if (Window == null || Window.ViewContent == null || Window.ViewContent.Project == project)
 				return;
-			while (editorExtension != null) {
-				try {
-					editorExtension.Dispose ();
-				} catch (Exception ex) {
-					LoggingService.LogError ("Exception while disposing extension:" + editorExtension, ex);
-				}
-				editorExtension = editorExtension.Next as TextEditorExtension;
-			}
-			editorExtension = null;
+			DetachExtensionChain ();
 			ISupportsProjectReload pr = GetContent<ISupportsProjectReload> ();
 			if (pr != null) {
+				// Unsubscribe project events
+				if (Window.ViewContent.Project != null)
+					Window.ViewContent.Project.Modified -= HandleProjectModified;
 				Window.ViewContent.Project = project;
 				pr.Update (project);
 			}
 			if (project != null)
 				project.Modified += HandleProjectModified;
-			OnDocumentAttached ();
+			InitializeExtensionChain ();
+
+			ListenToProjectLoad (project);
+		}
+
+		void ListenToProjectLoad (Project project)
+		{
+			if (currentWrapper != null) {
+				currentWrapper.Loaded -= HandleInLoadChanged;
+				currentWrapper = null;
+			}
+			if (project != null) {
+				var wrapper = TypeSystemService.GetProjectContentWrapper (project);
+				wrapper.Loaded += HandleInLoadChanged;
+				currentWrapper = wrapper;
+				RunWhenLoaded (delegate {
+					currentWrapper.RequestLoad ();
+				});
+			}
+			StartReparseThread ();
+		}
+
+		void HandleInLoadChanged (object sender, EventArgs e)
+		{
+			StartReparseThread ();
 		}
 
 		void HandleProjectModified (object sender, SolutionItemModifiedEventArgs e)
 		{
-			if (!e.Any (
-					x => x is SolutionItemModifiedEventInfo &&
-				(((SolutionItemModifiedEventInfo)x).Hint == "TargetFramework" ||
-				((SolutionItemModifiedEventInfo)x).Hint == "References")))
+			if (!e.Any (x => x.Hint == "TargetFramework" || x.Hint == "References"))
 				return;
 			StartReparseThread ();
 		}
@@ -662,27 +798,37 @@ namespace MonoDevelop.Ide.Gui
 			try {
 				string currentParseFile = FileName;
 				var editor = Editor;
-				if (editor == null)
+				if (editor == null || string.IsNullOrEmpty (currentParseFile))
 					return null;
+				TypeSystemService.AddSkippedFile (currentParseFile);
 				string currentParseText = editor.Text;
 				this.parsedDocument = TypeSystemService.ParseFile (Project, currentParseFile, editor.Document.MimeType, currentParseText);
 				if (Project == null && this.parsedDocument != null) {
-					singleFileContext = GetProjectContext ().UpdateProjectContent (singleFileContext.GetFile (currentParseFile), parsedDocument.ParsedFile);
+					singleFileContext = GetProjectContext ().AddOrUpdateFiles (parsedDocument.ParsedFile);
 				}
 			} finally {
+
 				OnDocumentParsed (EventArgs.Empty);
 			}
 			return this.parsedDocument;
 		}
 
-		static readonly Lazy<IUnresolvedAssembly> mscorlib = new Lazy<IUnresolvedAssembly> ( () => new CecilLoader ().LoadAssemblyFile (typeof (object).Assembly.Location));
-		static readonly Lazy<IUnresolvedAssembly> systemCore = new Lazy<IUnresolvedAssembly>( () => new CecilLoader ().LoadAssemblyFile (typeof (System.Linq.Enumerable).Assembly.Location));
-		static readonly Lazy<IUnresolvedAssembly> system = new Lazy<IUnresolvedAssembly>( () => new CecilLoader ().LoadAssemblyFile (typeof (System.Uri).Assembly.Location));
+		static readonly Lazy<IUnresolvedAssembly> mscorlib = new Lazy<IUnresolvedAssembly> ( () => new IkvmLoader ().LoadAssemblyFile (typeof (object).Assembly.Location));
+		static readonly Lazy<IUnresolvedAssembly> systemCore = new Lazy<IUnresolvedAssembly>( () => new IkvmLoader ().LoadAssemblyFile (typeof (System.Linq.Enumerable).Assembly.Location));
+		static readonly Lazy<IUnresolvedAssembly> system = new Lazy<IUnresolvedAssembly>( () => new IkvmLoader ().LoadAssemblyFile (typeof (System.Uri).Assembly.Location));
 
 		static IUnresolvedAssembly Mscorlib { get { return mscorlib.Value; } }
 		static IUnresolvedAssembly SystemCore { get { return systemCore.Value; } }
 		static IUnresolvedAssembly System { get { return system.Value; } }
 
+		public bool IsProjectContextInUpdate {
+			get {
+				if (currentWrapper == null)
+					return false;
+				return !currentWrapper.IsLoaded;
+			}
+		}
+
 		public virtual IProjectContent GetProjectContext ()
 		{
 			if (Project == null) {
@@ -691,7 +837,7 @@ namespace MonoDevelop.Ide.Gui
 					singleFileContext = singleFileContext.AddAssemblyReferences (new [] { Mscorlib, System, SystemCore });
 				}
 				if (parsedDocument != null)
-					return singleFileContext.UpdateProjectContent (null, parsedDocument.ParsedFile);
+					return singleFileContext.AddOrUpdateFiles (parsedDocument.ParsedFile);
 				return singleFileContext;
 			}
 			
@@ -699,36 +845,49 @@ namespace MonoDevelop.Ide.Gui
 		}
 		
 		uint parseTimeout = 0;
-		void StartReparseThread ()
+		object reparseLock = new object();
+		internal void StartReparseThread ()
 		{
-			// Don't directly parse the document because doing it at every key press is
-			// very inefficient. Do it after a small delay instead, so several changes can
-			// be parsed at the same time.
-			string currentParseFile = FileName;
-			CancelParseTimeout ();
-			
-			parseTimeout = GLib.Timeout.Add (ParseDelay, delegate {
-				var editor = Editor;
-				if (editor == null)
-					return false;
-				string currentParseText = editor.Text;
-				string mimeType = editor.Document.MimeType;
-				ThreadPool.QueueUserWorkItem (delegate {
-					var currentParsedDocument = TypeSystemService.ParseFile (Project, currentParseFile, mimeType, currentParseText);
-					Application.Invoke (delegate {
-						// this may be called after the document has closed, in that case the OnDocumentParsed event shouldn't be invoked.
-						if (isClosed)
+			lock (reparseLock) {
+				if (currentWrapper != null)
+					currentWrapper.EnsureReferencesAreLoaded ();
+
+				// Don't directly parse the document because doing it at every key press is
+				// very inefficient. Do it after a small delay instead, so several changes can
+				// be parsed at the same time.
+				string currentParseFile = FileName;
+				if (string.IsNullOrEmpty (currentParseFile))
+					return;
+				CancelParseTimeout ();
+				if (IsProjectContextInUpdate) {
+					return;
+				}
+				parseTimeout = GLib.Timeout.Add (ParseDelay, delegate {
+					var editor = Editor;
+					if (editor == null || IsProjectContextInUpdate) {
+						parseTimeout = 0;
+						return false;
+					}
+					string currentParseText = editor.Text;
+					string mimeType = editor.Document.MimeType;
+					ThreadPool.QueueUserWorkItem (delegate {
+						if (IsProjectContextInUpdate) {
 							return;
-						this.parsedDocument = currentParsedDocument;
-//						this.parsedDocument = currentParsedDocument;
-//						if (this.parsedDocument != null && !this.parsedDocument.HasErrors)
-//							this.lastErrorFreeParsedDocument = parsedDocument;
-						OnDocumentParsed (EventArgs.Empty);
+						}
+						TypeSystemService.AddSkippedFile (currentParseFile);
+						var currentParsedDocument = TypeSystemService.ParseFile (Project, currentParseFile, mimeType, currentParseText);
+						Application.Invoke (delegate {
+							// this may be called after the document has closed, in that case the OnDocumentParsed event shouldn't be invoked.
+							if (isClosed)
+								return;
+							this.parsedDocument = currentParsedDocument;
+							OnDocumentParsed (EventArgs.Empty);
+						});
 					});
+					parseTimeout = 0;
+					return false;
 				});
-				parseTimeout = 0;
-				return false;
-			});
+			}
 		}
 		
 		/// <summary>
@@ -755,7 +914,7 @@ namespace MonoDevelop.Ide.Gui
 				window.ViewContent.Project = null;
 		}
 		
-		protected virtual void OnDocumentParsed (EventArgs e)
+		void OnDocumentParsed (EventArgs e)
 		{
 			EventHandler handler = this.DocumentParsed;
 			if (handler != null)
@@ -770,7 +929,10 @@ namespace MonoDevelop.Ide.Gui
 		
 		public string[] CommentTags {
 			get {
-				return GetCommentTags (FileName);
+				if (IsFile)
+					return GetCommentTags (FileName);
+				else
+					return null;
 			}
 		}
 		
@@ -810,6 +972,14 @@ namespace MonoDevelop.Ide.Gui
 //				Editor.Options.TabsToSpaces, Editor.Options.TabSize, Editor.EolMarker);
 //		}
 
+		/// <summary>
+		/// If the document shouldn't restore the settings after the load it can be disabled with this method.
+		/// That is useful when opening a document and programmatically scrolling to a specified location.
+		/// </summary>
+		public void DisableAutoScroll ()
+		{
+			Mono.TextEditor.Utils.FileSettingsStore.Remove (FileName);
+		}
 	}
 	
 	
