@@ -33,25 +33,24 @@ using System.Collections;
 using System.IO;
 using System.Text;
 
-using Mono.Addins;
 using MonoDevelop.Core;
 using MonoDevelop.Projects;
 using MonoDevelop.Ide.Templates;
-using MonoDevelop.Ide.Gui;
-using MonoDevelop.Ide.Gui.Dialogs;
 
-using MonoDevelop.Components;
-using IconView = MonoDevelop.Components.IconView;
 using Gtk;
 using System.Collections.Generic;
 using MonoDevelop.Ide.Gui.Components;
+using System.Linq;
+using System.Reflection;
+using MonoDevelop.Components;
+using Mono.Addins;
 
 namespace MonoDevelop.Ide.Projects {
 	/// <summary>
 	/// This class displays a new project dialog and sets up and creates a a new project,
 	/// the project types are described in an XML options file
 	/// </summary>
-	public partial class NewProjectDialog: Gtk.Dialog
+	internal partial class NewProjectDialog: Gtk.Dialog
 	{
 		ArrayList alltemplates = new ArrayList();
 		List<Category> categories = new List<Category> ();
@@ -69,7 +68,8 @@ namespace MonoDevelop.Ide.Projects {
 		CombineEntryFeatureSelector featureList;
 		IWorkspaceFileObject newItem;
 		Category recentCategory;
-		List<string> recentIds = new List<string> ();
+		List<string> recentTemplates = new List<string> ();
+		bool disposeNewItem = true;
 			
 		public NewProjectDialog (SolutionFolder parentFolder, bool openCombine, string basePath)
 		{
@@ -86,7 +86,7 @@ namespace MonoDevelop.Ide.Projects {
 			this.openSolution = openCombine;
 			TransientFor = IdeApp.Workbench.RootWindow;
 			Title = newSolution ? GettextCatalog.GetString ("New Solution") : GettextCatalog.GetString ("New Project");
-
+			
 			InitializeTemplates ();
 			
 			if (!newSolution) {
@@ -95,6 +95,9 @@ namespace MonoDevelop.Ide.Projects {
 				chk_combine_directory.Hide ();
 				lbl_subdirectory.Hide ();
 			}
+
+			TreeIter iter;
+			ExpandCategory ("C#", out iter);
 		}
 		
 		public void SelectTemplate (string id)
@@ -104,11 +107,13 @@ namespace MonoDevelop.Ide.Projects {
 			SelectTemplate (iter, id);
 		}
 		
-		ProjectTemplate GetTemplate (string id)
+		ProjectTemplate GetTemplate (string language, string id)
 		{
 			foreach (ProjectTemplate template in ProjectTemplate.ProjectTemplates) {
-				if (template.Id == id)
-					return template;
+				if (template.Id == id) {
+					if (language == null || template.LanguageName == language)
+						return template;
+				}
 			}
 			return null;
 		}
@@ -148,14 +153,16 @@ namespace MonoDevelop.Ide.Projects {
 			} while (catStore.IterNext (ref trial));
 			return false;
 		}
-		
-		void SelectCategory (string category)
+
+		bool ExpandCategory (string category, out TreeIter result)
 		{
 			string[] cats = category.Split ('/');
 			
 			TreeIter iter;
-			if (!catStore.GetIterFirst (out iter))
-				return;
+			if (!catStore.GetIterFirst (out iter)) {
+				result = TreeIter.Zero;
+				return false;
+			}
 			
 			TreeIter nextIter = iter;
 			for (int i = 0; i < cats.Length; i++) {
@@ -170,7 +177,15 @@ namespace MonoDevelop.Ide.Projects {
 			}
 			
 			lst_template_types.ExpandToPath (catStore.GetPath (iter));
-			lst_template_types.Selection.SelectIter (iter);
+			result = iter;
+			return true;
+		}
+		
+		void SelectCategory (string category)
+		{
+			TreeIter iter;
+			if (ExpandCategory (category, out iter))
+				lst_template_types.Selection.SelectIter (iter);
 		}
 		
 		void InitializeView()
@@ -199,6 +214,10 @@ namespace MonoDevelop.Ide.Projects {
 				cat_text_render.Destroy ();
 				cat_text_render = null;
 			}
+
+			if (disposeNewItem && newItem != null)
+				newItem.Dispose ();
+
 			base.OnDestroyed ();
 		}
 		
@@ -302,8 +321,9 @@ namespace MonoDevelop.Ide.Projects {
 		{
 			if (!btn_new.Sensitive)
 				return;
+			btn_new.Sensitive = false;
+			
 			if (notebook.Page == 0) {
-				
 				if (!CreateProject ())
 					return;
 				
@@ -334,6 +354,7 @@ namespace MonoDevelop.Ide.Projects {
 						}
 					}
 					notebook.Page++;
+					btn_new.Sensitive = true;
 					btn_new.Label = Gtk.Stock.Ok;
 					return;
 				}
@@ -364,9 +385,25 @@ namespace MonoDevelop.Ide.Projects {
 				IdeApp.ProjectOperations.Save (parentFolder.ParentSolution);
 			else
 				IdeApp.ProjectOperations.Save (newItem);
-			
-			if (openSolution)
-				selectedItem.OpenCreatedSolution();
+
+			if (openSolution) {
+				var op = selectedItem.OpenCreatedSolution ();
+				op.Completed += delegate {
+					if (op.Success) {
+						var sol = IdeApp.Workspace.GetAllSolutions ().FirstOrDefault ();
+						if (sol != null)
+							InstallProjectTemplatePackages (sol);
+					}
+				};
+			}
+			else {
+				// The item is not a solution being opened, so it is going to be added to
+				// an existing item. In this case, it must not be disposed by the dialog.
+				disposeNewItem = false;
+				if (parentFolder != null)
+					InstallProjectTemplatePackages (parentFolder.ParentSolution);
+			}
+
 			Respond (ResponseType.Ok);
 		}
 		
@@ -374,11 +411,18 @@ namespace MonoDevelop.Ide.Projects {
 		{
 			if (templateView.CurrentlySelected != null) {
 				PropertyService.Set ("Dialogs.NewProjectDialog.LastSelectedCategory",  ((ProjectTemplate)templateView.CurrentlySelected).Category);
-				recentIds.Remove (templateView.CurrentlySelected.Id);
-				recentIds.Insert (0, templateView.CurrentlySelected.Id);
-				if (recentIds.Count > 15)
-					recentIds.RemoveAt (recentIds.Count - 1);
-				string strRecent = string.Join (",", recentIds.ToArray ());
+				string template;
+				// keep the old format if the language is not specified
+				if (String.IsNullOrEmpty(templateView.CurrentlySelected.LanguageName)) {
+					template = templateView.CurrentlySelected.Id;
+				} else { // use the newer format with language before id
+					template = templateView.CurrentlySelected.LanguageName + "/" + templateView.CurrentlySelected.Id;
+				}
+				recentTemplates.Remove (template);
+				recentTemplates.Insert (0, template);
+				if (recentTemplates.Count > 15)
+					recentTemplates.RemoveAt (recentTemplates.Count - 1);
+				string strRecent = string.Join (",", recentTemplates.ToArray ());
 				PropertyService.Set ("Dialogs.NewProjectDialog.RecentTemplates", strRecent);
 				PropertyService.SaveProperties ();
 				//PropertyService.Set("Dialogs.NewProjectDialog.LargeImages", ((RadioButton)ControlDictionary["largeIconsRadioButton"]).Checked);
@@ -386,16 +430,18 @@ namespace MonoDevelop.Ide.Projects {
 			
 			string solution = txt_subdirectory.Text;
 			string name     = txt_name.Text;
-			string location = entry_location.Path;
+			string location = ProjectLocation;
 
 			if(solution.Equals("")) solution = name; //This was empty when adding after first combine
-			
+
 			if (
-				!FileService.IsValidPath (solution) || 
+				(CreateSolutionDirectory &&
+				!FileService.IsValidPath (solution)) || 
 			    !FileService.IsValidFileName(name) ||
+				name.IndexOf (' ') >= 0 ||
 				!FileService.IsValidPath(location))
 			{
-				MessageService.ShowError (GettextCatalog.GetString ("Illegal project name.\nOnly use letters, digits, space, '.' or '_'."));
+				MessageService.ShowError (GettextCatalog.GetString ("Illegal project name.\nOnly use letters, digits, '.' or '_'."));
 				return false;
 			}
 
@@ -410,26 +456,40 @@ namespace MonoDevelop.Ide.Projects {
 			
 			if (templateView.CurrentlySelected == null || name.Length == 0)
 				return false;
-				
+
 			ProjectTemplate item = (ProjectTemplate) templateView.CurrentlySelected;
 			
 			try {
-				System.IO.Directory.CreateDirectory (ProjectLocation);
+				if (Directory.Exists (ProjectLocation)) {
+					var question = GettextCatalog.GetString ("Directory {0} already exists.\nDo you want to continue creating the project?", ProjectLocation);
+					var btn = MessageService.AskQuestion (question, AlertButton.No, AlertButton.Yes);
+					if (btn != AlertButton.Yes)
+						return false;
+				}
+
+				Directory.CreateDirectory (location);
 			} catch (IOException) {
-				MessageService.ShowError (GettextCatalog.GetString ("Could not create directory {0}. File already exists.", ProjectLocation));
+				MessageService.ShowError (GettextCatalog.GetString ("Could not create directory {0}. File already exists.", location));
 				return false;
 			} catch (UnauthorizedAccessException) {
-				MessageService.ShowError (GettextCatalog.GetString ("You do not have permission to create to {0}", ProjectLocation));
+				MessageService.ShowError (GettextCatalog.GetString ("You do not have permission to create to {0}", location));
 				return false;
 			}
-			
-			ProjectCreateInformation cinfo = CreateProjectCreateInformation ();
+
+			if (newItem != null) {
+				newItem.Dispose ();
+				newItem = null;
+			}
 			
 			try {
+				ProjectCreateInformation cinfo = CreateProjectCreateInformation ();
 				if (newSolution)
 					newItem = item.CreateWorkspaceItem (cinfo);
 				else
 					newItem = item.CreateProject (parentFolder, cinfo);
+			} catch (UserException ex) {
+				MessageService.ShowError (ex.Message, ex.Details);
+				return false;
 			} catch (Exception ex) {
 				MessageService.ShowException (ex, GettextCatalog.GetString ("The project could not be created"));
 				return false;
@@ -441,8 +501,8 @@ namespace MonoDevelop.Ide.Projects {
 		ProjectCreateInformation CreateProjectCreateInformation ()
 		{
 			ProjectCreateInformation cinfo = new ProjectCreateInformation ();
-			cinfo.SolutionPath = SolutionLocation;
-			cinfo.ProjectBasePath = ProjectLocation;
+			cinfo.SolutionPath = FileService.ResolveFullPath (SolutionLocation);
+			cinfo.ProjectBasePath = FileService.ResolveFullPath (ProjectLocation);
 			cinfo.ProjectName = txt_name.Text;
 			cinfo.SolutionName = CreateSolutionDirectory ? txt_subdirectory.Text : txt_name.Text;
 			cinfo.ParentFolder = parentFolder;
@@ -450,38 +510,72 @@ namespace MonoDevelop.Ide.Projects {
 			return cinfo;
 		}
 
+		void InstallProjectTemplatePackages (Solution sol)
+		{
+			if (!selectedItem.HasPackages ())
+				return;
+
+			foreach (ProjectTemplatePackageInstaller installer in AddinManager.GetExtensionObjects ("/MonoDevelop/Ide/ProjectTemplatePackageInstallers")) {
+				installer.Run (sol, selectedItem.PackageReferencesForCreatedProjects);
+			}
+		}
+
 		// icon view event handlers
 		void SelectedIndexChange(object sender, EventArgs e)
 		{
-			if (templateView.CurrentlySelected != null) {
-				ProjectTemplate ptemplate = (ProjectTemplate) templateView.CurrentlySelected;
-				lbl_template_descr.Text = StringParserService.Parse (ptemplate.Description);
-				labelTemplateTitle.Markup = "<b>" + GLib.Markup.EscapeText (ptemplate.Name) + "</b>";
+			try {
+				btn_new.Sensitive = true;
+				txt_name.Sensitive = true;
+				txt_subdirectory.Sensitive = true;
+				chk_combine_directory.Sensitive = true;
+				entry_location.Sensitive = true;
 				
-				if (ptemplate.SolutionDescriptor.EntryDescriptors.Length == 0) {
-					txt_subdirectory.Sensitive = false;
-					chk_combine_directory.Sensitive = false;
-					lbl_subdirectory.Sensitive = false;
-					btn_new.Label = Gtk.Stock.Ok;
-				} else {
-					txt_subdirectory.Sensitive = true;
-					chk_combine_directory.Sensitive = true;
-					lbl_subdirectory.Sensitive = true;
-					txt_subdirectory.Text = txt_name.Text;
+				if (templateView.CurrentlySelected != null) {
+					ProjectTemplate ptemplate = (ProjectTemplate) templateView.CurrentlySelected;
+					lbl_template_descr.Text = StringParserService.Parse (ptemplate.Description);
+					labelTemplateTitle.Markup = "<b>" + GLib.Markup.EscapeText (ptemplate.Name) + "</b>";
 					
-					ProjectCreateInformation cinfo = CreateProjectCreateInformation ();
-					if (ptemplate.HasItemFeatures (parentFolder, cinfo))
-						btn_new.Label = Gtk.Stock.GoForward;
-					else
+					if (ptemplate.SolutionDescriptor.EntryDescriptors.Length == 0) {
+						txt_subdirectory.Sensitive = false;
+						chk_combine_directory.Sensitive = false;
+						lbl_subdirectory.Sensitive = false;
 						btn_new.Label = Gtk.Stock.Ok;
+					} else {
+						lbl_subdirectory.Sensitive = true;
+						txt_subdirectory.Text = txt_name.Text;
+						
+						ProjectCreateInformation cinfo = CreateProjectCreateInformation ();
+						if (ptemplate.HasItemFeatures (parentFolder, cinfo))
+							btn_new.Label = Gtk.Stock.GoForward;
+						else
+							btn_new.Label = Gtk.Stock.Ok;
+					}
 				}
+				else {
+					lbl_template_descr.Text = String.Empty;
+					labelTemplateTitle.Text = "";
+				}
+				
+				PathChanged (null, null);
+				
+				btn_new.GrabDefault ();
+			} catch (Exception ex) {
+				txt_name.Sensitive = false;
+				btn_new.Sensitive = false;
+				txt_subdirectory.Sensitive = false;
+				chk_combine_directory.Sensitive = false;
+				entry_location.Sensitive = false;
+				
+				while (ex is TargetInvocationException)
+						ex = ((TargetInvocationException) ex).InnerException;
+				
+				if (ex is UserException) {
+					var user = (UserException) ex;
+					MessageService.ShowError (user.Message, user.Details);
+				} else {
+					MessageService.ShowException (ex);
+				};
 			}
-			else {
-				lbl_template_descr.Text = String.Empty;
-				labelTemplateTitle.Text = "";
-			}
-			
-			PathChanged (null, null);
 		}
 		
 		protected void cancelClicked (object o, EventArgs e)
@@ -491,7 +585,7 @@ namespace MonoDevelop.Ide.Projects {
 		
 		void ActivateIfReady ()
 		{
-			if (templateView.CurrentlySelected == null || txt_name.Text.Trim () == "" || (txt_subdirectory.Sensitive && chk_combine_directory.Active && txt_subdirectory.Text.Trim ().Length == 0))
+			if (templateView.CurrentlySelected == null || !txt_name.Sensitive || txt_name.Text.Trim () == "" || (txt_subdirectory.Sensitive && chk_combine_directory.Active && txt_subdirectory.Text.Trim ().Length == 0))
 				btn_new.Sensitive = false;
 			else
 				btn_new.Sensitive = true;
@@ -524,7 +618,7 @@ namespace MonoDevelop.Ide.Projects {
 			if (basePath == null)
 				basePath = IdeApp.ProjectOperations.ProjectsDefaultPath;
 				
-			entry_location.Path = basePath;
+			entry_location.Path = FileService.ResolveFullPath (basePath);
 			
 			PathChanged (null, null);
 			
@@ -558,6 +652,10 @@ namespace MonoDevelop.Ide.Projects {
 					return template;
 				}
 			}
+			
+			public bool DisplayCategory {
+				get; set;
+			}
 		}
 
 		private void InitializeTemplates ()
@@ -576,12 +674,24 @@ namespace MonoDevelop.Ide.Projects {
 			
 			recentCategory = new Category (GettextCatalog.GetString ("Recent"));
 			string strRecent = PropertyService.Get<string> ("Dialogs.NewProjectDialog.RecentTemplates", "");
-			recentIds = new List<string> (strRecent.Split (new char[] {','}, StringSplitOptions.RemoveEmptyEntries));
 
-			foreach (string id in recentIds) {
-				ProjectTemplate pt = GetTemplate (id);
+			recentTemplates = new List<string> ();
+			foreach (string template in strRecent.Split (new char[] {','}, StringSplitOptions.RemoveEmptyEntries)) {
+				// Some templates can't be identified using only id, sometimes language is needed too
+				string id, language = null;
+				// If the string was serialized with old format leave null value for the language variable
+				int separator = template.IndexOf ('/');
+				if (separator == -1) {
+					id = template;
+				} else { // In other case extract the language
+					language = template.Substring (0, separator);
+					id = template.Substring (separator + 1);
+				}
+				recentTemplates.Add (template);
+
+				ProjectTemplate pt = GetTemplate (language, id);
 				if (pt != null)
-					recentCategory.Templates.Add (new TemplateItem (pt));
+					recentCategory.Templates.Add (new TemplateItem (pt) { DisplayCategory = true });
 			}
 			
 			InitializeComponents ();
@@ -713,8 +823,8 @@ namespace MonoDevelop.Ide.Projects {
 				Model = templateStore;
 				
 				TreeViewColumn col = new TreeViewColumn ();
-				CellRendererIcon crp = new CellRendererIcon ();
-				crp.StockSize = (uint) Gtk.IconSize.Dnd;
+				CellRendererImage crp = new CellRendererImage ();
+				crp.StockSize = Gtk.IconSize.Dnd;
 				crp.Ypad = 2;
 				col.PackStart (crp, false);
 				col.AddAttribute (crp, "stock-id", 0);
@@ -751,8 +861,14 @@ namespace MonoDevelop.Ide.Projects {
 			public void Add (TemplateItem templateItem)
 			{
 				string name = GLib.Markup.EscapeText (templateItem.Name);
-				if (!string.IsNullOrEmpty (templateItem.Template.LanguageName))
-					name += "\n<span foreground='darkgrey'><small>" + templateItem.Template.LanguageName + "</small></span>";
+				string desc = null;
+				if (templateItem.DisplayCategory)
+					desc = templateItem.Template.Category;
+				else if (!string.IsNullOrEmpty (templateItem.Template.LanguageName))
+					desc = templateItem.Template.LanguageName;
+				
+				if (desc != null)
+					name += "\n<span foreground='darkgrey'><small>" + desc + "</small></span>";
 				templateStore.AppendValues (templateItem.Template.Icon.IsNull ? "md-project" : templateItem.Template.Icon.ToString (), name, templateItem.Template);
 			}
 			
