@@ -145,6 +145,18 @@ namespace MonoDevelop.Ide.Tasks
 			// Initialize with existing tags.
 			foreach (Task t in comments)
 				AddGeneratedTask (t);
+
+			view.Destroyed += delegate {
+				view.RowActivated -= OnRowActivated;
+				TaskService.CommentTasksChanged -= OnCommentTasksChanged;
+				CommentTag.SpecialCommentTagsChanged -= OnCommentTagsChanged;
+				IdeApp.Workspace.WorkspaceItemLoaded -= OnWorkspaceItemLoaded;
+				IdeApp.Workspace.WorkspaceItemUnloaded -= OnWorkspaceItemUnloaded;
+				comments.TasksAdded -= DispatchService.GuiDispatch<TaskEventHandler> (GeneratedTaskAdded);
+				comments.TasksRemoved -= DispatchService.GuiDispatch<TaskEventHandler> (GeneratedTaskRemoved);
+
+				PropertyService.PropertyChanged -= DispatchService.GuiDispatch<EventHandler<PropertyChangedEventArgs>> (OnPropertyUpdated);
+			};
 		}
 
 		void LoadColumnsVisibility ()
@@ -188,20 +200,40 @@ namespace MonoDevelop.Ide.Tasks
 			foreach (var sln in wob.GetAllSolutions ())
 				LoadSolutionContents (sln);
 		}
+
+		void UpdateCommentTagsForProject (Solution solution, Project project)
+		{
+			var ctx = TypeSystemService.GetProjectContentWrapper (project);
+			if (ctx == null)
+				return;
+			var tags = ctx.GetExtensionObject<ProjectCommentTags> ();
+			if (tags == null) {
+				tags = new ProjectCommentTags ();
+				ctx.UpdateExtensionObject (tags);
+				tags.Update (ctx.Project);
+			} else {
+				foreach (var kv in tags.Tags) {
+					UpdateCommentTags (solution, kv.Key, kv.Value);
+				}
+			}
+		}
 		
 		void LoadSolutionContents (Solution sln)
 		{
 			loadedSlns.Add (sln);
-			
-			// Load all tags that are stored in pidb files
-			foreach (Project p in sln.GetAllProjects ()) {
-				var pContext = TypeSystemService.GetProjectContext (p);
-				if (pContext == null)
-					continue;
-				foreach (ProjectFile file in p.Files) {
-					UpdateCommentTags (sln, file.Name, GetSpecialComments (pContext, file.Name));
+			System.Threading.ThreadPool.QueueUserWorkItem (delegate {
+				sln.SolutionItemAdded += delegate(object sender, SolutionItemChangeEventArgs e) {
+					var newProject = e.SolutionItem as Project;
+					if (newProject == null)
+						return;
+					UpdateCommentTagsForProject (sln, newProject);
+				};
+
+				// Load all tags that are stored in pidb files
+				foreach (Project p in sln.GetAllProjects ()) {
+					UpdateCommentTagsForProject (sln, p);
 				}
-			}
+			});
 		}
 		
 		
@@ -225,8 +257,11 @@ namespace MonoDevelop.Ide.Tasks
 			//because of parse queueing, it's possible for this event to come in after the solution is closed
 			//so we track which solutions are currently open so that we don't leak memory by holding 
 			// on to references to closed projects
-			if (e.Project != null && e.Project.ParentSolution != null && loadedSlns.Contains (e.Project.ParentSolution))
-				UpdateCommentTags (e.Project.ParentSolution, e.FileName, e.TagComments);
+			if (e.Project != null && e.Project.ParentSolution != null && loadedSlns.Contains (e.Project.ParentSolution)) {
+				Application.Invoke (delegate {
+					UpdateCommentTags (e.Project.ParentSolution, e.FileName, e.TagComments);
+				});
+			}
 		}
 		
 		void OnCommentTagsChanged (object sender, EventArgs e)
@@ -244,22 +279,26 @@ namespace MonoDevelop.Ide.Tasks
 			List<Task> newTasks = new List<Task> ();
 			if (tagComments != null) {  
 				foreach (Tag tag in tagComments) {
-					if (!priorities.ContainsKey (tag.Key))
-						continue;
-					
-					//prepend the tag if it's not already there
+					TaskPriority priority;
 					string desc = tag.Text.Trim ();
-					if (!desc.StartsWith (tag.Key)) {
-						if (desc.StartsWith (":"))
-							desc = tag.Key + desc;
-						else if (tag.Key.EndsWith (":"))
-							desc = tag.Key + " " + desc;
-						else
-							desc = tag.Key + ": " + desc;
+
+					if (!priorities.TryGetValue (tag.Key, out priority)) {
+						if (!Enum.TryParse (tag.Key, out priority))
+							priority = TaskPriority.High;
+					} else {
+						//prepend the tag if it's not already there
+						if (!desc.StartsWith (tag.Key, StringComparison.Ordinal)) {
+							if (desc.StartsWith (":", StringComparison.Ordinal))
+								desc = tag.Key + desc;
+							else if (tag.Key.EndsWith (":", StringComparison.Ordinal))
+								desc = tag.Key + " " + desc;
+							else
+								desc = tag.Key + ": " + desc;
+						}
 					}
 					
 					Task t = new Task (fileName, desc, tag.Region.BeginColumn, tag.Region.BeginLine,
-					                   TaskSeverity.Information, priorities[tag.Key], wob);
+					                   TaskSeverity.Information, priority, wob);
 					newTasks.Add (t);
 				}
 			}
@@ -469,15 +508,18 @@ namespace MonoDevelop.Ide.Tasks
 				if (doc != null && doc.HasProject && doc.Project is DotNetProject) {
 					string[] commentTags = doc.CommentTags;
 					if (commentTags != null && commentTags.Length == 1) {
-						string line = doc.Editor.GetLineText (task.Line);
-						int index = line.IndexOf (commentTags[0]);
-						if (index != -1) {
-							doc.Editor.SetCaretTo (task.Line, task.Column);
-							line = line.Substring (0, index);
-							var ls = doc.Editor.Document.GetLine (task.Line);
-							doc.Editor.Replace (ls.Offset, ls.Length, line);
-							comments.Remove (task);
-						}
+						doc.DisableAutoScroll ();
+						doc.RunWhenLoaded (() => {
+							string line = doc.Editor.GetLineText (task.Line);
+							int index = line.IndexOf (commentTags[0]);
+							if (index != -1) {
+								doc.Editor.SetCaretTo (task.Line, task.Column);
+								line = line.Substring (0, index);
+								var ls = doc.Editor.Document.GetLine (task.Line);
+								doc.Editor.Replace (ls.Offset, ls.Length, line);
+								comments.Remove (task);
+							}
+						}); 
 					}
 				}
 			}
