@@ -41,6 +41,7 @@ using System.Collections.Generic;
 using System.Collections;
 using MonoDevelop.Ide.Codons;
 using System.Xml;
+using System.Linq;
 
 namespace MonoDevelop.Ide.Templates
 {
@@ -358,7 +359,6 @@ namespace MonoDevelop.Ide.Templates
 					return DisplayBindingService.GetDefaultViewBinding (null, mimeType, null) != null;
 				}
 			}
-			return true;
 		}
 
 		protected virtual bool CreateFile (FileDescriptionTemplate newfile, SolutionItem policyParent, Project project, string directory, string language, string name)
@@ -376,7 +376,7 @@ namespace MonoDevelop.Ide.Templates
                 if (directory != null) {
                     string fileName = singleFile.SaveFile (policyParent, project, language, directory, name);
                     if (fileName != null) {
-                        IdeApp.Workbench.OpenDocument (fileName);
+						IdeApp.Workbench.OpenDocument (fileName, project);
                         return true;
                     }
 				} else {
@@ -407,7 +407,7 @@ namespace MonoDevelop.Ide.Templates
 
             //filter on conditions
             if (project != null) {
-				if (!string.IsNullOrEmpty(projecttype) && (projecttype != project.ProjectType))
+				if (!string.IsNullOrEmpty (projecttype) && project.GetProjectTypes ().All (p => p != projecttype))
                     return false;
 
                 foreach (FileTemplateCondition condition in conditions)
@@ -488,7 +488,7 @@ namespace MonoDevelop.Ide.Templates
         {
             //Template can match all CodeDom .NET languages with a "*"
             if (list.Contains ("*")) {
-                foreach (ILanguageBinding lb in LanguageBindingService.LanguageBindings) {
+                foreach (var lb in LanguageBindingService.LanguageBindings) {
                     IDotNetLanguageBinding dnlang = lb as IDotNetLanguageBinding;
                     if (dnlang != null && dnlang.GetCodeDomProvider () != null)
                         list.Add (dnlang.Language);
