@@ -39,6 +39,8 @@ using MonoDevelop.Ide.Gui;
 using MonoDevelop.Ide.StandardHeader;
 using System.Text;
 using MonoDevelop.Ide.Gui.Content;
+using MonoDevelop.Ide.CodeFormatting;
+using MonoDevelop.Projects.SharedAssetsProjects;
 
 namespace MonoDevelop.Ide.Templates
 {
@@ -55,7 +57,7 @@ namespace MonoDevelop.Ide.Templates
 		string customTool;
 		List<string> references = new List<string> ();
 		
-		public override void Load (XmlElement filenode)
+		public override void Load (XmlElement filenode, FilePath baseDirectory)
 		{
 			name = filenode.GetAttribute ("name");
 			defaultName = filenode.GetAttribute ("DefaultName");
@@ -132,7 +134,7 @@ namespace MonoDevelop.Ide.Templates
 						string res = netProject.AssemblyContext.GetAssemblyFullName (aref, netProject.TargetFramework);
 						res = netProject.AssemblyContext.GetAssemblyNameForVersion (res, netProject.TargetFramework);
 						if (!ContainsReference (netProject, res))
-							netProject.References.Add (new ProjectReference (ReferenceType.Gac, aref));
+							netProject.References.Add (new ProjectReference (ReferenceType.Package, aref));
 					}
 				}
 				
@@ -141,8 +143,28 @@ namespace MonoDevelop.Ide.Templates
 				return null;
 		}
 		
+		public override bool SupportsProject (Project project, string projectPath)
+		{
+			DotNetProject netProject = project as DotNetProject;
+			if (netProject != null) {
+				// Ensure that the references are valid inside the project's target framework.
+				foreach (string aref in references) {
+					string res = netProject.AssemblyContext.GetAssemblyFullName (aref, netProject.TargetFramework);
+					if (string.IsNullOrEmpty (res))
+						return false;
+					res = netProject.AssemblyContext.GetAssemblyNameForVersion (res, netProject.TargetFramework);
+					if (string.IsNullOrEmpty (res))
+						return false;
+				}
+			}
+			
+			return true;
+		}
+		
 		bool ContainsReference (DotNetProject project, string aref)
 		{
+			if (string.IsNullOrEmpty (aref))
+				return false;
 			string aname;
 			int i = aref.IndexOf (',');
 			if (i == -1)
@@ -150,8 +172,8 @@ namespace MonoDevelop.Ide.Templates
 			else
 				aname = aref.Substring (0, i);
 			foreach (ProjectReference pr in project.References) {
-				if (pr.ReferenceType == ReferenceType.Gac && (pr.Reference == aname || pr.Reference.StartsWith (aname + ",")) || 
-					pr.ReferenceType != ReferenceType.Gac && pr.Reference.Contains (aname))
+				if (pr.ReferenceType == ReferenceType.Package && (pr.Reference == aname || pr.Reference.StartsWith (aname + ",")) || 
+					pr.ReferenceType != ReferenceType.Package && pr.Reference.Contains (aname))
 					return true;
 			}
 			return false;
@@ -222,7 +244,7 @@ namespace MonoDevelop.Ide.Templates
 					fileName = fileName + defaultExtension;
 				}
 				else if (!string.IsNullOrEmpty  (language)) {
-					ILanguageBinding languageBinding = GetLanguageBinding (language);
+					var languageBinding = GetLanguageBinding (language);
 					fileName = languageBinding.GetFileName (fileName);
 				} 
 			}
@@ -243,12 +265,19 @@ namespace MonoDevelop.Ide.Templates
 			string content = CreateContent (project, tags, language);
 			content = StringParserService.Parse (content, tags);
 			string mime = DesktopService.GetMimeTypeForUri (fileName);
-			Formatter formatter = !String.IsNullOrEmpty (mime) ? TextFileService.GetFormatter (mime) : null;
+			CodeFormatter formatter = !string.IsNullOrEmpty (mime) ? CodeFormatterService.GetFormatter (mime) : null;
 			
-			if (formatter != null)
-				content = formatter.FormatText (policyParent != null ? policyParent.Policies : null, content);
+			if (formatter != null) {
+				var formatted = formatter.FormatText (policyParent != null ? policyParent.Policies : null, content);
+				if (formatted != null)
+					content = formatted;
+			}
 			
-			MemoryStream ms = new MemoryStream ();
+			var ms = new MemoryStream ();
+
+			var bom = Encoding.UTF8.GetPreamble ();
+			ms.Write (bom, 0, bom.Length);
+
 			byte[] data;
 			if (AddStandardHeader) {
 				string header = StandardHeaderService.GetHeader (policyParent, fileName, true);
@@ -256,14 +285,21 @@ namespace MonoDevelop.Ide.Templates
 				ms.Write (data, 0, data.Length);
 			}
 			
-			Mono.TextEditor.Document doc = new Mono.TextEditor.Document ();
+			Mono.TextEditor.TextDocument doc = new Mono.TextEditor.TextDocument ();
 			doc.Text = content;
 			
-			TextStylePolicy textPolicy = policyParent != null ? policyParent.Policies.Get<TextStylePolicy> ("text/plain") : MonoDevelop.Projects.Policies.PolicyService.GetDefaultPolicy<TextStylePolicy> ("text/plain");
+			TextStylePolicy textPolicy = policyParent != null ? policyParent.Policies.Get<TextStylePolicy> ("text/plain")
+				: MonoDevelop.Projects.Policies.PolicyService.GetDefaultPolicy<TextStylePolicy> ("text/plain");
 			string eolMarker = TextStylePolicy.GetEolMarker (textPolicy.EolMarker);
 			byte[] eolMarkerBytes = System.Text.Encoding.UTF8.GetBytes (eolMarker);
-			foreach (Mono.TextEditor.LineSegment line in doc.Lines) {
-				data = System.Text.Encoding.UTF8.GetBytes (doc.GetTextAt (line.Offset, line.EditableLength));
+			
+			var tabToSpaces = textPolicy.TabsToSpaces? new string (' ', textPolicy.TabWidth) : null;
+			
+			foreach (Mono.TextEditor.DocumentLine line in doc.Lines) {
+				var lineText = doc.GetTextAt (line.Offset, line.Length);
+				if (tabToSpaces != null)
+					lineText = lineText.Replace ("\t", tabToSpaces);
+				data = System.Text.Encoding.UTF8.GetBytes (lineText);
 				ms.Write (data, 0, data.Length);
 				ms.Write (eolMarkerBytes, 0, eolMarkerBytes.Length);
 			}
@@ -290,7 +326,8 @@ namespace MonoDevelop.Ide.Templates
 		// We supply defaults whenever it is possible, to avoid having unsubstituted tags. However,
 		// do not substitute blanks when a sensible default cannot be guessed, because they result
 		//in less obvious errors.
-		public virtual void ModifyTags (SolutionItem policyParent, Project project, string language, string identifier, string fileName, ref Dictionary<string,string> tags)
+		public virtual void ModifyTags (SolutionItem policyParent, Project project, string language,
+			string identifier, string fileName, ref Dictionary<string,string> tags)
 		{
 			DotNetProject netProject = project as DotNetProject;
 			string languageExtension = "";
@@ -302,7 +339,11 @@ namespace MonoDevelop.Ide.Templates
 			}
 			
 			//need a default namespace or if there is no project, substitutions can get very messed up
-			string ns = netProject != null ? netProject.GetDefaultNamespace (fileName) : "Application";
+			string ns;
+			if (project is IDotNetFileContainer)
+				ns = ((IDotNetFileContainer)project).GetDefaultNamespace (fileName);
+			else
+				ns = "Application";
 			
 			//need an 'identifier' for tag substitution, e.g. class name or page name
 			//if not given an identifier, use fileName
@@ -318,7 +359,6 @@ namespace MonoDevelop.Ide.Templates
 				tags ["FullName"] = ns.Length > 0 ? ns + "." + identifier : identifier;
 				
 				//some .NET languages may be able to use keywords as identifiers if they're escaped
-				//for simplicity, we escape the identifier in "Name" and provide "UnescapedName"
 				IDotNetLanguageBinding dnb = binding as IDotNetLanguageBinding;
 				if (dnb != null) {
 					System.CodeDom.Compiler.CodeDomProvider provider = dnb.GetCodeDomProvider ();
@@ -329,8 +369,18 @@ namespace MonoDevelop.Ide.Templates
 			}
 			
 			tags ["Namespace"] = ns;
-			if (project != null)
+			if (policyParent != null)
+				tags ["SolutionName"] = policyParent.Name;
+			if (project != null) {
 				tags ["ProjectName"] = project.Name;
+				tags ["SafeProjectName"] = CreateIdentifierName (project.Name);
+				var info = project.AuthorInformation ?? AuthorInformation.Default;
+				tags ["AuthorCopyright"] = info.Copyright;
+				tags ["AuthorCompany"] = info.Company;
+				tags ["AuthorTrademark"] = info.Trademark;
+				tags ["AuthorEmail"] = info.Email;
+				tags ["AuthorName"] = info.Name;
+			}
 			if ((language != null) && (language.Length > 0))
 				tags ["Language"] = language;
 			if (languageExtension.Length > 0)
@@ -366,7 +416,7 @@ namespace MonoDevelop.Ide.Templates
 		
 		protected ILanguageBinding GetLanguageBinding (string language)
 		{
-			ILanguageBinding binding = LanguageBindingService.GetBindingPerLanguageName (language);
+			var binding = LanguageBindingService.GetBindingPerLanguageName (language);
 			if (binding == null)
 				throw new InvalidOperationException ("Language '" + language + "' not found");
 			return binding;
