@@ -46,12 +46,11 @@ using MonoDevelop.Components.Commands;
 using MonoDevelop.Core;
 using MonoDevelop.Ide.Gui.Dialogs;
 using MonoDevelop.Ide.Gui;
-using MonoDevelop.Core.Execution;
 using MonoDevelop.Core.Instrumentation;
 using System.Diagnostics;
-using MonoDevelop.Projects;
 using System.Collections.Generic;
-using MonoDevelop.Core.LogReporting;
+using MonoDevelop.Ide.Extensions;
+using MonoDevelop.Components.Extensions;
 
 namespace MonoDevelop.Ide
 {
@@ -73,27 +72,48 @@ namespace MonoDevelop.Ide
 		
 		int Run (MonoDevelopOptions options)
 		{
+			LoggingService.LogInfo ("Starting {0} {1}", BrandingService.ApplicationName, IdeVersionInfo.MonoDevelopVersion);
+			LoggingService.LogInfo ("Running on {0}", IdeVersionInfo.GetRuntimeInfo ());
+
+			IdeApp.Customizer = options.IdeCustomizer ?? new IdeCustomizer ();
+			IdeApp.Customizer.Initialize ();
+
 			Counters.Initialization.BeginTiming ();
-			
-			if (options.LogCounters) {
-				string logFile = Path.Combine (Environment.CurrentDirectory, "monodevelop.clog");
+
+			if (options.PerfLog) {
+				string logFile = Path.Combine (Environment.CurrentDirectory, "monodevelop.perf-log");
 				LoggingService.LogInfo ("Logging instrumentation service data to file: " + logFile);
 				InstrumentationService.StartAutoSave (logFile, 1000);
 			}
+
+			//ensure native libs initialized before we hit anything that p/invokes
+			Platform.Initialize ();
 			
 			Counters.Initialization.Trace ("Initializing GTK");
+			if (Platform.IsWindows && !CheckWindowsGtk ())
+				return 1;
 			SetupExceptionManager ();
 			
 			try {
-				MonoDevelop.Ide.Gui.GLibLogging.Enabled = true;
+				GLibLogging.Enabled = true;
 			} catch (Exception ex) {
 				LoggingService.LogError ("Error initialising GLib logging.", ex);
 			}
-			
-			//OSXFIXME
+
+			SetupTheme ();
+
 			var args = options.RemainingArgs.ToArray ();
-			Gtk.Application.Init ("monodevelop", ref args);
-			
+			Gtk.Application.Init (BrandingService.ApplicationName, ref args);
+
+			LoggingService.LogInfo ("Using GTK+ {0}", IdeVersionInfo.GetGtkVersion ());
+
+			// XWT initialization
+			FilePath p = typeof(IdeStartup).Assembly.Location;
+			Assembly.LoadFrom (p.ParentDirectory.Combine ("Xwt.Gtk.dll"));
+			Xwt.Application.InitializeAsGuest (Xwt.ToolkitType.Gtk);
+			Xwt.Toolkit.CurrentEngine.RegisterBackend<IExtendedTitleBarWindowBackend,GtkExtendedTitleBarWindowBackend> ();
+			Xwt.Toolkit.CurrentEngine.RegisterBackend<IExtendedTitleBarDialogBackend,GtkExtendedTitleBarDialogBackend> ();
+
 			//default to Windows IME on Windows
 			if (Platform.IsWindows && Mono.TextEditor.GtkWorkarounds.GtkMinorVersion >= 16) {
 				var settings = Gtk.Settings.Default;
@@ -107,9 +127,10 @@ namespace MonoDevelop.Ide
 			EndPoint ep = null;
 			
 			DispatchService.Initialize ();
-			
+
 			// Set a synchronization context for the main gtk thread
 			SynchronizationContext.SetSynchronizationContext (new GtkSynchronizationContext ());
+			Runtime.MainSynchronizationContext = SynchronizationContext.Current;
 			
 			AddinManager.AddinLoadError += OnAddinError;
 			
@@ -128,29 +149,25 @@ namespace MonoDevelop.Ide
 				}
 			}
 			
-			DefaultTheme = Gtk.Settings.Default.ThemeName;
-			if (!string.IsNullOrEmpty (IdeApp.Preferences.UserInterfaceTheme))
-				Gtk.Settings.Default.ThemeName = IdeApp.Preferences.UserInterfaceTheme;
-			
-			//don't show the splash screen on the Mac, so instead we get the expected "Dock bounce" effect
-			//this also enables the Mac platform service to subscribe to open document events before the GUI loop starts.
-			if (Platform.IsMac)
-				options.NoLogo = true;
-			
-			IProgressMonitor monitor;
-			
-			if (options.NoLogo) {
-				monitor = new MonoDevelop.Core.ProgressMonitoring.ConsoleProgressMonitor ();
-			} else {
-				monitor = SplashScreenForm.SplashScreen;
-				SplashScreenForm.SplashScreen.ShowAll ();
-			}
-			
 			Counters.Initialization.Trace ("Initializing Runtime");
-			monitor.BeginTask (GettextCatalog.GetString ("Starting {0}", BrandingService.ApplicationName), 3);
-			monitor.Step (1);
 			Runtime.Initialize (true);
+
+			IdeApp.Customizer.OnCoreInitialized ();
+
+			Counters.Initialization.Trace ("Initializing theme");
+
+			DefaultTheme = Gtk.Settings.Default.ThemeName;
+			string theme = IdeApp.Preferences.UserInterfaceTheme;
+			if (string.IsNullOrEmpty (theme))
+				theme = DefaultTheme;
+			ValidateGtkTheme (ref theme);
+			if (theme != DefaultTheme)
+				Gtk.Settings.Default.ThemeName = theme;
 			
+			IProgressMonitor monitor = new MonoDevelop.Core.ProgressMonitoring.ConsoleProgressMonitor ();
+			
+			monitor.BeginTask (GettextCatalog.GetString ("Starting {0}", BrandingService.ApplicationName), 2);
+
 			//make sure that the platform service is initialised so that the Mac platform can subscribe to open-document events
 			Counters.Initialization.Trace ("Initializing Platform Service");
 			DesktopService.Initialize ();
@@ -190,35 +207,27 @@ namespace MonoDevelop.Ide
 				version += "." + Assembly.GetEntryAssembly ().GetName ().Version.Build;
 			if (Assembly.GetEntryAssembly ().GetName ().Version.Revision != 0)
 				version += "." + Assembly.GetEntryAssembly ().GetName ().Version.Revision;
-			
-			// System checks
-			if (!CheckBug77135 ())
-				return 1;
-			
-			if (!CheckQtCurve ())
-				return 1;
 
 			CheckFileWatcher ();
 			
 			Exception error = null;
 			int reportedFailures = 0;
-			
+
 			try {
 				Counters.Initialization.Trace ("Loading Icons");
 				//force initialisation before the workbench so that it can register stock icons for GTK before they get requested
 				ImageService.Initialize ();
 				
 				if (errorsList.Count > 0) {
-					if (monitor is SplashScreenForm)
-						SplashScreenForm.SplashScreen.Hide ();
 					AddinLoadErrorDialog dlg = new AddinLoadErrorDialog ((AddinError[]) errorsList.ToArray (typeof(AddinError)), false);
 					if (!dlg.Run ())
 						return 1;
-					if (monitor is SplashScreenForm)
-						SplashScreenForm.SplashScreen.Show ();
 					reportedFailures = errorsList.Count;
 				}
-				
+
+				if (!CheckSCPlugin ())
+					return 1;
+
 				// no alternative for Application.ThreadException?
 				// Application.ThreadException += new ThreadExceptionEventHandler(ShowErrorBox);
 
@@ -239,8 +248,11 @@ namespace MonoDevelop.Ide
 			
 			if (error != null) {
 				LoggingService.LogFatalError (null, error);
-				MessageService.ShowException (error,
-				                              GettextCatalog.GetString ("MonoDevelop failed to start. The following error has been reported: ") + error.Message);
+				string message = BrandingService.BrandApplicationName (GettextCatalog.GetString (
+					"MonoDevelop failed to start. The following error has been reported: {0}",
+					error.Message
+				));
+				MessageService.ShowException (error, message);
 				return 1;
 			}
 
@@ -268,18 +280,134 @@ namespace MonoDevelop.Ide
 			Counters.Initialization.EndTiming ();
 				
 			AddinManager.AddExtensionNodeHandler("/MonoDevelop/Ide/InitCompleteHandlers", OnExtensionChanged);
-			
+			StartLockupTracker ();
 			IdeApp.Run ();
+
+			IdeApp.Customizer.OnIdeShutdown ();
 			
 			// unloading services
 			if (null != socket_filename)
 				File.Delete (socket_filename);
-			
+			lockupCheckRunning = false;
 			Runtime.Shutdown ();
+
+			IdeApp.Customizer.OnCoreShutdown ();
+
 			InstrumentationService.Stop ();
+			AddinManager.AddinLoadError -= OnAddinError;
 			
 			return 0;
 		}
+
+		static DateTime lastIdle;
+		static bool lockupCheckRunning = true;
+
+		[Conditional("DEBUG")]
+		static void StartLockupTracker ()
+		{
+			if (Platform.IsWindows)
+				return;
+			GLib.Timeout.Add (2000, () => {
+				lastIdle = DateTime.Now;
+				return true;
+			});
+			lastIdle = DateTime.Now;
+			var lockupCheckThread = new Thread (delegate () {
+				while (lockupCheckRunning) {
+					const int waitTimeout = 5000;
+					const int maxResponseTime = 10000;
+					Thread.Sleep (waitTimeout); 
+					if ((DateTime.Now - lastIdle).TotalMilliseconds > maxResponseTime) {
+						var pid = Process.GetCurrentProcess ().Id;
+						Mono.Unix.Native.Syscall.kill (pid, Mono.Unix.Native.Signum.SIGQUIT); 
+						return;
+					}
+				}
+			});
+			lockupCheckThread.Name = "Lockup check";
+			lockupCheckThread.IsBackground = true;
+			lockupCheckThread.Start (); 
+		}
+
+		void SetupTheme ()
+		{
+			// Use the bundled gtkrc only if the Xamarin theme is installed
+			if (File.Exists (Path.Combine (Gtk.Rc.ModuleDir, "libxamarin.so")) || File.Exists (Path.Combine (Gtk.Rc.ModuleDir, "libxamarin.dll"))) {
+				var gtkrc = "gtkrc";
+				if (Platform.IsWindows) {
+					gtkrc += ".win32";
+					var osv = Environment.OSVersion.Version;
+					if (osv.Major == 6 && osv.Minor < 1)
+						gtkrc += "-vista";
+				} else if (Platform.IsMac) {
+					gtkrc += ".mac";
+				}
+				Environment.SetEnvironmentVariable ("GTK2_RC_FILES", PropertyService.EntryAssemblyPath.Combine (gtkrc));
+			}
+		}
+
+		[System.Runtime.InteropServices.DllImport("kernel32.dll", CharSet = System.Runtime.InteropServices.CharSet.Unicode, SetLastError = true)]
+		[return: System.Runtime.InteropServices.MarshalAs(System.Runtime.InteropServices.UnmanagedType.Bool)]
+		static extern bool SetDllDirectory (string lpPathName);
+
+		static bool CheckWindowsGtk ()
+		{
+			string location = null;
+			Version version = null;
+			Version minVersion = new Version (2, 12, 22);
+
+			using (var key = Microsoft.Win32.Registry.LocalMachine.OpenSubKey(@"SOFTWARE\Xamarin\GtkSharp\InstallFolder")) {
+				if (key != null)
+					location = key.GetValue (null) as string;
+			}
+			using (var key = Microsoft.Win32.Registry.LocalMachine.OpenSubKey (@"SOFTWARE\Xamarin\GtkSharp\Version")) {
+				if (key != null)
+					Version.TryParse (key.GetValue (null) as string, out version);
+			}
+
+			//TODO: check build version of GTK# dlls in GAC
+			if (version == null || version < minVersion || location == null || !File.Exists (Path.Combine (location, "bin", "libgtk-win32-2.0-0.dll"))) {
+				LoggingService.LogError ("Did not find required GTK# installation");
+				string url = "http://monodevelop.com/Download";
+				string caption = "Fatal Error";
+				string message =
+					"{0} did not find the required version of GTK#. Please click OK to open the download page, where " +
+					"you can download and install the latest version.";
+				if (DisplayWindowsOkCancelMessage (
+					string.Format (message, BrandingService.ApplicationName, url), caption)
+				) {
+					Process.Start (url);
+				}
+				return false;
+			}
+
+			LoggingService.LogInfo ("Found GTK# version " + version);
+
+			var path = Path.Combine (location, @"bin");
+			try {
+				if (SetDllDirectory (path)) {
+					return true;
+				}
+			} catch (EntryPointNotFoundException) {
+			}
+			// this shouldn't happen unless something is weird in Windows
+			LoggingService.LogError ("Unable to set GTK+ dll directory");
+			return true;
+		}
+
+		static bool DisplayWindowsOkCancelMessage (string message, string caption)
+		{
+			var name = typeof(int).Assembly.FullName.Replace ("mscorlib", "System.Windows.Forms");
+			var asm = Assembly.Load (name);
+			var md = asm.GetType ("System.Windows.Forms.MessageBox");
+			var mbb = asm.GetType ("System.Windows.Forms.MessageBoxButtons");
+			var okCancel = Enum.ToObject (mbb, 1);
+			var dr = asm.GetType ("System.Windows.Forms.DialogResult");
+			var ok = Enum.ToObject (dr, 1);
+
+			const BindingFlags flags = BindingFlags.InvokeMethod | BindingFlags.Public | BindingFlags.Static;
+			return md.InvokeMember ("Show", flags, null, null, new object[] { message, caption, okCancel }).Equals (ok);
+		}
 		
 		public bool Initialized {
 			get { return initialized; }
@@ -321,11 +449,11 @@ namespace MonoDevelop.Ide
 						continue;
 					file += c;
 				}
-				GLib.Idle.Add (delegate(){ return openFile (file); });
+				GLib.Idle.Add (() => OpenFile (file));
 			}
 		}
 
-		bool openFile (string file) 
+		static bool OpenFile (string file) 
 		{
 			if (string.IsNullOrEmpty (file))
 				return false;
@@ -348,25 +476,53 @@ namespace MonoDevelop.Ide
 					MonoDevelop.Projects.Services.ProjectService.IsSolutionItemFile (file)) {
 						IdeApp.Workspace.OpenWorkspaceItem (file);
 				} else {
-						IdeApp.Workbench.OpenDocument (file, line, column);
+					IdeApp.Workbench.OpenDocument (file, null, line, column, OpenDocumentOptions.DefaultInternal);
 				}
 			} catch {
 			}
 			IdeApp.Workbench.Present ();
 			return false;
 		}
-		
-		bool CheckQtCurve ()
+
+		internal static string[] gtkThemeFallbacks = new string[] {
+			"Xamarin",// the best!
+			"Gilouche", // SUSE
+			"Mint-X", // MINT
+			"Radiance", // Ubuntu 'light' theme (MD looks better with the light theme in 4.0 - if that changes switch this one)
+			"Clearlooks" // GTK theme
+		};
+
+		static void ValidateGtkTheme (ref string theme)
 		{
-			if (Gtk.Settings.Default.ThemeName == "QtCurve") {
-				string msg = "QtCurve theme not supported";
-				string desc = "Your system is using the QtCurve GTK+ theme. This theme is known to cause stability issues in MonoDevelop. Please select another theme in the GTK+ Theme Selector.\n\nIf you click on Proceed, MonoDevelop will switch to the default GTK+ theme.";
-				AlertButton res = MessageService.GenericAlert (Gtk.Stock.DialogWarning, msg, desc, AlertButton.Cancel, AlertButton.Proceed);
-				if (res == AlertButton.Cancel)
-					return false;
-				Gtk.Settings.Default.ThemeName = "Gilouche";
+			if (!MonoDevelop.Ide.Gui.OptionPanels.IDEStyleOptionsPanelWidget.IsBadGtkTheme (theme))
+				return;
+
+			var themes = MonoDevelop.Ide.Gui.OptionPanels.IDEStyleOptionsPanelWidget.InstalledThemes;
+
+			string fallback = gtkThemeFallbacks
+				.Select (fb => themes.FirstOrDefault (t => string.Compare (fb, t, StringComparison.OrdinalIgnoreCase) == 0))
+				.FirstOrDefault (t => t != null);
+
+			string message = "Theme Not Supported";
+
+			string detail;
+			if (themes.Count > 0) {
+				detail =
+					"Your system is using the '{0}' GTK+ theme, which is known to be very unstable. MonoDevelop will " +
+					"now switch to an alternate GTK+ theme.\n\n" +
+					"This message will continue to be shown at startup until you set a alternate GTK+ theme as your " +
+					"default in the GTK+ Theme Selector or MonoDevelop Preferences.";
+			} else {
+				detail =
+					"Your system is using the '{0}' GTK+ theme, which is known to be very unstable, and no other GTK+ " +
+					"themes appear to be installed. Please install another GTK+ theme.\n\n" +
+					"This message will continue to be shown at startup until you install a different GTK+ theme and " +
+					"set it as your default in the GTK+ Theme Selector or MonoDevelop Preferences.";
 			}
-			return true;
+
+			MessageService.GenericAlert (Gtk.Stock.DialogWarning, message, BrandingService.BrandApplicationName (detail), AlertButton.Ok);
+
+			theme = fallback ?? themes.FirstOrDefault () ?? theme;
 		}
 		
 		void CheckFileWatcher ()
@@ -380,7 +536,7 @@ namespace MonoDevelop.Ide
 						string msg = "Inotify watch limit is too low (" + n + ").\n";
 						msg += "MonoDevelop will switch to managed file watching.\n";
 						msg += "See http://monodevelop.com/Inotify_Watches_Limit for more info.";
-						LoggingService.LogWarning (msg);
+						LoggingService.LogWarning (BrandingService.BrandApplicationName (msg));
 						Runtime.ProcessService.EnvironmentVariableOverrides["MONO_MANAGED_WATCHER"] = 
 							Environment.GetEnvironmentVariable ("MONO_MANAGED_WATCHER");
 						Environment.SetEnvironmentVariable ("MONO_MANAGED_WATCHER", "1");
@@ -390,87 +546,61 @@ namespace MonoDevelop.Ide
 				LoggingService.LogWarning ("There was a problem checking whether to use managed file watching", e);
 			}
 		}
-		
-		bool CheckBug77135 ()
+
+		bool CheckSCPlugin ()
 		{
-			try {
-				// Check for bug 77135. Some versions of gnome-vfs2 and libgda
-				// make MD crash in the file open dialog or in FileIconLoader.
-				// Only in Suse.
-				
-				string path = "/etc/SuSE-release";
-				if (!File.Exists (path))
-					return true;
-					
-				// Only run the check for SUSE 10
-				StreamReader sr = File.OpenText (path);
-				string txt = sr.ReadToEnd ();
-				sr.Close ();
-				
-				if (txt.IndexOf ("SUSE LINUX 10") == -1)
-					return true;
-					
-				string current_libgda;
-				string current_gnomevfs;
-				string required_libgda = "1.3.91.5.4";
-				string required_gnomevfs = "2.12.0.9.2";
-				
-				StringWriter sw = new StringWriter ();
-				ProcessWrapper pw = Runtime.ProcessService.StartProcess ("rpm", "--qf %{version}.%{release} -q libgda", null, sw, null, null);
-				pw.WaitForOutput ();
-				current_libgda = sw.ToString ().Trim (' ','\n');
-				
-				sw = new StringWriter ();
-				pw = Runtime.ProcessService.StartProcess ("rpm", "--qf %{version}.%{release} -q gnome-vfs2", null, sw, null, null);
-				pw.WaitForOutput ();
-				current_gnomevfs = sw.ToString ().Trim (' ','\n');
-				
-				bool fail1 = Addin.CompareVersions (current_libgda, required_libgda) == 1;
-				bool fail2 = Addin.CompareVersions (current_gnomevfs, required_gnomevfs) == 1;
-				
-				if (fail1 || fail2) {
-					string msg = GettextCatalog.GetString ("Some packages installed in your system are not compatible with MonoDevelop:\n");
-					if (fail1)
-						msg += "\nlibgda " + current_libgda + " ("+ GettextCatalog.GetString ("version required: {0}", required_libgda) + ")";
-					if (fail2)
-						msg += "\ngnome-vfs2 " + current_gnomevfs + " ("+ GettextCatalog.GetString ("version required: {0}", required_gnomevfs) + ")";
-					msg += "\n\n";
-					msg += GettextCatalog.GetString ("You need to upgrade the previous packages to start using MonoDevelop.");
-					
-					SplashScreenForm.SplashScreen.Hide ();
-					Gtk.MessageDialog dlg = new Gtk.MessageDialog (null, Gtk.DialogFlags.Modal, Gtk.MessageType.Error, Gtk.ButtonsType.Ok, msg);
-					dlg.Run ();
-					dlg.Destroy ();
-					
-					return false;
-				} else
-					return true;
-			}
-			catch (Exception ex)
-			{
-				// Just ignore for now.
-				Console.WriteLine (ex);
-				return true;
+			if (Platform.IsMac && Directory.Exists ("/Library/Contextual Menu Items/SCFinderPlugin.plugin")) {
+				string message = "SCPlugin not supported";
+				string detail = "MonoDevelop has detected that SCPlugin (scplugin.tigris.org) is installed. " +
+				                "SCPlugin is a Subversion extension for Finder that is known to cause crashes in MonoDevelop and" +
+				                "other applications running on Mac OSX 10.9 (Mavericks) or upper. Please uninstall SCPlugin " +
+				                "before proceeding.";
+				var close = new AlertButton (BrandingService.BrandApplicationName (GettextCatalog.GetString ("Close MonoDevelop")));
+				var info = new AlertButton (GettextCatalog.GetString ("More Information"));
+				var cont = new AlertButton (GettextCatalog.GetString ("Continue Anyway"));
+				while (true) {
+					var res = MessageService.GenericAlert (Gtk.Stock.DialogWarning, message, BrandingService.BrandApplicationName (detail), info, cont, close);
+					if (res == close) {
+						LoggingService.LogInternalError ("SCPlugin detected", new Exception ("SCPlugin detected. Closing."));
+						return false;
+					}
+					if (res == info)
+						DesktopService.ShowUrl ("https://bugzilla.xamarin.com/show_bug.cgi?id=21755");
+					if (res == cont) {
+						bool exists = Directory.Exists ("/Library/Contextual Menu Items/SCFinderPlugin.plugin");
+						LoggingService.LogInternalError ("SCPlugin detected", new Exception ("SCPlugin detected. Continuing " + (exists ? "Installed." : "Uninstalled.")));
+						return true;
+					}
+				}
 			}
+			return true;
 		}
-		
-		void SetupExceptionManager ()
+
+		static void SetupExceptionManager ()
 		{
+			System.Threading.Tasks.TaskScheduler.UnobservedTaskException += (sender, e) => {
+				HandleException (e.Exception.Flatten (), false);
+				e.SetObserved ();
+			};
 			GLib.ExceptionManager.UnhandledException += delegate (GLib.UnhandledExceptionArgs args) {
 				HandleException ((Exception)args.ExceptionObject, args.IsTerminating);
 			};
 			AppDomain.CurrentDomain.UnhandledException += delegate (object sender, UnhandledExceptionEventArgs args) {
 				HandleException ((Exception)args.ExceptionObject, args.IsTerminating);
 			};
+			Xwt.Application.UnhandledException += (sender, e) => {
+				HandleException (e.ErrorException, false);
+			};
 		}
 		
-		void HandleException (Exception ex, bool willShutdown)
+		static void HandleException (Exception ex, bool willShutdown)
 		{
-			// Log the crash to the MonoDevelop.log file first:
-			LoggingService.LogError (string.Format ("An unhandled exception has occured. Terminating MonoDevelop? {0}", willShutdown), ex);
-			
-			// Pass it off to the reporting service now.
-			LogReportingService.ReportUnhandledException (ex, willShutdown);
+			var msg = String.Format ("An unhandled exception has occured. Terminating MonoDevelop? {0}", willShutdown);
+
+			if (willShutdown)
+				LoggingService.LogFatalError (msg, ex);
+			else
+				LoggingService.LogCriticalError (msg, ex);
 		}
 		
 		/// <summary>SDBM-style hash, bounded to a range of 1000.</summary>
@@ -486,30 +616,42 @@ namespace MonoDevelop.Ide
 			return (int)(hash % 1000);
 		}
 		
-		public static int Main (string[] args)
+		public static int Main (string[] args, IdeCustomizer customizer = null)
 		{
 			var options = MonoDevelopOptions.Parse (args);
 			if (options.ShowHelp || options.Error != null)
 				return options.Error != null? -1 : 0;
 			
 			LoggingService.Initialize (options.RedirectOutput);
-			
+
+			if (customizer == null)
+				customizer = LoadBrandingCustomizer ();
+			options.IdeCustomizer = customizer;
+
 			int ret = -1;
 			bool retry = false;
 			do {
 				try {
-					Runtime.SetProcessName (BrandingService.ApplicationName);
+					var exename = Path.GetFileNameWithoutExtension (Assembly.GetEntryAssembly ().Location);
+					if (!Platform.IsMac && !Platform.IsWindows)
+						exename = exename.ToLower ();
+					Runtime.SetProcessName (exename);
 					var app = new IdeStartup ();
 					ret = app.Run (options);
 					break;
 				} catch (Exception ex) {
 					if (!retry && AddinManager.IsInitialized) {
-						LoggingService.LogWarning (BrandingService.ApplicationName + " failed to start. Rebuilding addins registry.");
+						LoggingService.LogWarning (BrandingService.ApplicationName + " failed to start. Rebuilding addins registry.", ex);
 						AddinManager.Registry.Rebuild (new Mono.Addins.ConsoleProgressStatus (true));
-						LoggingService.LogInfo ("Addin registry rebuilt. Restarting MonoDevelop.");
+						LoggingService.LogInfo ("Addin registry rebuilt. Restarting {0}.", BrandingService.ApplicationName);
 						retry = true;
 					} else {
-						LoggingService.LogFatalError (BrandingService.ApplicationName + " failed to start. Some of the assemblies required to run MonoDevelop (for example gtk-sharp, gnome-sharp or gtkhtml-sharp) may not be properly installed in the GAC.", ex);
+						LoggingService.LogFatalError (
+							string.Format (
+								"{0} failed to start. Some of the assemblies required to run {0} (for example gtk-sharp)" +
+								"may not be properly installed in the GAC.",
+								BrandingService.ApplicationName
+							), ex);
 						retry = false;
 					}
 				} finally {
@@ -518,8 +660,34 @@ namespace MonoDevelop.Ide
 			}
 			while (retry);
 
+			LoggingService.Shutdown ();
+
 			return ret;
 		}
+
+		static IdeCustomizer LoadBrandingCustomizer ()
+		{
+			var pathsString = BrandingService.GetString ("CustomizerAssemblyPath");
+			if (string.IsNullOrEmpty (pathsString))
+				return null;
+
+			var paths = pathsString.Split (new [] {';'}, StringSplitOptions.RemoveEmptyEntries);
+			var type = BrandingService.GetString ("CustomizerType");
+			if (!string.IsNullOrEmpty (type)) {
+				foreach (var path in paths) {
+					var file = BrandingService.GetFile (path.Replace ('/',Path.DirectorySeparatorChar));
+					if (File.Exists (file)) {
+						Assembly asm = Assembly.LoadFrom (file);
+						var t = asm.GetType (type, true);
+						var c = Activator.CreateInstance (t) as IdeCustomizer;
+						if (c == null)
+							throw new InvalidOperationException ("Customizer class specific in the branding file is not an IdeCustomizer subclass");
+						return c;
+					}
+				}
+			}
+			return null;
+		}
 	}
 	
 	public class MonoDevelopOptions
@@ -527,18 +695,18 @@ namespace MonoDevelop.Ide
 		MonoDevelopOptions ()
 		{
 			IpcTcp = (PlatformID.Unix != Environment.OSVersion.Platform);
+			RedirectOutput = true;
 		}
 		
 		Mono.Options.OptionSet GetOptionSet ()
 		{
-			return new Mono.Options.OptionSet () {
-				{ "nologo", "Do not display splash screen.", s => NoLogo = true },
-				{ "ipc-tcp", "Use the Tcp channel for inter-process comunication.", s => IpcTcp = true },
-				{ "newwindow", "Do not open in an existing instance of " + BrandingService.ApplicationName, s => NewWindow = true },
+			return new Mono.Options.OptionSet {
+				{ "no-splash", "Do not display splash screen (deprecated).", s => {} },
+				{ "ipc-tcp", "Use the Tcp channel for inter-process communication.", s => IpcTcp = true },
+				{ "new-window", "Do not open in an existing instance of " + BrandingService.ApplicationName, s => NewWindow = true },
 				{ "h|?|help", "Show help", s => ShowHelp = true },
-				{ "clog", "Log internal counter data", s => LogCounters = true },
-				{ "clog-interval=", "Interval between counter logs (in milliseconds)", (int i) => LogCountersInterval = i },
-				{ "redirect-output", "Whether to redirect stdout/stderr to a log file", s => RedirectOutput = true },
+				{ "perf-log", "Enable performance counter logging", s => PerfLog = true },
+				{ "no-redirect", "Disable redirection of stdout/stderr to a log file", s => RedirectOutput = false },
 			};
 		}
 		
@@ -559,7 +727,7 @@ namespace MonoDevelop.Ide
 			}
 			
 			if (opt.ShowHelp) {
-				Console.WriteLine (BrandingService.ApplicationName + " " + BuildVariables.PackageVersionLabel);
+				Console.WriteLine (BrandingService.ApplicationName + " " + BuildInfo.VersionLabel);
 				Console.WriteLine ("Options:");
 				optSet.WriteOptionDescriptions (Console.Out);
 			}
@@ -567,15 +735,14 @@ namespace MonoDevelop.Ide
 			return opt;
 		}
 		
-		public bool NoLogo { get; set; }
 		public bool IpcTcp { get; set; }
 		public bool NewWindow { get; set; }
 		public bool ShowHelp { get; set; }
-		public bool LogCounters { get; set; }
-		public int LogCountersInterval { get; set; }
+		public bool PerfLog { get; set; }
 		public bool RedirectOutput { get; set; }
 		public string Error { get; set; }
 		public IList<string> RemainingArgs { get; set; }
+		public IdeCustomizer IdeCustomizer { get; set; }
 	}
 	
 	public class AddinError
