@@ -45,17 +45,17 @@ namespace MonoDevelop.Projects
 			Assert.AreEqual (1, sol.Items.Count);
 			Assert.IsTrue (sol.Items [0] is DotNetProject, "Project is DotNetProject");
 			DotNetProject project = (DotNetProject) sol.Items [0];
-			Assert.AreEqual ("csharp-console-mdp", project.Name);
+			Assert.AreEqual ("csharp-console", project.Name);
 			
 			// Check files
 			
 			Assert.AreEqual (2, project.Files.Count, "File count");
 			
-			ProjectFile file = project.GetProjectFile (Path.Combine (project.BaseDirectory, "Main.cs"));
-			Assert.AreEqual ("Main.cs", Path.GetFileName (file.Name));
+			ProjectFile file = project.GetProjectFile (Path.Combine (project.BaseDirectory, "Program.cs"));
+			Assert.AreEqual ("Program.cs", Path.GetFileName (file.Name));
 			Assert.AreEqual (BuildAction.Compile, file.BuildAction);
 			
-			file = project.GetProjectFile (Path.Combine (project.BaseDirectory, "AssemblyInfo.cs"));
+			file = project.GetProjectFile (Path.Combine (project.BaseDirectory, "Properties", "AssemblyInfo.cs"));
 			Assert.AreEqual ("AssemblyInfo.cs", Path.GetFileName (file.Name));
 			Assert.AreEqual (BuildAction.Compile, file.BuildAction);
 			
@@ -63,7 +63,7 @@ namespace MonoDevelop.Projects
 			
 			Assert.AreEqual (1, project.References.Count);
 			ProjectReference pr = project.References [0];
-			Assert.AreEqual (ReferenceType.Gac, pr.ReferenceType);
+			Assert.AreEqual (ReferenceType.Package, pr.ReferenceType);
 			Assert.AreEqual ("System, Version=2.0.0.0, Culture=neutral, PublicKeyToken=b77a5c561934e089", pr.Reference);
 			
 			// Configurations
@@ -109,15 +109,15 @@ namespace MonoDevelop.Projects
 			Assert.AreEqual (3, project.References.Count);
 			
 			ProjectReference pr = project.References [0];
-			Assert.AreEqual (ReferenceType.Gac, pr.ReferenceType);
+			Assert.AreEqual (ReferenceType.Package, pr.ReferenceType);
 			Assert.AreEqual ("System, Version=2.0.0.0, Culture=neutral, PublicKeyToken=b77a5c561934e089", pr.Reference);
 			
 			pr = project.References [1];
-			Assert.AreEqual (ReferenceType.Gac, pr.ReferenceType);
+			Assert.AreEqual (ReferenceType.Package, pr.ReferenceType);
 			Assert.AreEqual ("System.Data, Version=2.0.0.0, Culture=neutral, PublicKeyToken=b77a5c561934e089", pr.Reference);
 			
 			pr = project.References [2];
-			Assert.AreEqual (ReferenceType.Gac, pr.ReferenceType);
+			Assert.AreEqual (ReferenceType.Package, pr.ReferenceType);
 			Assert.AreEqual ("System.Xml, Version=2.0.0.0, Culture=neutral, PublicKeyToken=b77a5c561934e089", pr.Reference);
 			
 			// Configurations
@@ -151,10 +151,10 @@ namespace MonoDevelop.Projects
 			Assert.AreEqual (0, project.Configurations.Count);
 			
 			InitializeProject (dir, project, "TestProject");
-			project.References.Add (new ProjectReference (ReferenceType.Gac, "System, Version=2.0.0.0, Culture=neutral, PublicKeyToken=b77a5c561934e089"));
-			project.References.Add (new ProjectReference (ReferenceType.Gac, "System.Data, Version=2.0.0.0, Culture=neutral, PublicKeyToken=b77a5c561934e089"));
-			project.References.Add (new ProjectReference (ReferenceType.Gac, "System.Xml, Version=2.0.0.0, Culture=neutral, PublicKeyToken=b77a5c561934e089"));
-			project.Files.Add (new ProjectFile (Path.Combine (dir, "Main.cs")));
+			project.References.Add (new ProjectReference (ReferenceType.Package, "System, Version=2.0.0.0, Culture=neutral, PublicKeyToken=b77a5c561934e089"));
+			project.References.Add (new ProjectReference (ReferenceType.Package, "System.Data, Version=2.0.0.0, Culture=neutral, PublicKeyToken=b77a5c561934e089"));
+			project.References.Add (new ProjectReference (ReferenceType.Package, "System.Xml, Version=2.0.0.0, Culture=neutral, PublicKeyToken=b77a5c561934e089"));
+			project.Files.Add (new ProjectFile (Path.Combine (dir, "Program.cs")));
 			project.Files.Add (new ProjectFile (Path.Combine (dir, "Resource.xml"), BuildAction.EmbeddedResource));
 			project.Files.Add (new ProjectFile (Path.Combine (dir, "Excluded.xml"), BuildAction.Content));
 			ProjectFile pf = new ProjectFile (Path.Combine (dir, "Copy.xml"), BuildAction.Content);
@@ -174,7 +174,23 @@ namespace MonoDevelop.Projects
 			Assert.AreEqual (1, scDebug.Configurations.Count);
 			Assert.AreEqual (1, scRelease.Configurations.Count);
 			Assert.AreEqual (2, project.Configurations.Count);
-			
+			foreach (var v in project.Files) {
+				if (v.FilePath.FileName == "Program.cs") {
+				File.WriteAllText (v.FilePath,
+				                   @"
+using System;
+
+namespace Foo {
+	public class MainClass {
+		public static void Main (string [] args)
+		{
+		}
+	}
+}");
+				} else {
+					File.WriteAllText (v.FilePath, v.Name);
+				}
+			}
 			return sol;
 		}
 		
@@ -187,7 +203,7 @@ namespace MonoDevelop.Projects
 			
 			DotNetProject p = (DotNetProject) sol.Items [0];
 			Assert.AreEqual (5, p.Files.Count);
-			Assert.IsTrue (p.Files.GetFile (Path.Combine (p.BaseDirectory, "Main.cs")) != null);
+			Assert.IsTrue (p.Files.GetFile (Path.Combine (p.BaseDirectory, "Program.cs")) != null);
 			Assert.IsTrue (p.Files.GetFile (Path.Combine (p.BaseDirectory, "Resource.xml")) != null);
 			Assert.IsTrue (p.Files.GetFile (Path.Combine (p.BaseDirectory, "Excluded.xml")) != null);
 			Assert.IsTrue (p.Files.GetFile (Path.Combine (p.BaseDirectory, "Copy.xml")) != null);
@@ -209,6 +225,7 @@ namespace MonoDevelop.Projects
 			
 			DotNetProjectConfiguration pcDebug = project.AddNewConfiguration ("Debug") as DotNetProjectConfiguration;
 			CSharpCompilerParameters csparamsDebug = (CSharpCompilerParameters) pcDebug.CompilationParameters;
+			csparamsDebug.DebugType = "full";
 			pcDebug.OutputDirectory = Path.Combine (dir, "bin/Debug");
 			pcDebug.OutputAssembly = name;
 			pcDebug.DebugMode = true;
@@ -217,6 +234,7 @@ namespace MonoDevelop.Projects
 			
 			DotNetProjectConfiguration pcRelease = project.AddNewConfiguration ("Release") as DotNetProjectConfiguration;
 			CSharpCompilerParameters csparamsRelease = (CSharpCompilerParameters) pcRelease.CompilationParameters;
+			csparamsRelease.DebugType = "none";
 			pcRelease.OutputDirectory = Path.Combine (dir, "bin/Release");
 			pcRelease.OutputAssembly = name;
 			csparamsRelease.DefineSymbols = "TRACE";
@@ -226,101 +244,6 @@ namespace MonoDevelop.Projects
 			project.FileName = pfile;
 		}
 		
-		public static Solution CreateProjectWithFolders (string hint)
-		{
-			string dir = Util.CreateTmpDir (hint);
-			Directory.CreateDirectory (Util.Combine (dir, "console-project"));
-			Directory.CreateDirectory (Util.Combine (dir, "nested-solution1"));
-			Directory.CreateDirectory (Util.Combine (dir, "nested-solution1", "library1"));
-			Directory.CreateDirectory (Util.Combine (dir, "nested-solution1", "library2"));
-			Directory.CreateDirectory (Util.Combine (dir, "nested-solution2"));
-			Directory.CreateDirectory (Util.Combine (dir, "nested-solution2", "console-project2"));
-			Directory.CreateDirectory (Util.Combine (dir, "nested-solution2", "nested-solution3"));
-			Directory.CreateDirectory (Util.Combine (dir, "nested-solution2", "nested-solution3", "library3"));
-			Directory.CreateDirectory (Util.Combine (dir, "nested-solution2", "nested-solution3", "library4"));
-			
-			Solution sol = new Solution ();
-			sol.FileName = Path.Combine (dir, "nested-solutions-mdp");
-			SolutionConfiguration scDebug = sol.AddConfiguration ("Debug", true);
-			SolutionConfiguration scRelease = sol.AddConfiguration ("Release", true);
-			
-			DotNetProject project1 = CreateProject (Util.Combine (dir, "console-project"), "C#", "console-project");
-			project1.Files.Add (new ProjectFile (Path.Combine (project1.BaseDirectory, "Main.cs")));
-			project1.Files.Add (new ProjectFile (Path.Combine (project1.BaseDirectory, "AssemblyInfo.cs")));
-			sol.RootFolder.Items.Add (project1);
-			
-			// nested-solution1
-			
-			SolutionFolder folder1 = new SolutionFolder ();
-			sol.RootFolder.Items.Add (folder1);
-			folder1.Name = "nested-solution1";
-			
-			DotNetProject projectLib1 = CreateProject (Util.Combine (dir, "nested-solution1", "library1"), "C#", "library1");
-			projectLib1.References.Add (new ProjectReference (ReferenceType.Gac, "System, Version=2.0.0.0, Culture=neutral, PublicKeyToken=b77a5c561934e089"));
-			projectLib1.Files.Add (new ProjectFile (Path.Combine (projectLib1.BaseDirectory, "MyClass.cs")));
-			projectLib1.Files.Add (new ProjectFile (Path.Combine (projectLib1.BaseDirectory, "AssemblyInfo.cs")));
-			projectLib1.CompileTarget = CompileTarget.Library;
-			folder1.Items.Add (projectLib1);
-			
-			DotNetProject projectLib2 = CreateProject (Util.Combine (dir, "nested-solution1", "library2"), "C#", "library2");
-			projectLib2.References.Add (new ProjectReference (ReferenceType.Gac, "System, Version=2.0.0.0, Culture=neutral, PublicKeyToken=b77a5c561934e089"));
-			projectLib2.Files.Add (new ProjectFile (Path.Combine (projectLib2.BaseDirectory, "MyClass.cs")));
-			projectLib2.Files.Add (new ProjectFile (Path.Combine (projectLib2.BaseDirectory, "AssemblyInfo.cs")));
-			projectLib2.CompileTarget = CompileTarget.Library;
-			folder1.Items.Add (projectLib2);
-			
-			// nested-solution2
-
-			SolutionFolder folder2 = new SolutionFolder ();
-			folder2.Name = "nested-solution2";
-			sol.RootFolder.Items.Add (folder2);
-			
-			DotNetProject project2 = CreateProject (Util.Combine (dir, "nested-solution2", "console-project2"), "C#", "console-project2");
-			project2.Files.Add (new ProjectFile (Path.Combine (project2.BaseDirectory, "Main.cs")));
-			project2.Files.Add (new ProjectFile (Path.Combine (project2.BaseDirectory, "AssemblyInfo.cs")));
-			project2.References.Add (new ProjectReference (ReferenceType.Gac, "System, Version=2.0.0.0, Culture=neutral, PublicKeyToken=b77a5c561934e089"));
-			
-			// nested-solution3
-
-			SolutionFolder folder3 = new SolutionFolder ();
-			folder3.Name = "nested-solution3";
-			
-			DotNetProject projectLib3 = CreateProject (Util.Combine (dir, "nested-solution2", "nested-solution3", "library3"), "C#", "library3");
-			projectLib3.References.Add (new ProjectReference (ReferenceType.Gac, "System, Version=2.0.0.0, Culture=neutral, PublicKeyToken=b77a5c561934e089"));
-			projectLib3.Files.Add (new ProjectFile (Path.Combine (projectLib3.BaseDirectory, "MyClass.cs")));
-			projectLib3.Files.Add (new ProjectFile (Path.Combine (projectLib3.BaseDirectory, "AssemblyInfo.cs")));
-			projectLib3.CompileTarget = CompileTarget.Library;
-			folder3.Items.Add (projectLib3);
-			
-			DotNetProject projectLib4 = CreateProject (Util.Combine (dir, "nested-solution2", "nested-solution3", "library4"), "C#", "library4");
-			projectLib4.References.Add (new ProjectReference (ReferenceType.Gac, "System, Version=2.0.0.0, Culture=neutral, PublicKeyToken=b77a5c561934e089"));
-			projectLib4.Files.Add (new ProjectFile (Path.Combine (projectLib4.BaseDirectory, "MyClass.cs")));
-			projectLib4.Files.Add (new ProjectFile (Path.Combine (projectLib4.BaseDirectory, "AssemblyInfo.cs")));
-			projectLib4.CompileTarget = CompileTarget.Library;
-			folder3.Items.Add (projectLib4);
-			
-			folder2.Items.Add (folder3);
-			folder2.Items.Add (project2);
-			
-			string file = Path.Combine (dir, "TestSolution.sln");
-			sol.FileName = file;
-			
-			project1.References.Add (new ProjectReference (ReferenceType.Gac, "System, Version=2.0.0.0, Culture=neutral, PublicKeyToken=b77a5c561934e089"));
-			project1.References.Add (new ProjectReference (projectLib1));
-			project1.References.Add (new ProjectReference (projectLib2));
-			project1.References.Add (new ProjectReference (projectLib3));
-			project1.References.Add (new ProjectReference (projectLib4));
-			
-			project2.References.Add (new ProjectReference (projectLib3));
-			project2.References.Add (new ProjectReference (projectLib4));
-			
-			Assert.AreEqual (2, sol.Configurations.Count);
-			Assert.AreEqual (6, scDebug.Configurations.Count);
-			Assert.AreEqual (6, scRelease.Configurations.Count);
-			
-			return sol;
-		}
-		
 		public static void CheckGenericItemProject (string fileFormat)
 		{
 			Solution sol = new Solution ();
